
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#ifndef AUTH_H
#define AUTH_H

/*
 * addauth
 *
 * input:  key  - the preprocessed hmac key
 *	   msg  - the data to be authenticated
 *	   len  - the data length
 *	   mode - 1 for real authentication, else 0
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine adds a hmac signature in from of the
 *		given message or clears the hmac area.
 */

static inline int addauth(MD5HMDATA *key,void *msg,int len,int mode)
{
	int i;
	struct timeval tv;
	uint8_t hmac[MD5_SIZE];

	/* sanity check */

	if(UNLIKELY(len<MD5_SIZE+8))BAD("internal error");

	/* clear hmac area */

	memset(msg,0,MD5_SIZE+8);

	/* done if no authentication */

	if(!mode)return 0;

	/* get system time */

	if(UNLIKELY(gettimeofday(&tv,NULL)))BAD("gettimeofday failure");

	/* write time to authentication area */

#if ULONG_MAX == 0xffffffff
	for(i=0;i<4;i++,tv.tv_sec>>=8)((char *)msg)[MD5_SIZE+i]=tv.tv_sec&0xff;
#elif ULONG_MAX == 0xffffffffffffffff
	for(i=0;i<8;i++,tv.tv_sec>>=8)((char *)msg)[MD5_SIZE+i]=tv.tv_sec&0xff;
#else
#error what size is time_t made of?
#endif

	/* create hmac for message */

	md5hmac(msg,len,hmac,key);

	/* copy hmac to message */

	memcpy(msg,hmac,MD5_SIZE);

	/* signal success */

	return 0;
}

/*
 * chkauth
 *
 * input:  key    - the preprocessed hmac key
 *	   msg    - the message to be validated
 *	   len    - the message length
 *	   mode   - 1 if authentocationis enabled, else 0
 *	   window - the time delta window (0 if disabled)
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine validates a given message, if authentication
 *		is enabled. In case of successful authentication the
 *		time window, if specifed, is checked, too, to prevent replay
 *		attacks at least a bit.
 */

static inline int chkauth(MD5HMDATA *key,void *msg,int len,int mode,int window)
{
	int i;
	time_t src;
	struct timeval tv;
	uint8_t hmacmsg[MD5_SIZE];
	uint8_t hmaccmp[MD5_SIZE];

	/* sanity check */

	if(UNLIKELY(len<MD5_SIZE+8))BAD("internal error");

	/* clear authentication area and be done if authentication is
	   not enabled */

	if(!mode)
	{
		memset(msg,0,MD5_SIZE+8);
		return 0;
	}

	/* get current time if a time window is specified */

	if(LIKELY(window))if(UNLIKELY(gettimeofday(&tv,NULL)))
		BAD("gettimeofday failure");

	/* copy message hmac to save buffer */

	memcpy(hmacmsg,msg,MD5_SIZE);

	/* clear message authentication area */

	memset(msg,0,MD5_SIZE);

	/* create hmac of message */

	md5hmac(msg,len,hmaccmp,key);

	/* signal error in case of hmac mismatch */

	if(memcmp(hmaccmp,hmacmsg,MD5_SIZE))
	{
		DEBUG("hmac mismatch");
		return-1;
	}

	/* get time from message */

#if ULONG_MAX == 0xffffffff
	for(src=0,i=0;i<4;i++)
	{
		src<<=8;
		src|=((unsigned char *)msg)[MD5_SIZE+3-i];
	}
#elif ULONG_MAX == 0xffffffffffffffff
	for(src=0,i=0;i<8;i++)
	{
		src<<=8;
		src|=((unsigned char *)msg)[MD5_SIZE+7-i];
	}
#else
#error what size is time_t made of?
#endif

	/* signal error if time window checking is enabled and message time
	   is out of window bounds */

	if(LIKELY(window))if(src<tv.tv_sec-window||src>tv.tv_sec+window)
	{
		DEBUG("time window mismatch");
		return -1;
	}

	/* signal success */

	return 0;
}

/*
 * setauth
 *
 * input:  key     - the pointer to the preprocessed key area
 *	   keydata - the actual input key
 *	   len     - the length of the input key
 *
 * description:	This routine generates the preprocessed hmac key from the
 *		user supplied key.
 */

static inline void setauth(MD5HMDATA *key,void *hash,void *keydata,int len)
{
	md5hmkey(keydata,len,key);
	if(hash)md5hmac(keydata,len,hash,key);
}

/*
 * clrauth
 *
 * input: key - pointer to the proeprocessed key area
 *
 * description:	This routine removes the preprocessed key from memory.
 */

static inline void clrauth(MD5HMDATA *key,void *hash)
{
	memset(key,0,sizeof(MD5HMDATA));
	if(hash)memset(hash,0,MD5_SIZE);
}

#endif
