
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "log.h"
#include "caps.h"

#if defined(__linux__)

/*
 * dropcaps
 *
 * input:  how - the required remaining capabilities
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine drops all capabilities except the requested
 *		ones for the calling thread. The filesystem user and group
 *		id are switched to a non privileged user before capability
 *		dropping, too, if no root access if requested. The caller
 *		can never regain any of the dropped privileges.
 */

int dropcaps(int how)
{
	uid_t uid;
	gid_t gid;
	struct passwd *pw;
	struct group *gr;
	struct passwd pwrk;
	struct group gwrk;
	struct __user_cap_header_struct h;
	struct __user_cap_data_struct d[2];
	char mem[1024];

	/* reset capability header and all capability bits */

	memset(&h,0,sizeof(h));
	memset(d,0,sizeof(d));

	/* get capability version from kernel */

	capget(&h,NULL);
	if(!h.version)BAD("can't determine capabilities version");

	/* set capabilities as requested */

	switch(how)
	{
	/* needs broadcast/multicast, unprivileged fs ugid (fall through):
	   note that CAP_NET_BROADCAST is unused in the kernel but I really
	   can't guess it it is going to be removed or used in the
	   future, so let's play as safe as possible. */

	case NEED_CAST:
#ifdef CAP_NET_BROADCAST
		d[0].effective|=1<<CAP_NET_BROADCAST;
		d[0].permitted|=1<<CAP_NET_BROADCAST;
#endif

	/* needs no capability, unprivileged filesystem ugid (fall through) */

	case NEED_NONE:
		getpwnam_r(NOPRIVUSR,&pwrk,mem,sizeof(mem),&pw);
		if(UNLIKELY(!pw))BAD(NOPRIVUSR ": no such user");
		uid=pw->pw_uid;
		getgrnam_r(NOPRIVGRP,&gwrk,mem,sizeof(mem),&gr);
		if(UNLIKELY(!gr))BAD(NOPRIVGRP ": no such group");
		gid=gr->gr_gid;
		setfsgid(gid);
		setfsuid(uid);

	/* needs no capability, filesystem access as root */

	case NEED_ROOT:
		if(capset(&h,d))BAD("capset: %s",errstr(errno,mem));
		return 0;

	/* needs set*ugid, filesystem access as root */

	case NEED_UGID:
		d[0].effective|=1<<CAP_SETGID;
		d[0].permitted|=1<<CAP_SETGID;
		d[0].effective|=1<<CAP_SETUID;
		d[0].permitted|=1<<CAP_SETUID;
		if(capset(&h,d))BAD("capset: %s",errstr(errno,mem));
		return 0;

	/* sanity check */

	default:BAD("illegal value");
	}
}

#endif
