
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "clntlog.h"
#include "unix.h"
#include "tcp.h"
#include "tcp6.h"
#include "sctp.h"
#include "sctp6.h"
#include "netio.h"
#include "stamp.h"
#include "auth.h"
#include "realpath.h"
#include "version.h"

#define IGNORE_OPT	0		/* ignore option		*/
#define FORCE_LOCAL	1		/* force local compile		*/
#define SET_HINT_S	2		/* output is assembler		*/
#define SET_HINT_C	3		/* compile only is set		*/
#define SET_HINT_O	4		/* output file is specified	*/
#define SCAN_DETAILS	5		/* scan options within option	*/
#define CHK_WALOCAL	6		/* local if -Wa is forced local	*/

typedef struct
{
	const char * const option;	/* option specification		*/
	const int len;			/* comparison string length	*/
	const int skip;			/* option argument skip flag	*/
	const int decision;		/* what to do in case of option	*/
} GCCARGS;

typedef struct
{
	const char * const option;	/* option specification		*/
	const int len;			/* comparison string length	*/
	const int skip;			/* option argument skip flag	*/
} STRIPARGS;

typedef struct
{
	const char * const ext;		/* extension name		*/
	const int hints;		/* extension type identifier	*/
} SRCEXT;

/* compiler options to be processed */

static GCCARGS const args_m[]=
{
	{"-MD",		     0,0,IGNORE_OPT},
	{"-MMD",	     0,0,IGNORE_OPT},
	{"-MG",		     0,0,IGNORE_OPT},
	{"-MP",		     0,0,IGNORE_OPT},
	{"-MF",		     0,1,IGNORE_OPT},
	{"-MT",		     0,1,IGNORE_OPT},
	{"-MQ",		     0,1,IGNORE_OPT},
	{"-MF",		     3,0,IGNORE_OPT},
	{"-MT",		     3,0,IGNORE_OPT},
	{"-MQ",		     3,0,IGNORE_OPT},
	{"-M",		     2,0,FORCE_LOCAL},
	{"-march=native",    0,0,FORCE_LOCAL},
	{"-mtune=native",    0,0,FORCE_LOCAL},
	{NULL,		     0,0,0}
};

static GCCARGS const args_d[]=
{
	{"-dM,",	     3,0,FORCE_LOCAL},
	{"-dD,",	     3,0,FORCE_LOCAL},
	{"-dN,",	     3,0,FORCE_LOCAL},
	{"-dI,",	     3,0,FORCE_LOCAL},
	{"-dU,",	     3,0,FORCE_LOCAL},
	{NULL,		     0,0,0}
};

static GCCARGS const args_f[]=
{
	{"-fprofile-arcs",   0,0,FORCE_LOCAL},
	{"-ftest-coverage",  0,0,FORCE_LOCAL},
	{"-frepo",	     0,0,FORCE_LOCAL},
	{"-fdirectives-only",0,0,FORCE_LOCAL},
	{NULL,		     0,0,0}
};

static GCCARGS const args_x[]=
{
	{"-x",		     2,0,FORCE_LOCAL},
	{"-Xassembler",      0,0,FORCE_LOCAL},
	{"-Xpreprocessor",   0,0,FORCE_LOCAL},
	{NULL,		     0,0,0}
};

static GCCARGS const args_w[]=
{
	{"-Wa,",	     4,0,CHK_WALOCAL},
	{"-Wp,",	     4,0,SCAN_DETAILS},
	{NULL,		     0,0,0}
};

static GCCARGS const args_s[]=
{
	{"-specs=",	     7,0,FORCE_LOCAL},
	{"-S",		     0,0,SET_HINT_S},
	{NULL,		     0,0,0}
};

static GCCARGS const args_o[]=
{
	{"-o",		     0,1,SET_HINT_O},
	{"-o",		     2,0,SET_HINT_O},
	{NULL,		     0,0,0}
};

static GCCARGS const args_e[]=
{
	{"-E",		     0,0,FORCE_LOCAL},
	{NULL,		     0,0,0}
};

static GCCARGS const args_a[]=
{
	{"-a",		     0,0,FORCE_LOCAL},
	{NULL,		     0,0,0}
};

static GCCARGS const args_c[]=
{
	{"-c",		     0,0,SET_HINT_C},
	{NULL,		     0,0,0}
};

static GCCARGS const args_p[]=
{
	{NULL,		     0,0,0}
};

static GCCARGS const args_empty[]=
{
	{NULL,		     0,0,0}
};

static GCCARGS const * const argstab[32]=
{
	args_empty,
	args_a,
	args_empty,
	args_c,
	args_d,
	args_e,
	args_f,
	args_empty,
	args_empty,
	args_empty,
	args_empty,
	args_empty,
	args_empty,
	args_m,
	args_empty,
	args_o,
	args_p,
	args_empty,
	args_empty,
	args_s,
	args_empty,
	args_empty,
	args_empty,
	args_w,
	args_x,
	args_empty,
	args_empty,
	args_empty,
	args_empty,
	args_empty,
	args_empty,
	args_empty
};

/* compuler options to be removed for remote compile */

static STRIPARGS const stripargs_i[]=
{
	{"-I",		   0,1},
	{"-I",		   2,0},
	{"-idirafter",	   0,1},
	{"-include",	   0,1},
	{"-imacros",	   0,1},
	{"-iprefix",	   0,1},
	{"-iwithprefix",   0,1},
	{"-iwithprefixbefore",0,1},
	{"-isystem",	   0,1},
	{"-imultilib",	   0,1},
	{"-isysroot",	   0,1},
	{"-iquote",	   0,1},
	{"-iquote",	   7,0},
	{NULL,		   0,0}
};

static STRIPARGS const stripargs_w[]=
{
	{"-Wcomment",	   9,0},
	{"-Wtrigraphs",	   0,0},
	{"-Wundef",	   0,0},
	{"-Wunused-macros",0,0},
	{"-Wendif-labels", 0,0},
	{"-Wsystem-headers",0,0},
	{"-Wp,",	   4,0},
	{"-Wl,",	   4,0},
	{NULL,		   0,0}
};

static STRIPARGS const stripargs_s[]=
{
	{"-S",		   0,0},
	{"-s",		   0,0},
	{"-static",	   0,0},
	{"-shared",	   0,0},
	{"-shared-libgcc", 0,0},
	{"-static-libgcc", 0,0},
	{"-symbolic",	   0,0},
	{NULL,		   0,0}
};

static STRIPARGS const stripargs_l[]=
{
	{"-l",		   0,1},
	{"-l",		   2,0},
	{"-L",		   0,1},
	{"-L",		   2,0},
	{NULL,		   0,0}
};

static STRIPARGS const stripargs_n[]=
{
	{"-nostdinc",	   9,0},
	{"-nostartfiles",  0,0},
	{"-nodefaultlibs", 0,0},
	{"-nostdlib",	   0,0},
	{NULL,		   0,0}
};

static STRIPARGS const stripargs_f[]=
{
	{"-fworking-directory",0,0},
	{"-fpch-deps",	   0,0},
	{"-fpch-preprocess",0,0},
	{NULL,		   0,0}
};

static STRIPARGS const stripargs_t[]=
{
	{"-trigraphs",	   0,0},
	{"-T",		   0,1},
	{"-T",		   2,0},
	{NULL,		   0,0}
};

static STRIPARGS const stripargs_u[]=
{
	{"-U",		   0,1},
	{"-U",		   2,0},
	{"-u",		   0,1},
	{"-undef",	   0,0},
	{NULL,		   0,0}
};

static STRIPARGS const stripargs_m[]=
{
	{"-MF",		   0,1},
	{"-MT",		   0,1},
	{"-MQ",		   0,1},
	{"-M",		   2,0},
	{"--sysroot",	   0,1},
	{NULL,		   0,0}
};

static STRIPARGS const stripargs_d[]=
{
	{"-D",		   0,1},
	{"-D",		   2,0},
	{NULL,		   0,0}
};

static STRIPARGS const stripargs_o[]=
{
	{"-o",		   0,1},
	{"-o",		   2,0},
	{NULL,		   0,0}
};

static STRIPARGS const stripargs_p[]=
{
	{"-P",		   0,0},
	{"-pie",	   0,0},
	{NULL,		   0,0}
};

static STRIPARGS const stripargs_r[]=
{
	{"-remap",	   0,0},
	{"-rdynamic",	   0,0},
	{NULL,		   0,0}
};

static STRIPARGS const stripargs_v[]=
{
	{"-v",		   0,0},
	{NULL,		   0,0}
};

static STRIPARGS const stripargs_a[]=
{
	{"-A",		   2,0},
	{NULL,		   0,0}
};

static STRIPARGS const stripargs_c[]=
{
	{"-c",		   0,0},
	{NULL,		   0,0}
};

static STRIPARGS const stripargs_h[]=
{
	{"-H",		   0,0},
	{NULL,		   0,0}
};

static STRIPARGS const stripargs_x[]=
{
	{"-Xlinker",	   0,1},
	{NULL,		   0,0}
};

static STRIPARGS const stripargs_empty[]=
{
	{NULL,		   0,0}
};

static STRIPARGS const * const striptab[32]=
{
	stripargs_empty,
	stripargs_a,
	stripargs_empty,
	stripargs_c,
	stripargs_d,
	stripargs_empty,
	stripargs_f,
	stripargs_empty,
	stripargs_h,
	stripargs_i,
	stripargs_empty,
	stripargs_empty,
	stripargs_l,
	stripargs_m,
	stripargs_n,
	stripargs_o,
	stripargs_p,
	stripargs_empty,
	stripargs_r,
	stripargs_s,
	stripargs_t,
	stripargs_u,
	stripargs_v,
	stripargs_w,
	stripargs_x,
	stripargs_empty,
	stripargs_empty,
	stripargs_empty,
	stripargs_empty,
	stripargs_empty,
	stripargs_empty,
	stripargs_empty
};

/* source file extensions */

static SRCEXT const srcext_c[]=
{
	{"c",	HINT_CPP|HINT_STDC},
	{"cc",	HINT_CPP|HINT_CXX},
	{"cpp",	HINT_CPP|HINT_CXX},
	{"CPP",	HINT_CPP|HINT_CXX},
	{"cxx",	HINT_CPP|HINT_CXX},
	{"cp",	HINT_CPP|HINT_CXX},
	{"c++",	HINT_CPP|HINT_CXX},
	{"C",	HINT_CPP},
	{NULL,	0}
};

static SRCEXT const srcext_s[]=
{
	{"s",	HINT_ASM},
	{"S",	HINT_CPP|HINT_ASM},
	{"sx",	HINT_CPP|HINT_ASM},
	{NULL,	0}
};

static SRCEXT const srcext_i[]=
{
	{"i",	HINT_STDC},
	{"ii",	HINT_CXX},
	{NULL,	0}
};

static SRCEXT const srcext_m[]=
{
	{"m",	HINT_CPP|HINT_OBJC},
	{"mm",	HINT_CPP|HINT_OBJCXX},
	{"mi",	HINT_OBJC},
	{"mii",	HINT_OBJCXX},
	{"M",	HINT_CPP|HINT_OBJCXX},
	{NULL,	0}
};

static SRCEXT const srcext_other[]=
{
	{NULL,	0}
};

/* known object file extensions */

static char const * const objext[]=
{
	"o",
	"a",
	NULL
};

/* well known configure script source file names */

static struct
{
	const char * const name;	/* file name		*/
	const int len;			/* comparison length	*/
} const conf[]=
{
	{"conftest."    , 9},
	{"tmp.conftest.",13},
	{NULL           , 0}
};

int verbose=0;
char *in=NULL;
char *app=NETCC;

static int walocal=0;
static char *allowed=NULL;
static char *strip=NULL;
static int cpptmpfd=-1;
static int hints=0;
static pid_t cpppid=0;
static char *out=NULL;
static char *server=NULL;
static int timeout=0;
static int pipefd=-1;
static SHM *shmdta;

/*
 * dump
 *
 * input:  which - the type of command to be executed
 *	   argv  - NULL terminated argument array
 *
 * description: This routine logs the given command in debug mode.
 */

static int dump(char *which,char *argv[])
{
	int i;
	int l;
	char *m;

	/* create argument string from array */

	for(i=0,l=1;argv[i];i++)l+=strlen(argv[i])+1;
	if(UNLIKELY(!(m=malloc(l))))BAD("malloc failure");
	for(i=0,*m=0;argv[i];i++)
	{
		strcat(m,argv[i]);
		strcat(m," ");
	}

	/* output message */

	if(server)ERR("%s @ %s: %s",which,server,m);
	else ERR("%s @ local: %s",which,m);

	/* release working memory */

	free(m);

	/* signal success */

	return 0;
}

/*
 * location
 *
 * input:  argc - the argument array
 *	   argv - the argument count
 *
 * return: LOCAL for local compiles, ANYWHERE for local or remote compiles
 *
 * description:	This routine decides from the given command, if
 *		the compile job can be done local or remotely or
 *		if the compile job must be run locally.
 */

static int location(int argc,char *argv[])
{
	int i;
	int j;
	char *p;
	char *bfr;
	GCCARGS const *gccargs;
	SRCEXT const *srcext;

	/* local job if pathname given or cpp should be run */

	if(strchr(argv[0],'/'))return LOCAL;
	if(!strcmp(argv[0],CPP))return LOCAL;

	/* if allowed list is set local job, if not allowed remote tool */

	if(LIKELY(*allowed))
	{
		for(p=strtok(allowed,":");p;p=strtok(NULL,":"))
			if(!strcmp(p,argv[0]))break;
		if(!p)return LOCAL;
	}

	/* for all command arguments starting with a dash */

	for(i=1;i<argc;i++)if(argv[i][0]=='-')
	{
		/* search option in table */

		for(gccargs=argstab[argv[i][1]&0x1f],j=0;gccargs[j].option;j++)
		{
			if(!gccargs[j].len&&!strcmp(argv[i],gccargs[j].option))
				break;
			else if(gccargs[j].len>0&&!strncmp(argv[i],
				gccargs[j].option,gccargs[j].len))
				break;
			else if(gccargs[j].len<0&&strstr(argv[i],
				gccargs[j].option))
				break;
		}

		/* if the option was found */

		if(gccargs[j].option)
		{
			/* skip option arguments */

			i+=gccargs[j].skip;

			/* handle option descision */

			switch(gccargs[j].decision)
			{
			/* option needs local processing */

			case FORCE_LOCAL:
				return LOCAL;

			/* compile only option found */

			case SET_HINT_C:
				hints|=HINT_C;
				break;

			/* assembler output option found */

			case SET_HINT_S:
				hints|=HINT_S;
				break;

			/* target file name found */

			case SET_HINT_O:
				/* let the local compiler handle multiple
				   output specifications */

				if(UNLIKELY(hints&HINT_OUT))return LOCAL;

				/* memorize target pathname */

				hints|=HINT_OUT;
				out=&argv[i][0]+gccargs[j].len;

				/* let the local compiler handle stdout
				   as a target */

				if(UNLIKELY(!strcmp(out,"-")))return LOCAL;
				break;

			/* scan option details */

			case SCAN_DETAILS:
				/* duplicate option argument for processing */

				if(UNLIKELY(!(bfr=strdup(
					argv[i]+gccargs[j].len))))
						return LOCAL;

				/* for all options in argument */

				for(p=strtok(bfr,",");p;p=strtok(NULL,","))
					if(p[0]=='-')
				{
					/* search option in table */

					gccargs=argstab[p[1]&0x1f];
					for(j=0;gccargs[j].option;j++)
					{
						if(!gccargs[j].len&&
						  !strcmp(p,gccargs[j].option))
							break;
						else if(gccargs[j].len>0&&
						  !strncmp(p,gccargs[j].option,
						    gccargs[j].len))
							break;
						else if(gccargs[j].len<0&&
						  strstr(p,gccargs[j].option))
							break;
					}

					/* if option requires local compile
					   do signal local compile */

					if(gccargs[j].option)
						if(gccargs[j].decision==
							FORCE_LOCAL)
					{
						free(bfr);
						return LOCAL;
					}
				}

				/* release working memory */

				free(bfr);
				break;

			/* enforce local compile if user requested or
			   if no assembler option (not '-Wa,-') */

			case CHK_WALOCAL:
				if(argv[i][gccargs[j].len]!='-')return LOCAL;
				if(walocal)return LOCAL;
				break;
			}
		}
	}

	/* all command arguments not starting with a dash */

	else
	{
		/* compiles with option files must be done locally */

		if(argv[i][0]=='@')return LOCAL;

		/* if we habe a possible source file extension */

		if((p=strrchr(argv[i],'.')))
		{
			/* if we hit an object extension let the local
			   compiler sort this out, possibly linking */

			p++;
			for(j=0;objext[j];j++)if(!strcmp(p,objext[j]))
				return LOCAL;

			/* prepare extension lookup */

			switch(*p&0xdf)
			{
			case 'C':
				srcext=srcext_c;
				break;

			case 'S':
				srcext=srcext_s;
				break;

			case 'I':
				srcext=srcext_i;
				break;

			case 'M':
				srcext=srcext_m;
				break;

			default:srcext=srcext_other;
				break;
			}

			/* if we have a known source extension */

			for(j=0;srcext[j].ext;j++)if(!strcmp(p,srcext[j].ext))
			{
				/* run locally if compiling mutiple sources
				   at the same time */

				if(UNLIKELY(hints&HINT_IN))return LOCAL;

				/* mark source file found */

				hints|=HINT_IN|srcext[j].hints;

				/* compile 'configure' script sources
				   locally */

				for(j=0;conf[j].name;j++)
					if(!strncmp(argv[i],conf[j].name,
						conf[j].len))return LOCAL;

				/* memorize source file pathname */

				in=argv[i];
				break;
			}
		}
	}

	/* run locally if no -c or -S options seen */

	if(UNLIKELY(!(hints&(HINT_C|HINT_S))))return LOCAL;

	/* run locally if no input file found */

	if(UNLIKELY(!(hints&HINT_IN)))return LOCAL;

	/* run locally if only the assembler is to be invoked */

	if(UNLIKELY(hints&HINT_ASM))return LOCAL;

	/* everything else can run locally or remotely */

	return ANYWHERE;
}

/*
 * stderr_output
 *
 * input:  ptr - data to be output
 *	   len - data length
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine output the stderr output of a remote compile.
 */

static int stderr_output(char *ptr,int len)
{
	int olen;
	char *out;
	char *tmp;
	z_stream z;

	/* reset working memory pointer */

	out=NULL;

	/* proceed directly in case of uncompressed input */

	if(LIKELY(!(hints&HINT_ZLIB)))goto print;

	/* prepare zlib */

	memset(&z,0,sizeof(z));
	z.next_in=(void *)ptr;
	z.avail_in=len;

	/* init zlib */

	if(UNLIKELY(inflateInit2(&z,-15)!=Z_OK))BAD("inflateInit2 failure");

	/* prepare output buffer */

	if(UNLIKELY(!(out=malloc(8192))))
	{
		inflateEnd(&z);
		BAD("malloc failure");
	}
	olen=8192;
	z.next_out=(void *)out;
	z.avail_out=olen;

	/* decompress given data */

	while(1)switch(inflate(&z,Z_NO_FLUSH))
	{
	case Z_STREAM_END:
		len=olen-z.avail_out;
		if(UNLIKELY(inflateEnd(&z)!=Z_OK))
		{
			free(out);
			BAD("inflateEnd failure");
		}
		goto print;

	case Z_BUF_ERROR:
		if(UNLIKELY(!(tmp=realloc(out,olen+8192))))
		{
			free(out);
			inflateEnd(&z);
			BAD("malloc failure");
		}
		z.next_out=(void *)(tmp+((char *)z.next_out-out));
		out=tmp;
		z.avail_out+=8192;
		olen+=8192;

	case Z_OK:
		break;

	default:free(out);
		inflateEnd(&z);
		BAD("inflate failure");
	}

	/* output the resulting error message(s) to local stderr */

print:	for(tmp=out?out:ptr;len;tmp+=olen,len-=olen)
		if(UNLIKELY((olen=write(2,tmp,len))==-1))
	{
		olen=errno;
		if(out)free(out);
		BAD("write to stderr: %s",strerror(olen));
	}

	/* release allocated working memory */

	if(out)free(out);

	/* signal success */

	return 0;
}

/*
 * zfile
 *
 * input:  outfd
 *	   infd
 *	   len
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine reads "len" compressed bytes from "infd"
 *		and writes the decompressed data to "outfd".
 */

static int zfile(int outfd,int infd,int len)
{
	int i;
	int eos;
	struct pollfd p;
	z_stream z;
	char inp[65536];
	char out[65536];

	/* done if zero file length */

	if(!len)return 0;

	/* prepare working data */

	memset(&z,0,sizeof(z));

	p.fd=infd;
	p.events=POLLIN;

	eos=0;
	z.next_in=(void *)inp;

	/* prepare zlib */

	if(UNLIKELY(inflateInit2(&z,-15)!=Z_OK))BAD("inflateInit2 failure");

	/* for all compressed data to be processed */

	while(len)
	{
		/* wait for input data */

		if(UNLIKELY((i=poll(&p,1,NETWAIT*1000))!=1))
		{
			if(!i)i=EIO;
			else if((i=errno)==EINTR)continue;
			inflateEnd(&z);
			BAD("poll: %s",strerror(i));
		}

		if(UNLIKELY(!(p.revents&POLLIN)))
		{
			inflateEnd(&z);
			BAD("poll: %s",strerror(EIO));
		}

		/* read input data */

		switch((i=read(infd,z.next_in+z.avail_in,
			sizeof(inp)-z.avail_in)))
		{
		case 0:	errno=EIO;
		case -1:if((i=errno)==EINTR)continue;
			inflateEnd(&z);
			BAD("read: %s",strerror(i));

		default:len-=i;
			z.avail_in+=i;
			if(z.avail_in!=sizeof(inp))continue;
		}

		/* feed input data to zlib */

		do
		{
			/* reset output buffer */

			z.next_out=(void *)out;
			z.avail_out=sizeof(out);

			/* continue decompressing */

			switch(inflate(&z,Z_NO_FLUSH))
			{
			case Z_OK:
			case Z_BUF_ERROR:
				break;

			case Z_STREAM_END:
				if(!len)
				{
					eos=1;
					break;
				}
			default:BAD("inflate failure");
			}

			/* write data to output file */

			i=sizeof(out)-z.avail_out;
			if(UNLIKELY(write(outfd,out,i)!=i))
			{
				i=errno;
				inflateEnd(&z);
				BAD("write: %s",strerror(i));
			}

			/* done if end of compressed stream reached */

			if(eos)goto done;
		} while(z.avail_out!=sizeof(out));

		if(z.avail_in)memmove(inp,z.next_in,z.avail_in);
		z.next_in=(void *)inp;
	}

	/* for all pending output data exceeding the output buffer size */

	do
	{
		/* reset output buffer */

		z.next_out=(void *)out;
		z.avail_out=sizeof(out);

		/* continue decompressing */

		switch((i=inflate(&z,Z_NO_FLUSH)))
		{
		case Z_STREAM_END:
			eos=1;
		case Z_OK:
		case Z_BUF_ERROR:
			break;

		default:BAD("inflate failure");
		}

		/* write data to output file */

		i=sizeof(out)-z.avail_out;
		if(UNLIKELY(write(outfd,out,i)!=i))
		{
			i=errno;
			inflateEnd(&z);
			BAD("write: %s",strerror(i));
		}

		/* done if end of compressed stream reached */

		if(eos)goto done;
	} while(z.avail_out!=sizeof(out));

	/* handle final output data (if any) */

	z.next_out=(void *)out;
	z.avail_out=sizeof(out);

	if(UNLIKELY(inflate(&z,Z_FINISH)!=Z_STREAM_END))
	{
		inflateEnd(&z);
		BAD("inflate finish failure");
	}

	i=sizeof(out)-z.avail_out;
	if(UNLIKELY(write(outfd,out,i)!=i))
	{
		i=errno;
		inflateEnd(&z);
		BAD("write: %s",strerror(i));
	}

	/* deinit zlib */

done:	if(UNLIKELY(inflateEnd(&z)!=Z_OK))BAD("inflateEnd failure");

	/* signal success */

	return 0;
}

/*
 * compressed_cpp
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine reads cpp stdout and compresses to
 *		the temporary source file for remote compilation.
 */

static int compressed_cpp(void)
{
	int i;
	int ifree;
	char *iptr;
	z_stream z;
	char inp[65536];
	char out[65536];

	/* prepare zlib data */

	memset(&z,0,sizeof(z));
	z.data_type=Z_TEXT;

	/* init zlib */

	if(UNLIKELY(deflateInit2(&z,3,Z_DEFLATED,-15,8,Z_FILTERED)!=Z_OK))
		BAD("deflateInit2 failure");

	/* preset processing data */

	z.next_in=(void *)inp;
	z.next_out=(void *)out;
	z.avail_out=sizeof(out);

	iptr=inp;
	ifree=sizeof(inp);

	/* process whole cpp stdout */

	while(1)switch((i=read(pipefd,iptr,ifree)))
	{
	case -1:i=errno;
		deflateEnd(&z);
		BAD("cpp read: %s",strerror(i));

	case 0:	close(pipefd);
		pipefd=-1;
		z.avail_in=0;

		if(UNLIKELY(deflate(&z,Z_FINISH)!=Z_STREAM_END))
		{
			deflateEnd(&z);
			BAD("deflate finish failure");
		}

		i=(int)((char *)z.next_out-out);
		if(LIKELY(i))if(UNLIKELY(write(cpptmpfd,out,i)!=i))
		{
			i=errno;
			deflateEnd(&z);
			BAD("write: %s",strerror(i));
		}

		if(UNLIKELY(deflateEnd(&z)!=Z_OK))
			BAD("deflateEnd failure");

		if(UNLIKELY(lseek(cpptmpfd,0,SEEK_SET)))
			BAD("lseek: %s",strerror(errno));

		return 0;

	default:z.avail_in+=i;

		if(UNLIKELY(deflate(&z,Z_NO_FLUSH)!=Z_OK))
		{
			deflateEnd(&z);
			BAD("deflate failure");
		}

		if(UNLIKELY(z.avail_in))
		{
			if(inp!=(char *)z.next_in)
				memmove(inp,z.next_in,z.avail_in);
			z.next_in=(void *)inp;
			ifree=sizeof(inp)-z.avail_in;
			iptr=inp+z.avail_in;
		}
		else
		{
			z.next_in=(void *)inp;
			ifree=sizeof(inp);
			iptr=inp;
		}

		i=(int)((char *)z.next_out-out);
		if(LIKELY(i))if(UNLIKELY(write(cpptmpfd,out,i)!=i))
		{
			i=errno;
			deflateEnd(&z);
			BAD("write: %s",strerror(i));
		}

		z.next_out=(void *)out;
		z.avail_out=sizeof(out);
		break;
	}
}

/*
 * start_cpp
 *
 * input:  s      - socket handle to netccd or -1
 *	   argc   - argument count
 *	   argv   - argument array
 *	   tmpdir - temporary working directory
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine starts a local cpp as a preparation
 *		for a remote compile.
 */

static int start_cpp(int s,int argc,char *argv[],char *tmpdir)
{
	int i;
	int j;
	int p[2];
	char *ptr;
	char **nv;
	sigset_t n;
	sigset_t o;
	char dummy;

	/* sanity checking */

	if(argc<1)BAD("zero argument count starting preprocessor");

	/* handle aready preprocessed source file */

	if(UNLIKELY(!(hints&HINT_CPP)))
	{
		if((cpptmpfd=open(in,O_RDONLY))!=-1)return 0;
		BAD("can't open preprocessed source %s: %s",in,strerror(errno));
	}

	/* prepare temporary file pathname */

	if(UNLIKELY(!(nv=malloc((argc+1)*sizeof(char *)))))
		BAD("malloc failure");

	if(UNLIKELY(!(ptr=malloc(strlen(tmpdir)+18))))
	{
		free(nv);
		BAD("malloc failure");
	}
	sprintf(ptr,"%s/netcc-tmp-XXXXXX",tmpdir);

	/* create deleted temporary file */

	sigfillset(&n);
	sigprocmask(SIG_BLOCK,&n,&o);
	if(UNLIKELY((cpptmpfd=mkstemp(ptr))==-1))
	{
		i=errno;
		sigprocmask(SIG_SETMASK,&o,NULL);
		free(ptr);
		free(nv);
		BAD("mkstemp failure: %s",strerror(errno));
	}
	unlink(ptr);
	sigprocmask(SIG_SETMASK,&o,NULL);

	/* release work memory */

	free(ptr);

	/* convert compile command to cpp command */

	for(j=0,i=0;i<argc;i++)
	{
		if(!strcmp(argv[i],"-c"))nv[j++]="-E";
		else if(!strcmp(argv[i],"-S"))nv[j++]="-E";
		else if(!strcmp(argv[i],"-o"))i++;
		else if(!strncmp(argv[i],"-o",2));
		else nv[j++]=argv[i];
	}
	nv[j]=NULL;

	/* create stdout pipe for data compression if necessary */

	if(UNLIKELY(hints&HINT_ZLIB))if(UNLIKELY(pipe(p)))
	{
		free(nv);
		BAD("pipe failure: %s",strerror(errno));
	}

	/* wait for cpp slot if cpp running is managed by netccd */

	if(UNLIKELY(s!=-1))if(UNLIKELY(doread(s,&dummy,1)))
	{
		free(nv);
		BAD("daemon failed while waiting for CPP slot");
	}

	/*  handle debug output */

	if(UNLIKELY(verbose>1))dump("preprocessor",nv);

	/* start cpp */

	switch((cpppid=fork()))
	{
	case -1:i=errno;
		cpppid=0;
		free(nv);
		close(cpptmpfd);
		cpptmpfd=-1;
		if(UNLIKELY(hints&HINT_ZLIB))
		{
			close(p[0]);
			close(p[1]);
		}
		BAD("fork failure: %s",strerror(i));

	case 0:	if(UNLIKELY(hints&HINT_ZLIB))
		{
			close(cpptmpfd);
			close(p[0]);
			if(p[1]!=1)
			{
				dup2(p[1],1);
				close(p[1]);
			}
		}
		else if(cpptmpfd!=1)
		{
			dup2(cpptmpfd,1);
			close(cpptmpfd);
		}
		execvp(nv[0],nv);
		FAIL("can't execute %s: %s",nv[0],strerror(errno));

	default:if(UNLIKELY(hints&HINT_ZLIB))
		{
			close(p[1]);
			pipefd=p[0];
		}
		free(nv);
		return 0;
	}
}

/*
 * wait_cpp
 *
 * input:  s - socket handle to netccd or -1
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine waits for termination of a local cpp.
 *		If cpp exits with error execution is terminated.
 */

static int wait_cpp(int s)
{
	int res;
	char dummy;

	/* sanity check */

	if(!cpppid)return 0;

	/* wait for cpp to terminate */

	while(1)switch(waitpid(cpppid,&res,0))
	{
	/* error or cpp not yet erminated */

	case -1:if(errno!=EINTR)
		{
			cpppid=0;
			BAD("waitpid failure: %s",strerror(errno));
		}

	case 0:	continue;

	/* cpp did terminate */

	default:/* mark cpp as terminated, handle cpp errors */

		cpppid=0;
		if(res)
		{
			if(LIKELY(WIFEXITED(res)))exit(WEXITSTATUS(res));
			if(LIKELY(WIFSIGNALED(res)))raise(WTERMSIG(res));
			FAIL("local preprocessor termination %d not handled",
				res);
		}

		/* notify netccd that cpp did terminate if cpp is managed
		   by netccd */

		if(UNLIKELY(s!=-1))dowrite(s,&dummy,1);

		/* rewind */

		if(LIKELY(cpptmpfd!=-1))
			if(UNLIKELY(lseek(cpptmpfd,0,SEEK_SET)==-1))
				BAD("lseek on preprocessed source failed: %s",
					strerror(errno));

		/* signal success */

		return 0;
	}
}

/*
 * mkout
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine creates an output file name from the
 *		input file name.
 */

static int mkout(void)
{
	char *p;

	/* get pointer to file name */

	if(!(p=strrchr(in,'/')))p=in;
	else p++;

	/* duplicate file name */

	if(UNLIKELY(!(out=strdup(p))))BAD("malloc failure");

	/* handle names without extension */

	if(UNLIKELY(!(p=strrchr(out,'.'))))goto rarity;
	else if(UNLIKELY(!*(++p)))goto rarity;

	/* set up proper output filename */

	if(UNLIKELY(hints&HINT_S))strcpy(p,"s");
	else strcpy(p,"o");

	/* signal success */

	return 0;

	/* handle special file names */

rarity:	free(out);

	/* get pointer to file name */

	if(!(p=strrchr(in,'/')))p=in;
	else p++;

	/* allocate sufficient space for output file name */

	if(UNLIKELY(!(out=malloc(strlen(p)+3))))BAD("malloc failure");
	strcpy(out,p);

	/* prepare extension */

	if(UNLIKELY(!strchr(out,'.')))strcat(out,".");
	p=strrchr(out,'.')+1;

	/* set proper output file extension */

	if(UNLIKELY(hints&HINT_S))strcpy(p,"s");
	else strcpy(p,"o");

	/* signal success */

	return 0;
}

/*
 * netstart
 *
 * input:  addr6  - remote IPv6 address or NULL
 *	   addr   - remote IPv4 address or IPv6 local network interface index
 *	   port   - remote port
 *	   ttl    - network ttl
 *	   argc   - argument count
 *	   argv   - argument array
 *	   unique - remote system unique id
 *
 * return: socket handle in case of success, -1 in case of error
 *
 * description:	This routine connects to the remote compile server
 *		and sends the compile command to be executed.
 */

static int netstart(struct in6_addr *addr6,int addr,int port,int ttl,int argc,
	char *argv[],int unique)
{
	int i;
	int s;
	int count;
	uint32_t len;
	struct iovec *iov;
	char **args;
	THDR hdr;
	STRIPARGS const *stripargs;

	/* set up new argument array */

	if(UNLIKELY(!(args=malloc(sizeof(char *)*(argc+1)))))
		BAD("malloc failure");

	/* set up new argument array (strip unwanted options) */

	for(count=1,i=1,args[0]=argv[0];i<argc;i++)if(argv[i][0]=='-')
	{
		for(stripargs=striptab[argv[i][1]&0x1f],s=0;
			stripargs[s].option;s++)
		{
			if(!stripargs[s].len&&!strcmp(argv[i],
				stripargs[s].option))
				break;
			else if(stripargs[s].len>0&&!strncmp(argv[i],
				stripargs[s].option,stripargs[s].len))
				break;
			else if(stripargs[s].len<0&&strstr(argv[i],
				stripargs[s].option))
				break;
		}
		if(stripargs[s].option)i+=stripargs[s].skip;
		else args[count++]=argv[i];
	}
	else if(strcmp(argv[i],in))args[count++]=argv[i];

	/* strip the trashy <bla>-pc-linux-gnu- prefix for remote compile */

	if(strip)
	{
		i=strlen(strip);
		if(!strncmp(args[0],strip,i))args[0]+=i;
	}

	/* insane argument count sanity check */

	if(UNLIKELY(count>0xffff))
	{
		free(args);
		BAD("too many arguments");
	}

	/* debug output if requested */

	if(UNLIKELY(verbose>1))
	{
		args[count]=NULL;
		dump("remote compile",args);
	}

	/* prepare send */

	if(UNLIKELY(!(iov=malloc(sizeof(struct iovec)*(count+1)))))
	{
		free(args);
		BAD("malloc failure");
	}
	iov[0].iov_base=&hdr;
	iov[0].iov_len=sizeof(THDR);
	for(len=0,i=0;i<count;i++)
	{
		iov[i+1].iov_base=args[i];
		iov[i+1].iov_len=strlen(args[i])+1;
		len+=iov[i+1].iov_len;
	}
	i=len+sizeof(THDR);
	hdr.data[0]=htonl(hints&HINT_RMASK);
	hdr.data[1]=htonl(count|(timeout<<16));
	hdr.data[2]=htonl(len);
	hdr.data[3]=htonl(unique);

	/* connect to remote system */

	if(SWITCHES_SUPPORT_MLDV2(hints&HINT_IPV6))
	{
		if(UNLIKELY(ttl>1))addr=0;
		if(hints&HINT_SCTP)
		{
			if(hints&HINT_SCTPAUTH)
			{
				if(UNLIKELY((s=sctp6connect(addr6,port,addr,ttl,
					shmdta->hash,sizeof(shmdta->hash),
					(hints&HINT_SCTPBUG)?1:0))==-1))
					goto fail;
			}
			else if(UNLIKELY((s=sctp6connect(addr6,port,addr,ttl,
				NULL,0,0))==-1))goto fail;
		}
		else if(UNLIKELY((s=tcp6connect(addr6,port,addr,ttl))==-1))
			goto fail;
	}
	else
	{
		if(hints&HINT_SCTP)
		{
			if(hints&HINT_SCTPAUTH)
			{
				if(UNLIKELY((s=sctpconnect(addr,port,ttl,
					shmdta->hash,sizeof(shmdta->hash),
					(hints&HINT_SCTPBUG)?1:0))==-1))
					goto fail;
			}
			else if(UNLIKELY((s=sctpconnect(addr,port,ttl,
				NULL,0,0))==-1))goto fail;
		}
		else if(UNLIKELY((s=tcpconnect(addr,port,ttl))==-1))
		{
fail:			free(args);
			free(iov);
			BAD("can't connect to peer %s",server);
		}
	}

	/* handle cork */

	if(hints&HINT_USECORK)cork(s);

	/* authenticate data */

	if(UNLIKELY(addauth(&shmdta->key,&hdr,sizeof(THDR),hints&HINT_AUTH)))
	{
		close(s);
		free(args);
		free(iov);
		BAD("auth setup failure for peer %s",server);
	}

	/* send header and argument array */

	if(UNLIKELY(dowritev(s,iov,count+1,i)))
	{
		i=errno;
		close(s);
		free(args);
		free(iov);
		BAD("send flags and arguments to peer %s failed: %s",server,
			strerror(i));
	}

	/* handle cork */

	if(hints&HINT_USECORK)uncork(s);

	/* release work memory */

	free(args);
	free(iov);

	/* return socket handle */

	return s;
}

/*
 * network
 *
 * input:  s      - remote system socket handle
 *	   r      - remote compile result code
 *	   unique - remote system unique id
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine
 */

static int network(int s,int *r,int unique)
{
	int i;
	int fd=-1;
	int nosplice;
	int len;
	char *bfr=NULL;
	THDR hdr;
	struct stat stb;
	struct pollfd p;
	clock_t tps;
	clock_t start;
	clock_t now;
	clock_t pingwait;
	char dummy;

	/* handle cork */

	if(hints&HINT_USECORK)cork(s);

	/* send preprocesed code to remote */

	if(UNLIKELY(fstat(cpptmpfd,&stb)))
		BAD("fstat on preprocessed source failed: %s",strerror(errno));
	hdr.data[0]=htonl(stb.st_size);
	if(UNLIKELY(dowrite(s,hdr.data,sizeof(uint32_t))))
		BAD("send source length to peer %s failed: %s",server,
			strerror(errno));

	if(UNLIKELY(dosendfile(s,cpptmpfd,stb.st_size,hints&HINT_SCTP)))
		BAD("send source to peer %s failed: %s",server,strerror(errno));

	/* handle cork */

	if(hints&HINT_USECORK)uncork(s);

	/* tcp shutdown for writing */

	if(!(hints&HINT_SCTP))shutdown(s,SHUT_WR);

	/* select, if later we're using splice for file receive or not:
	   note that on linex as of kernel 3.0 splice with an SCTP
	   socket as source doesn't work */

	nosplice=!((hints&(HINT_SCTP|HINT_SPLICE))==HINT_SPLICE);

	/* in case of oob keepalive pings from remote */

	if(UNLIKELY((hints&(HINT_OOB|HINT_SCTP))==HINT_OOB))
	{
		/* prepare waiting data */

		tps=sysconf(_SC_CLK_TCK);
		start=timestamp();
		p.fd=s;
		p.events=POLLIN|POLLPRI;
		timeout*=tps;
		pingwait=(OOBINTERVAL+1)*tps;

		/* wait, until either a ping time out happens, a compile
		   timeout happens (if set) or the remote compile completes */

		while(1)
		{
			if(UNLIKELY(timeout))
			{
				now=timestamp()-start;
				if(now>timeout)goto tmo;
				now=timeout-now;
				if(now>pingwait)now=pingwait;
			}
			else now=pingwait;
			now/=tps;
			now*=1000;
			while(poll(&p,1,now)<0);
			if(p.revents&POLLPRI)
			{
				if(recv(s,&dummy,1,MSG_OOB)!=1)
					BAD("recv: %s",strerror(errno));
				if(!(p.revents&POLLIN))continue;
			}
			if(p.revents&POLLIN)break;
			BAD("missing ping from remote compile server");
		}
	}

	/* in case of a simple timeout */

	else if(UNLIKELY(timeout))
	{
		/* wait for remote compile to finsh or for timeout to happen */

		p.fd=s;
		p.events=POLLIN;
		if(UNLIKELY(poll(&p,1,timeout*1000)!=1))goto tmo;
		if(UNLIKELY(!(p.revents&POLLIN)))
		{
tmo:			BAD("timeout waiting for remote compiler");
		}
	}

	/* read compile result header from remote */

	if(UNLIKELY(doread(s,&hdr,sizeof(THDR))))
		BAD("receive status from peer %s failed: %s",server,
			strerror(errno));

	/* validate header authentication */

	if(UNLIKELY(chkauth(&shmdta->key,&hdr,sizeof(THDR),hints&HINT_AUTH,
		shmdta->window)))BAD("peer reply authentication failure");

	/* extract header data */

	*r=ntohl(hdr.data[0]);
	len=ntohl(hdr.data[1]);
	i=ntohl(hdr.data[2]);

	/* validate remote system unique id */

	if(UNLIKELY(ntohl(hdr.data[3])!=unique))BAD("peer unique id mismatch");

	/* handle output file according to remote compile result */

	if(UNLIKELY(hdr.data[0]))unlink(out);
	else if(UNLIKELY((fd=open(out,O_WRONLY|O_CREAT|O_TRUNC,0644))==-1))
		BAD("can't open output file %s: %s",out,strerror(errno));

	/* handle stderr from remote compile */

	if(len)
	{
		/* read remote stderr output to local buffer */

		if(UNLIKELY(!(bfr=malloc(len))))BAD("malloc failure");
		if(UNLIKELY(doread(s,bfr,len)))
		{
			i=errno;
			free(bfr);
			close(fd);
			BAD("receive stderr from peer %s failed: %s",server,
				strerror(i));
		}
	}

	/* read output file from remote, if there is any */

	if(UNLIKELY(hints&HINT_ZLIB))
	{
		if(UNLIKELY(zfile(fd,s,i)))
		{
			if(bfr)free(bfr);
			close(fd);
			unlink(out);
			BAD("receive compressed object from peer %s failed",
				server);
		}
	}
	else if(UNLIKELY(dorecvfile(fd,s,i,nosplice)))
	{
		i=errno;
		if(bfr)free(bfr);
		close(fd);
		unlink(out);
		BAD("receive object from peer %s failed: %s",server,
			strerror(i));
	}

	/* close local output file, if any */

	if(LIKELY(fd!=-1))close(fd);

	/* tcp shutdown for reading */

	if(!(hints&HINT_SCTP))shutdown(s,SHUT_RD);

	/* in case of stderr output from remote */

	if(bfr)
	{
		/* output to local stderr */

		stderr_output(bfr,len);
		free(bfr);
	}

	/* signal success */

	return 0;
}

/*
 * doclearenv
 *
 * description:	This routine removes the requested environment variables
 *		from the environment.
 */

static inline void doclearenv(void)
{
	char *str;
	char clrenv[CLRENV];

	if(LIKELY(!shmdta->clrenv[0]))return;
	strcpy(clrenv,shmdta->clrenv);
	for(str=strtok(clrenv,":");str;str=strtok(NULL,":"))unsetenv(str);
}

/*
 * usage
 *
 * description:	This routine prints usage information and terminates
 *		processing.
 */

static NORETURN void usage(void)
{
	/* WARNING: You are *NOT* allowed to remove copyrignt and version
	  from the usage message - you are also *NOT* allowed to remove
	  the usage function from the program in *ANY* way */

	fprintf(stderr,"Usage: " NETCC " compile-command compile-options\n"
		COPYRIGHT "\n"
		"Version: " VERSION "\n"
		"'" NETCC " -h' displays this help text\n"
		"Environment variables used:\n"
		SOCKNAME "   socket path (default: " SOCKET ")\n"
		LOGNAME "      logfile instead of standard error (default: unset)\n"
		TIMEOUT "  timeout for remote compiles if set (30-300)\n"
		ALLOWNAME "    allowed remote commands (default: " ALLOWEDRMT ")\n"
		VERBOSE "  enable verbose output if set\n"
		DBGFLAG "    enable debug output if set\n"
		LOCALFLAG "    force local compile if set\n"
		WALOCALFLAG "  force local compile if '-Wa' option given\n");
	exit(1);
}

/*
 * main
 *
 * input:  argc - the argument count
 *	   argv - the argument array
 *
 * return: 0 in case of success, a value > 0 in case of error
 *
 * description:	This routine handles local as well as remote compiles
 *		and interacts with the netccd daemon for load management.
 */

int main(int argc,char *argv[])
{
	int r;
	int s;
	int c=-1;
	int f=0;
	int l=0;
	int shmfd;
	char *str;
	char *p;
	char *sp=SOCKET;
	sigset_t t;
	sigset_t o;
	REQ req;
	MSG msg;
	struct in_addr srv;
	char name[INET6_ADDRSTRLEN];

	/* sanity check */

	if(UNLIKELY(argc<1))FAIL("no arguments given");

	/* get pointer to file name */

	if(!(p=strrchr(argv[0],'/')))p=argv[0];
	else p++;

	/* strip self from arguments, if necessary */

	if(!strcmp(p,NETCC))
	{
		argc--;
		argv++;
		if(UNLIKELY(argc<1))FAIL("no arguments given");
		if(UNLIKELY(argc==1))if(UNLIKELY(!strcmp(argv[0],"-h")))usage();
	}

	/* convert absolute paths to the shim directory from paranoid
	   setups to a relative tool path */

	if(UNLIKELY((p=strchr(argv[0],'/'))!=NULL))
		if(LIKELY((p=realpath(argv[0],NULL))!=NULL))
	{
		if((str=strrchr(p,'/')))if(!strcmp(str+1,NETCC))
			argv[0]=strrchr(argv[0],'/')+1;
		free(p);
	}

	/* assert that we have a command to execute */

	if(UNLIKELY(!argv[0][0]))FAIL("no command specified");

	/* retrieve environment flags */

	if(UNLIKELY(getenv(VERBOSE)!=NULL))verbose=1;
	if(UNLIKELY(getenv(DBGFLAG)!=NULL))verbose=2;
	if(UNLIKELY(getenv(LOCALFLAG)!=NULL))l=1;
	if(UNLIKELY(getenv(WALOCALFLAG)!=NULL))walocal=1;

	/* retrieve and check other environment values */

	if(UNLIKELY((p=getenv(TIMEOUT))!=NULL))
	{
		timeout=atoi(p);
		if(timeout<30||timeout>300)FAIL("illegal timeout value %s",p);
	}

	if((str=getenv(STRIPNAME)))strip=str;
	if((str=getenv(ALLOWNAME)))allowed=str;
	if((str=getenv(SOCKNAME)))sp=str;
	if(UNLIKELY(!*sp))FAIL("illegal socket pathname");

	/* sanity check for strip option */

	if(strip)if(UNLIKELY(!strcmp(argv[0],strip)))
		FAIL("no command specified");

	/* set up default allowed tool list, if no user specified list given */

	if(LIKELY(!allowed))allowed=ALLOWEDRMT;
	if(UNLIKELY(!(allowed=strdup(allowed))))FAIL("malloc failure");

	/* determine if forced local compile or anywhere compile */

	if(UNLIKELY(l))req.where=LOCAL;
	else req.where=location(argc,argv);
	req.extra=hints;
	free(allowed);

	/* connect to netccd */

	if(UNLIKELY((s=unixconnect(sp))==-1))FAIL("can't connect to daemon");

	/* send request */

	if(UNLIKELY(dowrite(s,&req,sizeof(req))))
		FAIL("can't send to daemon: %s",strerror(errno));

	/* get reply including shared memory fd */

	if(UNLIKELY(doreadfd(s,&msg,sizeof(msg),&shmfd)))
		FAIL("can't receive from daemon: %s",strerror(errno));
	fcntl(shmfd,F_SETFD,FD_CLOEXEC);

	/* access shared memory */

	if((shmdta=mmap(NULL,sizeof(SHM),PROT_READ,MAP_SHARED,shmfd,0))==
		MAP_FAILED)FAIL("can't access shared memory");

	/* update processing hints */

	hints|=msg.extra;

	/* clear requested environment */

	doclearenv();

	/* set up path defined by netccd */

	if(UNLIKELY(putenv(shmdta->path)))FAIL("can't setup PATH");

	/* convert cc->gcc and c++->g++ if requested */

	if(hints&HINT_ZAP)
	{
		if(!strcmp(argv[0],CC))argv[0]=GCC;
		else if(!strcmp(argv[0],CXX))argv[0]=GXX;
	}

	/* handle local compiles */

	if(!msg.addr)goto local;

	/* block unwanted signals */

	sigemptyset(&t);
	sigaddset(&t,SIGPIPE);
	sigaddset(&t,SIGCHLD);
	sigprocmask(SIG_BLOCK,&t,&o);
	sigaddset(&t,SIGALRM);
	sigaddset(&t,SIGUSR1);
	sigaddset(&t,SIGUSR2);
	sigaddset(&t,SIGCONT);
	sigaddset(&t,SIGSTOP);
	sigaddset(&t,SIGTSTP);
	sigaddset(&t,SIGTTIN);
	sigaddset(&t,SIGTTOU);
	sigprocmask(SIG_BLOCK,&t,&o);

	/* debug and error output support */

	if(SWITCHES_SUPPORT_MLDV2(hints&HINT_IPV6))
		server=(char *)inet_ntop(AF_INET6,&msg.addr6,name,sizeof(name));
	else
	{
		srv.s_addr=htonl(msg.addr);
		server=(char *)inet_ntop(AF_INET,&srv,name,sizeof(name));
	}

	/* do remote processing, retry locally in case of failure */

	if(UNLIKELY(start_cpp(UNLIKELY(msg.extra&HINT_CPP)?s:-1,argc,argv,
		shmdta->tmpdir+7)))REDO("can't run preprocessor");
	if(UNLIKELY((f=c=netstart(&msg.addr6,msg.addr,msg.port,msg.ttl,
		argc,argv,msg.unique))==-1))
			REDO("remote compile setup @ %s failed",server);
	if(UNLIKELY(!(hints&HINT_OUT)))if(UNLIKELY(mkout()))
		REDO("malloc failed");
	if(UNLIKELY(hints&HINT_ZLIB))if(UNLIKELY(compressed_cpp()))
		REDO("compression failed");
	if(UNLIKELY(wait_cpp(UNLIKELY(msg.extra&HINT_CPP)?s:-1)))
		REDO("preprocessor failure");
	if(UNLIKELY(network(c,&r,msg.unique)))
		REDO("remote compile @ %s failed",server);
	if(UNLIKELY(!(hints&HINT_OUT)))if(LIKELY(out!=NULL))free(out);
	close(cpptmpfd);
	close(c);
	close(s);

	/* restore old signal mask */

	sigprocmask(SIG_SETMASK,&o,NULL);

	/* simulate compiler exit for normal termination */

	if(LIKELY(WIFEXITED(r)))return WEXITSTATUS(r);

	/* retry locally if remote compiler gotr a signal */

	if(LIKELY(WIFSIGNALED(r)))
	{
		ERR("remote compiler @ %s got signal %d, trying local compile",
			server,WTERMSIG(r));
		goto redo2;
	}

	/* aiehhh, should not happen */

	FAIL("internal error: remote compile @ %s termination %d not handled",
		server,r);

	/* clean up remote processing */

redo:	if(pipefd!=-1)close(pipefd);
	if(cpptmpfd!=-1)
	{
		close(cpptmpfd);
		cpptmpfd=-1;
	}
	if(c!=-1)
	{
		if(!(hints&HINT_SCTP))shutdown(c,SHUT_RDWR);
		close(c);
	}
	close(s);
	if(!(hints&HINT_OUT))if(out)free(out);
	wait_cpp(-1);

	sigprocmask(SIG_SETMASK,&o,NULL);

	/* notify netccd that remote interworking failed */

	if(f==-1)
	{
		req.where=NOSERVICE;
		req.extra=msg.addr;
		req.addr6=msg.addr6;

		if(UNLIKELY((s=unixconnect(sp))==-1))
			FAIL("can't connect to daemon");

		if(UNLIKELY(dowrite(s,&req,sizeof(req))))
			FAIL("can't send to daemon: %s",strerror(errno));

		close(s);
	}

	/* retry with a local compile */

redo2:	req.where=LOCAL;

	if(UNLIKELY((s=unixconnect(sp))==-1))FAIL("can't connect to daemon");

	if(UNLIKELY(dowrite(s,&req,sizeof(req))))
		FAIL("can't send to daemon: %s",strerror(errno));
	if(UNLIKELY(doreadfd(s,&msg,sizeof(msg),&shmfd)))
		FAIL("can't receive from daemon: %s",strerror(errno));
	fcntl(shmfd,F_SETFD,FD_CLOEXEC);

	/* debug output if requested */

local:	if(UNLIKELY(verbose>1))
	{
		server=NULL;
		dump("compile",argv);
	}

	/* set up tmpdir environment variable */

	if(UNLIKELY(putenv(shmdta->tmpdir)))FAIL("can't set TMPDIR");

	/* compile locally, act according to compile result - we must do it
	   this way to keep the netccd socket handle to ourselves and
	   open during the compile - the close of the handle informs
	   netccd that the comile is complete */

	switch(fork())
	{
	case 0:	if(msg.nice)r=nice(msg.nice);
		execvp(argv[0],argv);
		FAIL("can't execute %s",argv[0]);
	case -1:FAIL("fork failed");
	default:while(1)
		{
			switch(wait(&r))
			{
			case -1:if(LIKELY(errno==EINTR))continue;
				FAIL("can't wait for compiler: %s",
					strerror(errno));
			}
			break;
		}
		close(s);
		if(LIKELY(WIFEXITED(r)))return WEXITSTATUS(r);
		if(LIKELY(WIFSIGNALED(r)))raise(WTERMSIG(r));
		FAIL("internal error: local compile termination %d not handled",
			r);
	}
}
