
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "unix.h"
#include "clntlog.h"
#include "netio.h"
#include "version.h"

int verbose=0;
char *in=NULL;
char *app=NETCCDCTL;

/*
 * usage
 *
 * description:	This routine prints a usage message and exits.
 */

static NORETURN void usage(void)
{
	/* WARNING: You are *NOT* allowed to remove copyrignt and version
	  from the usage message - you are also *NOT* allowed to remove
	  the usage function from the program in *ANY* way */

	fprintf(stderr,"Usage: " NETCCDCTL " OPTIONS\n"
		COPYRIGHT "\n"
		"Version: " VERSION "\n"
		"OPTTIONS:\n"
		"-s             socket-path (default " SOCKET ")\n"
		"-v             enable verbose output\n"
		"-c             enable client only mode\n"
		"-C             disable client only mode\n"
		"-l             enable local only mode\n"
		"-L             disable local only mode\n"
		"-b             use slow broadcast speed\n"
		"-B             use fast broadcast speed\n"
		"-e             use heavy load setting\n"
		"-E             use lighter load setting\n"
		"-z             disable remote compile transport compression\n"
		"-Z             enable remote compile transport compression\n"
		"-k             disable TCP_CORK data coalescence\n"
		"-K             enable TCP_CORK data coalescence\n"
		"-F cpu-speed   set slow/fast cpu speed boundary in MHz (default " STR(SLOWCPU) ")\n"
		"-n nice-value  set compile time nice value (0-20, default " STR(NICE) ")\n"
		"-i interface   update or change network interface\n"
		"-h             this help text\n");
	exit(1);
}

/*
 * main
 *
 * input:  argc - the argument counts
 *	   argv - argument pointer array
 *
 * return: 0 in case of success, a value > 0 in case of error
 *
 * description:	This routine connects to nettccd and has netccd modify
 *		the runtime configurable options.
 */

int main(int argc,char *argv[])
{
	int i;
	char *sp;
	REQ req;
	CTLDATA ctl;
	CTLSTATUS res;

	/* assert that no logging to file takes place */

	unsetenv(LOGNAME);

	/* data preset */

	memset(&ctl,0,sizeof(ctl));
	sp=SOCKET;
	req.where=CONTROL;

	/* process command line arguments */

	while((i=getopt(argc,argv,"s:cClLF:n:i:vhbBeEzZkK"))!=-1)switch(i)
	{
	case 's':
		if(!*optarg)FAIL("illegal socket path");
		sp=optarg;
		break;

	case 'c':
		ctl.map|=MAP_CLIENTONLY;
		ctl.clientonly=1;
		break;

	case 'C':
		ctl.map|=MAP_CLIENTONLY;
		ctl.clientonly=0;
		break;

	case 'l':
		ctl.map|=MAP_LOCALONLY;
		ctl.justlocal=1;
		break;

	case 'L':
		ctl.map|=MAP_LOCALONLY;
		ctl.justlocal=0;
		break;

	case 'b':
		ctl.map|=MAP_BCASTSPEED;
		ctl.slowbcast=1;
		break;

	case 'B':
		ctl.map|=MAP_BCASTSPEED;
		ctl.slowbcast=0;
		break;

	case 'e':
		ctl.map|=MAP_SYSLOAD;
		ctl.lightload=0;
		break;

	case 'E':
		ctl.map|=MAP_SYSLOAD;
		ctl.lightload=1;
		break;

	case 'z':
		ctl.map|=MAP_ZLIB;
		ctl.zlib=0;
		break;

	case 'Z':
		ctl.map|=MAP_ZLIB;
		ctl.zlib=1;
		break;

	case 'k':
		ctl.map|=MAP_CORK;
		ctl.cork=0;
		break;

	case 'K':
		ctl.map|=MAP_CORK;
		ctl.cork=1;
		break;

	case 'F':
		i=atoi(optarg);
		if(i<0)FAIL("illegal cpu speed");
		ctl.bogoslow=i;
		ctl.map|=MAP_FASTCPU;
		break;

	case 'n':
		i=atoi(optarg);
		if(i<0||i>20)FAIL("illegal nice value");
		ctl.nice=i;
		ctl.map|=MAP_NICE;
		break;

	case 'i':
		if(!*optarg||strlen(optarg)>=sizeof(ctl.dev))
			FAIL("illegal interface name");
		strcpy(ctl.dev,optarg);
		ctl.map|=MAP_DEVICE;
		break;

	case 'v':
		verbose=1;
		break;

	case 'h':
	default:usage();
	}

	if(optind<argc)usage();

	/* assert that we're root */

	if(UNLIKELY(getuid()||geteuid()||getgid()||getgid()))
		FAIL("must be root");

	/* communicate with netccd */

	if(UNLIKELY((i=unixconnect(sp))==-1))
		FAIL("can't connect to daemon");
	if(UNLIKELY(dowrite(i,&req,sizeof(req))))
		FAIL("can't send");
	if(UNLIKELY(dowrite(i,&ctl,sizeof(ctl))))
		FAIL("can't send");
	if(UNLIKELY(doread(i,&res,sizeof(res))))
		FAIL("can't receive");
	close(i);

	/* print any parameter change failures */

	if(res.map)
	{
		if(res.map&MAP_CLIENTONLY)ERR("can't change client only mode");
		if(res.map&MAP_LOCALONLY)ERR("can't change local only mode");
		if(res.map&MAP_ZLIB)ERR("can't change compression mode");
		if(res.map&MAP_CORK)ERR("can't change TCP_CORK mode");
		if(res.map&MAP_FASTCPU)ERR("can't set fast cpu frequency");
		if(res.map&MAP_NICE)ERR("can't set nice value");
		if(res.map&MAP_BCASTSPEED)ERR("can't set broadcast speed");
		if(res.map&MAP_SYSLOAD)ERR("can't set system load");
		if(res.map&MAP_DEVICE)
			ERR("can't change/update network interface");

		/* signal error */

		return 1;
	}

	/* signal success */

	return 0;
}
