
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "stats.h"
#include "ifc.h"
#include "unix.h"
#include "update.h"
#include "miniweb.h"
#include "remote.h"
#include "term.h"
#include "lock.h"
#include "log.h"
#include "safeapi.h"
#include "netio.h"
#include "sigs.h"
#include "auth.h"
#include "version.h"
#include "watch.h"
#include "stamp.h"
#include "ipc.h"
#include "fqdn.h"
#include "caps.h"

typedef struct _client
{
	struct _client *next;	/* double linked list prev element	*/
	struct _client *prev;	/* linked list next element		*/
	int addr;		/* IPv4 address or IPv6 interface index	*/
	int stamp;		/* used by udp functionality		*/
	int fd;			/* client socket handle			*/
	int cpp;		/* cpp limiting processing state	*/
	struct in6_addr a6;	/* IPv6 address				*/
} CLIENT;

typedef struct
{
	pthread_t h;
	pthread_mutex_t mtx;
	sem_t sem;
	int notify;
	int ipcinit;
	CLIENT *waithead;
	CLIENT *waittail;
	CLIENT *runhead;
	CLIENT *runtail;
} CPPQ;

static int running=1;
static int port=PORT;
static int niceval=NICE;
static int extra=0;
static int flags=0;
static int local=0;
static int maxjobs=MAXJOBS;
static int maxcpp=0;
static int auth=0;
static int v6=0;
static int ttl=1;
static int ipckey=0;
static int idx;
static int shmfd;
static SHM *shmdta;

/*
 * handler
 *
 * input:  depends on how signals are handled
 *
 * return: depends on how signals are handled
 *
 * description: This routine is the signal handler. It either
 *		get called traditionally when a signal is
 *		invoked or prefessably polls a file
 *		descriptor that becomes readable in case
 *		of a signal. It calls the termination notifier
 *		routine when a signal is catched.
 */

#if defined(__linux__) || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
static void *handler(void *data)
{
	struct pollfd p[2];

	dropcaps(NEED_NONE);

	p[0].fd=(int)((long)data);
	p[0].events=POLLIN;
	p[1].fd=termget();
	p[1].events=POLLIN;

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
	while(poll(p,2,-1)<=0);
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	running=0;
	termnotify();
	close(p[0].fd);
	pthread_exit(NULL);
}
#else
static void handler(int unused)
{
	running=0;
	termnotify();
}
#endif

/*
 * docontrol
 *
 * input: s - socket handle
 *
 * description:	This routine handles the modification requests of
 *		netccd-ctl.
 */

static void docontrol(int s)
{
	int uid;
	int gid;
	CTLDATA ctl;
	CTLSTATUS res;
	char mem[256];

	/* assert that the client is root */

	if(UNLIKELY(getpeerid(s,&uid,&gid)))return;

	if(UNLIKELY(uid||gid))
	{
		WARN("illegal control connect, uid=%d gid=%d",uid,gid);
		return;
	}

	/* read client request and reset result */

	if(UNLIKELY(doread(s,&ctl,sizeof(ctl))))
	{
		NOTICE("read control data failure: %s",errstr(errno,mem));
		return;
	}

	res.map=0;

	/* slow fast cpu threshold change */

	if(ctl.map&MAP_FASTCPU)
	{
		if(UNLIKELY(ctl.bogoslow<=0))
		{
			NOTICE("illegal cpu data %d supplied",ctl.bogoslow);
			res.map|=MAP_FASTCPU;
		}
		else if(UNLIKELY(update_mode(FAST_CPU_MHZ,ctl.bogoslow,NULL)))
			res.map|=MAP_FASTCPU;
	}

	/* client only mode change */

	if(ctl.map&MAP_CLIENTONLY)
	{
		if(UNLIKELY(ctl.clientonly<0||ctl.clientonly>1))
		{
			NOTICE("illegal client only flag data %d supplied",
				ctl.clientonly);
			res.map|=MAP_CLIENTONLY;
		}
		else if(UNLIKELY(update_mode(CLIENT_ONLY,ctl.clientonly,NULL)))
			res.map|=MAP_CLIENTONLY;
	}

	/* localonly mode change */

	if(ctl.map&MAP_LOCALONLY)
	{
		if(UNLIKELY(ctl.justlocal<0||ctl.justlocal>1))
		{
			NOTICE("illegal local only flag data %d supplied",
				ctl.justlocal);
			res.map|=MAP_LOCALONLY;
		}
		else if(UNLIKELY(update_mode(LOCAL_ONLY,ctl.justlocal,NULL)))
			res.map|=MAP_LOCALONLY;
	}

	/* transport data compression mode change */

	if(ctl.map&MAP_ZLIB)
	{
		if(UNLIKELY(ctl.zlib<0||ctl.zlib>1))
		{
			NOTICE("illegal zlib flag data %d supplied",ctl.zlib);
			res.map|=MAP_ZLIB;
		}
		else
		{
			if(!ctl.zlib)
			{
				flags&=~HINT_ZLIB;
				extra&=~HINT_ZLIB;
			}
			else
			{
				flags|=HINT_ZLIB;
				extra|=HINT_ZLIB;
			}
			if(UNLIKELY(update_mode(CFG_FLAGS,flags,NULL)))
				res.map|=MAP_ZLIB;
		}
	}

	/* TCP_CORK mode change */

	if(ctl.map&MAP_CORK)
	{
		if(UNLIKELY(ctl.cork<0||ctl.cork>1))
		{
			NOTICE("illegal cork flag data %d supplied",ctl.cork);
			res.map|=MAP_CORK;
		}
		else
		{
			if(!ctl.cork)
			{
				flags&=~HINT_USECORK;
				extra&=~HINT_USECORK;
			}
			else
			{
				flags|=HINT_USECORK;
				extra|=HINT_USECORK;
			}
			if(UNLIKELY(update_mode(CFG_FLAGS,flags,NULL)))
				res.map|=MAP_CORK;
		}
	}

	/* broadcast/multicast emit speed change */

	if(ctl.map&MAP_BCASTSPEED)
	{
		if(UNLIKELY(ctl.slowbcast<0||ctl.slowbcast>1))
		{
			NOTICE("illegal broadcast speed flag data %d supplied",
				ctl.slowbcast);
			res.map|=MAP_BCASTSPEED;
		}
		else if(UNLIKELY(update_mode(BCAST_SPEED,ctl.slowbcast,NULL)))
			res.map|=MAP_BCASTSPEED;
	}

	/* light load flag change */

	if(ctl.map&MAP_SYSLOAD)
	{
		if(UNLIKELY(ctl.lightload<0||ctl.lightload>1))
		{
			NOTICE("illegal system load flag data %d supplied",
				ctl.lightload);
			res.map|=MAP_SYSLOAD;
		}
		else if(UNLIKELY(update_mode(SYSTEM_LOAD,ctl.lightload,NULL)))
			res.map|=MAP_SYSLOAD;
	}

	/* nice value change */

	if(ctl.map&MAP_NICE)
	{
		if(UNLIKELY(ctl.nice<0||ctl.nice>20))
		{
			NOTICE("illegal nice value data %d supplied",
				ctl.nice);
			res.map|=MAP_NICE;
		}
		else if(UNLIKELY(remotemgr(RMT_NICE,NULL,ctl.nice)))
			res.map|=MAP_NICE;
		else niceval=ctl.nice;
	}

	/* network device change or update */

	if(ctl.map&MAP_DEVICE)
	{
		if(UNLIKELY(!*ctl.dev))
		{
			NOTICE("empty network device string supplied");
			res.map|=MAP_DEVICE;
		}
		else if(UNLIKELY(updatemgr(UPDATE_NETWORK,ctl.dev,0)))
			res.map|=MAP_DEVICE;
		else if(UNLIKELY(remotemgr(RMT_NETWORK,ctl.dev,0)))
			res.map|=MAP_DEVICE;
	}

	/* return result to client */

	if(UNLIKELY(dowrite(s,&res,sizeof(res))))
		NOTICE("send control result failure: %s",errstr(errno,mem));
}

/*
 * cpphandler
 *
 * input:  data - preset shared processing data
 *
 * return: always NULL
 *
 * description:	This routine is the cpp queueing thread in case of
 *		parallel cpp limiting.
 */

static void *cpphandler(void *data)
{
	CPPQ *cpp=(CPPQ *)data;
	CLIENT *e;
#if defined(__linux__)
	uint64_t ev;
#elif defined(__FreeBSD__)
	struct kevent ev;
#endif
	char dummy;

	/* thread setup */

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	dftblock(THREAD);
	dropcaps(NEED_NONE);

#if defined(__linux__)
	ev=1;
#elif defined(__FreeBSD__)
	EV_SET(&ev,0x101,EVFILT_USER,0,NOTE_TRIGGER,0,EMPTY);
#endif

	/* do until process termination */

	while(LIKELY(running))
	{
		/* wait for next preprocessing job */

		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
		sem_wait(&cpp->sem);
		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);

		/* wait for free cpp slot and adjust running count */

		ipc_lock(CPPCOND);
repeat:		if(ipc_sum(CPPCOUNT)>=maxcpp)
		{
			ipc_sigwait(CPPCOND);
			goto repeat;
		}
		ipc_adj(CPPCOUNT,1);
		ipc_unlock(CPPCOND);

		/* set cpp job to running and notify netcc */

		pthread_mutex_lock(&cpp->mtx);
		e=cpp->waithead;
		cpp->waithead=e->next;
		e->cpp=2;
		e->next=NULL;
		dowrite(e->fd,&dummy,1);

		/* move job to run queue */

		if(!cpp->runhead)cpp->runhead=cpp->runtail=e;
		else
		{
			e->prev=cpp->runtail;
			cpp->runtail->next=e;
			cpp->runtail=e;
		}

		/* notify client handler of queue change */

#if defined(__linux__)
		dummy=write(cpp->notify,&ev,sizeof(ev));
#elif defined(__FreeBSD__)
		kevent(cpp->notify,&ev,1,NULL,0,NULL);
#else
		dummy=write(cpp->notify,&dummy,1);
#endif
		pthread_mutex_unlock(&cpp->mtx);
	}

	/* terminate thread */

	pthread_exit(NULL);
}

/*
 * worker
 *
 * input: data - unix domain socket listening handle
 *
 * return: always NULL
 *
 * description:	This routine is the client connection handling thread.
 */

static void *worker(void *data)
{
	int sock=(int)((long)data);
	int jobs;
	int old;
	int queued;
	int n;
	int i;
#if defined(__linux__) || defined(__FreeBSD__)
	int notify;
#else
	int notify[2];
#endif
	CLIENT *waithead;
	CLIENT *waittail;
	CLIENT *lclwaithead;
	CLIENT *lclwaittail;
	CLIENT *runq;
	CLIENT *c;
	CLIENT *e;
	CLIENT *s;
#if defined(__linux__)
	struct pollfd pp[4];
	struct epoll_event *ep;
#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	struct pollfd pp[4];
	struct kevent *ep;
#else
	struct pollfd *pp;
#endif
	struct pollfd q;
	CPPQ cpp;
	MSG msg;
	REQ req;
	char mem[256];

	/* thread setup */

	dftblock(THREAD);
	dropcaps(NEED_NONE);

	/* reset failure shutdown relevant data */

	waithead=NULL;
	waittail=NULL;
	lclwaithead=NULL;
	lclwaittail=NULL;
	runq=NULL;
	c=NULL;
	cpp.waithead=NULL;
	cpp.waittail=NULL;
	cpp.runhead=NULL;
	cpp.runtail=NULL;
	cpp.ipcinit=0;
#if defined(__linux__) || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	pp[3].fd=-1;
#endif
#if defined(__linux__) || defined(__FreeBSD__)
	notify=-1;
#else
	notify[0]=notify[1]=-1;
#endif

	/* allocate polling memory for running jobs */

#if defined(__linux__)
	if(UNLIKELY(!(ep=malloc(maxjobs*sizeof(struct epoll_event)))))
		GOTO(fail,"malloc failure");
#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	if(UNLIKELY(!(ep=malloc(maxjobs*sizeof(struct kevent)))))
		GOTO(fail,"malloc failure");
#else
	if(UNLIKELY(!(pp=malloc((maxjobs+3)*sizeof(struct pollfd)))))
		GOTO(fail,"malloc failure");
#endif

	/* set up cpp job change notification handle */

#if defined(__linux__)
#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((notify=eventfd(0,EFD_CLOEXEC))==-1))
			GOTO(fail,"eventfd: %s",errstr(errno,mem));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((notify=eventfd(0,0))==-1))
		{
			i=errno;
			unlock();
			GOTO(fail,"eventfd: %s",errstr(i,mem));
		}
		fcntl(notify,F_SETFD,FD_CLOEXEC);
		unlock();
	}
#elif defined(__FreeBSD__)
	if(UNLIKELY((notify=kqueue())==-1))
		GOTO(fail,"kqueue: %s",errstr(errno,mem));
	EV_SET(&ep[0],0x101,EVFILT_USER,EV_ADD|EV_ENABLE|EV_CLEAR,
		NOTE_FFNOP,0,EMPTY);
	if(UNLIKELY(kevent(notify,ep,1,NULL,0,NULL)==-1))
		GOTO(fail,"kqueue: %s",errstr(errno,mem));
#else
#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY(pipe2(notify,O_CLOEXEC|O_NONBLOCK)))
			GOTO(fail,"pipe2: %s",errstr(errno,mem));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY(pipe(notify)))
		{
			i=errno;
			unlock();
			GOTO(fail,"pipe: %s",errstr(i,mem));
		}
		fcntl(notify[0],F_SETFD,FD_CLOEXEC);
		fcntl(notify[1],F_SETFD,FD_CLOEXEC);
		unlock();
		i=1;
		if(UNLIKELY(ioctl(notify[0],FIONBIO,&i)))
		{
			i=errno;
			GOTO(fail,"ioctl(FIONBIO): %s",errstr(i,mem));
		}
		if(UNLIKELY(ioctl(notify[1],FIONBIO,&i)))
		{
			i=errno;
			GOTO(fail,"ioctl(FIONBIO): %s",errstr(i,mem));
		}
	}
#endif

	/* init cpp thread interworking stuff */

	if(UNLIKELY(pthread_mutex_init(&cpp.mtx,NULL)))
		GOTO(fail,"pthread_mutex_init: %s",errstr(errno,mem));

	if(UNLIKELY(sem_init(&cpp.sem,0,0)))
	{
		i=errno;
		pthread_mutex_destroy(&cpp.mtx);
		GOTO(fail,"sem_init: %s",errstr(i,mem));
	}
	cpp.ipcinit=1;

#if defined(__linux__) || defined(__FreeBSD__)
	cpp.notify=notify;
#else
	cpp.notify=notify[1];
#endif

	/* get running jobs polling handles */

#if defined(__linux__)
#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((pp[3].fd=epoll_create1(EPOLL_CLOEXEC))==-1))
			GOTO(fail,"epoll_create1: %s",errstr(errno,mem));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((pp[3].fd=epoll_create(maxjobs))==-1))
		{
			i=errno;
			unlock();
			GOTO(fail,"epoll_create: %s",errstr(i,mem));
		}
		fcntl(pp[3].fd,F_SETFD,FD_CLOEXEC);
		unlock();
	}
	pp[3].events=POLLIN;
#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	if(UNLIKELY((pp[3].fd=kqueue())==-1))
		GOTO(fail,"kqueue: %s",errstr(i,mem));
	pp[3].events=POLLIN;
#endif

	/* start cpp queue handling thread */

	if(UNLIKELY(pthread_create(&cpp.h,NULL,cpphandler,&cpp)))
		GOTO(fail,"pthread_create: %s",errstr(i,mem));

	/* preset local data */

	pp[0].fd=sock;
	pp[0].events=POLLIN;
	pp[1].fd=termget();
	pp[1].events=POLLIN;
#if defined(__linux__) || defined(__FreeBSD__)
	pp[2].fd=notify;
#else
	pp[2].fd=notify[0];
#endif
	pp[2].events=POLLIN;
	q.events=POLLIN;
	msg.port=port;
	msg.ttl=ttl;
	msg.nice=niceval;
	old=0;
	jobs=0;
	queued=0;

	/* do until termination is requested */

	while(LIKELY(running))
	{
		/* process waiting job queues */

		while(waithead||lclwaithead)
		{
			/* set up local/anywhere preference for next job */

			if(local)n=1;
			else if(lclwaithead&&!waithead)n=1;
			else n=0;

			/* get target for next job */

			switch((msg.addr=select_service(queued,n,&msg.addr6,
				&msg.extra,&msg.unique)))
			{
			/* all busy, job can't run */

			case -1:goto out;

			/* job runs locally */

			case 0:	/* process local job queue */

				for(e=NULL;lclwaithead;e=NULL)
				{
					/* update queue stats */

					update_mode(QUEUED_JOBS,--queued,NULL);

					/* check if client is still alive */

					q.fd=lclwaithead->fd;
					if(UNLIKELY(poll(&q,1,0)==1))
					    if(LIKELY(q.revents&POLLIN))
						goto lclskip;

					/* send job data to client */

					msg.extra=extra;
					if(LIKELY(!dowritefd(lclwaithead->fd,
						&msg,sizeof(msg),shmfd)))
					{
						/* remove job from wait queue */

						DEBUG("job started");
						e=lclwaithead;
						e->addr=msg.addr;
						lclwaithead=e->next;
						e->cpp=0;
						break;
					}

					/* remove dead client */

lclskip:				INFO("client disconnect before "
						"service");
					e=lclwaithead;
					lclwaithead=e->next;
					close(e->fd);
					free(e);
				}

				/* if no job selected process anywhere queue */

				if(!e)for(e=NULL;waithead;e=NULL)
				{
					/* update queue stats */

					update_mode(QUEUED_JOBS,--queued,NULL);

					/* check if client is still alive */

					q.fd=waithead->fd;
					if(UNLIKELY(poll(&q,1,0)==1))
					    if(LIKELY(q.revents&POLLIN))
						goto anyskip;

					/* send job data to client */

					msg.extra=extra;
					if(LIKELY(!dowritefd(waithead->fd,
						&msg,sizeof(msg),shmfd)))
					{
						/* remove job from wait queue */

						DEBUG("job started");
						e=waithead;
						e->addr=msg.addr;
						waithead=e->next;
						e->cpp=0;
						break;
					}

					/* remove dead client */

anyskip:				INFO("client disconnect before "
						"service");
					e=waithead;
					waithead=e->next;
					close(e->fd);
					free(e);
				}

				/* done if there is no job available */

				if(UNLIKELY(!e))break;

				/* add job to poll queue */

#if defined(__linux__)
				ep[0].events=EPOLLIN;
				ep[0].data.ptr=e;
				if(UNLIKELY(epoll_ctl(pp[3].fd,EPOLL_CTL_ADD,
					e->fd,ep)))
				{
					WARN("epoll_ctl: %s",errstr(errno,mem));
					close(e->fd);
					free(e);
					continue;
				}
#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
				EV_SET(&ep[0],e->fd,EVFILT_READ,
					EV_ADD|EV_ENABLE,0,0,PUT(e));
				if(UNLIKELY(kevent(pp[3].fd,ep,1,NULL,0,NULL)
					==-1))
				{
					WARN("kevent: %s",errstr(errno,mem));
					close(e->fd);
					free(e);
					continue;
				}
#endif

				/* add job to run queue */

				e->next=runq;
				if(runq)runq->prev=e;
				runq=e;

				/* update job stats */

				update_mode(JOB_NEW,e->addr,&e->stamp);
				jobs++;

				/* try to handle next job */

				continue;

			/* job will compile remotely */

			default:/* done if anywhere queue is empty */

				if(!waithead)goto out;

				/* memorite IPv6 interface index */

				if(SWITCHES_SUPPORT_MLDV2(v6))msg.addr=idx;

				/* process anywhere job queue */

				for(e=NULL;waithead;e=NULL)
				{
					/* update queue stats */

					update_mode(QUEUED_JOBS,--queued,NULL);

					/* check if client is still alive */

					q.fd=waithead->fd;
					if(UNLIKELY(poll(&q,1,0)==1))
					    if(LIKELY(q.revents&POLLIN))
						goto dftskip;

					/* send job data to client */

					msg.extra|=extra;
					if(msg.extra&HINT_SCTP)
						msg.extra&=~HINT_USECORK;
					if(LIKELY(waithead->cpp))
					{
						if(UNLIKELY(maxcpp))
							msg.extra|=HINT_CPP;
						else waithead->cpp=0;
					}
					if(LIKELY(!dowritefd(waithead->fd,&msg,
						sizeof(msg),shmfd)))
					{
						/* remove job from wait queue */

						DEBUG("job started");
						e=waithead;
						if(SWITCHES_SUPPORT_MLDV2(v6))
						{
							e->addr=1;
							e->a6=msg.addr6;
						}
						else e->addr=msg.addr;
						waithead=e->next;
						break;
					}

					/* remove dead client */

dftskip:				INFO("client disconnect before "
						"service");
					e=waithead;
					waithead=e->next;
					close(e->fd);
					free(e);
				}

				/* done if no job is available */

				if(UNLIKELY(!e))break;

				/* in case of parallel cpp limiting */

				if(UNLIKELY(e->cpp))
				{
					/* add job to cpp wait queue */

					e->next=NULL;
					pthread_mutex_lock(&cpp.mtx);
					if(!cpp.waithead)
						cpp.waithead=cpp.waittail=e;
					else
					{
						cpp.waittail->next=e;
						cpp.waittail=e;
					}
					sem_post(&cpp.sem);
					pthread_mutex_unlock(&cpp.mtx);
				}

				/* no parallel cpp limiting */

				else
				{
					/* add job to poll queue */

#if defined(__linux__)
					ep[0].events=EPOLLIN;
					ep[0].data.ptr=e;
					if(UNLIKELY(epoll_ctl(pp[3].fd,
						EPOLL_CTL_ADD,e->fd,ep)))
					{
						WARN("epoll_ctl: %s",
							errstr(errno,mem));
						close(e->fd);
						free(e);
						continue;
					}
#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
					EV_SET(&ep[0],e->fd,EVFILT_READ,
						EV_ADD|EV_ENABLE,0,0,PUT(e));
					if(UNLIKELY(kevent(pp[3].fd,ep,1,NULL,
						0,NULL)==-1))
					{
						WARN("kevent: %s",
							errstr(errno,mem));
						close(e->fd);
						free(e);
						continue;
					}
#endif

					/* add job to run queue */

					e->next=runq;
					if(runq)runq->prev=e;
					runq=e;
				}

				/* update job stats */

				update_mode(JOB_NEW,e->addr,&e->stamp,&e->a6);
				jobs++;

				/* try to handle next job */

				continue;
			}
		}

		/* assert that at least one job is running */

out:		while((lclwaithead||waithead)&&!runq)
		{
			/* force local processing */

			msg.addr=0;

			/* process local job queue */

			for(e=NULL;lclwaithead;e=NULL)
			{
				/* update queue stats */

				update_mode(QUEUED_JOBS,--queued,NULL);

				/* check if client is still alive */

				q.fd=lclwaithead->fd;
				if(UNLIKELY(poll(&q,1,0)==1))
					if(LIKELY(q.revents&POLLIN))
						goto lclskip2;

				/* send job data to client */

				msg.extra=extra;
				if(LIKELY(!dowritefd(lclwaithead->fd,&msg,
					sizeof(msg),shmfd)))
				{
					/* remove job from wait queue */

					DEBUG("job started");
					e=lclwaithead;
					e->addr=msg.addr;
					lclwaithead=e->next;
					e->cpp=0;
					break;
				}

				/* remove dead client */

lclskip2:			INFO("client disconnect before service");
				e=lclwaithead;
				lclwaithead=e->next;
				close(e->fd);
				free(e);
			}

			/* if no job selected process anywhere queue */

			if(!e)for(e=NULL;waithead;e=NULL)
			{
				/* update queue stats */

				update_mode(QUEUED_JOBS,--queued,NULL);

				/* check if client is still alive */

				q.fd=waithead->fd;
				if(UNLIKELY(poll(&q,1,0)==1))
					if(LIKELY(q.revents&POLLIN))
						goto anyskip2;

				/* send job data to client */

				msg.extra=extra;
				if(LIKELY(!dowritefd(waithead->fd,&msg,
					sizeof(msg),shmfd)))
				{
					/* remove job from wait queue */

					DEBUG("job started");
					e=waithead;
					e->addr=msg.addr;
					waithead=e->next;
					e->cpp=0;
					break;
				}

				/* remove dead client */

anyskip2:			INFO("client disconnect before service");
				e=waithead;
				waithead=e->next;
				close(e->fd);
				free(e);
			}

			/* done if no job is available */

			if(UNLIKELY(!e))break;

			/* add job to poll queue */

#if defined(__linux__)
			ep[0].events=EPOLLIN;
			ep[0].data.ptr=e;
			if(UNLIKELY(epoll_ctl(pp[3].fd,EPOLL_CTL_ADD,e->fd,ep)))
			{
				WARN("epoll_ctl: %s",errstr(errno,mem));
				close(e->fd);
				free(e);
				continue;
			}
#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
			EV_SET(&ep[0],e->fd,EVFILT_READ,EV_ADD|EV_ENABLE,0,0,
				PUT(e));
			if(UNLIKELY(kevent(pp[3].fd,ep,1,NULL,0,NULL)==-1))
			{
				WARN("kevent: %s",errstr(errno,mem));
				close(e->fd);
				free(e);
				continue;
			}
#endif

			/* add job to run queue */

			e->next=NULL;
			runq=e;

			/* update job stats */

			update_mode(JOB_NEW,e->addr,&e->stamp);
			jobs++;
		}

		/* toggle network operating mode if required */

		if(UNLIKELY(!old&&jobs))update_mode(UPDATE_ENABLE,0,NULL);
		else if(UNLIKELY(old&&!jobs))update_mode(UPDATE_DISABLE,0,NULL);
		old=jobs;

		/* limit total amount of jobs to specified limit */

		if(queued+jobs==maxjobs)pp[0].events=0;
		else pp[0].events=POLLIN;

		/* wait for event */

#if defined(__linux__) || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
restart:	while(poll(pp,4,-1)<=0);
#else
restart:	for(i=3,e=runq;e;e=e->next,i++)
		{
			pp[i].events=POLLIN;
			pp[i].fd=e->fd;
		}

		while(poll(pp,i,-1)<=0);
#endif

		/* exit in case of termination request */

		if(UNLIKELY(pp[1].revents&POLLIN))break;

		/* cpp queue change notify */

		if(pp[2].revents&POLLIN)
		{
			/* remove jobs crom cpp out queue and add them to the
			   run queue, reset cpp change notification and add
			   jobs to the event fd - adjust cpp stats, too */

			pthread_mutex_lock(&cpp.mtx);
#if defined(__linux__)
			*mem=read(notify,mem,sizeof(uint64_t));
			for(e=cpp.runhead;e;)
			{
				ep[0].events=EPOLLIN;
				ep[0].data.ptr=e;
				if(UNLIKELY(epoll_ctl(pp[3].fd,EPOLL_CTL_ADD,
					e->fd,ep)))
				{
					WARN("epoll_ctl: %s",errstr(errno,mem));
					s=e;
					e=e->next;
					if(s==cpp.runhead)cpp.runhead=s->next;
					else
					{
						s->prev->next=s->next;
						if(s->next)
							s->next->prev=s->prev;
					}
					close(s->fd);
					DEBUG("job aborted");
					ipc_lock(CPPCOND);
					ipc_adj(CPPCOUNT,-1);
					ipc_bcast(CPPCOND);
					ipc_unlock(CPPCOND);
					update_mode(JOB_DONE,s->addr,&s->stamp,
						&s->a6);
					free(s);
					jobs--;
				}
				else e=e->next;
			}
#elif defined(__FreeBSD__)
			kevent(notify,NULL,0,ep,1,NULL);
			for(e=cpp.runhead;e;)
			{
				EV_SET(&ep[0],e->fd,EVFILT_READ,
					EV_ADD|EV_ENABLE,0,0,e);
				if(UNLIKELY(kevent(pp[3].fd,ep,1,NULL,0,NULL)
					==-1))
				{
					WARN("kevent: %s",errstr(errno,mem));
					s=e;
					e=e->next;
					if(s==cpp.runhead)cpp.runhead=s->next;
					else
					{
						s->prev->next=s->next;
						if(s->next)
							s->next->prev=s->prev;
					}
					close(s->fd);
					DEBUG("job aborted");
					ipc_lock(CPPCOND);
					ipc_adj(CPPCOUNT,-1);
					ipc_bcast(CPPCOND);
					ipc_unlock(CPPCOND);
					update_mode(JOB_DONE,s->addr,&s->stamp,
						&s->a6);
					free(s);
					jobs--;
				}
				else e=e->next;
			}
#elif defined(__OpenBSD__) || defined(__NetBSD__)
			while(read(notify[0],mem,1)==1);
			for(e=cpp.runhead;e;)
			{
				EV_SET(&ep[0],e->fd,EVFILT_READ,
					EV_ADD|EV_ENABLE,0,0,PUT(e));
				if(UNLIKELY(kevent(pp[3].fd,ep,1,NULL,0,NULL)
					==-1))
				{
					WARN("kevent: %s",errstr(errno,mem));
					s=e;
					e=e->next;
					if(s==cpp.runhead)cpp.runhead=s->next;
					else
					{
						s->prev->next=s->next;
						if(s->next)
							s->next->prev=s->prev;
					}
					close(s->fd);
					DEBUG("job aborted");
					ipc_lock(CPPCOND);
					ipc_adj(CPPCOUNT,-1);
					ipc_bcast(CPPCOND);
					ipc_unlock(CPPCOND);
					update_mode(JOB_DONE,s->addr,&s->stamp,
						&s->a6);
					free(s);
					jobs--;
				}
				else e=e->next;
			}
#else
			while(read(notify[0],mem,1)==1);
#endif
			if(cpp.runhead)
			{
				cpp.runtail->next=runq;
				if(runq)runq->prev=cpp.runtail;
				runq=cpp.runhead;
				cpp.runhead=NULL;
			}
			pthread_mutex_unlock(&cpp.mtx);

			/* we need to restart event wait */

			goto restart;
		}

		/* handle job events: cpp done, compile done, dead client */

#if defined(__linux__)
		if(pp[3].revents&POLLIN)
			for(n=epoll_wait(pp[3].fd,ep,maxjobs,0),i=0;i<n;i++)
		{
			s=ep[i].data.ptr;
			if(UNLIKELY(s->cpp==2))
			{
				ipc_lock(CPPCOND);
				ipc_adj(CPPCOUNT,-1);
				ipc_bcast(CPPCOND);
				ipc_unlock(CPPCOND);
				if(LIKELY(doread(s->fd,mem,1)))
				{
					s->cpp=0;
					continue;
				}
			}
			if(UNLIKELY(epoll_ctl(pp[3].fd,EPOLL_CTL_DEL,s->fd,ep)))
				WARN("epoll_ctl: %s",errstr(errno,mem));
			if(s==runq)runq=s->next;
			else
			{
				s->prev->next=s->next;
				if(s->next)s->next->prev=s->prev;
			}
			close(s->fd);
			DEBUG("job complete");
			update_mode(JOB_DONE,s->addr,&s->stamp,&s->a6);
			free(s);
			jobs--;
			break;
		}
#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
		if(pp[3].revents&POLLIN)for(n=kevent(pp[3].fd,NULL,0,
			ep,maxjobs,NULL),i=0;i<n;i++)
		{
			s=GET(ep[i].udata);
			if(UNLIKELY(s->cpp==2))
			{
				ipc_lock(CPPCOND);
				ipc_adj(CPPCOUNT,-1);
				ipc_bcast(CPPCOND);
				ipc_unlock(CPPCOND);
				if(LIKELY(doread(s->fd,mem,1)))
				{
					s->cpp=0;
					continue;
				}
			}
			EV_SET(&ep[0],s->fd,EVFILT_READ,EV_DELETE,0,0,PUT(s));
			if(UNLIKELY(kevent(pp[3].fd,ep,1,NULL,0,NULL)==-1))
				WARN("kevent: %s",errstr(errno,mem));
			if(s==runq)runq=s->next;
			else
			{
				s->prev->next=s->next;
				if(s->next)s->next->prev=s->prev;
			}
			close(s->fd);
			DEBUG("job complete");
			update_mode(JOB_DONE,s->addr,&s->stamp,&s->a6);
			free(s);
			jobs--;
			break;
		}
#else
		for(s=runq,n=3;n<i;n++)if(pp[n].revents&POLLIN)
		{
			while(s)if(pp[n].fd==s->fd)
			{
				if(UNLIKELY(s->cpp==2))
				{
					ipc_lock(CPPCOND);
					ipc_adj(CPPCOUNT,-1);
					ipc_bcast(CPPCOND);
					ipc_unlock(CPPCOND);
					if(LIKELY(doread(s->fd,mem,1)))
					{
						s->cpp=0;
						s=s->next;
						break;
					}
				}
				e=s;
				s=s->next;
				if(e==runq)runq=e->next;
				else
				{
					e->prev->next=e->next;
					if(e->next)e->next->prev=e->prev;
				}
				close(e->fd);
				DEBUG("job complete");
				update_mode(JOB_DONE,e->addr,&e->stamp,&e->a6);
				free(e);
				jobs--;
				break;
			}
			else s=s->next;
		}
#endif

		/* restart main loop if no new client pending */

		if(!(pp[0].revents&POLLIN))continue;

		/* get new client memory if necessary */

		if(!c)
		{
			if(UNLIKELY(!(c=malloc(sizeof(CLIENT)))))
			{
				NOTICE("malloc failure");
				usleep(20000);
				continue;
			}
			c->next=c->prev=NULL;
		}

		/* accept next client connection */

#ifdef SAFEAPI
		if(LIKELY(safeapi()))
		{
			if(UNLIKELY((c->fd=accept4(pp[0].fd,NULL,0,
				SOCK_CLOEXEC))==-1))continue;
		}
		else
#endif
		{
			shlock();
			if(UNLIKELY((c->fd=accept(pp[0].fd,NULL,0))==-1))
			{
				unlock();
				continue;
			}
			fcntl(c->fd,F_SETFD,FD_CLOEXEC);
			unlock();
		}

		/* make connection blocking */

		i=0;
		if(UNLIKELY(ioctl(c->fd,FIONBIO,&i)))
		{
			NOTICE("ioctl(FIONBIO): %s",errstr(errno,mem));
			close(c->fd);
			continue;
		}

		/* get client request */

		if(UNLIKELY(donetwait(c->fd)))
		{
			NOTICE("read wait: %s",errstr(errno,mem));
			close(c->fd);
			continue;
		}

		if(UNLIKELY(doread(c->fd,&req,sizeof(req))))
		{
			NOTICE("read request: %s",errstr(errno,mem));
			close(c->fd);
			continue;
		}

		/* blindly preset cpp processing mode */

		c->cpp=(req.extra&HINT_CPP)?1:0;

		/* process request type */

		switch(req.where)
		{
		/* local job */

		case LOCAL:
			DEBUG("job queued");
			if(!lclwaithead)lclwaithead=lclwaittail=c;
			else
			{
				lclwaittail->next=c;
				lclwaittail=c;
			}
			c=NULL;
			update_mode(QUEUED_JOBS,++queued,NULL);
			break;

		/* anywhere job */

		case ANYWHERE:
			DEBUG("job queued");
			if(!waithead)waithead=waittail=c;
			else
			{
				waittail->next=c;
				waittail=c;
			}
			c=NULL;
			update_mode(QUEUED_JOBS,++queued,NULL);
			break;

		/* broken remote service notification */

		case NOSERVICE:
			update_mode(SYSTEM_BROKEN,req.extra,NULL,&req.addr6);
			close(c->fd);
			break;

		/* netccd-ctl config change request */

		case CONTROL:
			docontrol(c->fd);

		/* unknown request */

		default:close(c->fd);
			break;
		}
	}

	/* terminate cpp queueing thread */

	pthread_cancel(cpp.h);
	pthread_join(cpp.h,NULL);

	/* assert that all threads are notified that we'll terminate */

fail:	termnotify();

	/* close job event handle and job event memory */

#if defined(__linux__) || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	if(pp[3].fd!=-1)close(pp[3].fd);
	if(ep)free(ep);
#else
	if(pp)free(pp);
#endif

	/* release new client memory */

	if(LIKELY(c!=NULL))free(c);

	/* emtpy cpp job wait queue */

	while(cpp.waithead)
	{
		e=cpp.waithead;
		cpp.waithead=e->next;
		close(e->fd);
		free(e);
	}

	/* emtpy cpp job run queue and update stats */

	while(cpp.runhead)
	{
		e=cpp.runhead;
		cpp.runhead=e->next;
		close(e->fd);
		free(e);
		ipc_lock(CPPCOND);
		ipc_adj(CPPCOUNT,-1);
		ipc_bcast(CPPCOND);
		ipc_unlock(CPPCOND);
	}

	/* emtpy local job queue */

	while(lclwaithead)
	{
		e=lclwaithead;
		lclwaithead=e->next;
		close(e->fd);
		free(e);
	}

	/* emtpy anywhere job queue */

	while(waithead)
	{
		e=waithead;
		waithead=e->next;
		close(e->fd);
		free(e);
	}

	/* empty run queue and update stats */

	while(runq)
	{
		e=runq;
		runq=e->next;
		close(e->fd);
		if(UNLIKELY(e->cpp==2))
		{
			ipc_lock(CPPCOND);
			ipc_adj(CPPCOUNT,-1);
			ipc_bcast(CPPCOND);
			ipc_unlock(CPPCOND);
		}
		free(e);
	}

	/* release thread interworking stuff (beware of *BSD bugs) */

	if(cpp.ipcinit)
	{
		pthread_mutex_destroy(&cpp.mtx);
#if !defined(__NetBSD__) && !defined(__OpenBSD__)
		sem_destroy(&cpp.sem);
#else
#warning broken sem_destroy, fix this system, please
#endif
	}

	/* close cpp thread job change notify handle(s) */

#if defined(__linux__) || defined(__FreeBSD__)
	if(notify!=-1)close(notify);
#else
	if(notify[0]!=-1)close(notify[0]);
	if(notify[1]!=-1)close(notify[1]);
#endif

	/* exit thread */

	pthread_exit(NULL);
}

/*
 * getosarch
 *
 * output: arch - system architecture
 *	   alen - arch buffer size
 *	   os   - system os
 *	   olen - os buffr size
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine retrieves and truncates the os name and
 *		system arch from the kernel.
 */

static inline int getosarch(char *arch,int alen,char *os,int olen)
{
	struct utsname bfr;

	if(UNLIKELY(uname(&bfr)))BAD("uname failed");
	strncpy(arch,bfr.machine,alen);
	arch[alen-1]=0;
	strncpy(os,bfr.sysname,olen);
	os[olen-1]=0;
	return 0;
}

/*
 * gccversion
 *
 * input:  s - user supplied major.minor gcc version string
 *
 * return: the 16 bit binary compiler version or -1 in case of error
 *
 * description:	This routine converts a user supplied gcc version string to
 *		the required 16 bit binary format.
 */

static inline int gccversion(char *s)
{
	char *p;
	char *q;
	char *m;
	int ma;
	int mi;

	p=strtok_r(s,".",&m);
	q=strtok_r(NULL,".",&m);

	if(UNLIKELY(!p||!q))return -1;

	ma=atoi(p);
	mi=atoi(q);

	if(UNLIKELY(ma<1||ma>255))return -1;
	if(UNLIKELY(mi<0||mi>255))return -1;

	return (ma<<8)|mi;
}

/*
 * dosetauth
 *
 * input:  src - the authentication key source
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine
 */

static inline int dosetauth(char *src)
{
	int len;
	int fd;
	char bfr[8192];

	/* set fd to stdin if requested by caller */

	if(!strcmp(src,"-"))fd=0;

	/* otherwise open requested file */

	else
	{
#ifdef SAFEAPI
		if(LIKELY(safeapi()))
		{
#ifndef SAFEUID
			shlock();
#endif
			if(UNLIKELY((fd=open(src,O_RDONLY|O_CLOEXEC))==-1))
			{
#ifndef SAFEUID
				unlock();
#endif
				BAD("can't access %s",src);
			}
#ifndef SAFEUID
			unlock();
#endif
		}
		else
#endif
		{
			shlock();
			if(UNLIKELY((fd=open(src,O_RDONLY))==-1))
			{
				unlock();
				BAD("can't access %s",src);
			}
			fcntl(fd,F_SETFD,FD_CLOEXEC);
			unlock();
		}
	}

	/* read key to buffer */

	if((len=read(fd,bfr,sizeof(bfr)))<=0)
	{
		if(fd>0)close(fd);
		BAD("can't read from %s",src);
	}

	/* close file, if not stdin */

	if(fd>0)close(fd);

	/* set up udp authentification */

	if(UNLIKELY(updatemgr(UPDATE_AUTH,bfr,len)))
	{
		memset(bfr,0,sizeof(bfr));
		BAD("internal error: can't set udp auth mode");
	}

	/* set up tcp authentication */

	if(UNLIKELY(remotemgr(RMT_AUTH,bfr,len)))
	{
		memset(bfr,0,sizeof(bfr));
		BAD("internal error: can't set transport auth mode");
	}

	/* set up authentication for client tcp connections */

	setauth(&shmdta->key,shmdta->hash,bfr,len);
	auth=1;

	/* clear key buffer */

	memset(bfr,0,sizeof(bfr));

	/* signal success */

	return 0;
}

/*
 * mkmapfile
 *
 * input:  tmp  - temporary files working directory
 *	   size - file size
 *
 * return: file descriptor in case of success, -1 in case of error
 *
 * description:	This routine
 */

static inline int mkmapfile(char *tmp,int size)
{
	int l;
	int fd;
	char *fn;
	char zero[8192];

	/* create pathname template */

	if(UNLIKELY(!(fn=malloc(strlen(tmp)+19))))BAD("malloc failure");
	strcpy(fn,tmp);
	strcat(fn,"/netccd-wrk-XXXXXX");

	/* create deleted temporary file */

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
#ifndef SAFEUID
		shlock();
#endif
		if(UNLIKELY((fd=mkostemp(fn,O_CLOEXEC))==-1))
		{
#ifndef SAFEUID
			unlock();
#endif
			free(fn);
			BAD("mkostemp failure");
		}
		unlink(fn);
		free(fn);
#ifndef SAFEUID
		unlock();
#endif
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((fd=mkstemp(fn))==-1))
		{
			free(fn);
			unlock();
			BAD("mkstemp failure");
		}
		unlink(fn);
		free(fn);
		fcntl(fd,F_SETFD,FD_CLOEXEC);
		unlock();
	}

	/* preset file to zero up to requested size */

	memset(zero,0,sizeof(zero));
	for(;size;size-=l)
	{
		l=size>sizeof(zero)?sizeof(zero):size;
		if(UNLIKELY(write(fd,zero,l)!=l))
		{
			close(fd);
			BAD("zero shm file");
		}
	}

	/* reset file pointer */

	if(lseek(fd,0,SEEK_SET))
	{
		close(fd);
		BAD("lseek");
	}

	/* return descriptor */

	return fd;
}

/*
 * addrchk
 *
 * input:  addr - address string
 *	   v6   - 1 if IPv6 address, 0 if IPv4 address
 *
 * return: 0 if valid address, else -1
 *
 * description:	This routine checks the given IP address for validity.
 *		The address must be a multicast address.
 */

static inline int addrchk(char *addr,int v6)
{
	union
	{
		struct in6_addr a6;
		struct in_addr a4;
	} a;

	if(UNLIKELY(inet_pton(v6?AF_INET6:AF_INET,addr,&a)!=1))return -1;
	if(v6)
	{
		if(a.a6.s6_addr[0]!=0xff)return -1;
	}
	else if((ntohl(a.a4.s_addr)&0xe0000000)!=0xe0000000)return -1;
	return 0;
}

/*
 * usage
 *
 * description:	This routine prints a usage message and terminates processing.
 */

static NORETURN void usage(void)
{
	/* WARNING: You are *NOT* allowed to remove copyrignt and version
	  from the usage message - you are also *NOT* allowed to remove
	  the usage function from the program in *ANY* way */

	fprintf(stderr,"Usage: " NETCCD " OPTIONS\n"
		COPYRIGHT "\n"
		"Version: " VERSION "\n"
		"OPTIONS:\n"
		"-i interface         network interface (default " DEVICE ")\n"
		"-s socket-path       unix socket (default " SOCKET ")\n"
		"-p port              network port (default " STR(PORT) ")\n"
		"-u compile-user      set remote compile user (default " RMTUSER ")\n"
		"-g compile-group     set remote compile group (default " RMTGROUP ")\n"
		"-n nice-value        set compile nice value (default " STR(NICE) ", range 0-20)\n"
		"-F cpu-speed         set slow/fast cpu speed boundary in MHz (default " STR(SLOWCPU) ")\n"
		"-M memory-limit      set free memory limit in MB (default " STR(MINFREE) ")\n"
		"-B                   activate slow broadcast/multicast speed\n"
		"-C                   enable client only mode\n"
		"-K                   enable TCP_CORK for remote compiles\n"
		"-r                   heavily limit the amount of running preprocessors\n"
		"-R                   lightly limit the amount of running preprocessors\n"
		"-E                   use more interactive friendly load limits\n"
		"-U socket-user       set unix socket owner (default root)\n"
		"-G socket-group      set unix socket group (default " ROOTGRP ")\n"
		"-l                   enable local only mode\n"
		"-w web-port          enable mini web server on specified port\n"
		"-t temp-directory    temporary files directory (default " TMPDIR ")\n"
		"-x exec-path         path used when executing compiles for remote peers\n"
		"-X default-gcc       name of the default compiler (default " GCC ")\n"
		"-q dump-option       option to retrieve compiler version (default " VOPT ")\n"
		"-V watched-file      file to be watched for compiler changes\n"
		"-j max-jobs          maximum queued jobs (default " STR(MAXJOBS) ")\n"
		"-a auth-source       use authentication from auth-source (- for stdin)\n"
		"-I window-time       maximum authentication time delta (0-300, default:" STR(WINDOW) ")\n"
		"-2                   use authenticated SCTP transport (requires -a and -T)\n"
		"-5                   as '-2' including fix for COOKIE ACK kernel auth bug\n"
		"-y                   manually mark the system as not hyperthreaded\n"
		"-y                   manually mark the system as not hyperthreaded\n"
		"-Y                   manually mark the system as hyperthreaded\n"
		"-o cpu-number        manually set the amount of available cpu cores\n"
		"-S cpu-speed         manually set the cpu clock speed in MHz\n"
		"-e env-list          colon separated list of environment variables to be unset\n"
		"-c compiler-version  user supplied compiler version (major.minor)\n"
		"-H host-identifier   set the host identifier (default is the fqdn of the host)\n"
		"-A system-arch       user supplied arch string (15 chars max)\n"
		"-O system-os         user supplied os string (15 chars max)\n"
		"-N                   prevent automatic architecture mixing\n"
		"-k ipc-key           enable system wide netccd interworking using ipc-key\n"
		"-b named-pipe        BSD workaround fifo (default " FIFO ")\n"
		"-f                   do not daemonize, log to stderr too\n"
		"-P pid-file          pid file (default " PIDFILE ")\n"
		"-0                   application level ping during remote compile\n"
		"-Q mcast-addr        user supplied multicast address\n"
		"-6                   use IPv6 instead of IPv4 (multicast uses " MCASTADDR6 ")\n"
		"-T                   use SCTP instead of TCP for data transport\n"
		"-m                   use multicast instead of broadcast (" MCASTADDR4 ")\n"
		"-W                   site wide mode (IPv4: implies -m, IPv6: " MSITEADDR6 ")\n"
		"-L                   site wide mode ttl (1-255, default 1)\n"
		"-z                   use compressed network data transfer\n"
		"-Z                   make netcc map cc/c++ to gcc/g++\n"
		"-J                   disable lockless operation mode\n"
		"-3                   allow zerocopy file receive in lockless mode\n"
		"-1                   prefer remote compiles to local compiles\n"
		"-8 tool-list         allowed tools for remote compiles (default " ALLOWEDRMT ")\n"
		"-4 membership        group membership and access (default 11)\n"
		"-v                   verbose output\n"
		"-D dumpdir           enable core dumps in specified directory\n"
		"-d                   debug mode (implies -v and -f)\n"
		"-h                   this help text\n");
	exit(1);
}

/*
 * main
 *
 * input:  argc - argument count
 *	   argv - argument array
 *
 * return: 0 in case of success, a value > 0 in case of error
 *
 * description:	This routine handles the netccd daemon startup and shutdown.
 */

int main(int argc,char *argv[])
{
	int rval;
	int val;
	int s;
	int fastcpu;
	int client;
	int webport;
	int backlog;
	int ccvers;
	int fg;
	int verbose;
	int debug;
	int minfree;
	int slowbcast;
	int cork;
	int light;
	int cpunum;
	int cpufreq;
	int sctp;
	int mcast;
	int zlib;
	int site;
	int nomix;
	int zap;
	int window;
	int unique;
	int htflag;
	int locking;
	int ping;
	int preferlocal;
	int sctpauth;
	int zerocopy;
	int grpval;
	int tstflg;
	char *user;
	char *group;
	char *rmtuser;
	char *rmtgroup;
	char *version;
	char *userarch;
	char *useros;
	char *sp;
	char *dev;
	char *tmp;
	char *pidfile;
	char *authsrc;
	char *path;
	char *gcc;
	char *watchfile;
	char *clrenv;
	char *bsdpipe;
	char *vopt;
	char *fqdn;
	char *mcaddr;
	char *dumpdir;
	char *allowed;
	char *str;
	char *save;
	FILE *fp;
	pthread_t wh;
	struct rlimit r;
	char arch[16];
	char os[16];

	/* block signals according to signal handling mode */

#if defined(__linux__) || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	dftblock(ALLSIGS);
#else
	sethandler(handler);
	dftblock(PROCESS);
#endif

	/* config presets */

	rval=1;
	site=0;
	zlib=0;
	mcast=LAN_BCAST;
	sctp=0;
	cpunum=-1;
	cpufreq=-1;
	light=0;
	cork=0;
	slowbcast=0;
	minfree=-1;
	debug=0;
	verbose=0;
	fg=0;
	ccvers=-1;
	webport=-1;
	fastcpu=-1;
	client=0;
	rmtuser=RMTUSER;
	rmtgroup=RMTGROUP;
	user=NULL;
	group=NULL;
	version=NULL;
	userarch=NULL;
	useros=NULL;
	sp=SOCKET;
	dev=DEVICE;
	tmp=TMPDIR;
	pidfile=PIDFILE;
	authsrc=NULL;
	path=DFTPATH;
	gcc=GCC;
	watchfile=NULL;
	nomix=0;
	clrenv=NULL;
	zap=0;
	window=WINDOW;
	bsdpipe=FIFO;
	vopt=VOPT;
	fqdn=NULL;
	mcaddr=NULL;
	htflag=-1;
	dumpdir=NULL;
	tstflg=0;
	locking=0;
	ping=0;
	preferlocal=1;
	allowed=ALLOWEDRMT;
	sctpauth=0;
	zerocopy=0;
	grpval=11;

	/* get a unique value */

	unique=(int)timestamp();

	/* process command line arguments */

	while((val=getopt(argc,argv,"i:s:p:u:g:n:F:CU:G:lw:t:j:c:A:O:fP:q:H:Q:y"
		"YvdhM:BKrREo:6S:TmzWL:a:I:x:X:V:Ne:Zk:b:D:J018:234:5"))!=-1)
		switch(val)
	{
	case 'i':
		dev=optarg;
		break;

	case 's':
		sp=optarg;
		break;

	case 'p':
		port=atoi(optarg);
		break;

	case 'u':
		rmtuser=optarg;
		break;

	case 'g':
		rmtgroup=optarg;
		break;

	case 'n':
		niceval=atoi(optarg);
		break;

	case 'F':
		fastcpu=atoi(optarg);
		tstflg|=1;
		break;

	case 'C':
		client=1;
		break;

	case 'U':
		user=optarg;
		break;

	case 'G':
		group=optarg;
		break;

	case 'l':
		local=1;
		break;

	case 'w':
		webport=atoi(optarg);
		tstflg|=2;
		break;

	case 't':
		tmp=optarg;
		break;

	case 'j':
		maxjobs=atoi(optarg);
		break;

	case 'c':
		version=optarg;
		break;

	case 'A':
		userarch=optarg;
		break;

	case 'O':
		useros=optarg;
		break;

	case 'f':
		fg=1;
		break;

	case 'P':
		pidfile=optarg;
		break;

	case 'v':
		verbose=1;
		break;

	case 'd':
		verbose=1;
		debug=1;
		fg=1;
		break;

	case 'M':
		minfree=atoi(optarg);
		tstflg|=4;
		break;

	case 'B':
		slowbcast=1;
		break;

	case 'K':
		cork=1;
		break;

	case 'r':
		maxcpp=1;
		break;

	case 'R':
		maxcpp=2;
		break;

	case 'E':
		light=1;
		break;

	case 'o':
		cpunum=atoi(optarg);
		tstflg|=8;
		break;

	case '6':
		v6=1;
		break;

	case 'S':
		cpufreq=atoi(optarg);
		tstflg|=16;
		break;

	case 'T':
		sctp=1;
		break;

	case 'm':
		mcast=LAN_MCAST;
		break;

	case 'z':
		zlib=1;
		break;

	case 'W':
		site=1;
		break;

	case 'L':
		ttl=atoi(optarg);
		break;

	case 'a':
		authsrc=optarg;
		break;

	case 'I':
		window=atoi(optarg);
		break;

	case 'x':
		path=optarg;
		break;

	case 'X':
		gcc=optarg;
		break;

	case 'V':
		watchfile=optarg;
		break;

	case 'N':
		nomix=1;
		break;

	case 'e':
		clrenv=optarg;
		break;

	case 'Z':
		zap=1;
		break;

	case 'k':
		ipckey=(int)strtol(optarg,NULL,0);
		tstflg|=32;
		break;

	case 'b':
		bsdpipe=optarg;
		break;

	case 'q':
		vopt=optarg;
		break;

	case 'H':
		fqdn=optarg;
		break;

	case 'Q':
		mcaddr=optarg;
		break;

	case 'y':
		htflag=0;
		break;

	case 'Y':
		htflag=1;
		break;

	case 'D':
		dumpdir=optarg;
		break;

	case 'J':
		locking=1;
		break;

	case '0':
		ping=1;
		break;

	case '1':
		preferlocal=0;
		break;

	case '8':
		allowed=optarg;
		break;

	case '2':
		sctpauth|=1;
		break;

	case '3':
		zerocopy=1;
		break;

	case '4':
		grpval=atoi(optarg);
		break;

	case '5':
		sctpauth|=2;
		break;

	case 'h':
	default:usage();
	}

	if(optind<argc)usage();

	/* open syslog */

	openlog(NETCCD,LOG_NDELAY|LOG_PID|(fg?LOG_PERROR:0),LOG_DAEMON);

	/* set syslog logging mask */

	if(LIKELY(!debug))setlogmask(LOG_UPTO(LOG_INFO));
	if(LIKELY(!verbose))setlogmask(LOG_UPTO(LOG_NOTICE));

	/* sanity check for root */

	if(UNLIKELY(getuid()||geteuid()||getgid()||getegid()))
		FAIL("must be root");

	/* yes, we really want that - child processes get 077 set just
	   before exec and all other stuff explicitely sets the
	   proper permissions which must not be messed up by umask */

	umask(0);

	/* basic setup */

	if(UNLIKELY(initlock()))FAIL("can't inizialize locking");
	if(LIKELY(!locking))if(UNLIKELY(enablesafeapi()))
			FAIL("can't select system api");
	if(!authsrc||!sctp)sctpauth=0;

	/* core dump handling */

	if((UNLIKELY(dumpdir!=NULL)))if(UNLIKELY(chdir(dumpdir)))
		FAIL("can't access core dump directory");
	r.rlim_cur=r.rlim_max=UNLIKELY(dumpdir!=NULL)?RLIM_INFINITY:0;
	if(UNLIKELY(setrlimit(RLIMIT_CORE,&r)))FAIL("can't set core dump mode");

	/* parameter checks */

	if(tstflg&1)if(UNLIKELY(fastcpu<0))FAIL("illegal fast cpu value");
	if(tstflg&2)if(UNLIKELY(webport<1||webport>65535))
		FAIL("illegal web port number");
	if(tstflg&4)if(UNLIKELY(minfree<1))FAIL("illegal free memory value");
	if(tstflg&8)if(UNLIKELY(cpunum<1))FAIL("illegal cpu number value");
	if(tstflg&16)if(UNLIKELY(cpufreq<1))FAIL("illegal cpu frequency value");
	if(tstflg&32)if(UNLIKELY(!ipckey||ipckey==-1))FAIL("illegal ipc key");
	if(UNLIKELY(window<0||window>300))FAIL("illegal window time value");
	if(UNLIKELY(ttl<1||ttl>255))FAIL("illegal ttl value");
	if(UNLIKELY(maxjobs<=0))FAIL("illegal queued job value");
	if(UNLIKELY(niceval<0||niceval>20))FAIL("illegal nice value");
	if(UNLIKELY(port<1||port>65535))FAIL("illegal port number");
	if(UNLIKELY(port==webport))
		FAIL("working and web port must not be identical");
	if(UNLIKELY(!*sp))FAIL("illegal socket path");
	if(UNLIKELY(!fg&&!*pidfile))FAIL("illegal pid file");
	if(UNLIKELY(!*dev))FAIL("illegal interface name");
	if(UNLIKELY(!*rmtuser))FAIL("illegal remote compile user");
	if(UNLIKELY(!*rmtgroup))FAIL("illegal remote compile group");
	if(user)if(UNLIKELY(!*user))FAIL("illegal socket user");
	if(group)if(UNLIKELY(!*group))FAIL("illegal socket group");
	if(UNLIKELY(!*tmp))FAIL("illegal temp dir path");
	if(UNLIKELY(!*path))FAIL("illegal path setting");

	if(UNLIKELY(strchr(gcc,'/')!=NULL))if(UNLIKELY(gcc[0]!='/'))
		FAIL("illegal default compiler specification");
	if(UNLIKELY(watchfile!=NULL))if(UNLIKELY(watchfile[0]!='/'))
		FAIL("illegal compiler version file specification");
	if(version)if(UNLIKELY((ccvers=gccversion(version))==-1))
		FAIL("can't process user defined gcc version");
	if(fqdn)if(UNLIKELY(setfqdn(fqdn)))FAIL("can't set host identifier");
	if(mcaddr)if(UNLIKELY(addrchk(mcaddr,v6)))
		FAIL("illegal multicast address specified");
	switch(grpval)
	{
	case 11:flags|=HINT_G0ACCESS;
		break;

	case 12:flags|=HINT_G1ACCESS;
		break;

	case 13:flags|=HINT_G0ACCESS|HINT_G1ACCESS;;
		break;

	case 21:flags|=HINT_GROUP1|HINT_G0ACCESS;
		break;

	case 22:flags|=HINT_GROUP1|HINT_G1ACCESS;
		break;

	case 23:flags|=HINT_GROUP1|HINT_G0ACCESS|HINT_G1ACCESS;
		break;

	default:FAIL("illegal group membership value %d",grpval);
	}

	/* shared memory setup (shared by all netcc instances) */

	if(UNLIKELY((shmfd=mkmapfile(tmp,sizeof(SHM)))==-1))
		FAIL("can't create file based shared memory");
	if(UNLIKELY((shmdta=mmap(NULL,sizeof(SHM),PROT_READ|PROT_WRITE,
		MAP_SHARED,shmfd,0))==MAP_FAILED))
			FAIL("can't create file based shared memory");
	shmdta->window=window;

	if(clrenv)
	{
		if(strlen(clrenv)>CLRENV-1)
			FAIL("environment clear list too long");
		strcpy(shmdta->clrenv,clrenv);
		for(str=strtok_r(clrenv,":",&save);str;str=strtok_r(NULL,":",
			&save))if(UNLIKELY(unsetenv(str)))
				FAIL("can't unset environment variable %s",str);
	}
	else shmdta->clrenv[0]=0;

	if(UNLIKELY(strlen(tmp)>CLTMPD-8))
		FAIL("temporary directory %s too long",tmp);
	strcpy(shmdta->tmpdir,"TMPDIR=");
	strcat(shmdta->tmpdir,tmp);
	if(UNLIKELY(putenv(shmdta->tmpdir)))FAIL("can't set TMPDIR");

	if(UNLIKELY(strlen(path)>CLPATH-6))FAIL("%s is too long",path);
	strcpy(shmdta->path,"PATH=");
	strcat(shmdta->path,path);
	if(UNLIKELY(putenv(shmdta->path)))FAIL("can't set PATH");

	/* implied parameter value setup */

	if(site&&!v6)mcast=1;
	if(!site)ttl=1;

	/* os and arch setup */

	if(UNLIKELY(getosarch(arch,sizeof(arch),os,sizeof(os))))
		FAIL("can't retrieve OS and ARCH");

	if(userarch)
	{
		if(UNLIKELY(strlen(userarch)>=sizeof(arch)))
			FAIL("user supplied ARCH too long");
		strcpy(arch,userarch);
	}

	if(useros)
	{
		if(UNLIKELY(strlen(useros)>=sizeof(os)))
			FAIL("user supplied OS too long");
		strcpy(os,useros);
	}

	/* get IPv6 interface index */

	if(SWITCHES_SUPPORT_MLDV2(v6))if(UNLIKELY((idx=getifindex(dev))==-1))
		FAIL("can't retrieve interface data for %s",dev);

	/* basic module configuration */

	if(LIKELY(*allowed))if(UNLIKELY(remotemgr(RMT_TOOLS,allowed,0)))
		FAIL("internal error: can't set allowed tools");
	if(UNLIKELY(remotemgr(RMT_ZCOPY,NULL,zerocopy)))
		FAIL("internal error: can't set zerocopy receive mode");
	if(UNLIKELY(remotemgr(RMT_USEV6,NULL,v6)))
		FAIL("internal error: can't set tcp IPv6 mode");
	if(UNLIKELY(remotemgr(RMT_SITE,NULL,site)))
		FAIL("internal error: can't set tcp IPv6 mode");
	if(UNLIKELY(remotemgr(RMT_SCTP,NULL,sctp)))
		FAIL("internal error: can't set sctp mode");
	if(UNLIKELY(remotemgr(RMT_SCTPAUTH,NULL,sctpauth)))
		FAIL("internal error: can't set sctp authentication mode");
	if(UNLIKELY(remotemgr(RMT_WINDOW,NULL,window)))
		FAIL("internal error: can't set transport auth window");
	if(UNLIKELY(remotemgr(RMT_UNIQUE,NULL,unique)))
		FAIL("internal error: can't set transport unique id");
	if(UNLIKELY(updatemgr(UPDATE_ARCH,arch,0)))
		FAIL("internal error: can't set ARCH");
	if(UNLIKELY(updatemgr(UPDATE_OS,os,0)))
		FAIL("internal error: can't set OS");
	if(UNLIKELY(updatemgr(UPDATE_MCADDR,mcaddr,0)))
		FAIL("internal error: can't set OS");
	if(UNLIKELY(updatemgr(UPDATE_NOMIX,NULL,nomix)))
		FAIL("internal error: can't set arch mixing mode");
	if(UNLIKELY(updatemgr(UPDATE_USEV6,NULL,v6)))
		FAIL("internal error: can't set udp IPv6 mode");
	if(UNLIKELY(updatemgr(UPDATE_SITE,NULL,site)))
		FAIL("internal error: can't set udp site wide mode");
	if(UNLIKELY(updatemgr(UPDATE_TTL,NULL,ttl)))
		FAIL("internal error: can't set udp ttl");
	if(UNLIKELY(updatemgr(UPDATE_MCAST,NULL,mcast)))
		FAIL("internal error: can't set udp multicast mode");
	if(UNLIKELY(updatemgr(UPDATE_WINDOW,NULL,window)))
		FAIL("internal error: can't set udp auth window");
	if(UNLIKELY(updatemgr(UPDATE_UNIQUE,NULL,unique)))
		FAIL("internal error: can't set udp unique id");
	if(UNLIKELY(updatemgr(UPDATE_USERMT,NULL,preferlocal)))
		FAIL("internal error: can't set local/remote preference");
	if(minfree!=-1)if(UNLIKELY(updatemgr(UPDATE_MEMFREE,NULL,minfree)))
		FAIL("internal error: can't set free memory limit");

	/* initialize resource stats handling */

	if(UNLIKELY(getstats(INIT_STATS,&cpunum,&cpufreq,&htflag)))
		FAIL("cant process /proc/cpuinfo");

	/* set up resource limits */

	if(UNLIKELY(getstats(NUMCPU_STATS,&val)))
		FAIL("can't get number of cpus");
	val<<=3;
	val+=64;
	backlog=val;
	if(UNLIKELY(getrlimit(RLIMIT_NPROC,&r)))
			FAIL("can't get RLIMIT_NPROC");
	if(val>r.rlim_cur||val>r.rlim_max)
	{
		if(val>r.rlim_cur)r.rlim_cur=val;
		if(val>r.rlim_max)r.rlim_max=val;
		if(UNLIKELY(setrlimit(RLIMIT_NPROC,&r)))
			FAIL("can't set RLIMIT_NPROC");
	}
	val+=maxjobs<<1;
	if(UNLIKELY(getrlimit(RLIMIT_NOFILE,&r)))
		FAIL("can't get RLIMIT_NOFILE");
	if(val>r.rlim_cur||val>r.rlim_max)
	{
		if(val>r.rlim_cur)r.rlim_cur=val;
		if(val>r.rlim_max)r.rlim_max=val;
		if(UNLIKELY(setrlimit(RLIMIT_NOFILE,&r)))
			FAIL("can't set RLIMIT_NOFILE");
	}

	/* calculate parallel cpp limit if requested */

	if(UNLIKELY(maxcpp))
	{
		val=maxcpp;
		if(UNLIKELY(getstats(NUMCPU_STATS,&maxcpp)))
			FAIL("can't get number of cpus");
		if(val==2)maxcpp<<=1;
		maxcpp+=2;
	}

	/* initialize authentication */

	if(authsrc)if(dosetauth(authsrc))FAIL("can't set up authentication");

	/* set up termination notification */

	if(UNLIKELY(terminit()))
		FJMP(f1,"termination initialization failure");

	/* if daemon mode */

	if(!fg)
	{
		/* open pid file */

#ifdef SAFEAPI
		if(LIKELY(safeapi()))
		{
			if(UNLIKELY((val=open(pidfile,
				O_WRONLY|O_CREAT|O_TRUNC|O_EXCL,0644))==-1))
					FJMP(f2,"can't access pid file");
		}
		else
#endif
		{
			shlock();
			if(UNLIKELY((val=open(pidfile,
				O_WRONLY|O_CREAT|O_TRUNC,0644))==-1))
			{
				unlock();
				FJMP(f2,"can't access pid file");
			}
			fcntl(val,F_SETFD,FD_CLOEXEC);
			unlock();
		}
		if(UNLIKELY(!(fp=fdopen(val,"w"))))
		{
			close(val);
			FJMP(f3,"can't access pid file");
		}

		/* daemonize */

		if(UNLIKELY(daemon(UNLIKELY(dumpdir!=NULL)?1:0,0)))
		{
			fclose(fp);
			FJMP(f3,"can't daemonize");
		}

		/* write daemon pid to file */

		fprintf(fp,"%d\n",getpid());
		fclose(fp);
	}

	/* join system wide netccd ipc */

	if(UNLIKELY(ipc_join(ipckey,bsdpipe,0)))
		FJMP(f3,"shared resource setup failure");

	/* create client socket */

	if(UNLIKELY((s=unixsocket(sp,user,group,maxjobs))==-1))
		FJMP(f4,"can't create AF_UNIX socket");

	/* start udp update thread */

	if(UNLIKELY(updatemgr(UPDATE_START,dev,port)))
		FJMP(f5,"udp handling thread setup failed");

	/* set up fast/slow threshold */

	if(fastcpu>=0)update_mode(FAST_CPU_MHZ,fastcpu,NULL);

	/* set up udp flags and client flags */

	extra=isclient()?HINT_ADDSWITCH:0;
	if(SWITCHES_SUPPORT_MLDV2(v6))extra|=HINT_IPV6;
	if(sctpauth)flags|=HINT_SCTPAUTH;
	if(sctpauth&2)flags|=HINT_SCTPBUG;
	if(sctp)flags|=HINT_SCTP;
	else if(cork)
	{
		extra|=HINT_USECORK;
		flags|=HINT_USECORK;
	}
	if(zlib)
	{
		extra|=HINT_ZLIB;
		flags|=HINT_ZLIB;
	}
	if(auth)extra|=HINT_AUTH;
	if(zap)extra|=HINT_ZAP;
	if(ping)extra|=HINT_OOB;
#ifdef SAFEAPI
	if(LIKELY(safeapi()))if(zerocopy)extra|=HINT_SPLICE;
#endif

	/* start mini web server if requested */

	if(webport!=-1)if(UNLIKELY(webmgr(
		SWITCHES_SUPPORT_MLDV2(v6)?WEB6_START:WEB_START,webport)))
	{
		termnotify();
		FJMP(f6,"web handling thread setup failed");
	}

	/* start remote compile server thread */

	if(UNLIKELY(remotemgr(RMT_START,dev,port,backlog,tmp,niceval,
		rmtuser,rmtgroup)))
	{
		termnotify();
		FJMP(f7,"tcp handling thread setup failed");
	}

	/* set up manually supplied gcc version or start gcc
	   monitoring thread */

	if(ccvers!=-1)update_mode(GCC_VERSION,ccvers,NULL);
	else if(UNLIKELY(watchmgr(WATCH_START,gcc,watchfile,vopt)))
	{
		termnotify();
		FJMP(f8,"gcc watcher thread setup failed");
	}

	/* start signal fd watching thread */

#if defined(__linux__) || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	if(UNLIKELY(sethandler(handler)))
	{
		termnotify();
		FJMP(f9,"signal handling thread setup failed");
	}
#endif

	/* set up final udp runtime data */

	update_mode(SYSTEM_LOAD,light,NULL);
	update_mode(CFG_FLAGS,flags,NULL);
	update_mode(BCAST_SPEED,slowbcast,NULL);
	update_mode(LOCAL_ONLY,local,NULL);
	update_mode(CLIENT_ONLY,client,NULL);

	/* handle client requests in local worker thread */

	if(UNLIKELY(pthread_create(&wh,NULL,worker,(void *)((long)s))))
	{
		termnotify();
		FJMP(f9,"worker thread start failed");
	}
	else
	{
		rval=0;
		dropcaps(NEED_ROOT);
		pthread_join(wh,NULL);
	}

	/* shut down and clean up */

f9:	watchmgr(WATCH_STOP,NULL,NULL,NULL);
f8:	remotemgr(RMT_STOP,NULL,0);
f7:	webmgr(WEB_STOP,0);
f6:	updatemgr(UPDATE_STOP,NULL,0);
f5:	unixclose(s,sp);
f4:	ipc_leave(0);
f3:
#ifndef SAFEUID
	shlock();
#endif
	if(!fg)unlink(pidfile);
#ifndef SAFEUID
	unlock();
#endif
f2:	termexit();
f1:	clrauth(&shmdta->key,shmdta->hash);
	exitlock();

	/* signal success or error */

	return rval;
}
