
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "fqdn.h"
#include "ifc.h"
#include "stats.h"
#include "unix.h"
#include "tcp.h"
#include "tcp6.h"
#include "udp.h"
#include "udp6.h"
#include "ugid.h"

#define TESTSOCK "datatest.sock"

#define IPCKEY 0x45236718

typedef struct
{
        pthread_mutex_t ctx;
        pthread_cond_t cnd;
} IPC;

int verbose=1;
char *in=NULL;
char *app="datatest";

/*
 * getosarch
 *
 * See daemon.c for details.
 */

static int getosarch(char *arch,int alen,char *os,int olen)
{
	struct utsname bfr;

	if(UNLIKELY(uname(&bfr)))return -1;
	strncpy(arch,bfr.machine,alen);
	arch[alen-1]=0;
	strncpy(os,bfr.sysname,olen);
	os[olen-1]=0;
	return 0;
}

/*
 * hmactest
 *
 * Test if the MD5-HMAC code works on this system.
 */

static int hmactest(void)
{
	uint8_t *key=(uint8_t *)"Jefe";
	uint8_t *data=(uint8_t *)"what do ya want for nothing?";
	char digest[16]=
	{
		0x75,0x0c,0x78,0x3e,0x6a,0xb0,0xb5,0x03,
		0xea,0xa8,0x6e,0x31,0x0a,0x5d,0xb7,0x38
	};
	MD5(cmp);
	MD5HMDATA hmkey;

	md5hmkey(key,4,&hmkey);
	md5hmac(data,28,cmp,&hmkey);

	return memcmp(digest,cmp,MD5_SIZE);
}

/*
 * main
 *
 * Perform various tests that must succeed as a base for a successful
 * port to a new platform.
 */

int main(int argc,char *argv[])
{
	int s;
	int c;
	int clnuid;
	int clngid;
	int addr;
	UNUSED(int bcast);
	char *dev=DEVICE;
	char *str;
	int val;
	struct in_addr a;
	struct pollfd p;
	pid_t p1;
	pid_t p2;
	pid_t p3;
	char *user;
	char *group;
	struct passwd *pw;
	struct group *gr;
	struct stat stb;
	char os[16];
	char arch[16];
	UDP6 udp6;
	UDP udp;
	int idx;
	struct in6_addr a6;

	unsetenv(LOGNAME);

	if(argc>1)dev=argv[1];

	if(!hmactest())printf("md5 hmac ok\n\n");
	else
	{
		printf("md5 hmac failure, I will not test anything else.\n");
		return 0;
	}

	str=getfqdn();
	printf("fqdn: %s\n",str?str:"not found");
	if(getosarch(arch,sizeof(arch),os,sizeof(os)))
		printf("can't get os and arch\n");
	else printf("os  : %s\narch: %s\n",os,arch);

	printf("\n");

	printf("IPv4 data for device %s:\n\n",dev);
	addr=val=getifcaddr(dev);
	a.s_addr=htonl(val);
	if(val==-1)str="not found";
	else str=inet_ntoa(a);
	printf("Address:   %s\n",str);
	bcast=val=getbrdaddr(dev);
	a.s_addr=htonl(val);
	if(val==-1)str="not found";
	else str=inet_ntoa(a);
	printf("Broadcast: %s\n",str);
	val=getnetmask(dev);
	a.s_addr=htonl(val);
	if(val==-1)str="not found";
	else str=inet_ntoa(a);
	printf("Netmask:   %s\n\n",str);

	printf("IPv6 data for device %s:\n\n",dev);
	idx=getifindex(dev);
	if(idx==-1)printf("interface index: not found\n\n");
	else printf("interface index: %d\n\n",idx);

	if(getstats(INIT_STATS,NULL,NULL,NULL))
	{
		printf("statistics init failed, can't continue\n");
		return 0;
	}

	printf("free memory (including buffers+caches): ");
	if(getstats(MEMFREE_STATS,&val))printf("not available\n");
	else printf("%d MB\n",val);

#ifdef __linux__
	printf("number of physical cores: ");
#else
	printf("number of cores: ");
#endif
	if(getstats(NUMCPU_STATS,&val))printf("not available\n");
	else printf("%d\n",val);

	printf("cpu clock speed: ");
	if(getstats(BOGOMIPS_STATS,&val))printf("not available\n");
	else printf("%d MHz\n",val>>1);

	printf("running processes (at least 1): ");
	if(getstats(RUNNING_STATS,&val))printf("not available\n");
	else printf("%d\n",val);

	switch((p1=fork()))
	{
	case -1:printf("can't fork, no more tests available\n");
		return 0;
	case 0:	while(1);
	}

	printf("running processes (at least 2): ");
	if(getstats(RUNNING_STATS,&val))printf("not available\n");
	else printf("%d\n",val);

	switch((p2=fork()))
	{
	case -1:printf("can't fork, no more tests available\n");
		kill(p1,SIGTERM);
		return 0;
	case 0:	while(1);
	}

	printf("running processes (at least 3): ");
	if(getstats(RUNNING_STATS,&val))printf("not available\n");
	else printf("%d\n",val);

	switch((p3=fork()))
	{
	case -1:printf("can't fork, no more tests available\n");
		kill(p1,SIGTERM);
		kill(p2,SIGTERM);
		return 0;
	case 0:	while(1);
	}

	printf("running processes (at least 4): ");
	if(getstats(RUNNING_STATS,&val))printf("not available\n");
	else printf("%d\n",val);

	kill(p1,SIGTERM);
	kill(p2,SIGTERM);
	kill(p3,SIGTERM);

	printf("\n");

	pw=getpwuid(getuid());
	gr=getgrgid(getgid());

	if(!pw||!gr)
	{
		printf("user/group lookup failure\n");
		return 0;
	}

	if(!getuid())
	{
		user="nobody";
		group="nobody";
	}
	else
	{
		user=strdup(pw->pw_name);
		group=strdup(gr->gr_name);
		if(!user||!group)
		{
			printf("malloc failure\n");
			return 0;
		}
	}

	if(!(s=unixsocket(TESTSOCK,user,group,128)))
	{
		printf("unix socket creation failure\n");
		return 0;
	}

	pw=getpwnam(user);
	gr=getgrnam(group);

	if(!pw||!gr)
	{
		printf("user/group lookup failure\n");
		return 0;
	}

	if(stat(TESTSOCK,&stb))
	{
		printf("can't stat unix socket\n");
		unixclose(s,TESTSOCK);
		return 0;
	}

	if((stb.st_mode&0777)!=0660)
	{
		printf("bad unix socket permissions\n");
		unixclose(s,TESTSOCK);
		return 0;
	}

	if(stb.st_uid!=pw->pw_uid)
	{
		printf("wrong unix socket owner\n");
		unixclose(s,TESTSOCK);
		return 0;
	}

	if(stb.st_gid!=pw->pw_gid)
	{
		printf("wrong unix socket group\n");
		unixclose(s,TESTSOCK);
		return 0;
	}

	switch((p1=fork()))
	{
	case -1:printf("can't fork, no more tests available\n");
		unixclose(s,TESTSOCK);
		return 0;
	case 0:	if(!getuid())
		{
			c=setresgid(gr->gr_gid,gr->gr_gid,gr->gr_gid);
			c=setresuid(pw->pw_uid,pw->pw_uid,pw->pw_uid);
		}
		if((c=unixconnect(TESTSOCK))==-1)
		{
			printf("can't connect to unix socket\n");
			return 0;
		}
		sleep(2);
		close(c);
		return 0;
	}

	printf("waiting for connect to unix socket\n");

	p.fd=s;
	p.events=POLLIN;
	if(poll(&p,1,1000)!=1)
	{
		printf("failed client connect to unix socket\n");
		unixclose(s,TESTSOCK);
		return 0;
	}

	if((c=accept(s,NULL,0))==-1)
	{
		printf("can't accept connection on unix socket\n");
		unixclose(s,TESTSOCK);
		return 0;
	}

	printf("connected\n");

	if(getpeerid(c,&clnuid,&clngid))
	{
		close(c);
		printf("can't get socket peer credentials\n");
		unixclose(s,TESTSOCK);
		return 0;
	}

	if(!clnuid&&!clngid)
	{
		printf("WARNING: can't get peer credentials on unix socket\n");
	}
	else if(clnuid!=pw->pw_uid||clngid!=gr->gr_gid)
	{
		close(c);
		printf("got wrong peer credentials on unix sockets\n");
		unixclose(s,TESTSOCK);
		return 0;
	}

	close(c);
	unixclose(s,TESTSOCK);

	printf("unix socket testing complete\n");

	printf("\n");
	if(addr==-1)
	{
		printf("IPv4 address not found, no further testing\n");
		return 0;
	}

	if((s=tcpsocket(NULL,PORT,128))==-1)
	{
		printf("can't listen on port %d\n",PORT);
		return 0;
	}

	switch((p1=fork()))
	{
	case -1:printf("can't fork, no more tests available\n");
		close(s);
		return 0;
	case 0:	if((c=tcpconnect(addr,PORT,1))==-1)
		{
			printf("can't connect to to tcp port %d\n",PORT);
			return 0;
		}
		sleep(2);
		close(c);
		return 0;
	}

	printf("waiting for connect to tcp port %d\n",PORT);

	p.fd=s;
	p.events=POLLIN;
	if(poll(&p,1,1000)!=1)
	{
		printf("failed client connect to tcp port %d\n",PORT);
		close(s);
		return 0;
	}

	if((c=accept(s,NULL,0))==-1)
	{
		printf("can't accept connection on tcp port %d\n",PORT);
		close(s);
		return 0;
	}

	printf("connected\n");

	close(c);
	close(s);
	printf("tcp testing complete\n");

	if(idx==-1)
	{
		printf("IPv6 interface index not found, no further testing\n");
		return 0;
	}

	if((s=tcp6socket(dev,PORT,128))==-1)
	{
		printf("can't listen on tcp6 port %d\n",PORT);
		return 0;
	}

	switch((p1=fork()))
	{
	case -1:printf("can't fork, no more tests available\n");
		close(s);
		return 0;
	case 0:	inet_pton(AF_INET6,"::1",&a6);
		if((c=tcp6connect(&a6,PORT,idx,1))==-1)
		{
			printf("can't connect to to tcp6 port %d\n",PORT);
			return 0;
		}
		sleep(2);
		close(c);
		return 0;
	}

	printf("waiting for connect to tcp6 port %d\n",PORT);

	p.fd=s;
	p.events=POLLIN;
	if(poll(&p,1,1000)!=1)
	{
		printf("failed client connect to tcp6 port %d\n",PORT);
		close(s);
		return 0;
	}

	if((c=accept(s,NULL,0))==-1)
	{
		printf("can't accept connection on tcp6 port %d\n",PORT);
		close(s);
		return 0;
	}

	printf("connected\n");

	close(c);
	close(s);
	printf("tcp6 testing complete\n");

	printf("\n");

	if((s=udpsocket(dev,PORT,&udp,LAN_BCAST,1,NULL))==-1)
	{
		printf("can't create udp socket on port %d\n",PORT);
		return 0;
	}

	udpclose(&udp);
	printf("udp broadcast socket create on port %d ok\n",PORT);

	if((s=udpsocket(dev,PORT,&udp,LAN_MCAST,1,NULL))==-1)
	{
		printf("can't create udp socket on port %d\n",PORT);
		return 0;
	}

	udpclose(&udp);
	printf("udp multicast socket create on port %d ok\n",PORT);

	if((s=udp6socket(dev,PORT,&udp6,LAN_MCAST,1,NULL))==-1)
	{
		printf("can't create udp6 socket on port %d\n",PORT);
		return 0;
	}

	udp6close(&udp6);
	printf("udp6 socket create on port %d ok\n",PORT);

	printf("\n");

	printf("finished checking, please validate above output\n");

	return 0;
}
