
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "log.h"
#include "realpath.h"
#include "find.h"

/*
 * find_path_file
 *
 * input:  fn   - the pathname to be processed.
 *	   mode - processing mode bits as specified in find.h
 *
 * return: the resulting pathname or NULL in case of error
 *
 * description:	This routine tries to find the specified file or
 *		symlink and returns either the absolute pathname
 *		or path of the given file or symlink.
 */

char *find_path_file(char *fn,int mode)
{
	char *p;
	char *ptr;
	char *wrk;
	char *path;
	char *save;
	struct stat stb;

	p=NULL;
	wrk=NULL;
	path=NULL;

	/* sanity checks */

	if(UNLIKELY(!fn||!*fn))GOTO(finish,"no input file specified");

	if(mode&FIND_ABSOLUTE)if(*fn!='/')GOTO(finish,"%s is not absolute",fn);

	/* in case of a path being specified */

	if(strchr(fn,'/'))
	{
		/* regular file mode */

		if(!(mode&FIND_LINK))
		{
			/* continue with common part if fn exists and is a
			   regular file */

			if(stat(fn,&stb))GOTO(finish,"%s not found\n",fn);
			if(S_ISREG(stb.st_mode))goto common;
			GOTO(finish,"%s is not a regular file",fn);
		}

		/* continue with common part if fn exists and is a regular
		   file */

		if(lstat(fn,&stb))GOTO(finish,"%s not found\n",fn);
		if(S_ISREG(stb.st_mode))goto common;
		if(!S_ISLNK(stb.st_mode))GOTO(finish,"%s is neither a regular "
			"file nor a symlink",fn);

		/* do common symlink processing */

		if(UNLIKELY(!(wrk=strdup(fn))))GOTO(finish,"malloc failure");
		goto comlnk;
	}

	/* get PATH from environment and set up working memory */

	if(UNLIKELY(!(path=getenv("PATH"))))
		GOTO(finish,"PATH not set");

	if(UNLIKELY(!(path=strdup(path))))
		GOTO(finish,"malloc failure");

	if(UNLIKELY(!(wrk=malloc(strlen(path)+strlen(fn)+2))))
		GOTO(finish,"malloc failure");

	/* for all path elements contained in PATH */

	for(ptr=strtok_r(path,":",&save);ptr;ptr=strtok_r(NULL,":",&save))
	{
		/* skip empty and relative entries */

		if(UNLIKELY(!*ptr||*ptr!='/'))continue;

		/* create pathname */

		strcpy(wrk,ptr);
		strcat(wrk,"/");
		strcat(wrk,fn);

		/* in regular file mode */

		if(!(mode&FIND_LINK))
		{
			/* common processing if fn exists and is a regular
			   file */

			if(stat(wrk,&stb))continue;
			if(!S_ISREG(stb.st_mode))continue;
			goto common;
		}

		/* common processing if fn exists and is a regular file */

		if(lstat(wrk,&stb))continue;
		if(S_ISREG(stb.st_mode))
		{
common:			if(UNLIKELY(!(p=realpath(wrk?wrk:fn,NULL))))
				GOTO(finish,"malloc failure");
			goto finish;
		}
		if(!S_ISLNK(stb.st_mode))continue;

		/* common symlink processing */

comlnk:		if(UNLIKELY((ptr=strrchr(wrk,'/'))==wrk))goto finish;
		*ptr++=0;

		/* get absolute path to symlink */

		if(UNLIKELY(!(save=realpath(wrk,NULL))))
			GOTO(finish,"malloc failure");

		/* create resulting absolute path */

		if(UNLIKELY(!(p=malloc(strlen(save)+strlen(ptr)+2))))
		{
			free(save);
			GOTO(finish,"malloc failure");
		}
		strcpy(p,save);
		strcat(p,"/");
		strcat(p,ptr);
		free(save);
		goto finish;
	}

	/* log a warning if fn doesn't exist */

	WARN("%s not found",fn);

	/* release allocated memory */

finish:	if(path)free(path);
	if(wrk)free(wrk);

	/* strip fn from pathname if path is requested */

	if(p&&!(mode&FIND_FILE))if(LIKELY((ptr=strrchr(p,'/'))!=NULL))
	{
		if(UNLIKELY(ptr==p))p[1]=0;
		else *ptr=0;
	}

	/* return result */

	return p;
}
