
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "log.h"
#include "ifc.h"
#include "lock.h"
#include "safeapi.h"

/*
 * getifdata
 *
 * input:  dev  - the device name
 *	   what - the ioctl to be executed
 *
 * return: the ioctl result or -1 in case of error
 *
 * description:	This routine executes the given network ioctl and returns
 *		the resulting networking value.
 */

#if defined(__linux__)
static int getifdata(char *dev,int what)
{
	int i;
	int fd;
	struct ifreq ifr;
	char mem[256];

	memset(&ifr,0,sizeof(ifr));
	ifr.ifr_addr.sa_family=AF_INET;
	strncpy(ifr.ifr_name,dev,sizeof(ifr.ifr_name));

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((fd=socket(AF_INET,SOCK_DGRAM|SOCK_CLOEXEC,0))==-1))
			BAD("socket: %s",errstr(errno,mem));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((fd=socket(AF_INET,SOCK_DGRAM,0))==-1))
		{
			i=errno;
			unlock();
			BAD("socket: %s",errstr(i,mem));
		}
		fcntl(fd,F_SETFD,FD_CLOEXEC);
		unlock();
	}

	if(UNLIKELY(ioctl(fd,what,&ifr)==-1))
	{
		i=errno;
		close(fd);
		BAD("ioctl: %s",errstr(i,mem));
	}

	close(fd);
	return ntohl(((struct sockaddr_in *)(&ifr.ifr_addr))->sin_addr.s_addr);
}
#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
static int getifdata(char *dev,unsigned long what)
{
	int i;
	int fd;
	struct ifreq ifr;
	char mem[256];

	memset(&ifr,0,sizeof(ifr));
	ifr.ifr_addr.sa_family=AF_INET;
	strncpy(ifr.ifr_name,dev,sizeof(ifr.ifr_name));

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((fd=socket(AF_INET,SOCK_DGRAM|SOCK_CLOEXEC,0))==-1))
			BAD("socket: %s",errstr(errno,mem));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((fd=socket(AF_INET,SOCK_DGRAM,0))==-1))
		{
			i=errno;
			unlock();
			BAD("socket: %s",errstr(i,mem));
		}
		fcntl(fd,F_SETFD,FD_CLOEXEC);
		unlock();
	}

	if(UNLIKELY(ioctl(fd,what,&ifr)==-1))
	{
		i=errno;
		close(fd);
		BAD("ioctl: %s",errstr(i,mem));
	}

	close(fd);
	return ntohl(((struct sockaddr_in *)(&ifr.ifr_addr))->sin_addr.s_addr);
}
#else
#error missing getifdata()
#endif

/*
 * getifindex
 *
 * input:  dev - the device name
 *
 * return: the device interface index or -1 in case of error
 *
 * description:	This routine returns the interface index of the given device.
 */

int getifindex(char *dev)
{
#if defined(__linux__)
	int i;
	int s;
	struct ifreq ifr;
	char mem[256];

	memset(&ifr,0,sizeof(ifr));
	ifr.ifr_addr.sa_family=AF_INET6;
	strncpy(ifr.ifr_name,dev,sizeof(ifr.ifr_name));

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY(s=socket(AF_INET6,SOCK_DGRAM|SOCK_CLOEXEC,0))==-1)
			BAD("socket: %s",errstr(errno,mem));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY(s=socket(AF_INET6,SOCK_DGRAM,0))==-1)
		{
			i=errno;
			unlock();
			BAD("socket: %s",errstr(i,mem));
		}
		fcntl(s,F_SETFD,FD_CLOEXEC);
		unlock();
	}

	if(UNLIKELY(ioctl(s,SIOCGIFINDEX,&ifr))==-1)
	{
		i=errno;
		close(s);
		BAD("ioctl: %s",errstr(i,mem));
	}

	close(s);
	return ifr.ifr_ifindex;
#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	return if_nametoindex(dev);
#else
#error missing getifindex()
#endif
}

/*
 * getifcaddr
 *
 * input:  dev - the device name
 *
 * return: the interface IPv4 address or -1 in case of error
 *
 * description:	This routine returns the IPv4 address of the given interface.
 */

int getifcaddr(char *dev)
{
	return getifdata(dev,SIOCGIFADDR);
}

/*
 * getbrdaddr
 *
 * input:  dev - the device name
 *
 * return: the interface IPv4 broadcast address or -1 in case of error
 *
 * description:	This routine returns the IPv4 broadcast address of the
 *		given interface.
 */

int getbrdaddr(char *dev)
{
	return getifdata(dev,SIOCGIFBRDADDR);
}

/*
 * getnetmask
 *
 * input:  dev - the device name
 *
 * return: the interface IPv4 netmask or -1 in case of error
 *
 * description:	This routine returns the IPv4 netmask of the given interface.
 */

int getnetmask(char *dev)
{
	return getifdata(dev,SIOCGIFNETMASK);
}
