
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

/******************************************************************************

To understand what's happening here you do need a bit on an overview:

We need

1. A system wide unique counter of compile jobs running
2. A system wide counter of cpp jobs running
3. A system wide (each thread of each process) pthread_cond_wait,
   pthread_cond_signal and pthread_cond_broadcast processing
4. In stand alone operation the above process wide (within this process)

For 1. und 2. either local memory is used or SYSV semaphores are "abused".
For 3. either the pthread_* functions are used or a workaround (ab)using
SYSV semaphores and named pipes is implemented for systems with an
incomplete pthread implementation.

To work everywhere the sequence, described with the pthread* functions must
always be:

Waiter:
pthread_mutex_lock(&mutex)
pthread_cond_wait(&condition,&mutex)
pthread_mutex_unlock(&mutex)

Notifier:
pthread_mutex_lock(&mutex)
pthread_cond_signal/broadcast(&condition)
pthread_mutex_unlock(&mutex)

Wahts's always needed to work is a thread based pthread_cleanup_push
and pthread_cleanup_pop functionality as a thread may be cancelled in
pthread_cond_wait which typically means a held mutex which must be
released.

For broken systems, the above mutex is replaced with a SYSV semaphore.
Another SYSV semaphore is abused to keep a total count of waiters.
All waiters wait on data in a named pipe. The notifier wakes one or
all waiters by writing either one byte or "amount of waiters" bytes
to the named pipe. The waiters do assert that the amount of bytes
in the named pipe is always correct.

Two more SYSV semaphores are used for all systems. The first is a
global lock for joining or leaving systems. The second is abused as
a system wide count of systmes having joined.

SYSV shared memory is used to flag ipc initialization for all systems.

For systems with a working pthread implementation the required mutex and
condition variables are also located in the SYSV shared memory.

Please note: SYSV shared memory and semaphores seem to be the only
common denominator across POSIX platforms for IPC between unrelated
processes. This is why this stuff is used here (oh well, I remember
using this stuff in the early 80s).

Another problem resulting from SYSV stuff is that it is subject to
uid/gid permission checking at access time. Thus process wide
uid/gid modification must run with an exclusive lock. The complicated
part to this is that we have threads waiting on SYSV semaphores while
holding a shared lock. Trying to get now an exclusive lock prevents
other shared lock requestors fro acquiring the lock, so the SYSV
semaphore is never upped and a deadlock occurs. Thus there is a
special shared lock for everything that does a "down" on a possibly
blocking SYSV semaphore. Exclusive lock requestors must first exclusively
get this lock to assert that nobody is waiting on a semaphore with
a common shared lock held. Then the common shared lock must be
exclusively acquired to assert that no SYSV related operation takes
place while the process uid/gid is temporarily changed.

IPC join/leave is a bit tricky, too. For this we need a SYSV semaphore
that gives us exclusive access for this action. First of all, we try
to exclusively create the semaphores and if we can, we set up all
necessary semaphore values until finally we set the setup semaphore to 1
(the default value for all semaphores at creation time is 0).
If we can't create the semaphores we try to access them. If this fails
we retry with semaphore creation. In acll case we now try to lock the
setup semaphore. If this fails with EIDRM another process has deleted
the IPC stuff before we could get hold of this, so we start all over with
semaphore creation. When we hold the system semaphore we do have exclusive
access and up the users counter (ok, semaphore abuse). We then create
and initialize the shared memory if necessary. At leave time we
acquire the system semaphore, down the users count and when the count
is zero we delete the shared memory and the semaphores.

As the remaining problem (a sad experinece from the old days) is
a segfaulting process leaving the shared memory and possibly the semaphores
in unusable state I do refrain from using ftok() and instead let the
user specify the SYSV IPC key manually. This way it is known which
values to lookup with ipcs and thus which values to use with ipcrm
to remove and thus reset the IPC resources.

******************************************************************************/

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "ipc.h"
#include "log.h"
#include "lock.h"
#include "safeapi.h"

#if defined(__linux__)
#define SYSSEM	4
#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
#define SYSSEM	4 + (2* MAXCOND)
#else
#error do not know required management semaphore count
#endif

typedef struct
{
	int initialized;
#if defined(__linux__)
	pthread_cond_t cnd[MAXCOND];
	pthread_mutex_t ctx[MAXCOND];
#endif
} IPC;

static int ipc=-1;
static int shm;
static int counters[MAXCOUNTERS];
static IPC *shmdta;
static pthread_mutex_t mtx=PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t ctx[MAXCOND];
static pthread_cond_t cnd[MAXCOND];
#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
static int nph[MAXCOND];
char *fifo[MAXCOND];
#endif
#ifndef SAFEUID
static pthread_rwlock_t ipcrw;
#endif

/*
 * ctxrelease
 *
 * input:  arg - either the held mutex or the semaphore base index
 *
 * description:	This routine cleans up whena a thread is cancelled
 *		during pthread_cond_wait.
 */

static void ctxrelease(void *arg)
{
	/* simple mutex release for modern systems */

#if defined(__linux__)
	pthread_mutex_unlock(arg);

	/* emulation cleanup for broken systems */

#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	int base;
	int v1;
	int v2;
	struct sembuf s;
	char mem[256];

	/* set up semaphore base index */

	base=(int)((long)arg);

	/* lock the condition semaphore */

	s.sem_num=base;
	s.sem_op=-1;
	s.sem_flg=SEM_UNDO;
#ifndef SAFEUID
	pthread_rwlock_rdlock(&ipcrw);
#endif
	if(UNLIKELY(semop(ipc,&s,1)==-1))
	{
#ifndef SAFEUID
		pthread_rwlock_unlock(&ipcrw);
#endif
		WARN("semop: %s",errstr(errno,mem));
	}
#ifndef SAFEUID
	else pthread_rwlock_unlock(&ipcrw);
#endif

	/* decrease the waiters count */

	s.sem_num=base+1;
	s.sem_op=-1;
	s.sem_flg=SEM_UNDO;
#ifndef SAFEUID
	shlock();
#endif
	if(UNLIKELY(semop(ipc,&s,1)==-1))
	{
#ifndef SAFEUID
		unlock();
#endif
		WARN("semop: %s",errstr(errno,mem));
	}
#ifndef SAFEUID
	else unlock();
#endif

	/* get the current waiters count */

#ifndef SAFEUID
	shlock();
#endif
	if(UNLIKELY((v1=semctl(ipc,base+1,GETVAL))==-1))
	{
#ifndef SAFEUID
		unlock();
#endif
		goto out;
	}
#ifndef SAFEUID
	else unlock();
#endif

	/* get the amount of wakeups from the named pipe */

	for(v2=0;read(nph[(base-2)>>1],mem,1)==1;v2++);

	/* assert that no more wakeups than waiters are in the pipe */

	if(v2>v1)v2=v1;

	/* restore pipe contents */

	for(v1=0;v1<v2;v1++)*mem=write(nph[(base-2)>>1],mem,1);

	/* unlock the condition semaphore */

out:	s.sem_num=base;
	s.sem_op=1;
	s.sem_flg=SEM_UNDO;
#ifndef SAFEUID
	shlock();
#endif
	if(UNLIKELY(semop(ipc,&s,1)==-1))
	{
#ifndef SAFEUID
		unlock();
#endif
		WARN("semop: %s",errstr(errno,mem));
	}
#ifndef SAFEUID
	else unlock();
#endif
#else
#error do not know how to implement ctxrelease
#endif
}

/*
 * ipc_join
 *
 * input:  key   - the SYSV ipc key to be used
 *	   np    - the broken system emulation named pipe basename
 *	   super - 1 if attaching as event waiting supervisor (vhost management)
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine either creates the IPC stuff or
 *		makes us a member using the IPC stuff.
 */

int ipc_join(int key,char *np,int super)
{
	int v;
	int h;
	int m;
	struct shmid_ds t;
	struct sembuf s;
#if defined(__linux__)
	pthread_mutexattr_t mattr;
	pthread_condattr_t cattr;
#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	int p[MAXCOND];
	struct stat stb;
#endif
	char mem[256];

	/* done if already initialized */

	if(UNLIKELY(ipc!=-1))return 0;

	/* zero key means stay within this process */

	if(LIKELY(!key))
	{
		/* reset counters */

		memset(counters,0,sizeof(counters));

		/* initialize mutexes and conditions */

		for(v=0;v<MAXCOND;v++)
		{
			if(UNLIKELY(pthread_mutex_init(&ctx[v],NULL)))
			{
				ERR("pthread_mutex_init: %s",errstr(errno,mem));
				goto lclfail;
			}

			if(UNLIKELY(pthread_cond_init(&cnd[v],NULL)))
			{
				ERR("pthread_cond_init: %s",errstr(errno,mem));
				pthread_mutex_destroy(&ctx[v]);
lclfail:			for(v--;v>=0;v--)
				{
					pthread_mutex_destroy(&ctx[v]);
					pthread_cond_destroy(&cnd[v]);
				}
				return -1;
			}
		}

		/* mark ipc as locally initialized */

		ipc=0;

		/* signal success */

		return 0;
	}

	/* sanity checking */

	else if(UNLIKELY(key==-1))BAD("illegal key -1");

	/* in emulation mode create required named pipe pathnames */

#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	if(UNLIKELY(!np))BAD("no fifo name");
	for(v=0;v<MAXCOND;v++)
	{
		p[v]=-1;
		if(UNLIKELY(!(fifo[v]=malloc(strlen(np)+7))))
		{
			for(v--;v>=0;v--)free(fifo[v]);
			BAD("malloc failure");
		}
		sprintf(fifo[v],"%s.%d",np,v+1);
	}
#endif

	/* initialize special shared/exclusive lock for semaphore waiters */

#ifndef SAFEUID
	if(UNLIKELY(pthread_rwlock_init(&ipcrw,NULL)))
	{
		h=errno;
		for(v=0;v<MAXCOND;v++)free(fifo[v]);
		BAD("pthread_rwlock_init: %s",errstr(h,mem));
	}
#endif

	/* make shared memory as not accessed */

	m=-1;

	/* prepare first semaphore operation */

	s.sem_num=0;
	s.sem_op=-1;
	s.sem_flg=SEM_UNDO;

	/* repeat, until IPC is set up or an utter failure occurs */

	while(1)
	{
		/* try to create the semaphore array */

#ifndef SAFEUID
		pthread_rwlock_rdlock(&ipcrw);
#endif
		if((LIKELY(h=semget(key,2*MAXCOUNTERS+SYSSEM,
			IPC_CREAT|IPC_EXCL|0600))==-1))
		{
			/* real failure, abort */

			if(UNLIKELY(errno!=EEXIST))
			{
				v=errno;
#ifndef SAFEUID
				pthread_rwlock_unlock(&ipcrw);
#endif
#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
				for(v=0;v<MAXCOND;v++)free(fifo[v]);
#endif
#ifndef SAFEUID
				pthread_rwlock_destroy(&ipcrw);
#endif
				BAD("semget: %s",errstr(v,mem));
			}

			/* try to access semaphore array */

			if(UNLIKELY((h=semget(key,2*MAXCOUNTERS+SYSSEM,0))==-1))
			{
				v=errno;
#ifndef SAFEUID
				pthread_rwlock_unlock(&ipcrw);
#endif

				/* real failure, abort */

				if(UNLIKELY(v!=ENOENT))
				{
#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
					for(v=0;v<MAXCOND;v++)free(fifo[v]);
#endif
#ifndef SAFEUID
					pthread_rwlock_destroy(&ipcrw);
#endif
					BAD("semget: %s",errstr(v,mem));
				}

				/* ok, let's try creating this stuff again */

				continue;
			}

			/* semaphore accessed, skip global init */

			goto common;
		}

		/* preset all condition semaphore values */

#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
		for(v=0;v<MAXCOND;v++)
			if(UNLIKELY(semctl(h,2+(v<<1),SETVAL,1)==-1))
				WARN("semctl: %s",errstr(errno,mem));
#endif

		/* up setup semaphore as the array is initialized */

		if(UNLIKELY(semctl(h,0,SETVAL,1)==-1))
			WARN("semctl: %s",errstr(errno,mem));

		/* try to acquire the setup semaphore */

common:		if(UNLIKELY(semop(h,&s,1)==-1))
		{
			v=errno;
#ifndef SAFEUID
			pthread_rwlock_unlock(&ipcrw);
#endif
			/* real failure, abort */

			if(UNLIKELY(v!=EIDRM))
			{
#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
				for(v=0;v<MAXCOND;v++)free(fifo[v]);
#endif
#ifndef SAFEUID
				pthread_rwlock_destroy(&ipcrw);
#endif
				BAD("semop: %s",errstr(v,mem));
			}

			/* semaphore deleted, retry with create */

			continue;
		}

		/* we exclusively hold the semaphore, process supervisor
		   mode if requested */

		if(super)
		{
			/* check that there's no supervisor attached */

			if(UNLIKELY(semctl(h,3,GETVAL)))
				GOTO(fail2,"supervisor already running or "
					"semctl failed");

			/* reset the notification semaphore */

			if(UNLIKELY(semctl(h,2,SETVAL,0)))
				GOTO(fail2,"semctl: %s",errstr(errno,mem));

			/* set supervisor available indicator */

			s.sem_num=3;
			s.sem_op=1;
			if(UNLIKELY(semop(h,&s,1)==-1))
				GOTO(fail2,"semop: %s",errstr(errno,mem));

		}

		/* try to access the shared memory and create it if it
		   doesn't exist */

		if(UNLIKELY((m=shmget(key,sizeof(IPC),IPC_CREAT|0600))==-1))
			GOTO(fail2,"shmget: %s",errstr(errno,mem));

		/* attach the shared memory */

		if(UNLIKELY((shmdta=shmat(m,NULL,0))==(void *)-1))
		{
			/* notify */

			WARN("shmat: %s",errstr(errno,mem));

			/* common cleanup */

			if(0)
			{
fail:;
#ifndef SAFEUID
				pthread_rwlock_rdlock(&ipcrw);
#endif
				if(super)
				{
					s.sem_num=3;
					s.sem_op=-1;
					semop(h,&s,1);
					semctl(h,2,SETVAL,0);
				}
			}

			/* if we're the last IPC user */

fail2:			if(!semctl(h,1,GETVAL))
			{
				/* remove shared memory */

				if(m!=-1)
				{
					memset(&t,0,sizeof(t));
					shmctl(m,IPC_RMID,&t);
				}

				/* remove semaphore array */

				semctl(h,0,IPC_RMID);
			}

			/* other users are around */

			else
			{
				/* release setup semaphore */

				s.sem_op=1;
				semop(h,&s,1);
			}
#ifndef SAFEUID
			pthread_rwlock_unlock(&ipcrw);
#endif

			/* close and remove emulation named pipes */

#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
			for(v=0;v<MAXCOND;v++)if(p[v]!=-1)close(p[v]);
#ifndef SAFEUID
			shlock();
#endif
			for(v=0;v<MAXCOND;v++)
			{
				if(!stat(fifo[v],&stb))if(S_ISFIFO(stb.st_mode))
					unlink(fifo[v]);
				free(fifo[v]);
			}
#ifndef SAFEUID
			unlock();
#endif
#endif
			/* destroy special lock */

#ifndef SAFEUID
			pthread_rwlock_destroy(&ipcrw);
#endif

			/* signal big bad utter failure */

			return -1;
		}
#ifndef SAFEUID
		pthread_rwlock_unlock(&ipcrw);
#endif
		/* if the shared stuff is not yet initialized */

		if(UNLIKELY(!(shmdta->initialized)))
		{
			/* do pthread specific setup for modern systems */

#if defined(__linux__)
			v=PTHREAD_PROCESS_SHARED;
			if(UNLIKELY(pthread_mutexattr_init(&mattr)))
				GOTO(fail,"pthread_mutexattr_init: %s",
					errstr(errno,mem));
			if(UNLIKELY(pthread_mutexattr_setpshared(&mattr,v)))
				GOTO(fail,"pthread_mutexattr_setpshared: %s",
					errstr(errno,mem));
			if(UNLIKELY(pthread_condattr_init(&cattr)))
				GOTO(fail,"pthread_condattr_init: %s",
					errstr(errno,mem));
			if(UNLIKELY(pthread_condattr_setpshared(&cattr,v)))
				GOTO(fail,"pthread_condattr_setpshared: %s",
					errstr(errno,mem));
			for(v=0;v<MAXCOND;v++)
			{
				if(UNLIKELY(pthread_mutex_init(&shmdta->ctx[v],
					&mattr)))GOTO(fail,"pthread_mutex_init:"
						" %s",errstr(errno,mem));
				if(UNLIKELY(pthread_cond_init(&shmdta->cnd[v],
					&cattr)))GOTO(fail,"pthread_cond_init:"
						" %s",errstr(errno,mem));
			}
			if(UNLIKELY(pthread_mutexattr_destroy(&mattr)))
				WARN("pthread_mutexattr_destroy: %s",
					errstr(errno,mem));
			if(UNLIKELY(pthread_condattr_destroy(&cattr)))
				WARN("pthread_condattr_destroy: %s",
					errstr(errno,mem));

			/* do emulation pipe setup for broken systems */

#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
#ifndef SAFEUID
			shlock();
#endif
			for(v=0;v<MAXCOND;v++)if(!stat(fifo[v],&stb))
			{
				if(UNLIKELY(!S_ISFIFO(stb.st_mode)))
				{
#ifndef SAFEUID
					unlock();
#endif
					GOTO(fail,"%s: not a fifo",fifo[v]);
				}
			}
			else
			{
#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
				if(UNLIKELY(mkfifo(fifo[v],0600)))
#else
				if(UNLIKELY(mknod(fifo[v],S_IFIFO|0600,0)))
#endif
				{
					v=errno;
#ifndef SAFEUID
					unlock();
#endif
					GOTO(fail,"mknod: %s",errstr(v,mem));
				}
			}
#ifndef SAFEUID
			unlock();
#endif
#endif
			/* mark shared stuff as initialized */

			shmdta->initialized=1;
		}
		break;
	}

	/* access the emulation pipes for broken systems */

#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
#ifndef SAFEUID
		shlock();
#endif
		for(v=0;v<MAXCOND;v++)if(UNLIKELY((p[v]=open(fifo[v],
			O_RDWR|O_NONBLOCK|O_CLOEXEC))==-1))
		{
			v=errno;
#ifndef SAFEUID
			unlock();
#endif
			GOTO(fail,"open: %s",errstr(v,mem));
		}
#ifndef SAFEUID
		unlock();
#endif
	}
	else
#endif
	{
		shlock();
		for(v=0;v<MAXCOND;v++)
		{
			if(UNLIKELY((p[v]=open(fifo[v],O_RDWR|O_NONBLOCK))==-1))
			{
				v=errno;
				unlock();
				GOTO(fail,"open: %s",errstr(v,mem));
			}
			fcntl(p[v],F_SETFD,FD_CLOEXEC);
		}
		unlock();
	}
#endif

	/* increase the IPC users count */

	s.sem_num=1;
	s.sem_op=1;
#ifndef SAFEUID
	shlock();
#endif
	if(UNLIKELY(semop(h,&s,1)==-1))
	{
		v=errno;
#ifndef SAFEUID
		unlock();
#endif
		BAD("semop: %s",errstr(v,mem));
	}

	/* up the IPC setup semaphore */

	s.sem_num=0;
	if(UNLIKELY(semop(h,&s,1)==-1))
	{
		v=errno;
#ifndef SAFEUID
		unlock();
#endif
		BAD("semop: %s",errstr(v,mem));
	}
#ifndef SAFEUID
	unlock();
#endif

	/* make emulation pipe handles accessible */

#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	for(v=0;v<MAXCOND;v++)nph[v]=p[v];
#endif

	/* make SYSV handles accessible */

	shm=m;
	ipc=h;

	/* signal success */

	return 0;
}

/*
 * ipc_leave
 *
 * input:  super - 1 if detaching as event waiting supervisor
 *
 * return: always 0
 *
 * description:	This routine either removes us as a IPC member or
 *		if we were the last user removes the IPC stuff.
 */

int ipc_leave(int super)
{
	int v;
	int i;
	struct sembuf s;
	struct shmid_ds t;
	char mem[256];

	/* done if ipc deinit done */

	if(ipc==-1)return 0;

	/* in process wide mode */

	if(!ipc)
	{
		/* destroy mutexes and conditions */

		for(i=0;i<MAXCOND;i++)
		{
			if(UNLIKELY(pthread_cond_destroy(&cnd[i])))
				WARN("pthread_cond_destroy: %s",
					errstr(errno,mem));
			if(UNLIKELY(pthread_mutex_destroy(&ctx[i])))
				WARN("pthread_mutex_destroy: %s",
					errstr(errno,mem));
		}

		/* mark ipc as deinitalized */

		ipc=-1;

		/* signal success */

		return 0;
	}

	/* acquire IPC setup semaphore */

	s.sem_num=0;
	s.sem_op=-1;
	s.sem_flg=SEM_UNDO;
#ifndef SAFEUID
	pthread_rwlock_rdlock(&ipcrw);
#endif
	if(UNLIKELY(semop(ipc,&s,1)==-1))WARN("semop: %s",errstr(errno,mem));

	/* if supervisor is detaching */

	if(super)
	{
		/* mark supervisor as not available and clear events */

		s.sem_num=3;
		if(UNLIKELY(semop(ipc,&s,1)==-1))
			WARN("semop: %s",errstr(errno,mem));
		if(UNLIKELY(semctl(ipc,2,SETVAL,0)))
			WARN("semctl: %s",errstr(errno,mem));
	}

	/* remove us from the users counter */

	s.sem_num=1;
	if(UNLIKELY(semop(ipc,&s,1)==-1))WARN("semop: %s",errstr(errno,mem));

	/* get remaining users count */

	if(UNLIKELY((v=semctl(ipc,1,GETVAL))==-1))
		WARN("semctl: %s",errstr(errno,mem));

#ifndef SAFEUID
	pthread_rwlock_unlock(&ipcrw);
	shlock();
#endif

	/* no user pthread cleanup for modern systems */

#if defined(__linux__)
	if(!v)for(i=0;i<MAXCOND;i++)
	{
		if(UNLIKELY(pthread_cond_destroy(&shmdta->cnd[i])))
			WARN("pthread_cond_destroy: %s",errstr(errno,mem));
		if(UNLIKELY(pthread_mutex_destroy(&shmdta->ctx[i])))
			WARN("pthread_mutex_destroy: %s",errstr(errno,mem));
	}
#endif

	/* named pipe emulation cleanup for broken systems */

#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	for(i=0;i<MAXCOND;i++)
	{
		close(nph[i]);
		if(!v)unlink(fifo[i]);
		free(fifo[i]);
	}
#endif
	/* destroy special shared/exclusive lock for semaphore waiters */

#ifndef SAFEUID
	pthread_rwlock_destroy(&ipcrw);
#endif

	/* detach from shared memory */

	if(UNLIKELY(shmdt(shmdta)))WARN("shmdt: %s",errstr(errno,mem));

	/* if no more users */

	if(!v)
	{
		/* destroy shared memory */

		memset(&t,0,sizeof(t));
		if(UNLIKELY(shmctl(shm,IPC_RMID,&t)))
			WARN("shmctl: %s",errstr(errno,mem));

		/* destroy semaphore array */

		if(UNLIKELY(semctl(ipc,0,IPC_RMID)==-1))
			WARN("semctl: %s",errstr(errno,mem));
	}

	/* still users left */

	else
	{
		/* up IPC setup semaphore */

		s.sem_num=0;
		s.sem_op=1;
		if(UNLIKELY(semop(ipc,&s,1)==-1))
			WARN("semop: %s",errstr(errno,mem));
	}
#ifndef SAFEUID
	unlock();
#endif
	/* mark ipc as deinitialized */

	ipc=-1;

	/* signal success */

	return 0;
}

/*
 * ipc_waitevent
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine waits for a counter modification event.
 *		It has to be used only by the supervisor.
 */

int ipc_waitevent(void)
{
	struct sembuf s;
	char mem[256];

	/* processing mode */

	switch(ipc)
	{
	/* not usable in single process and not initialized mode */

	case 0:
	case -1:return -1;

	/* real ipc mode */

	default:s.sem_num=2;
		s.sem_op=-1;
		s.sem_flg=0;
		if(UNLIKELY(semop(ipc,&s,1)==-1))
			BAD("semop: %s",errstr(errno,mem));
		return 0;
	}
}

/*
 * ipc_postevent
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine posts a counter modification event.
 *		It has to be used only by the supervisor and for
 *		supervisor termination.
 */

int ipc_postevent(void)
{
	struct sembuf s;
	char mem[256];

	/* processing mode */

	switch(ipc)
	{
	/* not usable in single process and not initialized mode */

	case 0:
	case -1:return -1;

	/* real ipc mode */

	default:s.sem_num=2;
		s.sem_op=1;
		s.sem_flg=0;
		if(UNLIKELY(semop(ipc,&s,1)==-1))
			BAD("semop: %s",errstr(errno,mem));
		return 0;
	}
}

/*
 * ipc_adj
 *
 * input: idx - the counter index
 *	  val - the adjustment value
 *
 * description:	This routine adjusts a global counter by the given
 *		adjustment. It is the callers resonsibility that
 *		the resulting value doesn't drop below zero as
 *		in this case the counter shows its true origin
 *		which may be being a semaphore.
 *		Note that when the external management semaphore
 *		is set the external notify semaphore will be
 *		upped after the value change in real ipc mode.
 */

void ipc_adj(int idx,int val)
{
	int r;
	struct sembuf s;
	char mem[256];

	/* processing mode */

	switch(ipc)
	{
	/* single process mode */

	case 0:	pthread_mutex_lock(&mtx);
		counters[idx]+=val;
		pthread_mutex_unlock(&mtx);

	/* not initialized */

	case -1:return;

	/* real IPC mode: adjust the semaphore by the specified value */

	default:s.sem_num=(idx<<1)+SYSSEM;
		s.sem_op=val;
		s.sem_flg=SEM_UNDO;
#ifndef SAFEUID
		shlock();
#endif
		if(UNLIKELY(semop(ipc,&s,1)==-1))
			WARN("semop: %s",errstr(errno,mem));
		if(UNLIKELY((r=semctl(ipc,3,GETVAL))==-1))
			WARN("semctl: %s",errstr(errno,mem));
		if(r>0)
		{
			s.sem_num=2;
			s.sem_op=1;
			s.sem_flg=0;
			if(UNLIKELY(semop(ipc,&s,1)==-1))
				WARN("semop: %s",errstr(errno,mem));
		}
#ifndef SAFEUID
		unlock();
#endif
	}
}

/*
 * ipc_superadj
 *
 * input: idx - the counter index
 *	  val - the adjustment value
 *
 * description:	This routine adjusts a global counter by the given
 *		adjustment. It is the callers resonsibility that
 *		the resulting value doesn't drop below zero as
 *		in this case the counter shows its true origin
 *		which is a semaphore. Must only be called by supervisor.
 */

void ipc_superadj(int idx,int val)
{
	struct sembuf s;
	char mem[256];

	/* processing mode */

	switch(ipc)
	{
	/* single process mode or not initialized */

	case 0:
	case -1:return;

	/* real IPC mode: adjust the semaphore by the specified value */

	default:s.sem_num=(idx<<1)+1+SYSSEM;
		s.sem_op=val;
		s.sem_flg=SEM_UNDO;
#ifndef SAFEUID
		shlock();
#endif
		if(UNLIKELY(semop(ipc,&s,1)==-1))
			WARN("semop: %s",errstr(errno,mem));
#ifndef SAFEUID
		unlock();
#endif
	}
}

/*
 * ipc_get
 *
 * input:  idx - the counter index
 *
 * return: the current counter value which may be -1 in case of error
 *
 * description:	This routine returns the current value of the
 *		specified global counter.
 */

int ipc_get(int idx)
{
	int r;
	char mem[256];

	/* processing mode */

	switch(ipc)
	{
	/* single process mode */

	case 0:	pthread_mutex_lock(&mtx);
		r=counters[idx];
		pthread_mutex_unlock(&mtx);
		return r;

	/* not initialized */

	case -1:return 0;

	/* real IPC mode: return current semaphore value */

	default:
#ifdef SAFEUID
		if(UNLIKELY((r=semctl(ipc,(idx<<1)+SYSSEM,GETVAL))==-1))
			WARN("semctl: %s",errstr(errno,mem));
#else
		shlock();
		if(UNLIKELY((r=semctl(ipc,(idx<<1)+SYSSEM,GETVAL))==-1))
		{
			r=errno;
			unlock();
			WARN("semctl: %s",errstr(r,mem));
			return -1;
		}
		unlock();
#endif
		return r;
	}
}

/*
 * ipc_sum
 *
 * input:  idx - the counter index
 *
 * return: the current counters value which may be -1 in case of error
 *
 * description:	This routine returns the summed current values of the
 *		specified global counter.
 */

int ipc_sum(int idx)
{
	int r1;
	int r2;
	char mem[256];

	/* processing mode */

	switch(ipc)
	{
	/* single process mode */

	case 0:	pthread_mutex_lock(&mtx);
		r1=counters[idx];
		pthread_mutex_unlock(&mtx);
		return r1;

	/* not initialized */

	case -1:return 0;

	/* real IPC mode: return current semaphore value */

	default:
#ifdef SAFEUID
		if(UNLIKELY((r1=semctl(ipc,(idx<<1)+SYSSEM,GETVAL))==-1))
			WARN("semctl: %s",errstr(errno,mem));
		if(UNLIKELY((r2=semctl(ipc,(idx<<1)+1+SYSSEM,GETVAL))==-1))
			WARN("semctl: %s",errstr(errno,mem));
		if(UNLIKELY(r1==-1||r2==-1))return -1;
		else return r1+r2;
#else
		shlock();
		if(UNLIKELY((r1=semctl(ipc,(idx<<1)+SYSSEM,GETVAL))==-1))
		{
			r1=errno;
			unlock();
			WARN("semctl: %s",errstr(r1,mem));
			return -1;
		}
		if(UNLIKELY((r2=semctl(ipc,(idx<<1)+1+SYSSEM,GETVAL))==-1))
		{
			r1=errno;
			unlock();
			WARN("semctl: %s",errstr(r1,mem));
			return -1;
		}
		unlock();
		return r1+r2;
#endif
	}
}

/*
 * ipc_lock
 *
 * input: idx - the mutex index
 *
 * description:	This routine locks the desired global mutex.
 */

void ipc_lock(int idx)
{
#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	int i;
	struct sembuf s;
	char mem[256];
#endif

	/* processing mode */

	switch(ipc)
	{
	/* single process mode (fall through) */

	case 0:	pthread_mutex_lock(&ctx[idx]);

	/* not initialized */

	case -1:return;

	/* real IPC mode for modern systems */

#if defined(__linux__)
	default:pthread_mutex_lock(&shmdta->ctx[idx]);

	/* named pipe emulation IPC mode for broken systems */

#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	default:s.sem_num=(idx<<1)+2;
		s.sem_op=-1;
		s.sem_flg=SEM_UNDO;
#ifndef SAFEUID
		pthread_rwlock_rdlock(&ipcrw);
#endif
		if(UNLIKELY(semop(ipc,&s,1)==-1))
		{
			i=errno;
#ifndef SAFEUID
			pthread_rwlock_unlock(&ipcrw);
#endif
			WARN("semop: %s",errstr(i,mem));
		}
#ifndef SAFEUID
		else pthread_rwlock_unlock(&ipcrw);
#endif
#else
#error do not know how to implement ipc_lock
#endif
	}
}

/*
 * ipc_unlock
 *
 * input: idx - the mutex index
 *
 * description:	This routine unlocks the desired global mutex.
 */

void ipc_unlock(int idx)
{
#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	int i;
	struct sembuf s;
	char mem[256];
#endif

	/* processing mode */

	switch(ipc)
	{
	/* single process mode (fall through) */

	case 0:	pthread_mutex_unlock(&ctx[idx]);

	/* not initialized */

	case -1:return;

	/* real IPC mode for modern systems */

#if defined(__linux__)
	default:pthread_mutex_unlock(&shmdta->ctx[idx]);

	/* named pipe emulation IPC mode for broken systems */

#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	default:s.sem_num=(idx<<1)+2;
		s.sem_op=1;
		s.sem_flg=SEM_UNDO;
#ifndef SAFEUID
		shlock();
#endif
		if(UNLIKELY(semop(ipc,&s,1)==-1))
		{
			i=errno;
#ifndef SAFEUID
			unlock();
#endif
			WARN("semop: %s",errstr(i,mem));
		}
#ifndef SAFEUID
		else unlock();
#endif
#else
#error do not know how to implement ipc_unlock
#endif
	}
}

/*
 * ipc_bcast
 *
 * input: idx - the conditon index which has to get broadcasted
 *
 * description:	This routine broadcasts the specified condition.
 */

void ipc_bcast(int idx)
{
#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	int i;
	char dummy;
#endif

	/* processing mode */

	switch(ipc)
	{
	/* single process mode (fall through) */

	case 0:	pthread_cond_broadcast(&cnd[idx]);

	/* not initialized */

	case -1:return;

	/* real IPC mode for modern systems */

#if defined(__linux__)
	default:pthread_cond_broadcast(&shmdta->cnd[idx]);

	/* named pipe emulation IPC mode for broken systems */

#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	default:
#ifndef SAFEUID
		shlock();
#endif
		/* notify all registered waiters */

		if((i=semctl(ipc,(idx<<1)+3,GETVAL))>0)
			while(i--)dummy=write(nph[idx],&dummy,1);
#ifndef SAFEUID
		unlock();
#endif
#else
#error do not know how to implement ipc_signal
#endif
	}
}

/*
 * ipc_signal
 *
 * input: idx - the conditon index which has to get signalled
 *
 * description:	This routine signals the specified condition.
 */

void ipc_signal(int idx)
{
#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	char dummy;
#endif

	/* processing mode */

	switch(ipc)
	{
	/* single process mode (fall through) */

	case 0:	pthread_cond_signal(&cnd[idx]);

	/* not initialized */

	case -1:return;

	/* real IPC mode for modern systems */

#if defined(__linux__)
	default:pthread_cond_signal(&shmdta->cnd[idx]);

	/* named pipe emulation IPC mode for broken systems */

#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	default:
#ifndef SAFEUID
		shlock();
#endif
		/* notify a single registered waiter */

		if(semctl(ipc,(idx<<1)+3,GETVAL)>0)
			dummy=write(nph[idx],&dummy,1);
#ifndef SAFEUID
		unlock();
#endif
#else
#error do not know how to implement ipc_signal
#endif
	}
}

/*
 * ipc_sigwait
 *
 * input: idx - the conditon index which has to be waited on
 *
 * description:	This routine waits on the specified condition.
 */

void ipc_sigwait(int idx)
{
	int state;
#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	int base;
	int i;
	int r;
	struct sembuf s;
	struct pollfd p;
	char mem[256];
#endif

	/* processing mode */

	switch(ipc)
	{
	/* single process mode (fall through) */

	case 0:	pthread_cleanup_push(ctxrelease,&ctx[idx]);
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,&state);
		pthread_cond_wait(&cnd[idx],&ctx[idx]);
		pthread_setcancelstate(state,NULL);
		pthread_cleanup_pop(0);

	/* not initialized */

	case -1:return;

	/* real IPC mode for modern systems */

#if defined(__linux__)
	default:pthread_cleanup_push(ctxrelease,&shmdta->ctx[idx]);
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,&state);
		pthread_cond_wait(&shmdta->cnd[idx],&shmdta->ctx[idx]);
		pthread_setcancelstate(state,NULL);
		pthread_cleanup_pop(0);

	/* named pipe emulation IPC mode for broken systems */

#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	default:/* set up semaphore base index */

		base=(idx<<1)+2;

		/* increase waiters count */

		s.sem_num=base+1;
		s.sem_op=1;
		s.sem_flg=SEM_UNDO;
#ifndef SAFEUID
		shlock();
#endif
		if(UNLIKELY(semop(ipc,&s,1)==-1))
		{
			i=errno;
#ifndef SAFEUID
			unlock();
#endif
			WARN("semop: %s",errstr(i,mem));
		}
#ifndef SAFEUID
		else unlock();
#endif
		/* install cancellation handler */

		pthread_cleanup_push(ctxrelease,(void *)((long)base));

		/* here's the big bad loop: release semaphore */

retry:		s.sem_num=base;
		s.sem_op=1;
		s.sem_flg=SEM_UNDO;
#ifndef SAFEUID
		shlock();
#endif
		if(UNLIKELY(semop(ipc,&s,1)==-1))
		{
			i=errno;
#ifndef SAFEUID
			unlock();
#endif
			WARN("semop: %s",errstr(i,mem));
		}
#ifndef SAFEUID
		else unlock();
#endif

		/* enable cancellation */

		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,&state);

		/* wait for notification */

		p.fd=nph[idx];
		p.events=POLLIN;
		r=poll(&p,1,-1);

		/* disable cancellation */

		pthread_setcancelstate(state,NULL);

		/* acquire semaphore */

		s.sem_num=base;
		s.sem_op=-1;
		s.sem_flg=SEM_UNDO;
#ifndef SAFEUID
		pthread_rwlock_rdlock(&ipcrw);
#endif
		if(UNLIKELY(semop(ipc,&s,1)==-1))
		{
			i=errno;
#ifndef SAFEUID
			pthread_rwlock_unlock(&ipcrw);
#endif
			WARN("semop: %s",errstr(i,mem));
		}
#ifndef SAFEUID
		else pthread_rwlock_unlock(&ipcrw);
#endif

		/* retry if pipe is empty */

		if(r!=1||!(p.revents&POLLIN))goto retry;

		/* retry if some other thread did grab the notification */

		if(read(nph[idx],mem,1)!=1)goto retry;

		/* remove cancellation handler */

		pthread_cleanup_pop(0);

		/* decrease waiters count */

		s.sem_num=base+1;
		s.sem_op=-1;
		s.sem_flg=SEM_UNDO;
#ifndef SAFEUID
		shlock();
#endif
		if(UNLIKELY(semop(ipc,&s,1)==-1))
		{
			i=errno;
#ifndef SAFEUID
			unlock();
#endif
			WARN("semop: %s",errstr(i,mem));
		}
#ifndef SAFEUID
		else unlock();
#endif
		/* get current waiters count */
#ifndef SAFEUID
		shlock();
#endif
		if(UNLIKELY((i=semctl(ipc,base+1,GETVAL))==-1))
		{
			i=errno;
#ifndef SAFEUID
			unlock();
#endif
			WARN("semctl: %s",errstr(i,mem));
			return;
		}
#ifndef SAFEUID
		else unlock();
#endif
		/* count amount of sognals in named pipe */

		for(r=0;read(nph[idx],mem,1)==1;r++);

		/* assert that there are no more signals than waiters */

		if(r>i)r=i;

		/* restore named pipe contents */

		for(i=0;i<r;i++)*mem=write(nph[idx],mem,1);
#else
#error do not know how to implement ipc_sigwait
#endif
	}
}

#ifndef SAFEUID

/*
 * ipc_needex
 *
 * description:	This routine acquires exclusively the lock for
 *		operations that may block on a SYSV semaphore.
 */

void ipc_needex(void)
{
	pthread_rwlock_wrlock(&ipcrw);
}

/*
 * ipc_exdone
 *
 * description:	This routine releases the exclusive lock for
 *		operations that may block on a SYSV semaphore.
 */

void ipc_exdone(void)
{
	pthread_rwlock_unlock(&ipcrw);
}

#endif
