
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "log.h"
#include "lock.h"

/******************************************************************************

Actually this source consists only of a global variable of lock debugging
is not used. The macros in the header which result in pthread_rwlock
operations are used in this case.

In debug mode real routines are used. Though it would have been easier
to use __thread this construct is not available everywhere so better
stick to pthread_get/setspecific for debug mode which is anyway slow.

******************************************************************************/

#ifndef LOCKDEBUG

pthread_rwlock_t rw;

#else

typedef struct
{
	int rd;	/* shared lock count - 0 or 1		*/
	int wr;	/* exclusive lock count - 0 or 1	*/
} LCKDTA;

static pthread_rwlock_t rw;
static pthread_key_t key;

/*
 * getdata
 *
 * return: pointer to debug data or NULL
 *
 * description:	This routine returns a pointer to thread specific lock
 *		debug data or NULL in case of error.
 */

static LCKDTA *getdata(void)
{
	LCKDTA *data;

	/* if thread specific data do not exist */

	if(UNLIKELY(!(data=pthread_getspecific(key))))
	{
		/* create and preset data */

		if(UNLIKELY(!(data=malloc(sizeof(LCKDTA)))))
			WARN("malloc failure");
		else
		{
			memset(data,0,sizeof(LCKDTA));
			if(pthread_setspecific(key,data))
				WARN("pthread_setspecific");
		}
	}

	/* return thread specific data pointer */

	return data;
}

/*
 * initlock
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine initialized the lock as well as
 *		the debugging functionality.
 */

int initlock(void)
{
	/* create a key and use free as the release function */

	if(pthread_key_create(&key,free))
		BAD("pthread_key_create");

	/* initialize the rwlock */

	if(UNLIKELY(pthread_rwlock_init(&rw,NULL)))
		BAD("pthread_rwlock_init");

	/* signal success */

	return 0;
}

/*
 * exitlock
 *
 * return: always 0
 *
 * description:	This routine deinitializes the rwlock and the
 *		debugging functionality.
 */

int exitlock(void)
{
	/* destroy the rwlock */

	if(UNLIKELY(pthread_rwlock_destroy(&rw)))
		WARN("pthread_rwlock_destroy");

	/* delete the thread specific key */

	if(UNLIKELY(pthread_key_delete(key)))
		WARN("pthread_key_delete");

	/* signal success */

	return 0;
}

/*
 * _shlock
 *
 * input: file - caller source file
 *	  line - caller source line
 *
 * description:	This routine acquires a shared lock with lock debugging
 *		enabled. It must be called only through the proper
 *		macro defined in the header.
 */

void _shlock(char *file,int line)
{
	LCKDTA *data;

	/* get thread specific data */

	if(LIKELY((data=getdata())!=NULL))
	{
		/* do checks */

		if(UNLIKELY(data->rd))WARN("double shlock: %s %d",file,line);
		if(UNLIKELY(data->wr))WARN("shlock while exlock held: %s %d",
			file,line);

		/* adjust data */

		if(LIKELY(!data->rd&&!data->wr))data->rd++;
	}

	/* lockless mode debug */

#if LOCKDEBUG == 2
	WARN("shlock: %s %d",file,line);
#endif

	/* do actual operation */

	pthread_rwlock_rdlock(&rw);
}

/*
 * _exlock
 *
 * input: file - caller source file
 *	  line - caller source line
 *
 * description:	This routine acquires an exclusive lock with lock debugging
 *		enabled. It must be called only through the proper
 *		macro defined in the header.
 */

void _exlock(char *file,int line)
{
	LCKDTA *data;

	/* get thread specific data */

	if(LIKELY((data=getdata())!=NULL))
	{
		/* do checks */

		if(UNLIKELY(data->wr))WARN("double exlock: %s %d",file,line);
		if(UNLIKELY(data->rd))WARN("exlock while shlock held: %s %d",
			file,line);

		/* adjust data */

		if(LIKELY(!data->rd&&!data->wr))data->wr++;
	}

	/* lockless mode debug */

#if LOCKDEBUG == 2
	WARN("exlock: %s %d",file,line);
#endif

	/* do actual operation */

	pthread_rwlock_wrlock(&rw);
}

/*
 * _unlock
 *
 * input: file - caller source file
 *	  line - caller source line
 *
 * description:	This routine releases a lock with lock debugging
 *		enabled. It must be called only through the proper
 *		macro defined in the header.
 */

void _unlock(char *file,int line)
{
	LCKDTA *data;

	/* get thread specific data */

	if(LIKELY((data=getdata())!=NULL))
	{
		/* do checks */

		if(UNLIKELY(!data->rd&&!data->wr))WARN("double unlock: %s %d",
			file,line);

		/* adjust data */

		if(LIKELY(data->rd))data->rd--;
		else if(LIKELY(data->wr))data->wr--;
	}

	/* lockless mode debug */

#if LOCKDEBUG == 2
	WARN("unlock: %s %d",file,line);
#endif

	/* do actual operation */

	pthread_rwlock_unlock(&rw);
}

#endif
