
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "miniweb.h"
#include "tcp.h"
#include "tcp6.h"
#include "update.h"
#include "term.h"
#include "lock.h"
#include "log.h"
#include "safeapi.h"
#include "netio.h"
#include "sigs.h"
#include "caps.h"

#define NOSTATS "HTTP/1.0 503 Service Temporarily Unavailable\r\n" \
	"Content-Type: text/html; charset=iso-8859-1\r\n\r\n" \
	"<html><head><title>Service Temporarily Unavailable</title></head>" \
	"<body>Service temporarily unavailable. Please try later.</body></html>"

#define STATS "HTTP/1.0 200 OK\r\n" \
	"Content-Type: text/html; charset=iso-8859-1\r\n\r\n"

static char *nostats=NOSTATS;
static char *stats=STATS;
static int tcp;
static int running=0;
static pthread_t h;
static sem_t sem;

typedef struct
{
	int s;
	char *b;
} SERVE;

typedef struct
{
	int s;
	pthread_attr_t attr;
} SERVICE;

/*
 * serveclean
 *
 * input: data - the current serving data
 *
 * description:	This routine cleans up after a cancellation of a
 *		web serving thread. As currently these threads
 *		are not cancelled it is currently not used.
 */

static void serveclean(void *data)
{
	SERVE *srv=(SERVE *)data;

	/* close socket handle, release memory and adjust serving semaphore */

	close(srv->s);
	if(srv->b)free(srv->b);
	sem_post(&sem);
}

/*
 * serve
 *
 * input:  data - the socket handle
 *
 * return: always NULL
 *
 * description:	This is a web serving thread that sends the status page
 *		to the connected client.
 */

static void *serve(void *data)
{
	SERVE srv;

	/* thread default setup */

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);

	dftblock(THREAD);
	dropcaps(NEED_NONE);

	/* data setup */

	srv.s=(int)((long)data);
	srv.b=NULL;

	/* install cancellation handler */

	pthread_cleanup_push(serveclean,&srv);

	/* get status page */

	if(UNLIKELY(!(srv.b=htmlstats())))
	{
		/* send 503 if status page is unavailable */

		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
		dowrite(srv.s,nostats,strlen(nostats));
		shutdown(srv.s,SHUT_RDWR);
		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
		goto fail;
	}

	/* send status page to client */

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

	if(UNLIKELY(dowrite(srv.s,stats,strlen(stats))))goto out;
	dowrite(srv.s,srv.b,strlen(srv.b));
	shutdown(srv.s,SHUT_RDWR);

out:	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);

	/* release working memory */

	free(srv.b);

	/* close client socket handle */

fail:	close(srv.s);

	/* adjust serving count */

	sem_post(&sem);

	/* remove cancellation handler */

	pthread_cleanup_pop(0);

	/* terminate thread */

	pthread_exit(NULL);
}

/*
 * serverclean
 *
 * input: data - server thread data pointer
 *
 * description:	This routine is the cancellation clean up for the
 *		web server thread.
 */

static void serverclean(void *data)
{
	SERVICE *srv=(SERVICE *)data;

	/* close client socket handle, destroy thread attributes */

	if(srv->s!=-1)close(srv->s);
	pthread_attr_destroy(&srv->attr);
}

/*
 * server
 *
 * input:  unused - unused
 *
 * return: always NULL
 *
 * description:	This routine is the web server thread.
 */

static void *server(void *unused)
{
	int i;
	SERVICE srv;
	struct pollfd p[2];
	pthread_t h;
	char mem[256];

	/* thread default setup */

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);

	dftblock(THREAD);
	dropcaps(NEED_NONE);

	/* data setup */

	srv.s=-1;
	pthread_attr_init(&srv.attr);
	pthread_attr_setdetachstate(&srv.attr,PTHREAD_CREATE_DETACHED);

	p[0].fd=tcp;
	p[0].events=POLLIN;
	p[1].fd=termget();
	p[1].events=POLLIN;

	/* install cancellation handler */

	pthread_cleanup_push(serverclean,&srv);

	/* enable thread cancellation */

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

	/* run until process termination is pending */

	while(LIKELY(running))
	{
		/* wait for client or termination notify */

		switch(poll(p,2,-1))
		{
		/* event pending (fall through) */

		default:/* termination pending, loop */

			if(UNLIKELY(p[1].revents&POLLIN))continue;

			/* client connect pending */

			if(p[0].revents&POLLIN)break;

		/* nothing to do, repeat */

		case 0:
		case -1:continue;
		}

		/* disable cancellation */

		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);

		/* get client connection */

#ifdef SAFEAPI
		if(LIKELY(safeapi()))
		{
			if(UNLIKELY((srv.s=accept4(tcp,NULL,NULL,SOCK_CLOEXEC))
				==-1))
			{
				pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,
					NULL);
				continue;
			}
		}
		else
#endif
		{
			shlock();
			if(UNLIKELY((srv.s=accept(tcp,NULL,NULL))==-1))
			{
				unlock();
				pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,
					NULL);
				continue;
			}
			fcntl(srv.s,F_SETFD,FD_CLOEXEC);
			unlock();
		}

		/* enable cancellation */

		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

		/* make client socket blocking */

		i=0;
		if(UNLIKELY(ioctl(srv.s,FIONBIO,&i)))
			JUMP(fail,"ioctl(FIONBIO): %s",errstr(errno,mem));

		/* set up socket keepalive */

#ifdef TCP_KEEPCNT
		i=3;
		if(UNLIKELY(setsockopt(srv.s,IPPROTO_TCP,TCP_KEEPCNT,&i,
			sizeof(i))))
				JUMP(fail,"setsockopt(TCP_KEEPCNT): %s",
					errstr(errno,mem));
#endif

#ifdef TCP_KEEPIDLE
		i=30;
		if(UNLIKELY(setsockopt(srv.s,IPPROTO_TCP,TCP_KEEPIDLE,&i,
			sizeof(i))))
				JUMP(fail,"setsockopt(TCP_KEEPIDLE): %s",
					errstr(errno,mem));
#endif

#ifdef TCP_KEEPINTVL
		i=15;
		if(UNLIKELY(setsockopt(srv.s,IPPROTO_TCP,TCP_KEEPINTVL,&i,
			sizeof(i))))
				JUMP(fail,"setsockopt(TCP_KEEPINTVL): %s",
					errstr(errno,mem));
#endif

#if !defined(TCP_KEEPCNT) || !defined(TCP_KEEPIDLE) || !defined(TCP_KEEPINTVL)
#warning only crippled keepalive support available
#endif

		i=1;
		if(UNLIKELY(setsockopt(srv.s,SOL_SOCKET,SO_KEEPALIVE,&i,
			sizeof(i))))
				JUMP(fail,"setsockopt(SO_KEEPALIVE): %s",
					errstr(errno,mem));

		/* wait for a free serving slot */

		sem_wait(&sem);

		/* disable cancellation */

		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);

		/* start serving thread */

		if(UNLIKELY(pthread_create(&h,&srv.attr,serve,
			(void *)((long)srv.s))))
		{
			NOTICE("pthread_create: %s",errstr(errno,mem));
			sem_post(&sem);
fail:			close(srv.s);
		}

		/* reset connection handle */

		srv.s=-1;

		/* enable cancellation */

		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
	}

	/* clean up */

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	pthread_cleanup_pop(0);
	pthread_attr_destroy(&srv.attr);

	/* terminate */

	pthread_exit(NULL);
}

/*
 * webmgr
 *
 * input:  mode - processing mode
 *	   port - listening port
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine is the mini web server module management
 *		routine.
 */

int webmgr(int mode,int port)
{
	int i;
	char mem[256];

	switch(mode)
	{
	/* start web server */

	case WEB_START:
	case WEB6_START:

		/* done if already running */

		if(UNLIKELY(running))return 0;

		/* initialize serving count semaphore */

		if(UNLIKELY(sem_init(&sem,0,4)))
			BAD("sem_init: %s",errstr(errno,mem));

		/* get listening socket */

		switch(mode)
		{
		case WEB6_START:
			tcp=tcp6socket(NULL,port,4);
			break;
		case WEB_START:
			tcp=tcpsocket(NULL,port,4);
			break;
		}
		if(UNLIKELY(tcp==-1))
		{
			sem_destroy(&sem);
			BAD("create tcp socket failure");
		}

		/* mark server as running */

		running=1;

		/* start web server thread */

		if(UNLIKELY(pthread_create(&h,NULL,server,NULL)))
		{
			i=errno;
			running=0;
			close(tcp);
			sem_destroy(&sem);
			NOTICE("pthread_create: %s",errstr(i,mem));
		}

		/* signal success */

		return 0;

	/* web server termination */

	case WEB_STOP:

		/* done if already stopped */

		if(UNLIKELY(!running))return 0;

		/* reset running state */

		running=0;

		/* cancel and join web server thread */

		pthread_cancel(h);
		pthread_join(h,NULL);

		/* close listening socket */

		close(tcp);

		/* destroy the semaphore (beware of *BSD bug) */

#if !defined(__NetBSD__) && !defined(__OpenBSD__)
		sem_destroy(&sem);
#else
#warning broken sem_destroy, fix this system, please
#endif
		/* signal success */

		return 0;

	/* sanity check */

	default:BAD("internal error");
	}
}
