
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#ifndef NETIO_H
#define NETIO_H

/*
 * netwait
 *
 * input:  fd - the descriptor to be processed
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine waits for a limited amount of time for data
 *		to become available on the given fd.
 */

static inline int donetwait(int fd)
{
	struct pollfd p;

	p.fd=fd;
	p.events=POLLIN;
	switch(poll(&p,1,NETWAIT*1000))
	{
	default:if(p.revents&POLLIN)return 0;
	case 0:	errno=EIO;
	case -1:return -1;
	}
}

/*
 * dowritev
 *
 * input:  fd     - the target descriptor
 *	   iov    - the io vector array
 *	   iovcnt - the amount of io vectors
 *	   len    . the total data length
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine sends a message to the given fd.
 */

static inline int dowritev(int fd,struct iovec *iov,int iovcnt,int len)
{
	int l;

	if(UNLIKELY(!len))goto out;
	while(1)
	{
		if(UNLIKELY((l=writev(fd,iov,iovcnt))==-1))return -1;
		if(LIKELY(!(len-=l)))goto out;
		for(;l>=iov[0].iov_len;iov++,iovcnt--)l-=iov[0].iov_len;
		iov[0].iov_base+=l;
		iov[0].iov_len-=l;
	}
out:	return 0;
}

/*
 * dowrite
 *
 * input:  fd    - the target descriptor
 *	   buf   - the data to be written
 *	   count - the data length
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine writes the given data to the given descriptor.
 */

static inline int dowrite(int fd,void *buf,size_t count)
{
	size_t l;

	for(;count;count-=l)if(UNLIKELY((l=write(fd,buf,count))==-1))return -1;
	else buf=(void *)(((char *)buf)+l);
	return 0;
}

/*
 * doread
 *
 * input:  fd    - the target descriptor
 *	   buf   - the buffer for the data to be read
 *	   count - the data length
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine reads the given data from the given descriptor.
 *		The strart delay is infinite, then the data must flow at
 *		a defined minimal rate.
 */

static inline int doread(int fd,void *buf,size_t count)
{
	size_t l;
	struct pollfd p;

	for(p.fd=fd,p.events=POLLIN;count;)switch((l=read(fd,buf,count)))
	{
	case 0:	errno=EIO;
	case -1:return -1;
	default:buf=(void *)(((char *)buf)+l);
		if((count-=l))switch(poll(&p,1,NETWAIT*1000))
		{
		default:if(p.revents&POLLIN)break;
		case 0:	errno=EIO;
		case -1:return -1;
		}
	}
	return 0;
}

/*
 * docopyfile
 *
 * input:  out   - target fd
 *	   in    - source fd
 *	   count - data length
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine copies the specified amount of data from the
 *		source fd to the destination fd.
 */

static inline int docopyfile(int out,int in,size_t count)
{
	size_t l;
	char bfr[65536];

	for(;count;count-=l)
	{
		l=LIKELY(count>sizeof(bfr))?sizeof(bfr):count;
		if(UNLIKELY(doread(in,bfr,l)))return -1;
		if(UNLIKELY(dowrite(out,bfr,l)))return -1;
	}
	return 0;
}

/*
 * dorecvfile
 *
 * input:  out   - the destination fd
 *	   in    - the source fd
 *	   count - the amount of data to be sent
 *	   sctp  - 1 if sctp is in use (allow for sctp broken splice())
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine receives a file on a best effort base
 *		using splice where not broken. This stuff is highly
 *		OS dependent.
 */

#if defined(__linux__)

static inline int dorecvfile(int out,int in,size_t count,int sctp)
{
	int r;
	int w;
	int p[2];
	struct pollfd pp;

	/* this is no problem: for netccd sctp will always be 1 if !SAFEAPI */

#ifdef SAFEAPI
	if(sctp||pipe2(p,O_CLOEXEC))goto out;
#else
	if(sctp||pipe(p))goto out;
#endif
	/* this will block even if SPLICE_F_NONBLOCK as the socket blocks */

	for(pp.fd=in,pp.events=POLLIN;count;)switch((r=splice(in,NULL,p[1],NULL,
		count>65536?65536:count,SPLICE_F_MOVE|SPLICE_F_NONBLOCK)))
	{
	case 0:	errno=EIO;
	case -1:goto fail;
	default:for(count-=r;r;r-=w)if(UNLIKELY((w=splice(p[0],NULL,out,NULL,r,
			SPLICE_F_MOVE|SPLICE_F_NONBLOCK))==-1))goto fail;
		if(count)switch(poll(&pp,1,NETWAIT*1000))
		{
		default:if(pp.revents&POLLIN)break;
		case 0: errno=EIO;
fail:		case -1:r=errno;
			close(p[0]);
			close(p[1]);
			errno=r;
			return -1;
		}
	}

	close(p[0]);
	close(p[1]);
	return 0;

out:	return docopyfile(out,in,count);
}

#else

#define dorecvfile(a,b,c,d) docopyfile(a,b,c)

#endif

/*
 * dosendfile
 *
 * input:  out   - the destination fd
 *	   in    - the source fd
 *	   count - the amount of data to be sent
 *	   sctp  - 1 if sctp is in use (allow for sctp broken sendfile())
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine sends the given file on a best effort base
 *		using sendfile implementations where available and not broken.
 *		This stuff is highly OS dependent.
 */

#if defined(__linux__)

#define dosendfile(a,b,c,d) _dosendfile(a,b,c)

static inline int _dosendfile(int out,int in,size_t count)
{
	off_t offset;

	for(offset=0;offset!=count;)
		if(UNLIKELY(sendfile(out,in,&offset,count-offset)==-1))
			return -1;
	return 0;
}

#elif defined(__FreeBSD__)

#warning no sendfile available for sctp, using fallback

static inline int dosendfile(int out,int in,size_t count,int sctp)
{
	off_t offset;
	off_t sent;

	if(sctp)return docopyfile(out,in,count);
	for(offset=0;count;offset+=sent,count-=sent)
		if(UNLIKELY(sendfile(in,out,offset,count,NULL,&sent,0)==-1))
			return -1;
	return 0;
}

#else

#warning no sendfile implementation, using fallback

#define dosendfile(a,b,c,d) docopyfile(a,b,c)

#endif

/*
 * dowritefd
 *
 * input:  s    - the descriptor the message is to be sent over
 *	   data - the message to be passed
 *	   len  - the message length
 *	   fd   - the file descriptor to be passed
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine sends the given data as well as the given
 *		file descriptor across a unix domain socket.
 */

static inline int dowritefd(int s,void *data,int len,int fd)
{
	struct iovec iov;
	struct msghdr hdr;
	struct cmsghdr cmsg[2];

	iov.iov_base=data;
	iov.iov_len=len;
	hdr.msg_name=NULL;
	hdr.msg_namelen=0;
	hdr.msg_iov=&iov;
	hdr.msg_iovlen=1;
	hdr.msg_control=(void *)&cmsg;
	hdr.msg_controllen=sizeof(struct cmsghdr)+sizeof(int);
	hdr.msg_flags=0;
	cmsg[0].cmsg_len=hdr.msg_controllen;
	cmsg[0].cmsg_level=SOL_SOCKET;
	cmsg[0].cmsg_type=SCM_RIGHTS;
	*((int *)(&cmsg[1]))=fd;

	if(sendmsg(s,&hdr,0)<0)return -1;
	return 0;
}

/*
 * doreadfd
 *
 * input:  s    - the descriptor the message is to be read from
 *	   data - the message buffer
 *	   len  - the message buffer size
 *	   fd   - storage location for the passed file descriptor
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine receives data as well as a file descriptor
 *		over a unix domain socket.
 */

static inline int doreadfd(int s,void *data,int len,int *fd)
{
	int l;
	struct iovec iov;
	struct msghdr hdr;
	struct cmsghdr cmsg[2];

	iov.iov_base=data;
	iov.iov_len=len;
	hdr.msg_name=NULL;
	hdr.msg_namelen=0;
	hdr.msg_iov=&iov;
	hdr.msg_iovlen=1;
	hdr.msg_control=(void *)&cmsg;
	hdr.msg_controllen=sizeof(struct cmsghdr)+sizeof(int);
	hdr.msg_flags=0;

	if((l=recvmsg(s,&hdr,0))==-1)return -1;
	if(l!=len)goto fail;
	if(cmsg[0].cmsg_len!=sizeof(struct cmsghdr)+sizeof(int))goto fail;
	if(cmsg[0].cmsg_level!=SOL_SOCKET||cmsg[0].cmsg_type!=SCM_RIGHTS)
		goto fail;
	*fd=*((int *)(&cmsg[1]));
	return 0;

fail:	errno=EIO;
	return -1;
}

#endif
