
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

/******************************************************************************

Note:
The whole ifdef mess in rmtcompile() can't be helped as long as so called
POSIX operating systems stick to the 80s state of art. We live in a world
of multicore systems and threads, friends, preforking is an ancient art!

******************************************************************************/

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "remote.h"
#include "update.h"
#include "ifc.h"
#include "tcp.h"
#include "tcp6.h"
#include "sctp.h"
#include "sctp6.h"
#include "lock.h"
#include "term.h"
#include "log.h"
#include "safeapi.h"
#include "netio.h"
#include "sigs.h"
#include "ipc.h"
#include "auth.h"
#include "stamp.h"
#include "ugid.h"
#include "caps.h"

/* there won't be an answer, let it be... */

#if !defined(CLOCK_MONOTONIC) || !defined(SIGEV_THREAD)
#warning system supports only ancient timers, must cripple functionality
#endif

typedef struct
{
	int valid;
	pthread_t tid;
} NOTIFY;

typedef struct _child
{
	struct _child *next;
	pthread_t tid;
} CHILD;

typedef struct
{
	int s;
	int ifd;
	int ofd;
	int efd;
	int pipefd[2];
	int job;
	int zlib;
	int ping;
#if defined(CLOCK_MONOTONIC) && defined(SIGEV_THREAD)
	int timer;
	timer_t td;
#endif
	pid_t pid;
	char *bfr;
	char **argv;
	char *wrk;
	char *in;
	char *out;
	char *err;
	char *outmem;
	pthread_t h;
	z_stream z;
} RMTDATA;

static char *toollist[MAXTOOLS+1]={NULL};
static int sctp=0;
static int v6=0;
static int running=0;
static int site=0;
static int auth=0;
static int unique=0;
static int sctpauth=0;
static int zerocopy=0;
static int window=WINDOW;
static int nosplice;
static int tcp;
static int self;
static int net;
static int mask;
static int niceval;
static int notify[2];
static uid_t uid;
static gid_t gid;
static pthread_t h;
static char *tmp;
static char *swtch1;
static char *swtch2;
static CHILD *head;
static CHILD *tail;
static int total;
static int delay;
static clock_t base;
static pthread_mutex_t mtx=PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t qtx=PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t ntx=PTHREAD_MUTEX_INITIALIZER;
static MD5HMDATA key;
static MD5(hash);

/*
 * zfile
 *
 * input:  rmt - required data in a cancel friendly way
 *	   len - the length of data to be processed
 *
 * return: 0 in case of success, -1 in casse of error
 *
 * description:	This routine decompresses a data stream
 *		of given length from a descriptor to
 *		an open file.
 */

static int zfile(RMTDATA *rmt,int len)
{
	int i;
	int eos;
	struct pollfd p;
	char in[65536];
	char out[65536];

	/* data setup */

	memset(&rmt->z,0,sizeof(rmt->z));

	p.fd=rmt->s;
	p.events=POLLIN;

	eos=0;
	rmt->z.next_in=(void *)in;

	/* decompress init */

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	if(UNLIKELY(inflateInit2(&rmt->z,-15)!=Z_OK))
	{
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
		BAD("inflateInit2 failure");
	}
	rmt->zlib=1;
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

	/* for all compressed data */

	while(len)
	{
		/* wait for data to be available */

		if(UNLIKELY((i=poll(&p,1,NETWAIT*1000))!=1))
		{
			if(!i)i=EIO;
			else if((i=errno)==EINTR)continue;
			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
			inflateEnd(&rmt->z);
			rmt->zlib=0;
			pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
			BAD("poll: %s",errstr(i,in));
		}

		if(UNLIKELY(!(p.revents&POLLIN)))
		{
			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
			inflateEnd(&rmt->z);
			rmt->zlib=0;
			pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
			BAD("poll: %s",errstr(EIO,in));
		}

		/* read next data chunk */

		switch((i=read(rmt->s,rmt->z.next_in+rmt->z.avail_in,
			sizeof(in)-rmt->z.avail_in)))
		{
		case 0:	errno=EIO;
		case -1:if((i=errno)==EINTR)continue;
			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
			inflateEnd(&rmt->z);
			rmt->zlib=0;
			pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
			BAD("read: %s",errstr(i,in));

		default:len-=i;
			rmt->z.avail_in+=i;
			if(rmt->z.avail_in!=sizeof(in))continue;
		}

		/* decompress */

		do
		{
			rmt->z.next_out=(void *)out;
			rmt->z.avail_out=sizeof(out);

			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
			switch(inflate(&rmt->z,Z_NO_FLUSH))
			{
			case Z_OK:
			case Z_BUF_ERROR:
				break;

			case Z_STREAM_END:
				if(!len)
				{
					eos=1;
					break;
				}
			default:pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,
					NULL);
				BAD("inflate failure");
			}
			pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

			i=sizeof(out)-rmt->z.avail_out;
			if(UNLIKELY(write(rmt->ifd,out,i)!=i))
			{
				i=errno;
				pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,
					NULL);
				inflateEnd(&rmt->z);
				rmt->zlib=0;
				pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,
					NULL);
				BAD("write: %s",errstr(i,in));
			}

			if(eos)goto done;
		} while(rmt->z.avail_out!=sizeof(out));

		/* move remaining bits to buffer start and continue */

		if(rmt->z.avail_in)memmove(in,rmt->z.next_in,rmt->z.avail_in);
		rmt->z.next_in=(void *)in;
	}

	/* process remaining data, if any */

	do
	{
		rmt->z.next_out=(void *)out;
		rmt->z.avail_out=sizeof(out);

		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
		switch((i=inflate(&rmt->z,Z_NO_FLUSH)))
		{
		case Z_STREAM_END:
			eos=1;
		case Z_OK:
		case Z_BUF_ERROR:
			break;

		default:pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
			BAD("inflate failure");
		}
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

		i=sizeof(out)-rmt->z.avail_out;
		if(UNLIKELY(write(rmt->ifd,out,i)!=i))
		{
			i=errno;
			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
			inflateEnd(&rmt->z);
			rmt->zlib=0;
			pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
			BAD("write: %s",errstr(i,in));
		}

		if(eos)goto done;
	} while(rmt->z.avail_out!=sizeof(out));

	/* now finish decompression for good */

	rmt->z.next_out=(void *)out;
	rmt->z.avail_out=sizeof(out);

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	if(UNLIKELY(inflate(&rmt->z,Z_FINISH)!=Z_STREAM_END))
	{
		inflateEnd(&rmt->z);
		rmt->zlib=0;
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
		BAD("inflate finish failure");
	}
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

	i=sizeof(out)-rmt->z.avail_out;
	if(UNLIKELY(write(rmt->ifd,out,i)!=i))
	{
		i=errno;
		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
		inflateEnd(&rmt->z);
		rmt->zlib=0;
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
		BAD("write: %s",errstr(i,in));
	}

	/* clean up */

done:	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	rmt->zlib=0;
	if(UNLIKELY(inflateEnd(&rmt->z)!=Z_OK))
	{
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
		BAD("inflateEnd failure");
	}
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

	/* signal success */

	return 0;
}

/*
 * pipecompress
 *
 * input:  rmt - required data in a cancel friendly way
 *
 * return: 1 in case of compressed data, 0 if no data, -1 in casse of error
 *
 * description: This routine compresses data from a pipe to an open file.
 */

static int pipecompress(RMTDATA *rmt)
{
	int i;
	int ifree;
	int data;
	char *iptr;
	char in[65536];
	char out[65536];

	/* data setup */

	data=0;
	memset(&rmt->z,0,sizeof(rmt->z));
	rmt->z.data_type=Z_TEXT;

	/* compression init */

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);

	if(UNLIKELY(deflateInit2(&rmt->z,3,Z_DEFLATED,-15,8,Z_FILTERED)!=Z_OK))
	{
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
		JUMP(out,"deflateInit2 failure");
	}
	rmt->zlib=-1;

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

	/* compression data setup */

	rmt->z.next_in=(void *)in;
	rmt->z.next_out=(void *)out;
	rmt->z.avail_out=sizeof(out);

	iptr=in;
	ifree=sizeof(in);

	/* compress while there's data available */

	while(1)switch((i=read(rmt->pipefd[0],iptr,ifree)))
	{
	/* read error */

	case -1:JUMP(out,"uncompressed read: %s",errstr(errno,in));

	/* pipe closed, EOF */

	case 0:	rmt->z.avail_in=0;

		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
		if(UNLIKELY(deflate(&rmt->z,Z_FINISH)!=Z_STREAM_END))
		{
			pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
			JUMP(out,"deflate finish failure");
		}
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

		i=(int)((char *)rmt->z.next_out-out);
		if(LIKELY(i))if(UNLIKELY(write(rmt->efd,out,i)!=i))
			JUMP(out,"write: %s",errstr(errno,in));

		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
		rmt->zlib=0;
		if(UNLIKELY(deflateEnd(&rmt->z)!=Z_OK))
		{
			pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
			JUMP(out,"deflateEnd failure");
		}
		close(rmt->pipefd[0]);
		rmt->pipefd[0]=-1;
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

		return data;

	/* more data to come */

	default:data=1;
		rmt->z.avail_in+=i;

		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
		if(UNLIKELY(deflate(&rmt->z,Z_NO_FLUSH)!=Z_OK))
		{
			pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
			JUMP(out,"deflate failure");
		}
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

		if(UNLIKELY(rmt->z.avail_in))
		{
			if(in!=(char *)rmt->z.next_in)
				memmove(in,rmt->z.next_in,rmt->z.avail_in);
			rmt->z.next_in=(void *)in;
			ifree=sizeof(in)-rmt->z.avail_in;
			iptr=in+rmt->z.avail_in;
		}
		else
		{
			rmt->z.next_in=(void *)in;
			ifree=sizeof(in);
			iptr=in;
		}

		i=(int)((char *)rmt->z.next_out-out);
		if(LIKELY(i))if(UNLIKELY(write(rmt->efd,out,i)!=i))
			JUMP(out,"write: %s",errstr(errno,in));

		rmt->z.next_out=(void *)out;
		rmt->z.avail_out=sizeof(out);
		break;
	}

	/* error cleanup */

out:	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	if(rmt->zlib)
	{
		deflateEnd(&rmt->z);
		rmt->zlib=0;
	}
	close(rmt->pipefd[0]);
	rmt->pipefd[0]=-1;
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

	/* signal error */

	return -1;
}

/*
 * objcompress
 *
 * input:  rmt - required data in a cancel friendly way
 *	   len - the length of data to be processed
 *
 * return: compressed data length or -1 in casse of error
 *
 * description:	This routine compresses a file of given
 *		from a descriptor to memory.
 */

static int objcompress(RMTDATA *rmt,int len)
{
	int i;
	int ifree;
	int outlen;
	char *iptr;
	char *tmp;
	char in[65536];

	/* data setup */

	memset(&rmt->z,0,sizeof(rmt->z));
	rmt->z.data_type=Z_BINARY;

	/* compression init */

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);

	if(UNLIKELY(deflateInit2(&rmt->z,3,Z_DEFLATED,-15,8,Z_FILTERED)!=Z_OK))
	{
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
		JUMP(out,"deflateInit2 failure");
	}
	rmt->zlib=-1;

	/* allocate first memory chunk */

	if(UNLIKELY(!(rmt->outmem=malloc(65536))))
	{
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
		JUMP(out,"malloc failure");
	}
	outlen=65536;

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

	/* compression data setup */

	rmt->z.next_in=(void *)in;
	rmt->z.next_out=(void *)rmt->outmem;
	rmt->z.avail_out=outlen;

	iptr=in;
	ifree=sizeof(in);

	/* compress complete file */

	while(len)switch((i=read(rmt->ofd,iptr,ifree)))
	{
	case 0:	errno=EIO;
	case -1:JUMP(out,"uncompressed read: %s",errstr(errno,in));

	default:rmt->z.avail_in+=i;
		len-=i;

		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);

repeat:		switch(deflate(&rmt->z,Z_NO_FLUSH))
		{
		case Z_BUF_ERROR:
			if(UNLIKELY(!(tmp=realloc(rmt->outmem,outlen+65536))))
			{
				pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,
					NULL);
				JUMP(out,"malloc failure");
			}
			rmt->z.next_out=(void *)(tmp+((char *)rmt->z.next_out-
				rmt->outmem));
			rmt->outmem=tmp;
			rmt->z.avail_out+=65536;
			outlen+=65536;
			goto repeat;

		case Z_OK:
			pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

			if(UNLIKELY(rmt->z.avail_in))
			{
				if(in!=(char *)rmt->z.next_in)memmove(in,
					rmt->z.next_in,rmt->z.avail_in);
				rmt->z.next_in=(void *)in;
				ifree=sizeof(in)-rmt->z.avail_in;
				iptr=in+rmt->z.avail_in;
			}
			else
			{
				rmt->z.next_in=(void *)in;
				ifree=sizeof(in);
				iptr=in;
			}
			break;

		default:pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
			JUMP(out,"deflate failure");
		}
	}

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);

	/* compress remaining bits of file */

rep2:	switch(deflate(&rmt->z,Z_FINISH))
	{
	case Z_BUF_ERROR:
		if(UNLIKELY(!(tmp=realloc(rmt->outmem,outlen+65536))))
		{
			pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
			JUMP(out,"malloc failure");
		}
		rmt->z.next_out=(void *)(tmp+((char *)rmt->z.next_out-
				rmt->outmem));
		rmt->outmem=tmp;
		rmt->z.avail_out+=65536;
		outlen+=65536;

	case Z_OK:
		goto rep2;

	case Z_STREAM_END:
		break;

	default:pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
		JUMP(out,"deflate finish failure");
	}

	/* finish compression for good */

	outlen-=rmt->z.avail_out;
	rmt->zlib=0;

	if(UNLIKELY(deflateEnd(&rmt->z)!=Z_OK))
	{
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
		JUMP(out,"deflateEnd failure");
	}

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

	/* return compressed length */

	return outlen;

	/* error cleanup */

out:	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	if(rmt->outmem)
	{
		free(rmt->outmem);
		rmt->outmem=NULL;
	}
	if(rmt->zlib)
	{
		deflateEnd(&rmt->z);
		rmt->zlib=0;
	}
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

	/* signal error */

	return -1;
}

/*
 * pinger
 *
 * input:  data - the socket descriptor to send to
 *
 * return: always NULL
 *
 * description: This is a fully cancellable thread that sends OOB messages
 *		at compiled in intervals. Only required for broken systems
 *		with crippled TCP keepalive.
 */

static void *pinger(void *data)
{
	int s=(int)((long)data);
	struct timespec tv;
	char dummy;

	dropcaps(NEED_NONE);

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
	while(1)
	{
		tv.tv_sec=OOBINTERVAL;
		tv.tv_nsec=0;
		nanosleep(&tv,NULL);
		dummy=send(s,&dummy,1,MSG_OOB);
	}
	pthread_exit(NULL);
}

/*
 * objcompress
 *
 * input:  rmt - required data in a cancel friendly way
 *
 * description:	This routine cleans up after thread cancellation.
 */

static void rmtclean(void *data)
{
	RMTDATA *rmt=(RMTDATA *)data;
#ifndef SAFEUID
	sigstore(o);
#endif
	struct stat stb;

	/* log a little */

	NOTICE("forcibly cancelling remote compilation");

	/* cancel timer if one is running */

#if defined(CLOCK_MONOTONIC) && defined(SIGEV_THREAD)
	if(rmt->timer)timer_delete(rmt->td);
#endif

	/* cancel braindead system pinger thread, if there is any */

	if(rmt->ping)
	{
		pthread_cancel(rmt->h);
		pthread_join(rmt->h,NULL);
	}

	/* shutdown tcp connection */

	if(!sctp)shutdown(rmt->s,SHUT_RDWR);

	/* close remote connection */

	close(rmt->s);

	/* kill any running child process */

	if(rmt->pid!=-1)
	{
		kill(rmt->pid,SIGKILL);
		waitpid(rmt->pid,NULL,0);
	}

	/* update job stats */

	if(rmt->job)update_mode(JOB_DONE,0,NULL);

	/* clean up zlib if it is in use */

	switch(rmt->zlib)
	{
		case 1:	inflateEnd(&rmt->z);
			break;
		case -1:deflateEnd(&rmt->z);
			break;
	}

	/* close open pipe handles */

	if(rmt->pipefd[0]!=-1)close(rmt->pipefd[0]);
	if(rmt->pipefd[1]!=-1)close(rmt->pipefd[1]);

	/* remove temporary files and directory */

#ifdef SAFEUID
	setfsgid(gid);
	setfsuid(uid);
#else
	ipc_needex();
	exlock();
	block(PROCESS,&o);
	if(setresgid(gid,gid,0))goto nogid;
	if(setresuid(uid,uid,0))goto nouid;
#endif
	if(rmt->ifd!=-1)
	{
		unlink(rmt->in);
		close(rmt->ifd);
	}
	if(rmt->ofd!=-1)
	{
		unlink(rmt->out);
		close(rmt->ofd);
	}
	if(rmt->efd!=-1)
	{
		unlink(rmt->err);
		close(rmt->efd);
	}
	if(rmt->wrk)
	{
		if(rmdir(rmt->wrk))
		{
			if(rmt->out)if(!stat(rmt->out,&stb))unlink(rmt->out);
			if(rmdir(rmt->wrk))
			{
				if(rmt->out)if(!stat(rmt->out,&stb))
					unlink(rmt->out);
				rmdir(rmt->wrk);
			}
		}
		free(rmt->wrk);
	}
#ifdef SAFEUID
	setfsuid(0);
	setfsgid(0);
#else
	setresuid(0,0,0);
nouid:	setresgid(0,0,0);
nogid:	unblock(PROCESS,&o);
	unlock();
	ipc_exdone();
#endif

	/* free working memory */

	if(rmt->in)free(rmt->in);
	if(rmt->out)free(rmt->out);
	if(rmt->err)free(rmt->err);
	if(rmt->argv)free(rmt->argv);
	if(rmt->bfr)free(rmt->bfr);
	if(rmt->outmem)free(rmt->outmem);
}

/*
 * killer
 *
 * input: pid - the child process to be killed
 *
 * description: This routine is called when the timer expires. It kills
 *		the given child process.
 */

#if defined(CLOCK_MONOTONIC) && defined(SIGEV_THREAD)

#if defined(__NetBSD__)

static void killer(union sigval *pid)
{
	kill(pid->sival_int,SIGKILL);
}

#else

static void killer(union sigval pid)
{
	kill(pid.sival_int,SIGKILL);
}

#endif

#endif

/*
 * rmtfinal
 *
 * input: unused - unused
 *
 * description:	This routine is the cancellation handler for the
 *		final rmtcompile() stage. It just releases the
 *		held mutex.
 */

static void rmtfinal(void *unused)
{
	pthread_mutex_unlock(&mtx);
}

/*
 * rmtcompile
 *
 * input:  data - the peer socket descriptor
 *
 * return: always NULL
 *
 * description:	This routine is a thread that performs a compile on
 *		behalf of a remote peer.
 */

static void *rmtcompile(void *data)
{
	int i;
	int l;
	int status;
	int flags;
	int total;
	int len;
	int tmo;
	char *p;
#if defined(CLOCK_MONOTONIC) && defined(SIGEV_THREAD)
	struct sigevent tev;
	struct itimerspec it;
#endif
	RMTDATA rmt;
	NOTIFY n;
#ifndef SAFEUID
	sigstore(o);
#endif
	struct stat stb;
	THDR hdr;
	char mem[256];

	/* default thread setup */

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);

	dftblock(THREAD);
	dropcaps(NEED_UGID);

	pthread_cleanup_push(rmtclean,&rmt);

	/* data setup */

	rmt.s=(int)((long)data);
	rmt.ifd=-1;
	rmt.ofd=-1;
	rmt.efd=-1;
	rmt.pipefd[0]=-1;
	rmt.pipefd[1]=-1;
	rmt.pid=-1;
	rmt.bfr=NULL;
	rmt.argv=NULL;
	rmt.wrk=NULL;
	rmt.in=NULL;
	rmt.out=NULL;
	rmt.err=NULL;
	rmt.outmem=NULL;
	rmt.job=0;
	rmt.zlib=0;
	rmt.ping=0;
#if defined(CLOCK_MONOTONIC) && defined(SIGEV_THREAD)
	rmt.timer=0;
#endif

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

	/* read header from peer and perform basic checks */

	if(UNLIKELY(donetwait(rmt.s)))
		JUMP(out,"wait for header from client: %s",errstr(errno,mem));

	if(UNLIKELY(doread(rmt.s,&hdr,sizeof(hdr))))
		JUMP(out,"read header from client: %s",errstr(errno,mem));

	if(UNLIKELY(chkauth(&key,&hdr,sizeof(THDR),auth,window)))
		JUMP(out,"client authentication failure");

	flags=ntohl(hdr.data[0]);
	total=ntohl(hdr.data[1]);
	len=ntohl(hdr.data[2]);

	tmo=(total>>16)&0xffff;
	total&=0xffff;

	if(UNLIKELY(ntohl(hdr.data[3])!=unique))JUMP(out,"unique id mismatch");

	/* if requested start braindead system pinger thread */

	if(UNLIKELY((flags&HINT_OOB)&&!sctp))
	{
		if(LIKELY(!pthread_create(&rmt.h,NULL,pinger,
			(void *)((long)rmt.s))))rmt.ping=1;
		else JUMP(out,"pthread_create: %s",errstr(errno,mem));
	}

	/* allocate argument memory, read and process arguments */

	if(UNLIKELY(!(p=rmt.bfr=malloc(len))))JUMP(out,"malloc failure");
	if(UNLIKELY(doread(rmt.s,p,len)))
		JUMP(out,"read argv from peer: %s",errstr(errno,mem));

	if(UNLIKELY(!(rmt.argv=malloc(sizeof(char *)*(total+6)))))
		JUMP(out,"malloc failure");

	for(p=rmt.bfr,i=0;i<total;i++)
	{
		rmt.argv[i]=p;
		while(*p++);
	}

	/* handle biarch cross compile flags addition */

	if(flags&HINT_ADDSWITCH)if(swtch1&&swtch2)
	{
		for(l=0;l<i;l++)if(!strcmp(rmt.argv[l],swtch1))break;
		else if(!strcmp(rmt.argv[l],swtch2))break;
		if(l==i)rmt.argv[i++]=swtch1;
	}

	/* set up actual compile mode */

	if(UNLIKELY(flags&HINT_S))rmt.argv[i++]="-S";
	else rmt.argv[i++]="-c";

	/* abort if the remote command request contains a path */

	if(UNLIKELY(strchr(rmt.argv[0],'/')!=NULL))
		JUMP(out,"request contains executable path");

	/* check for allowed tools */

	if(LIKELY(toollist[0]!=NULL))
	{
		for(l=0;toollist[l];l++)
			if(!strcmp(rmt.argv[0],toollist[l]))break;
		if(UNLIKELY(!toollist[l]))
			JUMP(out,"illegal tool %s requested",rmt.argv[0]);
	}

	/* create temporary directory and all required files within this
	   directory. Do this as the user the compiler will run under.
	   I'm not going to comment the ifdef mess further. Modernize
	   your systems. Thats all, folks.*/

	if(UNLIKELY(!(rmt.wrk=malloc(strlen(tmp)+19))))
		JUMP(out,"malloc failure");
	sprintf(rmt.wrk,"%s/netccd-tmp-XXXXXX",tmp);

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);

#ifdef SAFEUID
#ifdef SAFEAPI
	if(UNLIKELY(!safeapi()))shlock();
#else
	shlock();
#endif
	setfsgid(gid);
	setfsuid(uid);
#else
	ipc_needex();
	exlock();
	block(PROCESS,&o);
	if(UNLIKELY(setresgid(gid,gid,0)))
	{
		i=errno;
		unblock(PROCESS,&o);
		unlock();
		ipc_exdone();
		free(rmt.wrk);
		rmt.wrk=NULL;
		JUMP(out,"setresgid: %s",errstr(i,mem));
	}

	if(UNLIKELY(setresuid(uid,uid,0)))
	{
		i=errno;
		setresgid(0,0,0);
		unblock(PROCESS,&o);
		unlock();
		ipc_exdone();
		free(rmt.wrk);
		rmt.wrk=NULL;
		JUMP(out,"setresuid: %s",errstr(i,mem));
	}
#endif

	if(UNLIKELY(!mkdtemp(rmt.wrk)))
	{
		i=errno;
#ifdef SAFEUID
		setfsuid(0);
		setfsgid(0);
#ifdef SAFEAPI
		if(UNLIKELY(!safeapi()))unlock();
#else
		unlock();
#endif
#else
		setresuid(0,0,0);
		setresgid(0,0,0);
		unblock(PROCESS,&o);
		unlock();
		ipc_exdone();
#endif
		free(rmt.wrk);
		rmt.wrk=NULL;
		JUMP(out,"mkdtemp: %s",errstr(i,mem));
	}

	l=strlen(rmt.wrk);

	if(flags&HINT_STDC)p="i";
	else if(flags&HINT_CXX)p="ii";
	else if(flags&HINT_OBJC)p="mi";
	else if(flags&HINT_OBJCXX)p="mii";
	else
	{
#ifdef SAFEUID
		setfsuid(0);
		setfsgid(0);
#ifdef SAFEAPI
		if(UNLIKELY(!safeapi()))unlock();
#else
		unlock();
#endif
#else
		setresuid(0,0,0);
		setresgid(0,0,0);
		unblock(PROCESS,&o);
		unlock();
		ipc_exdone();
#endif
		JUMP(out,"internal error: flags=%04x",flags);
	}

	if(UNLIKELY(!(rmt.in=malloc(l+12))))
	{
#ifdef SAFEUID
		setfsuid(0);
		setfsgid(0);
#ifdef SAFEAPI
		if(UNLIKELY(!safeapi()))unlock();
#else
		unlock();
#endif
#else
		setresuid(0,0,0);
		setresgid(0,0,0);
		unblock(PROCESS,&o);
		unlock();
		ipc_exdone();
#endif
		JUMP(out,"malloc failure");
	}
	sprintf(rmt.in,"%s/netccd.%s",rmt.wrk,p);

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((rmt.ifd=open(rmt.in,
			O_WRONLY|O_CREAT|O_TRUNC|O_CLOEXEC,0600))==-1))
		{
			i=errno;
#ifdef SAFEUID
			setfsuid(0);
			setfsgid(0);
#else
			setresuid(0,0,0);
			setresgid(0,0,0);
			unblock(PROCESS,&o);
			unlock();
			ipc_exdone();
#endif
			JUMP(out,"open %s: %s",rmt.in,errstr(i,mem));
		}
	}
	else
#endif
	{
		if(UNLIKELY((rmt.ifd=open(rmt.in,
			O_WRONLY|O_CREAT|O_TRUNC,0600))==-1))
		{
			i=errno;
#ifdef SAFEUID
			setfsuid(0);
			setfsgid(0);
#else
			setresuid(0,0,0);
			setresgid(0,0,0);
			unblock(PROCESS,&o);
#endif
			unlock();
#ifndef SAFEUID
			ipc_exdone();
#endif
			JUMP(out,"open %s: %s",rmt.in,errstr(i,mem));
		}
		fcntl(rmt.ifd,F_SETFD,FD_CLOEXEC);
	}

	if(UNLIKELY(flags&HINT_S))p="s";
	else p="o";
	if(UNLIKELY(!(rmt.out=malloc(l+10))))
	{
#ifdef SAFEUID
		setfsuid(0);
		setfsgid(0);
#ifdef SAFEAPI
		if(UNLIKELY(!safeapi()))unlock();
#else
		unlock();
#endif
#else
		setresuid(0,0,0);
		setresgid(0,0,0);
		unblock(PROCESS,&o);
		unlock();
		ipc_exdone();
#endif
		JUMP(out,"malloc failure");
	}
	sprintf(rmt.out,"%s/netccd.%s",rmt.wrk,p);

	rmt.argv[i++]="-o";
	rmt.argv[i++]=rmt.out;
	rmt.argv[i++]=rmt.in;
	rmt.argv[i]=NULL;

	if(UNLIKELY(!(rmt.err=malloc(l+12))))
	{
#ifdef SAFEUID
		setfsuid(0);
		setfsgid(0);
#ifdef SAFEAPI
		if(UNLIKELY(!safeapi()))unlock();
#else
		unlock();
#endif
#else
		setresuid(0,0,0);
		setresgid(0,0,0);
		unblock(PROCESS,&o);
		unlock();
		ipc_exdone();
#endif
		JUMP(out,"malloc failure");
	}
	sprintf(rmt.err,"%s/netccd.log",rmt.wrk);

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((rmt.efd=open(rmt.err,
			O_RDWR|O_CREAT|O_TRUNC|O_CLOEXEC,0600))==-1))
		{
			i=errno;
#ifdef SAFEUID
			setfsuid(0);
			setfsgid(0);
#else
			setresuid(0,0,0);
			setresgid(0,0,0);
			unblock(PROCESS,&o);
			unlock();
			ipc_exdone();
#endif
			JUMP(out,"open %s: %s",rmt.err,errstr(i,mem));
		}
	}
	else
#endif
	{
		if(UNLIKELY((rmt.efd=open(rmt.err,
			O_RDWR|O_CREAT|O_TRUNC,0600))==-1))
		{
			i=errno;
#ifdef SAFEUID
			setfsuid(0);
			setfsgid(0);
#else
			setresuid(0,0,0);
			setresgid(0,0,0);
			unblock(PROCESS,&o);
#endif
			unlock();
#ifndef SAFEUID
			ipc_exdone();
#endif
			JUMP(out,"open %s: %s",rmt.err,errstr(i,mem));
		}
		fcntl(rmt.efd,F_SETFD,FD_CLOEXEC);
	}
#ifdef SAFEUID
	setfsuid(0);
	setfsgid(0);
#else
	setresuid(0,0,0);
	setresgid(0,0,0);
	unblock(PROCESS,&o);
#endif
	if(UNLIKELY(flags&HINT_ZLIB))
	{
#ifdef SAFEAPI
		if(LIKELY(safeapi()))
		{
			if(UNLIKELY(pipe2(rmt.pipefd,O_CLOEXEC)))
			{
#ifdef SAFEUID
				JUMP(out,"pipe2 failure: %s",errstr(errno,mem));
#else
				i=errno;
				unlock();
				ipc_exdone();
				JUMP(out,"pipe2 failure: %s",errstr(i,mem));
#endif
			}
		}
		else
#endif
		{
			if(UNLIKELY(pipe(rmt.pipefd)))
			{
				i=errno;
				unlock();
#ifndef SAFEUID
				ipc_exdone();
#endif
				JUMP(out,"pipe failure: %s",errstr(i,mem));
			}
			fcntl(rmt.pipefd[0],F_SETFD,FD_CLOEXEC);
			fcntl(rmt.pipefd[1],F_SETFD,FD_CLOEXEC);
		}
	}
#if defined(SAFEAPI) && defined(SAFEUID)
	if(UNLIKELY(!safeapi()))unlock();
#else
	unlock();
#endif
#ifndef SAFEUID
	ipc_exdone();
#endif
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

	/* read preprocessed file length from peer */

	if(UNLIKELY(doread(rmt.s,&len,sizeof(uint32_t))))
		JUMP(out,"read source length from peer: %s",errstr(errno,mem));
	len=ntohl(len);

	/* read preprocessed file from peer */

	if(UNLIKELY(flags&HINT_ZLIB))
	{
		if(UNLIKELY(zfile(&rmt,len)))goto out;
	}
	else if(UNLIKELY(dorecvfile(rmt.ifd,rmt.s,len,nosplice)))
		JUMP(out,"read source from peer: %s",errstr(errno,mem));

	/* shutdown the reading side of the TCP connection */

	if(!sctp)shutdown(rmt.s,SHUT_RD);

	/* update job stats and wait for free slot, then fork */

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	update_mode(JOB_PENDING,0,NULL);
	rmt.job=1;
#if defined(SAFEAPI) && defined(SAFEUID)
	if(UNLIKELY(!safeapi()))exlock();
#else
#ifndef SAFEUID
	ipc_needex();
#endif
	exlock();
#endif
	switch((rmt.pid=fork()))
	{
	/* fork failure */

	case -1:
#if defined(SAFEAPI) && defined(SAFEUID)
		if(UNLIKELY(!safeapi()))unlock();
#else
		unlock();
#ifndef SAFEUID
		ipc_exdone();
#endif
#endif
		update_mode(JOB_DONE,0,NULL);
		rmt.job=0;
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
		JUMP(out,"fork failure");

	/* client: set up descriptors and uid/gid, exec compiler */

	case 0:	if(UNLIKELY(flags&HINT_ZLIB))
		{
			if(rmt.pipefd[1]!=2)
			{
#ifdef SAFEAPI
				if(LIKELY(safeapi()))
					dup3(rmt.pipefd[1],2,0);
				else
#endif
				{
					dup2(rmt.pipefd[1],2);
					fcntl(2,F_SETFD,0);
				}
				close(rmt.pipefd[1]);
			}
			else fcntl(2,F_SETFD,0);
		}
		else if(rmt.efd!=2)
		{
#ifdef SAFEAPI
			if(LIKELY(safeapi()))
				dup3(rmt.efd,2,0);
			else
#endif
			{
				dup2(rmt.efd,2);
				fcntl(2,F_SETFD,0);
			}
			close(rmt.efd);
		}
		else fcntl(2,F_SETFD,0);
#ifdef SAFEAPI
		if(LIKELY(safeapi()))
		{
			close(0);
			open("/dev/null",O_RDONLY);
			close(1);
			open("/dev/null",O_WRONLY);
		}
		else
#endif
		{
			close(0);
			open("/dev/null",O_RDONLY);
			fcntl(0,F_SETFD,0);
			close(1);
			open("/dev/null",O_WRONLY);
			fcntl(1,F_SETFD,0);
		}
		if(niceval)i=nice(niceval);
		if(gid)i=setresgid(gid,gid,gid);
		if(uid)i=setresuid(uid,uid,uid);
		umask(077);
		execvp(rmt.argv[0],rmt.argv);
		ERR("can't execute %s: %s",rmt.argv[0],errstr(errno,mem));
		exit(255);

	/* parent */

	default:
#if defined(SAFEAPI) && defined(SAFEUID)
		if(UNLIKELY(!safeapi()))unlock();
#else
		unlock();
#ifndef SAFEUID
		ipc_exdone();
#endif
#endif

		/* start compile timeout timer if requested by peer */

#if defined(CLOCK_MONOTONIC) && defined(SIGEV_THREAD)
		if(tmo)
		{
			it.it_interval.tv_sec=0;
			it.it_interval.tv_nsec=0;
			it.it_value.tv_sec=tmo;
			it.it_value.tv_nsec=0;
			tev.sigev_notify=SIGEV_THREAD;
			tev.sigev_value.sival_int=rmt.pid;
			tev.sigev_notify_function=killer;
			tev.sigev_notify_attributes=NULL;
			if(UNLIKELY(timer_create(CLOCK_MONOTONIC,&tev,&rmt.td)))
				WARN("timer_create: %s",errstr(errno,mem));
			else if(UNLIKELY(timer_settime(rmt.td,0,&it,NULL)))
			{
				WARN("timer_settime: %s",errstr(errno,mem));
				timer_delete(rmt.td);
			}
			else rmt.timer=1;
		}
#endif
		/* compress compiler stderr if necessary */

		if(UNLIKELY(flags&HINT_ZLIB))
		{
			close(rmt.pipefd[1]);
			rmt.pipefd[1]=-1;
			pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
			if(UNLIKELY((l=pipecompress(&rmt))==-1))goto out;
		}
		else pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

		/* wait for compiler termination */

		while(UNLIKELY((i=waitpid(rmt.pid,&status,0))<=0))
			if(UNLIKELY(i==-1&&errno!=EINTR))
		{
			i=errno;
			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
			update_mode(JOB_DONE,0,NULL);
			rmt.job=0;
			pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
			JUMP(out,"waitpid: %s",errstr(i,mem));
		}
		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);

		/* cancel compile timout timer, if any */

#if defined(CLOCK_MONOTONIC) && defined(SIGEV_THREAD)
		if(rmt.timer)
		{
			timer_delete(rmt.td);
			rmt.timer=0;
		}
#endif
		/* update job stats */

		update_mode(JOB_DONE,0,NULL);
		rmt.job=0;
		rmt.pid=-1;
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
		break;
	}

	/* compile error, object file length is zero */

	if(UNLIKELY(status))len=0;

	/* compile success: go again though the ifdef mess as the user
	   the compiler was run under to access the object file - no
	   further comments, see above */

	else
	{
		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
#ifdef SAFEUID
#ifdef SAFEAPI
		if(UNLIKELY(!safeapi()))shlock();
#else
		shlock();
#endif
		setfsgid(gid);
		setfsuid(uid);
#else
		ipc_needex();
		exlock();
		block(PROCESS,&o);
		if(UNLIKELY(setresgid(gid,gid,0)))
		{
			i=errno;
			unblock(PROCESS,&o);
			unlock();
			ipc_exdone();
			JUMP(out,"setresgid: %s",errstr(i,mem));
		}
		if(UNLIKELY(setresuid(uid,uid,0)))
		{
			i=errno;
			setresgid(0,0,0);
			unblock(PROCESS,&o);
			unlock();
			ipc_exdone();
			JUMP(out,"setresuid: %s",errstr(i,mem));
		}
#endif
#ifdef SAFEAPI
		if(LIKELY(safeapi()))
		{
			if(UNLIKELY((rmt.ofd=open(rmt.out,
				O_RDONLY|O_CLOEXEC))==-1))
			{
				i=errno;
#ifdef SAFEUID
				setfsuid(0);
				setfsgid(0);
#else
				setresuid(0,0,0);
				setresgid(0,0,0);
				unblock(PROCESS,&o);
				unlock();
				ipc_exdone();
#endif
				JUMP(out,"open %s: %s",rmt.out,errstr(i,mem));
			}
		}
		else
#endif
		{
			if(UNLIKELY((rmt.ofd=open(rmt.out,O_RDONLY))==-1))
			{
				i=errno;
#ifdef SAFEUID
				setfsuid(0);
				setfsgid(0);
#else
				setresuid(0,0,0);
				setresgid(0,0,0);
				unblock(PROCESS,&o);
#endif
				unlock();
#ifndef SAFEUID
				ipc_exdone();
#endif
				JUMP(out,"open %s: %s",rmt.out,errstr(i,mem));
			}
			fcntl(rmt.ofd,F_SETFD,FD_CLOEXEC);
		}
#ifdef SAFEUID
		setfsuid(0);
		setfsgid(0);
#else
		setresuid(0,0,0);
		setresgid(0,0,0);
		unblock(PROCESS,&o);
#endif
#if defined(SAFEAPI) && defined(SAFEUID)
		if(UNLIKELY(!safeapi()))unlock();
#else
		unlock();
#ifndef SAFEUID
		ipc_exdone();
#endif
#endif
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
		if(UNLIKELY(fstat(rmt.ofd,&stb)))
			JUMP(out,"fstat on %s: %s",rmt.out,errstr(errno,mem));
		len=stb.st_size;
	}

	/* rewind the stderr file nad get file length */

	if(UNLIKELY(lseek(rmt.efd,0,SEEK_SET)))
		JUMP(out,"lseek on %s: %s",rmt.err,errstr(errno,mem));
	if(UNLIKELY(fstat(rmt.efd,&stb)))
		JUMP(out,"fstat on %s: %s",rmt.err,errstr(errno,mem));

	/* compress object file to memory in compression mode */

	if(UNLIKELY(flags&HINT_ZLIB))
	{
		if(!l)stb.st_size=0;
		if(LIKELY(len))if(UNLIKELY((len=objcompress(&rmt,len))==-1))
			goto out;
	}

	/* preset result header */

	hdr.data[0]=htonl(status);
	hdr.data[1]=htonl(stb.st_size);
	hdr.data[2]=htonl(len);
	hdr.data[3]=htonl(unique);

	/* handle TCP_CORK */

	if(flags&HINT_USECORK)cork(rmt.s);

	/* handle authentication */

	if(UNLIKELY(addauth(&key,&hdr,sizeof(THDR),auth)))
		JUMP(out,"add authentication to peer reply");

	/* cancel braindead system pinger thread, if it is running */

	if(UNLIKELY(rmt.ping))
	{
		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
		pthread_cancel(rmt.h);
		pthread_join(rmt.h,NULL);
		rmt.ping=0;
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
	}

	/* send header to peer */

	if(UNLIKELY(dowrite(rmt.s,&hdr,sizeof(THDR))))
		JUMP(out,"send header to peer: %s",errstr(errno,mem));

	/* send stderr to peer */

	if(UNLIKELY(dosendfile(rmt.s,rmt.efd,stb.st_size,sctp)))
		JUMP(out,"send stderr to peer: %s",errstr(errno,mem));

	/* send object to peer */

	if(UNLIKELY(flags&HINT_ZLIB))
	{
		if(dowrite(rmt.s,rmt.outmem,len))
			JUMP(out,"send object to peer: %s",errstr(errno,mem));
	}
	else if(UNLIKELY(dosendfile(rmt.s,rmt.ofd,len,sctp)))
		JUMP(out,"send object to peer: %s",errstr(errno,mem));

	/* handle TCP_CORK */

	if(flags&HINT_USECORK)uncork(rmt.s);

	/* clean up */

out:	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);

	/* kill compiler watchdog timer */

#if defined(CLOCK_MONOTONIC) && defined(SIGEV_THREAD)
	if(rmt.timer)timer_delete(rmt.td);
#endif

	/* kill braindead system pinger thread */

	if(UNLIKELY(rmt.ping))
	{
		pthread_cancel(rmt.h);
		pthread_join(rmt.h,NULL);
	}

	/* shutdown TCP connection */

	if(!sctp)shutdown(rmt.s,SHUT_WR);

	/* close peer connection */

	close(rmt.s);

	/* kill child process */

	if(UNLIKELY(rmt.pid!=-1))
	{
		kill(rmt.pid,SIGKILL);
		waitpid(rmt.pid,NULL,0);
	}

	/* close compression pipe handles */

	if(UNLIKELY(rmt.pipefd[0]!=-1))close(rmt.pipefd[0]);
	if(UNLIKELY(rmt.pipefd[1]!=-1))close(rmt.pipefd[1]);

	/* clean up temporary files and directory as the user the compiler
	   was run under. Again, I won't comment this ifdef mess. */

#ifdef SAFEUID
	setfsgid(gid);
	setfsuid(uid);
#else
	ipc_needex();
	exlock();
	block(PROCESS,&o);
	if(UNLIKELY(setresgid(gid,gid,0)))goto nogid;
	if(UNLIKELY(setresuid(uid,uid,0)))goto nouid;
#endif
	if(LIKELY(rmt.ifd!=-1))
	{
		unlink(rmt.in);
		close(rmt.ifd);
	}
	if(LIKELY(rmt.ofd!=-1))
	{
		unlink(rmt.out);
		close(rmt.ofd);
	}
	if(LIKELY(rmt.efd!=-1))
	{
		unlink(rmt.err);
		close(rmt.efd);
	}
	if(LIKELY(rmt.wrk!=NULL))
	{
		if(UNLIKELY(rmdir(rmt.wrk)))
		{
			if(rmt.out)if(!stat(rmt.out,&stb))unlink(rmt.out);
			rmdir(rmt.wrk);
		}
		free(rmt.wrk);
	}
#ifdef SAFEUID
	setfsuid(0);
	setfsgid(0);
#else
	setresuid(0,0,0);
nouid:	setresgid(0,0,0);
nogid:	unblock(PROCESS,&o);
	unlock();
	ipc_exdone();
#endif

	/* release working memory */

	if(LIKELY(rmt.in!=NULL))free(rmt.in);
	if(LIKELY(rmt.out!=NULL))free(rmt.out);
	if(LIKELY(rmt.err!=NULL))free(rmt.err);
	if(LIKELY(rmt.argv!=NULL))free(rmt.argv);
	if(LIKELY(rmt.bfr!=NULL))free(rmt.bfr);
	if(UNLIKELY(rmt.outmem!=NULL))free(rmt.outmem);

	/* remove cleanup handler */

	pthread_cleanup_pop(0);

	/* now this is interesting: we need to notify our invoking thread
	   that we're done in a safe way. Thus we must hold a mutex that
	   is held during cancellation. Then we must accept any pending
	   cancellation. If we pass this point we can inform the starting
	   thread that we're finished and then need to release the mutex. */

	pthread_mutex_lock(&mtx);
	pthread_cleanup_push(rmtfinal,NULL);
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
	pthread_testcancel();
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	pthread_cleanup_pop(0);
	n.valid=1;
	n.tid=pthread_self();
	dowrite(notify[1],&n,sizeof(n));
	pthread_mutex_unlock(&mtx);

	/* exit thread */

	pthread_exit(NULL);
}

/*
 * docancel
 *
 * description:	This routine cancels all running compiles for remote
 *		peers in a safe way.
 */

static void docancel(void)
{
	CHILD *c;
	CHILD **e;
	struct pollfd p;
	NOTIFY n;

	/* prepare to poll the compile termination pipe */

	p.fd=notify[0];
	p.events=POLLIN;

	/* acquire necessary mutexes */

	pthread_mutex_lock(&qtx);
	pthread_mutex_lock(&mtx);

	/* handle all pending compile terminations */

	while(poll(&p,1,0)==1)if(LIKELY(p.revents&POLLIN))
	{
		if(UNLIKELY(doread(notify[0],&n,sizeof(n))))continue;
		if(LIKELY(n.valid))for(e=&head;*e;)
		{
			if((*e)->tid==n.tid)
			{
				c=*e;
				*e=c->next;
				free(c);
				if(c==tail)tail=(CHILD *)e;
				total--;
				pthread_join(n.tid,NULL);
				break;
			}
			else e=&(*e)->next;
		}
	}

	/* cancel all compiles still running in a safe way (the might be
	   waiting in update_mode() for a free slot) */

	for(c=head;c;c=c->next)safe_cancel(c->tid);

	/* release the compile thread finishing mutex so that the threads
	   can cancel right now */

	pthread_mutex_unlock(&mtx);

	/* wait for all threads to complete cancellation and clean up */

	while(head)
	{
		c=head;
		head=c->next;
		pthread_join(c->tid,NULL);
		free(c);
		total--;
	}

	/* release thread queue mutex */

	pthread_mutex_unlock(&qtx);
}

/*
 * rmtworker
 *
 * input:  unused - unused
 *
 * return: always NULL
 *
 * description:	This routine is the remote compile master thread that
 *		accepts remote connections and starts remote compile
 *		threads.
 */

static void *rmtworker(void *unused)
{
	int i;
	int s;
	socklen_t l;
	struct pollfd p[3];
	struct sockaddr_in a;
	struct sockaddr_in6 a6;
	CHILD *c;
	CHILD *d;
	CHILD **e;
	NOTIFY n;
	char mem[256];

	/* default thread setup */

	dftblock(THREAD);
	dropcaps(NEED_UGID);

	/* data setup */

	p[0].fd=tcp;
	p[0].events=POLLIN;
	p[1].fd=termget();
	p[1].events=POLLIN;
	p[2].fd=notify[0];
	p[2].events=POLLIN;
	c=NULL;

	/* while the whole process is running */

	while(LIKELY(running))
	{
		/* wait for event */

		switch(poll(p,3,-1))
		{
		/* event available (fall through) */

		default:/* exit if termination is pending */

			if(UNLIKELY(p[1].revents&POLLIN))continue;

			/* ((child thread) termination) notification */

			if(p[2].revents&POLLIN)
			{
				/* done if termination pending */

				if(!running)continue;

				/* remove finished compile from queue */

				pthread_mutex_lock(&qtx);
				if(UNLIKELY(doread(notify[0],&n,sizeof(n))))
				{
					pthread_mutex_unlock(&qtx);
					continue;
				}
				if(LIKELY(n.valid))for(e=&head;*e;)
				{
					if((*e)->tid==n.tid)
					{
						DEBUG("worker thread complete");
						d=*e;
						*e=d->next;
						free(d);
						if(d==tail)tail=(CHILD *)e;
						total--;
						pthread_join(n.tid,NULL);
						break;
					}
					else e=&(*e)->next;
				}
				pthread_mutex_unlock(&qtx);
			}

			/* new peer waiting to get connected */

			if(p[0].revents&POLLIN)break;

		/* repeat wait for no event or error */

		case 0:
		case -1:continue;
		}

		/* get new child thread storage */

		if(!c)
		{
			c=malloc(sizeof(CHILD));
			if(UNLIKELY(!c))
			{
				NOTICE("malloc failure");
				usleep(20000);
				continue;
			}
			c->next=NULL;
		}

		/* accept connection from peer, do basic validy check
		   and respect the connection hold off delay set when
		   the interface changes */

		if(SWITCHES_SUPPORT_MLDV2(v6))
		{
			l=sizeof(a6);
#ifdef SAFEAPI
			if(LIKELY(safeapi()))
			{
				if(UNLIKELY((s=accept4(tcp,
					(struct sockaddr *)(&a6),&l,
					SOCK_CLOEXEC))==-1))continue;
			}
			else
#endif
			{
				shlock();
				if(UNLIKELY((s=accept(tcp,
					(struct sockaddr *)(&a6),&l))==-1))
				{
					unlock();
					continue;
				}
				fcntl(s,F_SETFD,FD_CLOEXEC);
				unlock();
			}

			if(UNLIKELY(l!=sizeof(a6)))goto fail;

			if(UNLIKELY(a6.sin6_family!=AF_INET6))goto fail;

	                pthread_mutex_lock(&ntx);
			if(UNLIKELY(delay))
			{
				if(timestamp()-base<delay)
				{
	                		pthread_mutex_unlock(&ntx);
					goto fail;
				}
				delay=0;
			}
			if(LIKELY(!site))
			{
				if(!sctp)if(UNLIKELY(a6.sin6_scope_id!=net))
				{
	                		pthread_mutex_unlock(&ntx);
					goto fail;
				}
				if(UNLIKELY(!linklocal(a6.sin6_addr)))
				{
	                		pthread_mutex_unlock(&ntx);
					goto fail;
				}
			}
	                pthread_mutex_unlock(&ntx);
		}
		else
		{
			l=sizeof(a);
#ifdef SAFEAPI
			if(LIKELY(safeapi()))
			{
				if(UNLIKELY((s=accept4(tcp,
					(struct sockaddr *)(&a),&l,
					SOCK_CLOEXEC))==-1))continue;
			}
			else
#endif
			{
				shlock();
				if(UNLIKELY((s=accept(tcp,
					(struct sockaddr *)(&a),&l))==-1))
				{
					unlock();
					continue;
				}
				fcntl(s,F_SETFD,FD_CLOEXEC);
				unlock();
			}

			if(UNLIKELY(l!=sizeof(a)))goto fail;

			if(UNLIKELY(a.sin_family!=AF_INET))goto fail;

	                pthread_mutex_lock(&ntx);
			if(UNLIKELY(delay))
			{
				if(timestamp()-base<delay)
				{
	                		pthread_mutex_unlock(&ntx);
					goto fail;
				}
				delay=0;
			}
			if(LIKELY(!site))if(UNLIKELY((
				ntohl(a.sin_addr.s_addr)&mask)!=net))
			{
	                	pthread_mutex_unlock(&ntx);
				goto fail;
			}
	                pthread_mutex_unlock(&ntx);
		}

		/* make peer socket blocking */

		i=0;
		if(UNLIKELY(ioctl(s,FIONBIO,&i)))
			JUMP(fail,"ioctl(FIONBIO): %s",errstr(errno,mem));

		/* no TCP keepalive setup for SCTP */

		if(sctp)goto skip;

#ifdef TCP_KEEPCNT
		i=3;
		if(UNLIKELY(setsockopt(s,IPPROTO_TCP,TCP_KEEPCNT,&i,sizeof(i))))
			JUMP(fail,"setsockopt(TCP_KEEPCNT): %s",
				errstr(errno,mem));
#endif

#ifdef TCP_KEEPIDLE
		i=30;
		if(UNLIKELY(setsockopt(s,IPPROTO_TCP,TCP_KEEPIDLE,&i,
			sizeof(i))))
				JUMP(fail,"setsockopt(TCP_KEEPIDLE): %s",
					errstr(errno,mem));
#endif

#ifdef TCP_KEEPINTVL
		i=15;
		if(UNLIKELY(setsockopt(s,IPPROTO_TCP,TCP_KEEPINTVL,&i,
			sizeof(i))))
				JUMP(fail,"setsockopt(TCP_KEEPINTVL): %s",
					errstr(errno,mem));
#endif

#if !defined(TCP_KEEPCNT) || !defined(TCP_KEEPIDLE) || !defined(TCP_KEEPINTVL)
#warning only crippled keepalive support available
#endif

		i=1;
		if(UNLIKELY(setsockopt(s,SOL_SOCKET,SO_KEEPALIVE,&i,sizeof(i))))
			JUMP(fail,"setsockopt(SO_KEEPALIVE): %s",
				errstr(errno,mem));

		/* create child thread handling the remote compile */

skip:		if(UNLIKELY(pthread_create(&c->tid,NULL,rmtcompile,
			(void *)((long)s))))
		{
			NOTICE("pthread_create: %s",errstr(errno,mem));
fail:			close(s);
			continue;
		}

		DEBUG("worker thread started");

		/* enqueue child thread */

		pthread_mutex_lock(&qtx);
		if(!head)head=tail=c;
		else
		{
			tail->next=c;
			tail=c;
		}
		total++;
		pthread_mutex_unlock(&qtx);
		c=NULL;
	}

	/* release allocated memory */

	if(c)free(c);

	/* cancel all remote compiles still running */

	docancel();

	/* exit thread */

	pthread_exit(NULL);
}

/*
 * remotemgr
 *
 * input:  mode - the action to be processed
 *	   ptr  - action dependent data
 *	   val  - action dependent data
 *	   ...  - more action dependent data
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine is the remote compile module management routine.
 */

int remotemgr(int mode,char *ptr,int val,...)
{
	int newnet;
	int newself;
	int newmask;
	int backlog;
	int nval;
	char *p;
	char *save;
	char *tmpdir;
	char *user;
	char *group;
	struct passwd *pw;
	struct group *gr;
	NOTIFY n;
#if !defined(__OpenBSD__)
	struct passwd pwrk;
	struct group gwrk;
#endif
	va_list ap;
	char mem[1024];

	switch(mode)
	{
	/* start remote compile service */

	case RMT_START:

		/* sanity check */

		if(UNLIKELY(running))return 0;

		/* get additional argumnets */

                va_start(ap,val);
                backlog=va_arg(ap,int);
                tmpdir=va_arg(ap,char *);
                nval=va_arg(ap,int);
                user=va_arg(ap,char *);
                group=va_arg(ap,char *);
		va_end(ap);

		/* argument sanity check */

		if(UNLIKELY(!ptr||!tmpdir||!user||!group))
			BAD("internal error");
		if(UNLIKELY(val<1||val>65535))BAD("internal error");
		if(UNLIKELY(backlog<1))BAD("internal error");
		if(UNLIKELY(nval<0||nval>20))BAD("internal error");

		/* oh, why don't they just call it BrokenBSD... */

#if defined(__OpenBSD__)
#warning broken getpwnam_r and getgrnam_r, fix this system, please
		ipc_needex();
		exlock();
		if(UNLIKELY(!(pw=getpwnam(user))))
		{
			unlock();
			ipc_exdone();
			BAD("%s: no such user",user);
		}
		uid=pw->pw_uid;
		if(UNLIKELY(!(gr=getgrnam(group))))
		{
			unlock();
			ipc_exdone();
			BAD("%s: no such group",group);
		}
		gid=gr->gr_gid;
		unlock();
		ipc_exdone();
#else
		getpwnam_r(user,&pwrk,mem,sizeof(mem),&pw);
		if(UNLIKELY(!pw))BAD("%s: no such user",user);
		uid=pw->pw_uid;
		getgrnam_r(group,&gwrk,mem,sizeof(mem),&gr);
		if(UNLIKELY(!gr))BAD("%s: no such group",group);
		gid=gr->gr_gid;
#endif
		/* data setup */

		swtch1=getswitch1();
		swtch2=getswitch2();
		tmp=tmpdir;
		niceval=nval;
		head=NULL;
		tail=NULL;
		total=0;
		delay=0;
		nosplice=1;
#ifdef SAFEAPI
		if(LIKELY(safeapi()))if(!sctp&&zerocopy)nosplice=0;
#endif

		/* create listening socket */

		if(SWITCHES_SUPPORT_MLDV2(v6))
		{
                	if(UNLIKELY((net=getifindex(ptr))==-1))
				BAD("can't get interface index of %s",ptr);
			if(sctp)
			{
#if defined(__linux__)
				p=ptr;
#else
				p=NULL;
#endif
				if(sctpauth)
				{
					if(UNLIKELY((tcp=sctp6socket(p,val,
						backlog,hash,sizeof(hash),
						sctpauth>>1))==-1))BAD(
							"can't create "
							"listening socket");
				}
				else if(UNLIKELY((tcp=sctp6socket(p,val,
					backlog,NULL,0,0))==-1))
					BAD("can't create listening socket");
			}
			else if(UNLIKELY((tcp=tcp6socket(ptr,val,backlog))
				==-1))
				BAD("can't create listening socket");
		}
		else
		{
                	if(UNLIKELY((self=getifcaddr(ptr))==-1))
				BAD("can't get ip of %s",ptr);
                	if(UNLIKELY((mask=getnetmask(ptr))==-1))
				BAD("can't get netmask of %s",ptr);
			net=self&mask;
			if(sctp)
			{
#if defined(__linux__)
				p=ptr;
#else
				p=NULL;
#endif
				if(sctpauth)
				{
					if(UNLIKELY((tcp=sctpsocket(p,val,
						backlog,hash,sizeof(hash),
						sctpauth>>1))==-1))BAD(
							"can't create "
							"listening socket");
				}
				else if(UNLIKELY((tcp=sctpsocket(p,val,
					backlog,NULL,0,0))==-1))
					BAD("can't create listening socket");
			}
			else if(UNLIKELY((tcp=tcpsocket(NULL,val,backlog))
				==-1))BAD("can't create listening socket");
		}

		/* create thread termination notification pipe */

#ifdef SAFEAPI
		if(LIKELY(safeapi()))
		{
			if(UNLIKELY(pipe2(notify,O_CLOEXEC)))
			{
				newself=errno;
				close(tcp);
				BAD("pipe2: %s",errstr(newself,mem));
			}
		}
		else
#endif
		{
			shlock();
			if(UNLIKELY(pipe(notify)))
			{
				newself=errno;
				close(tcp);
				unlock();
				BAD("pipe: %s",errstr(newself,mem));
			}
			fcntl(notify[0],F_SETFD,FD_CLOEXEC);
			fcntl(notify[1],F_SETFD,FD_CLOEXEC);
			unlock();
		}

		/* start listening thread */

		running=1;
		if(UNLIKELY(pthread_create(&h,NULL,rmtworker,NULL)))
		{
			newself=errno;
			running=0;
			close(tcp);
			close(notify[0]);
			close(notify[1]);
			BAD("pthread_create: %s",errstr(newself,mem));
		}

		/* signal success */

		return 0;

	/* stop remote compile service */

	case RMT_STOP:

		/* sanity check */

		if(UNLIKELY(!running))return 0;

		/* mark termination as pending */

		running=0;

		/* wake up listening thread */

		n.valid=0;
		dowrite(notify[1],&n,sizeof(n));

		/* wait for listening thread termination */

		pthread_join(h,NULL);

		/* clean up */

		close(tcp);
		close(notify[0]);
		close(notify[1]);
		clrauth(&key,hash);

		/* signal success */

		return 0;

	/* network config change */

	case RMT_NETWORK:

		/* sanity check */

                if(UNLIKELY(!running))BAD("internal error");

		/* get new network data, update in case of change */

		if(SWITCHES_SUPPORT_MLDV2(v6))
		{
                	if(UNLIKELY((newnet=getifindex(ptr))==-1))
				BAD("can't get interface index of %s",ptr);
                	pthread_mutex_lock(&ntx);
			if(net!=newnet)
			{
				net=newnet;
				docancel();
#if defined(__linux__)
				if(sctp)if(UNLIKELY(sctp6ifset(tcp,ptr)))
					ERR("can't ajust address list");
#endif
				n.valid=0;
				dowrite(notify[1],&n,sizeof(n));
			}
                	pthread_mutex_unlock(&ntx);
		}
		else
		{
                	if(UNLIKELY((newself=getifcaddr(ptr))==-1))
				BAD("can't get ip of %s",ptr);
                	if(UNLIKELY((newmask=getnetmask(ptr))==-1))
				BAD("can't get netmask of %s",ptr);
                	pthread_mutex_lock(&ntx);
                	if(self!=newself||mask!=newmask)
                	{
                	        self=newself;
                	        mask=newmask;
				net=self&mask;
				docancel();
#if defined(__linux__)
				if(sctp)if(UNLIKELY(sctpifset(tcp,ptr)))
					ERR("can't ajust address list");
#endif
				n.valid=0;
				dowrite(notify[1],&n,sizeof(n));
			}
                	pthread_mutex_unlock(&ntx);
		}

		/* signal success */

		return 0;

	/* change nice value */

	case RMT_NICE:
		if(UNLIKELY(val<0||val>20))BAD("internal error");
		shlock();
		niceval=val;
		unlock();
		return 0;

	/* IPv6 mode selection */

	case RMT_USEV6:
                if(UNLIKELY(running))BAD("internal error");
		if(UNLIKELY(val<0||val>1))BAD("internal error");
		v6=val;
		return 0;

	/* SCTP mode selection */

	case RMT_SCTP:
                if(UNLIKELY(running))BAD("internal error");
		if(UNLIKELY(val<0||val>1))BAD("internal error");
		sctp=val;
		return 0;

	/* sitewide mode selection */

	case RMT_SITE:
                if(UNLIKELY(running))BAD("internal error");
		if(UNLIKELY(val<0||val>1))BAD("internal error");
		site=val;
		return 0;

	/* authentication setup */

	case RMT_AUTH:
		if(UNLIKELY(running))BAD("internal error");
		if(UNLIKELY(val<1))BAD("internal error");
		setauth(&key,hash,ptr,val);
		auth=1;
		return 0;

	/* authentication time window setup */

	case RMT_WINDOW:
                if(UNLIKELY(running))BAD("internal error");
		if(UNLIKELY(val<0||val>300))BAD("internal error");
		window=val;
		return 0;

	/* kill all running remote compile jobs */

	case RMT_KILL:
                if(UNLIKELY(!running))BAD("internal error");
               	pthread_mutex_lock(&ntx);
		docancel();
		base=timestamp();
		delay=sysconf(_SC_CLK_TCK);
		n.valid=0;
		dowrite(notify[1],&n,sizeof(n));
               	pthread_mutex_unlock(&ntx);
		return 0;

	/* unique id setup */

	case RMT_UNIQUE:
                if(UNLIKELY(running))BAD("internal error");
		unique=val;
		return 0;

	/* allowed tools list */

	case RMT_TOOLS:
                if(UNLIKELY(running))BAD("internal error");
		if(UNLIKELY(!ptr))BAD("internal error");
		if(UNLIKELY(!(ptr=strdup(ptr))))BAD("malloc failure");
		for(val=0,p=strtok_r(ptr,":",&save);p;
			p=strtok_r(NULL,":",&save))
		{
			if(UNLIKELY(!*p))continue;
			if(UNLIKELY(!(toollist[val]=strdup(p))))
			{
				free(ptr);
				BAD("malloc failure");
			}
			if(val++==MAXTOOLS)BAD("too many allowed tools");
		}
		toollist[val]=NULL;
		free(ptr);
		return 0;

	/* sctp authentication */

	case RMT_SCTPAUTH:
                if(UNLIKELY(running))BAD("internal error");
		sctpauth=val;
		return 0;

	/* zerocopy file receive authentication */

	case RMT_ZCOPY:
                if(UNLIKELY(running))BAD("internal error");
		zerocopy=val;
		return 0;

	/* sanity check */

	default:BAD("internal error");
	}
}
