
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "sctp6.h"
#include "log.h"
#include "ifc.h"
#include "lock.h"
#include "safeapi.h"
#include "sctp_auth.h"

#ifdef SCTP_AVAILABLE

#if defined(__linux__)
#define PROTO SOL_SCTP
#elif defined(__FreeBSD__)
#define PROTO IPPROTO_SCTP
#else
#error do not know how to set SCTP protocol for getsockopt()/setsockopt()
#endif

/*
 * getaddrlist
 *
 * input:  dev - the device to be processed
 *
 * output: list - the list of IPv6 addresses of the device
 *
 * return: the amount of list entries or -1 in case of error
 *
 * description:	This routine creates an array of all IPv6 addresses
 *		assigned to the specified device.
 */

static int getaddrlist(char *dev,struct sockaddr_in6 **list)
{
	int total=0;
	int alloc=0;
	int x;
	int idx;
	struct ifaddrs *p;
	struct ifaddrs *e;
	struct sockaddr_in6 *l=NULL;
	struct sockaddr_in6 *tmp;
	char mem[256];

	/* get device index of specified device */

	if(UNLIKELY((idx=getifindex(dev))==-1))return -1;

	/* get all current interface data */

	if(UNLIKELY(getifaddrs(&p)))BAD("getifaddrs: %s",errstr(errno,mem));

	/* forr all IPv6 addresses of the specified device */

	for(e=p;e;e=e->ifa_next)if(!strcmp(dev,e->ifa_name))if(e->ifa_addr)
		if(e->ifa_addr->sa_family==AF_INET6)
	{
		/* enlarge array if necessary */

		if(total==alloc)
		{
			alloc+=4;
			if(UNLIKELY(!(tmp=realloc(l,
				alloc*sizeof(struct sockaddr_in6)))))
			{
				x=errno;
				if(l)free(l);
				freeifaddrs(p);
				BAD("realloc: %s",errstr(x,mem));
			}
			else l=tmp;
		}

		/* set up IPv6 address structure */

		memset(&l[total],0,sizeof(struct sockaddr_in6));
		l[total].sin6_family=AF_INET6;
		l[total].sin6_scope_id=idx;
		l[total++].sin6_addr=
			((struct sockaddr_in6 *)e->ifa_addr)->sin6_addr;
	}

	/* release interface data memory */

	freeifaddrs(p);

	/* return list to caller */

	*list=l;

	/* return entries in list */

	return total;
}

/*
 * getladdrlist
 *
 * input:  s - the listening socket to be processed
 *
 * output: list - the list of IPv6 addresses the socket is bound to
 *
 * return: the mount of list entries or -1 in case of an error
 *
 * description:	This routine returns all IPv6 addresses bound to a listening
 *		socket.
 */

static int getladdrlist(int s,struct sockaddr_in6 **list)
{
	int total=0;
	int alloc=0;
	int num;
	int x;
	char *ptr;
	struct sockaddr *la;
	struct sockaddr_in6 *l=NULL;
	struct sockaddr_in6 *tmp;
	char mem[256];

	/* get complete list of bound addresses */

	if(UNLIKELY((num=sctp_getladdrs(s,0,&la))==-1))
		BAD("sctp_getladdrs: %s",errstr(errno,mem));

	/* process packed list (aiyeeehh, a *PACKED* list!!!) */

	for(ptr=(char *)la,x=0;x<num;x++)
		switch(((struct sockaddr *)ptr)->sa_family)
	{
	/* IPv6 address */

	case AF_INET6:
		/* assert sufficient space in result array */

		if(total==alloc)
		{
			alloc+=4;
			if(UNLIKELY(!(tmp=realloc(l,
				alloc*sizeof(struct sockaddr_in6)))))
			{
				x=errno;
				if(l)free(l);
				sctp_freeladdrs(la);
				BAD("realloc: %s",errstr(x,mem));
			}
			else l=tmp;
		}

		/* copy entry to return array */

		memcpy(&l[total++],ptr,sizeof(struct sockaddr_in6));

		/* skip processed entry */

		ptr+=sizeof(struct sockaddr_in6);
		break;

	/* IPv4 address, just skip over  */

	case AF_INET:
		ptr+=sizeof(struct sockaddr_in);
		break;

	/* unknown address type of unknown size */

	default:if(l)free(l);
		sctp_freeladdrs(la);
		BAD("unknown family %d in list",
			((struct sockaddr *)ptr)->sa_family);
	}

	/* free darn packed array */

	sctp_freeladdrs(la);

	/* return list to caller */

	*list=l;

	/* return amount of list entries */

	return total;
}

/*
 * sctp6socket
 *
 * input:  dev     - the device to listen on or NULL
 *	   port    - the listening port
 *	   backlog - the listen backlog
 *	   key     - sctp hmac authentication key or NULL
 *	   keylen  - sctp hmac authentication key length or 0
 *	   kernel_bug_workaround - do not authenticate COOKIE ACK if 1
 *
 * return: a socket descriptor in case of success or -1 in case of error
 *
 * description:	This routine creates an IPv6 SCTP listening socket.
 */

int sctp6socket(char *dev,int port,int backlog,void *key,int keylen,
	int kernel_bug_workaround)
{
	int s;
	int x;
	int idx;
	int total;
	socklen_t l;
	struct sockaddr_in6 a;
	struct sockaddr_in6 *list;
	struct sctp_rtoinfo rto;
	struct sctp_paddrparams ap;
	char mem[256];

	/* sanity check */

	if(UNLIKELY(port<1||port>65535))BAD("internal error");

	/* get interface index if bound to device */

	if(dev)
	{
		if(UNLIKELY((idx=getifindex(dev))==-1))return -1;
		if(UNLIKELY((total=getaddrlist(dev,&list))==-1))return -1;
	}
	else
	{
		total=0;
		idx=0;
		list=NULL;
	}

	/* create socket */

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((s=socket(AF_INET6,SOCK_STREAM|SOCK_CLOEXEC,
			IPPROTO_SCTP)))==-1)BAD("socket: %s",errstr(errno,mem));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((s=socket(AF_INET6,SOCK_STREAM,IPPROTO_SCTP)))==-1)
		{
			x=errno;
			unlock();
			BAD("socket: %s",errstr(x,mem));
		}
		fcntl(s,F_SETFD,FD_CLOEXEC);
		unlock();
	}

	/* make socket reusable */

	x=1;
	if(UNLIKELY(setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&x,sizeof(x))))
	{
		x=errno;
		close(s);
		BAD("setsockopt(SO_REUSEADDR): %s",errstr(x,mem));
	}

	/* set up retransmission timeouts */

	l=sizeof(rto);
	memset(&rto,0,sizeof(rto));
	if(UNLIKELY(getsockopt(s,PROTO,SCTP_RTOINFO,&rto,&l)))
	{
		x=errno;
		close(s);
		BAD("getsockopt(SCTP_RTOINFO): %s",errstr(x,mem));
	}

	rto.srto_initial=3000;
	rto.srto_min=1000;
	rto.srto_max=10000;

	if(UNLIKELY(setsockopt(s,PROTO,SCTP_RTOINFO,&rto,sizeof(rto))))
	{
		x=errno;
		close(s);
		BAD("setsockopt(SCTP_RTOINFO): %s",errstr(x,mem));
	}

	/* set up heartbeat */

	l=sizeof(ap);
	memset(&ap,0,sizeof(ap));
	if(UNLIKELY(getsockopt(s,PROTO,SCTP_PEER_ADDR_PARAMS,&ap,&l)))
	{
		x=errno;
		close(s);
		BAD("getsockopt(SCTP_PEER_ADDR_PARAMS): %s",errstr(x,mem));
	}

	ap.spp_hbinterval=20000;
	ap.spp_pathmaxrxt=5;
#ifdef SPP_SACKDELAY_ENABLE
	ap.spp_sackdelay=200;
	ap.spp_flags|=SPP_HB_ENABLE|SPP_SACKDELAY_ENABLE;
	ap.spp_flags&=~SPP_HB_DISABLE|SPP_HB_DEMAND|SPP_SACKDELAY_DISABLE|
		SPP_HB_TIME_IS_ZERO;
#else
	ap.spp_flags|=SPP_HB_ENABLE;
	ap.spp_flags&=~SPP_HB_DISABLE|SPP_HB_DEMAND|SPP_HB_TIME_IS_ZERO;
#endif

	if(UNLIKELY(setsockopt(s,PROTO,SCTP_PEER_ADDR_PARAMS,&ap,
		sizeof(ap))))
	{
		x=errno;
		close(s);
		BAD("setsockopt(SCTP_PEER_ADDR_PARAMS): %s",errstr(x,mem));
	}

	/* enforce IPv6 mode if bound to device */

	if(dev)x=1;
	else x=0;
	if(UNLIKELY(setsockopt(s,IPPROTO_IPV6,IPV6_V6ONLY,&x,sizeof(x))))
	{
		x=errno;
		close(s);
		BAD("setsockopt(IPV6_V6ONLY): %s",errstr(x,mem));
	}

	/* make socket nonblocking */

	x=1;
	if(UNLIKELY(ioctl(s,FIONBIO,&x)))
	{
		x=errno;
		close(s);
		BAD("ioctl(FIONBIO): %s",errstr(x,mem));
	}

	/* set up transport authentication */

	if(key&&keylen)if(sctp_server_auth(s,key,keylen,kernel_bug_workaround))
	{
		close(s);
		BAD("sctp authentication setup failure");
	}

	/* bind socket */

	if(!total)
	{
		memset(&a,0,sizeof(a));
		a.sin6_family=AF_INET6;
		a.sin6_addr=in6addr_any;
		a.sin6_port=htons(port);
		a.sin6_scope_id=idx;

#ifndef SAFEUID
		shlock();
#endif
		if(UNLIKELY(bind(s,(struct sockaddr *)(&a),sizeof(a))))
		{
			x=errno;
#ifndef SAFEUID
			unlock();
#endif
			close(s);
			BAD("bind: %s",errstr(x,mem));
		}
#ifndef SAFEUID
		unlock();
#endif
	}
	else for(x=0;x<total;x++)
	{
		memset(&a,0,sizeof(a));
		a.sin6_family=AF_INET6;
		a.sin6_addr=list[x].sin6_addr;
		a.sin6_port=htons(port);
		a.sin6_scope_id=idx;

#ifndef SAFEUID
		shlock();
#endif
		if(UNLIKELY(sctp_bindx(s,(struct sockaddr *)(&a),1,
			SCTP_BINDX_ADD_ADDR)))
		{
			x=errno;
#ifndef SAFEUID
			unlock();
#endif
			close(s);
			free(list);
			BAD("sctp_bindx SCTP_BINDX_ADD_ADDR: %s",errstr(x,mem));
		}
#ifndef SAFEUID
		unlock();
#endif
	}

	if(list)free(list);

	/* make socket listen */

	if(UNLIKELY(listen(s,backlog)))
	{
		x=errno;
		close(s);
		BAD("listen: %s",errstr(x,mem));
	}

	/* return socket descriptor */

	return s;
}

/*
 * sctp6ifset
 *
 * input:  fd  - the listening socket descriptor
 *	   dev - the device to listen on
 *
 * return: 0 in case of success or -1 in case of error
 *
 * description: This routine changes the addresses of an IPv6 SCTP
 *		listening socket.
 */

int sctp6ifset(int fd,char *dev)
{
	int num;
	int total;
	int i;
	int x;
	struct sockaddr_in6 *newlist;
	struct sockaddr_in6 *list;
	char mem[256];

	/* get list of new IPv6 addresses */

	if(UNLIKELY((total=getaddrlist(dev,&newlist))==-1))return -1;
	if(UNLIKELY(!total))BAD("sctp6ifset with no new address to switch to");

	/* get list of bound IPv6 addresses */

	if(UNLIKELY((num=getladdrlist(fd,&list))==-1))
	{
		free(newlist);
		return -1;
	}

	/* remove identical entries from both lists */

	for(i=0;i<num;i++)if(LIKELY(list[i].sin6_family==AF_INET6))
		for(x=0;x<total;x++)if(LIKELY(newlist[x].sin6_family==AF_INET6))
			if(UNLIKELY(!memcmp(&list[i].sin6_addr,
				&newlist[x].sin6_addr,16)))
	{
		newlist[x].sin6_family=0;
		list[i].sin6_family=0;
		break;
	}

	/* add all remaining new addresses to the socket */

	for(x=0;x<total;x++)if(LIKELY(newlist[x].sin6_family==AF_INET6))
	{
#ifndef SAFEUID
		shlock();
#endif
		if(UNLIKELY(sctp_bindx(fd,(struct sockaddr *)&newlist[x],1,
			SCTP_BINDX_ADD_ADDR)))
		{
			x=errno;
#ifndef SAFEUID
			unlock();
#endif
			free(newlist);
			if(list)free(list);
			BAD("sctp_bindx SCTP_BINDX_ADD_ADDR: %s",errstr(x,mem));
		}
#ifndef SAFEUID
		unlock();
#endif
	}

	/* remove all old bound addresses from the socket */

	for(i=0;i<num;i++)if(LIKELY(list[i].sin6_family==AF_INET6))
	{
#ifndef SAFEUID
		shlock();
#endif
		if(UNLIKELY(sctp_bindx(fd,(struct sockaddr *)&list[i],1,
			SCTP_BINDX_REM_ADDR)))
		{
			x=errno;
#ifndef SAFEUID
			unlock();
#endif
			free(newlist);
			if(list)free(list);
			BAD("sctp_bindx SCTP_BINDX_REM_ADDR: %s",errstr(x,mem));
		}
#ifndef SAFEUID
		unlock();
#endif
	}

	/* release allocated memory */

	free(newlist);
	if(list)free(list);

	/* signal success */

	return 0;
}


#else

int sctp6socket(char *dev,int port,int backlog,void *key,int keylen,
	int kernel_bug_workaround)
{
	BAD("sctp not available on this system");
}

int sctp6ifset(int fd,char *dev)
{
	BAD("sctp not available on this system");
}

#endif
