
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#ifndef SCTP_AUTHCODE_H
#define SCTP_AUTHCODE_H

#ifdef SCTP_AVAILABLE

#if defined(__linux__)
#define PROTO SOL_SCTP
#elif defined(__FreeBSD__)
#define PROTO IPPROTO_SCTP
#else
#error do not know how to set SCTP protocol for getsockopt()/setsockopt()
#endif

/* SCTP chunks to be authenticated */

static const uint8_t const authchunks[]=
{
	0,3,4,5,7,8,9,10,11,12,13
};

static const uint8_t const authchunks_buggy_kernel[]=
{
	0,3,4,5,7,9,10,12,13
};

#ifdef errstr
#define SCTP_SET_AUTH sctp_server_auth
#else
#define SCTP_SET_AUTH sctp_client_auth
#endif

/*
 * sctp_server_auth
 * sctp_client_auth
 *
 * input:  s   - the sctp socket handle
 *	   key - the authentication key
 *	   len - the key length (> 0 && < 64k)
 *	   kernel_bug_workaround - do not authenticate COOKIE ACK if 1
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine sets up SCTP authenticatrion.
 */

int SCTP_SET_AUTH(int s,void *key,int len,int kernel_bug_workaround)
{
	int i;
	struct sctp_authkey *k;
	struct sctp_authkeyid id;
	struct sctp_authchunk a;
#ifdef errstr
	char mem[256];
#define ERRTXT(a) errstr(a,mem)
#else
#define ERRTXT(a) strerror(a)
#endif
	/* sanity checks */

	if(!key||!len||len>0xffff)BAD("internal error");

	/* set up key */

	if(UNLIKELY((k=malloc(sizeof(struct sctp_authkey)+len))==NULL))
		BAD("malloc failure");

	memset(k,0,sizeof(struct sctp_authkey));
	memcpy(k->sca_key,key,len);
	k->sca_keynumber=1;
#if defined(__linux__)
	k->sca_keylength=len;
#endif

	if(UNLIKELY(setsockopt(s,PROTO,SCTP_AUTH_KEY,
		k,sizeof(struct sctp_authkey)+len)))
	{
		i=errno;
		memset(k->sca_key,0,len);
		free(k);
		BAD("setsockopt(SCTP_AUTH_KEY): %s",ERRTXT(i));
	}

	memset(k->sca_key,0,len);
	free(k);

	/* set key as active key */

	memset(&id,0,sizeof(id));
	id.scact_keynumber=1;
	if(UNLIKELY(setsockopt(s,PROTO,SCTP_AUTH_ACTIVE_KEY,
		&id,sizeof(struct sctp_authkeyid))))
		BAD("setsockopt(SCTP_AUTH_ACTIVE_KEY): %s",ERRTXT(errno));

	/* select chunks to be authenticated (ABORT is a bad idea to
	   authenticate) */

	if(kernel_bug_workaround)for(i=0;i<sizeof(authchunks_buggy_kernel);i++)
	{
		a.sauth_chunk=authchunks_buggy_kernel[i];
		if(UNLIKELY(setsockopt(s,PROTO,SCTP_AUTH_CHUNK,
			&a,sizeof(struct sctp_authchunk))))
			BAD("setsockopt(SCTP_AUTH_CHUNK): %s",ERRTXT(errno));
	}
	else for(i=0;i<sizeof(authchunks);i++)
	{
		a.sauth_chunk=authchunks[i];
		if(UNLIKELY(setsockopt(s,PROTO,SCTP_AUTH_CHUNK,
			&a,sizeof(struct sctp_authchunk))))
			BAD("setsockopt(SCTP_AUTH_CHUNK): %s",ERRTXT(errno));
	}

	/* signal success */

	return 0;
}

#endif

#endif
