
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "clntlog.h"
#include "sctp.h"
#include "sctp_auth.h"

#ifdef SCTP_AVAILABLE

#if defined(__linux__)
#define PROTO SOL_SCTP
#elif defined(__FreeBSD__)
#define PROTO IPPROTO_SCTP
#else
#error do not know how to set SCTP protocol for getsockopt()/setsockopt()
#endif

/*
 * sctpconnect
 *
 * input:  addr   - target address
 *	   port   - target port
 *	   ttl    - IPv4 ttl, must be 1 for local LAN mode
 *	   key    - sctp hmac authentication key or NULL
 *	   keylen - sctp hmac authentication key length or 0
 *	   kernel_bug_workaround - do not authenticate COOKIE ACK if 1
 *
 * return: a socket descriptor or -1 in case of error
 *
 * description: This routine connects to a remote system using IPv4 and SCTP.
 */

int sctpconnect(int addr,int port,int ttl,void *key,int keylen,
	int kernel_bug_workaround)
{
	int s;
	int x;
	socklen_t l;
	struct sockaddr_in a;
	struct sctp_rtoinfo rto;
	struct sctp_paddrparams ap;

	/* sanity check */

	if(UNLIKELY(port<1||port>65535))BAD("internal error");

	/* create socket */

	if(UNLIKELY((s=socket(AF_INET,SOCK_STREAM,IPPROTO_SCTP))==-1))
		BAD("socket: %s",strerror(errno));
	fcntl(s,F_SETFD,FD_CLOEXEC);

	/* make socket blocking */

	x=0;
	if(UNLIKELY(ioctl(s,FIONBIO,&x)))
		BADJUMP(fail,"ioctl(FIONBIO): %s",strerror(errno));

	/* set up retransmission timeouts */

	l=sizeof(rto);
	memset(&rto,0,sizeof(rto));
	if(UNLIKELY(getsockopt(s,PROTO,SCTP_RTOINFO,&rto,&l)))
		BADJUMP(fail,"getsockopt(SCTP_RTOINFO): %s",strerror(errno));

	rto.srto_initial=3000;
	rto.srto_min=1000;
	rto.srto_max=10000;

	if(UNLIKELY(setsockopt(s,PROTO,SCTP_RTOINFO,&rto,sizeof(rto))))
		BADJUMP(fail,"setsockopt(SCTP_RTOINFO): %s",strerror(errno));

	/* set up heartbeat */

	l=sizeof(ap);
	memset(&ap,0,sizeof(ap));
	if(UNLIKELY(getsockopt(s,PROTO,SCTP_PEER_ADDR_PARAMS,&ap,&l)))
		BADJUMP(fail,"getsockopt(SCTP_PEER_ADDR_PARAMS): %s",
			strerror(errno));

	ap.spp_hbinterval=20000;
	ap.spp_pathmaxrxt=5;
#ifdef SPP_SACKDELAY_ENABLE
	ap.spp_sackdelay=200;
	ap.spp_flags|=SPP_HB_ENABLE|SPP_SACKDELAY_ENABLE;
	ap.spp_flags&=~SPP_HB_DISABLE|SPP_HB_DEMAND|SPP_SACKDELAY_DISABLE|
		SPP_HB_TIME_IS_ZERO;
#else
	ap.spp_flags|=SPP_HB_ENABLE;
	ap.spp_flags&=~SPP_HB_DISABLE|SPP_HB_DEMAND|SPP_HB_TIME_IS_ZERO;
#endif

	if(UNLIKELY(setsockopt(s,PROTO,SCTP_PEER_ADDR_PARAMS,&ap,
		sizeof(ap))))
		BADJUMP(fail,"setsockopt(SCTP_PEER_ADDR_PARAMS): %s",
			strerror(errno));

	/* set up ttl */

	x=ttl;
	if(UNLIKELY(setsockopt(s,IPPROTO_IP,IP_TTL,&x,sizeof(x))))
		BADJUMP(fail,"setsockopt(IP_TTL): %s",strerror(errno));

	/* set up transport authentication */

	if(key&&keylen)if(sctp_client_auth(s,key,keylen,kernel_bug_workaround))
	{
		close(s);
		BAD("sctp authentication setup failure");
	}

	/* connect to remote */

	memset(&a,0,sizeof(a));
	a.sin_family=AF_INET;
	a.sin_addr.s_addr=htonl(addr);
	a.sin_port=htons(port);

	if(UNLIKELY(connect(s,(struct sockaddr *)(&a),sizeof(a))))
	{
		BADMSG("connect: %s",strerror(errno));
fail:		close(s);
		return -1;
	}

	/* return socket descriptor */

	return s;
}

#else

int sctpconnect(int addr,int port,int ttl,void *key,int keylen,
	int kernel_bug_workaround)
{
	BADMSG("sctp not available on this system");
	return -1;
}

#endif
