
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "sctp_om.h"
#include "log.h"
#include "ifc.h"
#include "lock.h"
#include "safeapi.h"
#include "sctp_auth.h"

#ifdef SCTP_AVAILABLE

#if defined(__linux__)
#define PROTO SOL_SCTP
#elif defined(__FreeBSD__)
#define PROTO IPPROTO_SCTP
#else
#error do not know how to set SCTP protocol for getsockopt()/setsockopt()
#endif

/*
 * sctp_omsocket
 *
 * input:  dev     - the device to listen on or NULL
 *	   port    - the listening port
 *	   backlog - the listen backlog
 *	   key     - sctp hmac authentication key or NULL
 *	   keylen  - sctp hmac authentication key length or 0
 *	   kernel_bug_workaround - do not authenticate COOKIE ACK if 1
 *
 * return: a socket descriptor in case of success or -1 in case of error
 *
 * description: This routine creates an IPv4 SCTP one to many listening socket.
 */

int sctp_omsocket(char *dev,int port,int backlog,void *key,int keylen,
	int kernel_bug_workaround)
{
	int addr;
	int s;
	int x;
	socklen_t l;
	struct sockaddr_in a;
	struct sctp_rtoinfo rto;
	struct sctp_paddrparams ap;
	char mem[256];

	/* sanity check */

	if(UNLIKELY(port<1||port>65535))BAD("internal error");

	/* get interface address if bound to device */

	if(LIKELY(!dev))addr=0;
	else if(UNLIKELY((addr=getifcaddr(dev))==-1))return -1;

	/* create socket */

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((s=socket(AF_INET,SOCK_SEQPACKET|SOCK_CLOEXEC,
			IPPROTO_SCTP)))==-1)BAD("socket: %s",errstr(errno,mem));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((s=socket(AF_INET,SOCK_SEQPACKET,IPPROTO_SCTP)))
			==-1)
		{
			x=errno;
			unlock();
			BAD("socket: %s",errstr(x,mem));
		}
		fcntl(s,F_SETFD,FD_CLOEXEC);
		unlock();
	}

	/* make socket reusable */

	x=1;
	if(UNLIKELY(setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&x,sizeof(x))))
	{
		x=errno;
		close(s);
		BAD("setsockopt(SO_REUSEADDR): %s",errstr(x,mem));
	}

	/* make socket nonblocking */

	x=1;
	if(UNLIKELY(ioctl(s,FIONBIO,&x)))
	{
		x=errno;
		close(s);
		BAD("ioctl(FIONBIO): %s",errstr(x,mem));
	}

	/* set up retransmission timeouts */

	l=sizeof(rto);
	memset(&rto,0,sizeof(rto));
	if(UNLIKELY(getsockopt(s,PROTO,SCTP_RTOINFO,&rto,&l)))
	{
		x=errno;
		close(s);
		BAD("getsockopt(SCTP_RTOINFO): %s",errstr(x,mem));
	}

	rto.srto_initial=300;
	rto.srto_min=100;
	rto.srto_max=1000;

	if(UNLIKELY(setsockopt(s,PROTO,SCTP_RTOINFO,&rto,sizeof(rto))))
	{
		x=errno;
		close(s);
		BAD("setsockopt(SCTP_RTOINFO): %s",errstr(x,mem));
	}

	/* set up heartbeat */

	l=sizeof(ap);
	memset(&ap,0,sizeof(ap));
	if(UNLIKELY(getsockopt(s,PROTO,SCTP_PEER_ADDR_PARAMS,&ap,&l)))
	{
		x=errno;
		close(s);
		BAD("getsockopt(SCTP_PEER_ADDR_PARAMS): %s",errstr(x,mem));
	}

	ap.spp_hbinterval=2000;
	ap.spp_pathmaxrxt=3;
#ifdef SPP_SACKDELAY_ENABLE
	ap.spp_sackdelay=20;
	ap.spp_flags|=SPP_HB_ENABLE|SPP_SACKDELAY_ENABLE;
	ap.spp_flags&=~SPP_HB_DISABLE|SPP_HB_DEMAND|SPP_SACKDELAY_DISABLE|
		SPP_HB_TIME_IS_ZERO;
#else
	ap.spp_flags|=SPP_HB_ENABLE;
	ap.spp_flags&=~SPP_HB_DISABLE|SPP_HB_DEMAND|SPP_HB_TIME_IS_ZERO;
#endif

	if(UNLIKELY(setsockopt(s,PROTO,SCTP_PEER_ADDR_PARAMS,&ap,
		sizeof(ap))))
	{
		x=errno;
		close(s);
		BAD("setsockopt(SCTP_PEER_ADDR_PARAMS): %s",errstr(x,mem));
	}

	/* set up ttl */

	x=1;
	if(UNLIKELY(setsockopt(s,IPPROTO_IP,IP_TTL,&x,sizeof(x))))
	{
		x=errno;
		close(s);
		BAD("setsockopt(IP_TTL): %s",errstr(x,mem));
	}

	/* set up transport authentication */

	if(key&&keylen)if(sctp_server_auth(s,key,keylen,kernel_bug_workaround))
	{
		close(s);
		BAD("sctp authentication setup failure");
	}

	/* bind the socket */

	memset(&a,0,sizeof(a));
	a.sin_family=AF_INET;
	a.sin_addr.s_addr=addr?htonl(addr):INADDR_ANY;
	a.sin_port=htons(port);

#ifndef SAFEUID
	shlock();
#endif
	if(UNLIKELY(bind(s,(struct sockaddr *)(&a),sizeof(a))))
	{
		x=errno;
#ifndef SAFEUID
		unlock();
#endif
		close(s);
		BAD("bind: %s",errstr(x,mem));
	}
#ifndef SAFEUID
	unlock();
#endif

	/* make socket listen */

	if(UNLIKELY(listen(s,backlog)))
	{
		x=errno;
		close(s);
		BAD("listen: %s",errstr(x,mem));
	}

	/* return socket descriptor */

	return s;
}

#else

int sctp_omsocket(char *dev,int port,int backlog,void *key,int keylen,
	int kernel_bug_workaround)
{
	BAD("sctp not available on this system");
}

#endif
