
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "sctp.h"
#include "log.h"
#include "ifc.h"
#include "lock.h"
#include "safeapi.h"
#include "sctp_auth.h"

#ifdef SCTP_AVAILABLE

#if defined(__linux__)
#define PROTO SOL_SCTP
#elif defined(__FreeBSD__)
#define PROTO IPPROTO_SCTP
#else
#error do not know how to set SCTP protocol for getsockopt()/setsockopt()
#endif

/*
 * getladdrlist
 *
 * input:  s - the listening socket to be processed
 *
 * output: list - the list of IPv4 addresses the socket is bound to
 *
 * return: the mount of list entries or -1 in case of an error
 *
 * description:	This routine returns all IPv4 addresses bound to a listening
 *		socket.
 */

static int getladdrlist(int s,struct sockaddr_in **list)
{
	int total=0;
	int alloc=0;
	int num;
	int x;
	char *ptr;
	struct sockaddr *la;
	struct sockaddr_in *l=NULL;
	struct sockaddr_in *tmp;
	char mem[256];

	/* get complete list of bound addresses */

	if(UNLIKELY((num=sctp_getladdrs(s,0,&la))==-1))
		BAD("sctp_getladdrs: %s",errstr(errno,mem));

	/* process packed list (aiyeeehh, a *PACKED* list!!!) */

	for(ptr=(char *)la,x=0;x<num;x++)
		switch(((struct sockaddr *)ptr)->sa_family)
	{
	/* IPv4 address */

	case AF_INET:
		/* assert sufficient space in result array */

		if(total==alloc)
		{
			alloc+=4;
			if(UNLIKELY(!(tmp=realloc(l,
				alloc*sizeof(struct sockaddr_in)))))
			{
				x=errno;
				if(l)free(l);
				sctp_freeladdrs(la);
				BAD("realloc: %s",errstr(x,mem));
			}
			else l=tmp;
		}

		/* copy entry to return array */

		memcpy(&l[total++],ptr,sizeof(struct sockaddr_in));

		/* skip processed entry */

		ptr+=sizeof(struct sockaddr_in);
		break;

	/* IPv6 address, just skip over  */

	case AF_INET6:
		ptr+=sizeof(struct sockaddr_in6);
		break;

	/* unknown address type of unknown size */

	default:if(l)free(l);
		sctp_freeladdrs(la);
		BAD("unknown family %d in list",
			((struct sockaddr *)ptr)->sa_family);
	}

	/* free darn packed array */

	sctp_freeladdrs(la);

	/* return list to caller */

	*list=l;

	/* return amount of list entries */

	return total;
}

/*
 * sctpsocket
 *
 * input:  dev     - the device to listen on or NULL
 *	   port    - the listening port
 *	   backlog - the listen backlog
 *	   key     - sctp hmac authentication key or NULL
 *	   keylen  - sctp hmac authentication key length or 0
 *	   kernel_bug_workaround - do not authenticate COOKIE ACK if 1
 *
 * return: a socket descriptor in case of success or -1 in case of error
 *
 * description: This routine creates an IPv4 SCTP listening socket.
 */

int sctpsocket(char *dev,int port,int backlog,void *key,int keylen,
	int kernel_bug_workaround)
{
	int addr;
	int s;
	int x;
	socklen_t l;
	struct sockaddr_in a;
	struct sctp_rtoinfo rto;
	struct sctp_paddrparams ap;
	char mem[256];

	/* sanity check */

	if(UNLIKELY(port<1||port>65535))BAD("internal error");

	/* get interface address if bound to device */

	if(LIKELY(!dev))addr=0;
	else if(UNLIKELY((addr=getifcaddr(dev))==-1))return -1;

	/* create socket */

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((s=socket(AF_INET,SOCK_STREAM|SOCK_CLOEXEC,
			IPPROTO_SCTP)))==-1)BAD("socket: %s",errstr(errno,mem));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((s=socket(AF_INET,SOCK_STREAM,IPPROTO_SCTP)))==-1)
		{
			x=errno;
			unlock();
			BAD("socket: %s",errstr(x,mem));
		}
		fcntl(s,F_SETFD,FD_CLOEXEC);
		unlock();
	}

	/* make socket reusable */

	x=1;
	if(UNLIKELY(setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&x,sizeof(x))))
	{
		x=errno;
		close(s);
		BAD("setsockopt(SO_REUSEADDR): %s",errstr(x,mem));
	}

	/* make socket nonblocking */

	x=1;
	if(UNLIKELY(ioctl(s,FIONBIO,&x)))
	{
		x=errno;
		close(s);
		BAD("ioctl(FIONBIO): %s",errstr(x,mem));
	}

	/* set up retransmission timeouts */

	l=sizeof(rto);
	memset(&rto,0,sizeof(rto));
	if(UNLIKELY(getsockopt(s,PROTO,SCTP_RTOINFO,&rto,&l)))
	{
		x=errno;
		close(s);
		BAD("getsockopt(SCTP_RTOINFO): %s",errstr(x,mem));
	}

	rto.srto_initial=3000;
	rto.srto_min=1000;
	rto.srto_max=10000;

	if(UNLIKELY(setsockopt(s,PROTO,SCTP_RTOINFO,&rto,sizeof(rto))))
	{
		x=errno;
		close(s);
		BAD("setsockopt(SCTP_RTOINFO): %s",errstr(x,mem));
	}

	/* set up heartbeat */

	l=sizeof(ap);
	memset(&ap,0,sizeof(ap));
	if(UNLIKELY(getsockopt(s,PROTO,SCTP_PEER_ADDR_PARAMS,&ap,&l)))
	{
		x=errno;
		close(s);
		BAD("getsockopt(SCTP_PEER_ADDR_PARAMS): %s",errstr(x,mem));
	}

	ap.spp_hbinterval=20000;
	ap.spp_pathmaxrxt=5;
#ifdef SPP_SACKDELAY_ENABLE
	ap.spp_sackdelay=200;
	ap.spp_flags|=SPP_HB_ENABLE|SPP_SACKDELAY_ENABLE;
	ap.spp_flags&=~SPP_HB_DISABLE|SPP_HB_DEMAND|SPP_SACKDELAY_DISABLE|
		SPP_HB_TIME_IS_ZERO;
#else
	ap.spp_flags|=SPP_HB_ENABLE;
	ap.spp_flags&=~SPP_HB_DISABLE|SPP_HB_DEMAND|SPP_HB_TIME_IS_ZERO;
#endif

	if(UNLIKELY(setsockopt(s,PROTO,SCTP_PEER_ADDR_PARAMS,&ap,
		sizeof(ap))))
	{
		x=errno;
		close(s);
		BAD("setsockopt(SCTP_PEER_ADDR_PARAMS): %s",errstr(x,mem));
	}

	/* set up transport authentication */

	if(key&&keylen)if(sctp_server_auth(s,key,keylen,kernel_bug_workaround))
	{
		close(s);
		BAD("sctp authentication setup failure");
	}

	/* bind the socket */

	memset(&a,0,sizeof(a));
	a.sin_family=AF_INET;
	a.sin_addr.s_addr=addr?htonl(addr):INADDR_ANY;
	a.sin_port=htons(port);

#ifndef SAFEUID
	shlock();
#endif
	if(UNLIKELY(bind(s,(struct sockaddr *)(&a),sizeof(a))))
	{
		x=errno;
#ifndef SAFEUID
		unlock();
#endif
		close(s);
		BAD("bind: %s",errstr(x,mem));
	}
#ifndef SAFEUID
	unlock();
#endif

	/* make socket listen */

	if(UNLIKELY(listen(s,backlog)))
	{
		x=errno;
		close(s);
		BAD("listen: %s",errstr(x,mem));
	}

	/* return socket descriptor */

	return s;
}

/*
 * sctpifset
 *
 * input:  fd  - the listening socket descriptor
 *	   dev - the device to listen on
 *
 * return: 0 in case of success or -1 in case of error
 *
 * description: This routine changes the address of an IPv4 SCTP
 *		listening socket.
 */

int sctpifset(int fd,char *dev)
{
	int addr;
	int num;
	int flag;
	int i;
	int x;
	struct sockaddr_in *list;
	struct sockaddr_in a;
	char mem[256];

	/* get address of new interface */

	if(UNLIKELY((addr=getifcaddr(dev))==-1))return -1;
	addr=htonl(addr);

	/* get list of addresses bound to socket */

	if(UNLIKELY((num=getladdrlist(fd,&list))==-1))return -1;

	/* check, if desired address is already bound */

	for(flag=0,i=0;i<num;i++)if(LIKELY(list[i].sin_family==AF_INET))
		if(list[i].sin_addr.s_addr==addr)
	{
		flag=1;
		break;
	}

	/* if new address is not yet bound */

	if(!flag)
	{
		/* prepare address (port 0 means recycle existing port) */

		memset(&a,0,sizeof(a));
		a.sin_family=AF_INET;
		a.sin_addr.s_addr=addr;

		/* bind new address to socket */

#ifndef SAFEUID
		shlock();
#endif
		if(UNLIKELY(sctp_bindx(fd,(struct sockaddr *)&a,1,
			SCTP_BINDX_ADD_ADDR)))
		{
			x=errno;
#ifndef SAFEUID
			unlock();
#endif
			if(list)free(list);
			BAD("sctp_bindx SCTP_BINDX_ADD_ADDR: %s",
				errstr(x,mem));
		}
#ifndef SAFEUID
		unlock();
#endif
	}

	/* remove all old addresses from socket */

	for(i=0;i<num;i++)if(UNLIKELY(list[i].sin_family!=AF_INET)||
		((struct sockaddr_in *)&list[i])->sin_addr.s_addr!=addr)
	{
#ifndef SAFEUID
		shlock();
#endif
		if(UNLIKELY(sctp_bindx(fd,(struct sockaddr *)&list[i],1,
			SCTP_BINDX_REM_ADDR)))
		{
			x=errno;
#ifndef SAFEUID
			unlock();
#endif
			if(list)free(list);
			BAD("sctp_bindx SCTP_BINDX_REM_ADDR: %s",
				errstr(x,mem));
		}
#ifndef SAFEUID
		unlock();
#endif
	}

	/* free bound addresses list */

	if(list)free(list);

	/* signal success */

	return 0;
}

#else

int sctpsocket(char *dev,int port,int backlog,void *key,int keylen,
	int kernel_bug_workaround)
{
	BAD("sctp not available on this system");
}

int sctpifset(int fd,char *dev)
{
	BAD("sctp not available on this system");
}

#endif
