
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "sigs.h"
#include "log.h"
#include "lock.h"
#include "safeapi.h"

/*
 * dftblock
 *
 * input:  how - the processing mode requested
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine installs signal blocking as requesed by
 *		caller. See the OpenBSD rant below for a "special"
 *		OpenBSD "feature".
 */

int dftblock(int how)
{
#if defined(__OpenBSD__)
	int i;
#endif
	sigset_t s;
	char mem[256];

	/* selet all signals and then remove core dumping signals */

	sigfillset(&s);
	sigdelset(&s,SIGILL);
	sigdelset(&s,SIGABRT);
	sigdelset(&s,SIGFPE);
	sigdelset(&s,SIGSEGV);
	sigdelset(&s,SIGBUS);
	sigdelset(&s,SIGSYS);

	switch(how)
	{
	/* process setup, old style signal handler way */

	case PROCESS:

		/* block all signals in set */

#if !defined(__OpenBSD__)
		if(UNLIKELY(sigprocmask(SIG_SETMASK,&s,NULL)))
			BAD("sigprocmask: %s",errstr(errno,mem));
#else
#warning broken sigprocmask, fix this system, please
#endif

		/* remove termination signals from set */

		sigdelset(&s,SIGTERM);
		sigdelset(&s,SIGINT);
		sigdelset(&s,SIGQUIT);
		sigdelset(&s,SIGHUP);
#if defined(__OpenBSD__)

		/* OpenBSD a.k.a. BrokenBSD special: when one links
		   against the BrokenBSD pthread library all signal
		   blocking is gone for good. Do these "we're so
		   secure" folks ever do any regression test or are
		   threads supposed to be so insecure that one must
		   not use them?!? Anyway all signals need to be
		   ignored for BrokenBSD. All? Well, SIGCHLD can't
		   be ignored or the waitpid stuff will fail. Cute.
		   One can only hope that this causes no EINTR as
		   in this case I'm giving up on BrokenBSD. I will
		   *NOT* clutter the code with EINTR checks. Note
		   that even if this mess would be fixed today
		   the below code has to stay for years for the
		   older systems around. */

		for(i=1;i<_NSIG;i++)if(sigismember(&s,i))if(i!=SIGCHLD)
			signal(i,SIG_IGN);

		/* now we do an utter no-op, BrokenBSD style */

		if(UNLIKELY(sigprocmask(SIG_SETMASK,&s,NULL)))
			BAD("sigprocmask: %s",errstr(errno,mem));
#else
		/* set thead sigmask to allow termination signals */

		if(UNLIKELY(pthread_sigmask(SIG_SETMASK,&s,NULL)))
			BAD("pthread_sigmask: %s",errstr(errno,mem));
#endif
		return 0;

	/* process setup, new style signal fd way */

	case ALLSIGS:
#if defined(__OpenBSD__)
		/* see above rant about BrokenBSD crazyness why this
		   has to be done on special request by a single os */

		for(i=1;i<_NSIG;i++)if(sigismember(&s,i))if(i!=SIGCHLD)
			signal(i,SIG_IGN);
#endif
		/* block all signals */

		if(UNLIKELY(sigprocmask(SIG_SETMASK,&s,NULL)))
			BAD("sigprocmask: %s",errstr(errno,mem));

	/* per thread setup, except for BrokenBSD */

#if !defined(__OpenBSD__)
	case THREAD:

		/* block thread signal delivery */

		if(UNLIKELY(pthread_sigmask(SIG_SETMASK,&s,NULL)))
			BAD("pthread_sigmask: %s",errstr(errno,mem));
#else
#warning broken pthread_sigmask, fix this system, please
#endif
	}

	/* signal success (how does one return an atomic bomb for BrokenBSD?) */

	return 0;
}

/*
 * sethandler
 *
 * input:  routine - either old style signal routine or signal fd thread
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine installs either an old style signal handler
 *		for the four termination signals or for the new style
 *		way a signal thread is created which is passed a
 *		file descriptor that becomes readable when a termination
 *		signal is pending.
 */

int sethandler(void *routine)
{
#if defined(__linux__)
	int fd;
	int i;
	pthread_t h;
	sigset_t s;
	char mem[256];

	/* create a signal set containign the termination signals */

	sigemptyset(&s);
	sigaddset(&s,SIGTERM);
	sigaddset(&s,SIGQUIT);
	sigaddset(&s,SIGINT);
	sigaddset(&s,SIGHUP);

	/* get a file descriptor for the signals */

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((fd=signalfd(-1,&s,SFD_CLOEXEC))==-1))
			BAD("signalfd: %s",errstr(errno,mem));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((fd=signalfd(-1,&s,0))==-1))
		{
			i=errno;
			unlock();
			BAD("signalfd: %s",errstr(i,mem));
		}
		fcntl(fd,F_SETFD,FD_CLOEXEC);
		unlock();
	}

	/* create signal waiting thread */

	if(UNLIKELY(pthread_create(&h,NULL,routine,(void *)((long)fd))))
	{
		i=errno;
		close(fd);
		BAD("pthread_create: %s",errstr(i,mem));
	}
#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	int fd;
	int i;
	pthread_t h;
	struct kevent ev[4];
	char mem[256];

	/* kreate a kqueue */

	if(UNLIKELY((fd=kqueue())==-1))BAD("kqueue: %s",errstr(errno,mem));

	/* add the four termination signals to the kqueue */

	EV_SET(&ev[0],SIGTERM,EVFILT_SIGNAL,EV_ADD|EV_ENABLE,0,0,EMPTY);
	EV_SET(&ev[1],SIGQUIT,EVFILT_SIGNAL,EV_ADD|EV_ENABLE,0,0,EMPTY);
	EV_SET(&ev[2],SIGINT,EVFILT_SIGNAL,EV_ADD|EV_ENABLE,0,0,EMPTY);
	EV_SET(&ev[3],SIGHUP,EVFILT_SIGNAL,EV_ADD|EV_ENABLE,0,0,EMPTY);
	if(kevent(fd,ev,4,NULL,0,NULL)==-1)
	{
		i=errno;
		close(fd);
		BAD("kevent: %s",errstr(i,mem));
	}

	/* create signal waiting thread */

	if(UNLIKELY(pthread_create(&h,NULL,routine,(void *)((long)fd))))
	{
		i=errno;
		close(fd);
		BAD("pthread_create: %s",errstr(i,mem));
	}
#else
	/* set up old style signal handler */

	signal(SIGTERM,routine);
	signal(SIGQUIT,routine);
	signal(SIGINT,routine);
	signal(SIGHUP,routine);
#endif
	/* signal success */

	return 0;
}

#if !defined(__linux__) && !defined(__FreeBSD__) && !defined(__OpenBSD__) && !defined(__NetBSD__)

/*
 * block
 *
 * input:  how - process/thread processing mode
 *	   m   - old signal set memory
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine blocks all signals on best effort base.
 *		It is only required for old style signal handling.
 */

int block(int how,sigset_t *m)
{
	sigset_t s;
	char mem[256];

	sigfillset(&s);
	if(how)
	{
		if(UNLIKELY(sigprocmask(SIG_BLOCK,&s,m)))
			BAD("sigprocmask: %s",errstr(errno,mem));
	}
	else if(UNLIKELY(pthread_sigmask(SIG_BLOCK,&s,m)))
		BAD("pthread_sigmask: %s",errstr(errno,mem));
	return 0;
}

/*
 * unblock
 *
 * input:  how - process/thread processing mode
 *	   m   - old signal set memory
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine restores the old signal blocking mask.
 *		It is only required for old style signal handling.
 */

int unblock(int how,sigset_t *m)
{
	char mem[256];

	if(how)
	{
		if(UNLIKELY(sigprocmask(SIG_SETMASK,m,NULL)))
			BAD("sigprocmask: %s",errstr(errno,mem));
	}
	else if(UNLIKELY(pthread_sigmask(SIG_SETMASK,m,NULL)))
		BAD("pthread_sigmask: %s",errstr(errno,mem));
	return 0;
}

#endif
