
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "log.h"
#include "stats.h"
#include "lock.h"
#include "common.h"
#include "safeapi.h"

static int memfree=0;
static int procs=0;
static int cpunum=1;
static int realcores=-1;
static int avgmips=1000;

static pthread_mutex_t mtx=PTHREAD_MUTEX_INITIALIZER;

#if defined(__linux__)

/*
 * trim
 *
 * input: s - the string to be trimmed
 *
 * return: the trimmed string
 *
 * description: This routine removes leading and trailing spaces of the
 *		given string. The pointer returned points to some place
 *		in the delivered source string.
 */

static inline char *trim(char *s)
{
	char *p;

	/* skip leading blanks and tabs */

	while(*s==' '||*s=='\t')s++;

	/* find end of string */

	for(p=s;*p;p++);

	/* remove trailng blanks and tabs */

	for(;p!=s;p--)if(p[-1]!=' '&&p[-1]!='\t')
	{
		*p=0;
		break;
	}

	/* return trimmed result */

	return s;
}

#endif

/*
 * cpudata
 *
 * input:  usercores - -1 if system determined, else user specified core count
 *	   userfreq  - -1 if system determined, else user specified cpu freq
 *	   userht    - 1 if ht cpu, 0 if no ht cpu, -1 if system determined
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine sets up the amount of physical cores and the
 *		average core frequency. If not specified it tries to
 *		detect hyperthreaded cpus and removes hyperthreaded cores
 *		from the core count.
 */

static int cpudata(int usercores,int userfreq,int userht)
{
#if defined(__linux__)

	int fd;
	int phys;
	int cores;
	int cpunum;
	long cpumap;
	FILE *fp;
	char *name;
	char *value;
	char *save;
	double bogomips;
	char bfr[2048];

	/* preset data */

	phys=-1;
	cores=0;
	cpunum=0;
	cpumap=0;
	bogomips=0;

	/* sanity checks */

	if(UNLIKELY(!usercores))BAD("can't set zero cpus");
	if(UNLIKELY(!userfreq))BAD("can't set zero cpu speed");

	/* everything user specified, set data and be done */

	if(UNLIKELY(usercores>0&&userfreq>0))
	{
		pthread_mutex_lock(&mtx);
		realcores=usercores;
		avgmips=userfreq<<1;
		pthread_mutex_unlock(&mtx);
		return 0;
	}

	/* access /proc/cpuinfo */

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((fd=open("/proc/cpuinfo",O_RDONLY|O_CLOEXEC))==-1))
			BAD("open /proc/cpuinfo: %s",errstr(errno,bfr));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((fd=open("/proc/cpuinfo",O_RDONLY))==-1))
		{
			phys=errno;
			unlock();
			BAD("open /proc/cpuinfo: %s",errstr(phys,bfr));
		}
		fcntl(fd,F_SETFD,FD_CLOEXEC);
		unlock();
	}
	if(UNLIKELY(!(fp=fdopen(fd,"r"))))
	{
		phys=errno;
		unlock();
		close(fd);
		BAD("open /proc/cpuinfo: %s",errstr(phys,bfr));
	}

	/* process the whole file */

	while(LIKELY(fgets(bfr,sizeof(bfr),fp)!=NULL))
	{
		/* next core */

		if(UNLIKELY(!strcmp(bfr,"\n")))
		{
			phys=-1;
			cpunum++;
			continue;
		}

		/* get name: value pairs */

		name=strtok_r(bfr,":\n",&save);
		value=strtok_r(NULL,":\n",&save);

		if(UNLIKELY(!name))
		{
			fclose(fp);
			BAD("missing name");
		}

		if(!value)continue;

		/* remove leading and trailing blank characters */

		name=trim(name);
		value=trim(value);

		/* assemble relevant data */

		if(!strcmp(name,"bogomips"))bogomips+=atof(value);
		else if(!strcmp(name,"physical id"))phys=atoi(value);
		else if(!strcmp(name,"cpu cores")&&phys!=-1)
		{
			if(!(cpumap&(1<<phys)))
			{
				cores+=atoi(value);
				cpumap|=1<<phys;
			}
		}
	}

	/* close file */

	fclose(fp);

	/* setup data */

	pthread_mutex_lock(&mtx);
	if(UNLIKELY(usercores>0))realcores=usercores;
	else if(LIKELY(userht==-1))realcores=cores;
	else
	{
		realcores=cpunum;
		if(userht==1&&realcores>1)realcores>>=1;
	}
	if(UNLIKELY(userfreq>0))avgmips=userfreq<<1;
	else avgmips=(int)(bogomips/cpunum+0.5);
	pthread_mutex_unlock(&mtx);

	/* signal success */

	return 0;

#elif defined(__FreeBSD__)

	int nrcpu;
	int freq;
	size_t len;
	char bfr[256];

	/* sanity check */

	if(UNLIKELY(!usercores))BAD("can't set zero cpus");
	if(UNLIKELY(!userfreq))BAD("can't set zero cpu speed");

	/* everything user specified, set data and be done */

	if(UNLIKELY(usercores>0&&userfreq>0))
	{
		pthread_mutex_lock(&mtx);
		realcores=usercores;
		avgmips=userfreq<<1;
		pthread_mutex_unlock(&mtx);
		return 0;
	}

	/* get core amount */

	len=sizeof(nrcpu);
	if(UNLIKELY(sysctlbyname("hw.ncpu",&nrcpu,&len,NULL,0)))
		BAD("sysctl(hw.ncpu): %s",errstr(errno,bfr));

	/* get core frequency */

	len=sizeof(freq);
	if(UNLIKELY(sysctlbyname("dev.cpu.0.freq",&freq,&len,NULL,0)))
		BAD("sysctl(dev.cpu.0.freq): %s",errstr(errno,bfr));

	/* setup data */

	pthread_mutex_lock(&mtx);
	if(UNLIKELY(usercores>0))realcores=usercores;
	else if(LIKELY(userht!=1))realcores=nrcpu;
	else
	{
		realcores=nrcpu;
		if(realcores>1)realcores>>=1;
	}
	if(UNLIKELY(userfreq>0))avgmips=userfreq<<1;
	else avgmips=freq<<1;
	pthread_mutex_unlock(&mtx);

	/* signal success */

	return 0;

#elif defined(__OpenBSD__)

	int nrcpu;
	int freq;
	int mib[2];
	size_t len;
	char bfr[256];

	/* sanity check */

	if(UNLIKELY(!usercores))BAD("can't set zero cpus");
	if(UNLIKELY(!userfreq))BAD("can't set zero cpu speed");

	/* everything user specified, set data and be done */

	if(UNLIKELY(usercores>0&&userfreq>0))
	{
		pthread_mutex_lock(&mtx);
		realcores=usercores;
		avgmips=userfreq<<1;
		pthread_mutex_unlock(&mtx);
		return 0;
	}

	/* get core amount */

	mib[0]=CTL_HW;
	mib[1]=HW_NCPUFOUND;
	len=sizeof(nrcpu);
	if(UNLIKELY(sysctl(mib,2,&nrcpu,&len,NULL,0)))
		BAD("sysctl(hw.ncpufound): %s",errstr(errno,bfr));

	/* get core frequency */

	mib[0]=CTL_HW;
	mib[1]=HW_CPUSPEED;
	len=sizeof(freq);
	if(UNLIKELY(sysctl(mib,2,&freq,&len,NULL,0)))
		BAD("sysctl(hw.cpuspeed): %s",errstr(errno,bfr));

	/* setup data */

	pthread_mutex_lock(&mtx);
	if(UNLIKELY(usercores>0))realcores=usercores;
	else if(LIKELY(userht!=1))realcores=nrcpu;
	else
	{
		realcores=nrcpu;
		if(realcores>1)realcores>>=1;
	}
	if(UNLIKELY(userfreq>0))avgmips=userfreq<<1;
	else avgmips=freq<<1;
	pthread_mutex_unlock(&mtx);

	/* signal success */

	return 0;

#elif defined(__NetBSD__)
#warning anyone tell me something better than machdep.tsc_freq for cpu frequency

	int nrcpu;
	long long freq;
	size_t len;
	char bfr[256];

	/* sanity check */

	if(UNLIKELY(!usercores))BAD("can't set zero cpus");
	if(UNLIKELY(!userfreq))BAD("can't set zero cpu speed");

	/* everything user specified, set data and be done */

	if(UNLIKELY(usercores>0&&userfreq>0))
	{
		pthread_mutex_lock(&mtx);
		realcores=usercores;
		avgmips=userfreq<<1;
		pthread_mutex_unlock(&mtx);
		return 0;
	}

	/* get core amount */

	len=sizeof(nrcpu);
	if(UNLIKELY(sysctlbyname("hw.ncpu",&nrcpu,&len,NULL,0)))
		BAD("sysctl(hw.ncpu): %s",errstr(errno,bfr));

	/* get core frequency */

	len=sizeof(freq);
	if(UNLIKELY(sysctlbyname("machdep.tsc_freq",&freq,&len,NULL,0)))
		BAD("sysctl(machdep.tsc_freq): %s",errstr(errno,bfr));
	freq/=1000000;

	/* setup data */

	pthread_mutex_lock(&mtx);
	if(UNLIKELY(usercores>0))realcores=usercores;
	else if(LIKELY(userht!=1))realcores=nrcpu;
	else
	{
		realcores=nrcpu;
		if(realcores>1)realcores>>=1;
	}
	if(UNLIKELY(userfreq>0))avgmips=userfreq<<1;
	else avgmips=(int)(freq<<1);
	pthread_mutex_unlock(&mtx);

	/* signal success */

	return 0;

#else
#error system specific implementation required for cpudata()
#endif
}

/*
 * memstat
 *
 * return: the amount of ree memory in MB or -1 in case of error
 *
 * description:	This routine retrieves the amount of free memory.
 */

static int memstat(void)
{
#if defined(__linux__)

	int fd;
	int l;
	UNUSED(long total);
	long mfree;
	long buffers;
	long cached;
	char *ptr;
	char bfr[256];

	/* open /proc/meminfo */

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((fd=open("/proc/meminfo",O_RDONLY|O_CLOEXEC))==-1))
			BAD("open /proc/meminfo: %s",errstr(errno,bfr));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((fd=open("/proc/meminfo",O_RDONLY))==-1))
		{
			l=errno;
			unlock();
			BAD("open /proc/meminfo: %s",errstr(l,bfr));
		}
		fcntl(fd,F_SETFD,FD_CLOEXEC);
		unlock();
	}

	/* read relevant data to buffer */

	l=read(fd,bfr,sizeof(bfr)-1);

	/* close file */

	close(fd);

	/* sanity check */

	if(UNLIKELY(l<sizeof(bfr)-1))BAD("short read");

	/* terminate buffer */

	bfr[l]=0;

	/* scan buffer data */

	for(ptr=bfr;*ptr&&(*ptr<'0'||*ptr>'9');ptr++);
	total=atol(ptr);
	for(;*ptr&&*ptr!='\n';ptr++);

	for(;*ptr&&(*ptr<'0'||*ptr>'9');ptr++);
	mfree=atol(ptr);
	for(;*ptr&&*ptr!='\n';ptr++);

	for(;*ptr&&(*ptr<'0'||*ptr>'9');ptr++);
	buffers=atol(ptr);
	for(;*ptr&&*ptr!='\n';ptr++);

	for(;*ptr&&(*ptr<'0'||*ptr>'9');ptr++);
	cached=atol(ptr);

	/* return amount of free memory */

	return (int)((mfree+buffers+cached+512)>>10);

#elif defined(__FreeBSD__)

	int pgsz;
	int mfr;
	int mca;
	int mbu;
	size_t len;
	char bfr[256];

	/* get memory data */

	len=sizeof(pgsz);
	if(UNLIKELY(sysctlbyname("vm.stats.vm.v_page_size",&pgsz,&len,NULL,0)))
		BAD("sysctl(vm.stats.vm.v_page_size): %s",errstr(errno,bfr));

	len=sizeof(mfr);
	if(UNLIKELY(sysctlbyname("vm.stats.vm.v_free_count",&mfr,&len,NULL,0)))
		BAD("sysctl(vm.stats.vm.v_free_count): %s",errstr(errno,bfr));

	len=sizeof(mca);
	if(UNLIKELY(sysctlbyname("vm.stats.vm.v_cache_count",&mca,&len,NULL,0)))
		BAD("sysctl(vm.stats.vm.v_cache_count): %s",errstr(errno,bfr));

	len=sizeof(mbu);
	if(UNLIKELY(sysctlbyname("vfs.bufspace",&mbu,&len,NULL,0)))
		BAD("sysctl(vfs.bufspace): %s",errstr(errno,bfr));

	/* return amount of free memory */

	return (int)(((long long)mfr*(long long)pgsz+
		(long long)mca*(long long)pgsz+(long long)mbu)>>20);

#elif defined(__OpenBSD__)

	int pgsz;
	struct vmtotal totals;
	size_t len;
	int mib[2];
	char bfr[256];

	/* get memory data */

	mib[0]=CTL_HW;
	mib[1]=HW_PAGESIZE;
	len=sizeof(pgsz);
	if(UNLIKELY(sysctl(mib,2,&pgsz,&len,NULL,0)))
		BAD("sysctl(hw.pagesize): %s",errstr(errno,bfr));

	mib[0]=CTL_VM;
	mib[1]=VM_METER;
	len=sizeof(totals);
	if(UNLIKELY(sysctl(mib,2,&totals,&len,NULL,0)))
		BAD("sysctl(vm.vmmeter): %s",errstr(errno,bfr));

	/* return amount of free memory */

	return (int)(((long long)totals.t_free*(long long)pgsz)>>20);

#elif defined(__NetBSD__)

	struct uvmexp_sysctl uvm;
	size_t len;
	char bfr[256];

	/* get memory data */

	len=sizeof(uvm);
	if(UNLIKELY(sysctlbyname("vm.uvmexp2",&uvm,&len,NULL,0)))
		BAD("sysctl(vm.uvmexp2): %s",errstr(errno,bfr));

	/* return amount of free memory */

	return (int)((((long long)uvm.free+(long long)uvm.inactive+
		(long long)uvm.filepages)*(long long)uvm.pagesize)>>20);

#else
#error system specific implementation required for memstat()
#endif
}

/*
 * procstat
 *
 * return: the amount of running/runnable processes or -1 in case of error
 *
 * description:	This routine returns the amount of running/runnable processes.
 *		Porting hint: no average, this *MUST* be the current system
 *		state.
 */

static int procstat(void)
{
#if defined(__linux__)

	int fd;
	int l;
	char *p;
	char bfr[256];

	/* open /proc/loadavg */

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((fd=open("/proc/loadavg",O_RDONLY|O_CLOEXEC))==-1))
			BAD("open /proc/loadavg: %s",errstr(errno,bfr));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((fd=open("/proc/loadavg",O_RDONLY))==-1))
		{
			l=errno;
			unlock();
			BAD("open /proc/loadavg: %s",errstr(l,bfr));
		}
		fcntl(fd,F_SETFD,FD_CLOEXEC);
		unlock();
	}

	/* read relevant data */

	l=read(fd,bfr,sizeof(bfr)-1);

	/* close file */

	close(fd);

	/* terminate buffer */

	bfr[l]=0;

	/* sanity check */

	if(UNLIKELY(!(p=strchr(bfr,'/'))))BAD("bad data");

	/* get amount of runnable processes */

	while(p!=bfr)
	{
		if(p[-1]>='0'&&p[-1]<='9')p--;
		else break;
	}

	/* return amount of runnable processes */

	return atoi(p);

#elif defined(__FreeBSD__)

	struct vmtotal totals;
	size_t len;
	char bfr[256];

	/* get system data */

	len=sizeof(totals);
	if(UNLIKELY(sysctlbyname("vm.vmtotal",&totals,&len,NULL,0)))
		BAD("sysctl(vm.vmtotal): %s",errstr(errno,bfr));

	/* return amount of runnable processes */

	return totals.t_rq;

#elif defined(__OpenBSD__)

	struct vmtotal totals;
	size_t len;
	int mib[2];
	char bfr[256];

	/* get system data */

	mib[0]=CTL_VM;
	mib[1]=VM_METER;
	len=sizeof(totals);
	if(UNLIKELY(sysctl(mib,2,&totals,&len,NULL,0)))
		BAD("sysctl(vm.vmmeter): %s",errstr(errno,bfr));

	/* return amount of runnable processes */

	return totals.t_rq;

#elif defined(__NetBSD__)

	struct vmtotal totals;
	size_t len;
	char bfr[256];

	/* get system data */

	len=sizeof(totals);
	if(UNLIKELY(sysctlbyname("vm.vmmeter",&totals,&len,NULL,0)))
		BAD("sysctl(vm.vmmeter): %s",errstr(errno,bfr));

	/* return amount of runnable processes */

	return totals.t_rq;

#else
#error system specific implementation required for procstat()
#endif
}

/*
 * cpucount
 *
 * return: the amount of currently active (i.e. not disabled) cores.
 */

static int cpucount(void)
{
#if defined(__linux__) || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	int n;

	/* get number of online cores */

	n=sysconf(_SC_NPROCESSORS_ONLN);

	/* return sanity checked value */

	if(realcores>0)n=n>realcores?realcores:n;
	return n<=0?1:n;
#else
#error system specific implementation required for cpucount()
#endif
}

/*
 * getstats
 *
 * input:  what - the action to be taken
 *	   val  - pointer to returned value
 *	   ...  - depends upon action taken
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine initializes and returns system statistics
 *		(ok, no statistics but current value snapshots).
 */

int getstats(int what,int *val,...)
{
	int v;
	int *p;
	int *h;
	va_list ap;

	/* handle requested action */

	switch(what)
	{
	/* initalization */

	case INIT_STATS:

		/* get cpu data */

		va_start(ap,val);
		p=va_arg(ap,void *);
		h=va_arg(ap,void *);
		v=cpudata(val?*val:-1,p?*p:-1,h?*h:-1);
		va_end(ap);

		/* signal error if cpu data retrieval/setup failed */

		if(UNLIKELY(v))return -1;
		break;

	case MEMFREE_STATS:

		/* get free memory, return last successful retrieved value
		   in case of retrieval problem */

		if(LIKELY((v=memstat())!=-1))
		{
			pthread_mutex_lock(&mtx);
			memfree=v;
		}
		else pthread_mutex_lock(&mtx);
		if(LIKELY(val!=NULL))*val=memfree;
		pthread_mutex_unlock(&mtx);
		break;

	case RUNNING_STATS:

		/* get runnable processes amount, return last successful
		   retrieved value in case of retrieval problem */

		if(LIKELY((v=procstat())!=-1))
		{
			pthread_mutex_lock(&mtx);
			procs=v;
		}
		else pthread_mutex_lock(&mtx);
		if(LIKELY(val!=NULL))*val=procs;
		pthread_mutex_unlock(&mtx);
		break;

	case NUMCPU_STATS:

		/* get online core count, return last successful
		   retrieved value in case of retrieval problem */

		if(LIKELY((v=cpucount())!=-1))
		{
			pthread_mutex_lock(&mtx);
			cpunum=v;
		}
		else pthread_mutex_lock(&mtx);
		if(LIKELY(val!=NULL))*val=cpunum;
		pthread_mutex_unlock(&mtx);
		break;

	case BOGOMIPS_STATS:

		/* return cpu bogomips (freq*2) value */

		if(LIKELY(val!=NULL))*val=avgmips;
		break;

	/* sanity check */

	default:BAD("internal error");
	}

	/* signal success */

	return 0;
}
