
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "clntlog.h"
#include "tcp6.h"

/*
 * tcp6connect
 *
 * input:  addr - remote system address
 *	   port - remote system port
 *	   idx  - interface index in link local mode
 *	   ttl  - hoplimit, must be 1 for link local mode
 *
 * return: a socket descriptor or -1 in case of error
 *
 * description:	This routine connects to a remote system using IPv6 and TCP.
 */

int tcp6connect(struct in6_addr *addr,int port,int idx,int ttl)
{
	int s;
	int x;
	struct sockaddr_in6 a;

	/* sanity check */

	if(UNLIKELY(port<1||port>65535))BAD("internal error");

	/* create socket */

	if(UNLIKELY((s=socket(AF_INET6,SOCK_STREAM,0))==-1))
		BAD("socket: %s",strerror(errno));
	fcntl(s,F_SETFD,FD_CLOEXEC);

	/* make socket blocking */

	x=0;
	if(UNLIKELY(ioctl(s,FIONBIO,&x)))
		BADJUMP(fail,"ioctl(FIONBIO): %s",strerror(errno));

	/* set up tcp keepalive */

#ifdef TCP_KEEPCNT
	x=3;
	if(UNLIKELY(setsockopt(s,IPPROTO_TCP,TCP_KEEPCNT,&x,sizeof(x))))
		BADJUMP(fail,"setsockopt(TCP_KEEPCNT): %s",strerror(errno));
#endif

#ifdef TCP_KEEPIDLE
	x=30;
	if(UNLIKELY(setsockopt(s,IPPROTO_TCP,TCP_KEEPIDLE,&x,sizeof(x))))
		BADJUMP(fail,"setsockopt(TCP_KEEPIDLE): %s",strerror(errno));
#endif

#ifdef TCP_KEEPINTVL
	x=15;
	if(UNLIKELY(setsockopt(s,IPPROTO_TCP,TCP_KEEPINTVL,&x,sizeof(x))))
		BADJUMP(fail,"setsockopt(TCP_KEEPINTVL): %s",strerror(errno));
#endif

#if !defined(TCP_KEEPCNT) || !defined(TCP_KEEPIDLE) || !defined(TCP_KEEPINTVL)
#warning only crippled keepalive support available
#endif

	x=1;
	if(UNLIKELY(setsockopt(s,SOL_SOCKET,SO_KEEPALIVE,&x,sizeof(x))))
		BADJUMP(fail,"setsockopt(SO_KEEPALIVE): %s",strerror(errno));

	/* keep out of band data out of band */

	x=0;
	if(UNLIKELY(setsockopt(s,SOL_SOCKET,SO_OOBINLINE,&x,sizeof(x))))
		BADJUMP(fail,"setsockopt(SO_OOBINLINE): %s",strerror(errno));

	/* set up hoplimit */

	x=ttl;
	if(UNLIKELY(setsockopt(s,IPPROTO_IPV6,IPV6_UNICAST_HOPS,&x,sizeof(x))))
		BADJUMP(fail,"setsockopt(IPV6_UNICAST_HOPS): %s",
			strerror(errno));

	/* enforce IPv6 usage */

	x=1;
	if(UNLIKELY(setsockopt(s,IPPROTO_IPV6,IPV6_V6ONLY,&x,sizeof(x))))
		BADJUMP(fail,"setsockopt(IPV6_V6ONLY): %s",strerror(errno));

	/* connect to remote */

	memset(&a,0,sizeof(a));
	a.sin6_family=AF_INET6;
	a.sin6_addr=*addr;
	a.sin6_port=htons(port);
	a.sin6_scope_id=UNLIKELY(ttl>1)?0:idx;

	if(UNLIKELY(connect(s,(struct sockaddr *)(&a),sizeof(a))))
	{
		BADMSG("connect: %s",strerror(errno));
fail:		close(s);
		return -1;
	}

	/* return socket descriptor */

	return s;
}
