
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "tcp6.h"
#include "log.h"
#include "ifc.h"
#include "lock.h"
#include "safeapi.h"

/*
 * tcp6socket
 *
 * input:  dev      - the listening interface or NULL
 *	   port     - the listening port
 *	   backlock - the listen backlog
 *
 * return: a socket descriptor or -1 in case of error
 *
 * description:	This routine creates an IPv6 TCP listening socket.
 */

int tcp6socket(char *dev,int port,int backlog)
{
	int s;
	int x;
	int idx;
	struct sockaddr_in6 a;
	char mem[256];

	/* sanity check */

	if(UNLIKELY(port<1||port>65535))BAD("internal error");

	/* get interface index in link local mode */

	if(!dev)idx=0;
	else if(UNLIKELY((idx=getifindex(dev))==-1))return -1;

	/* create socket */

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((s=socket(AF_INET6,SOCK_STREAM|SOCK_CLOEXEC,0)))
			==-1)BAD("socket: %s",errstr(errno,mem));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((s=socket(AF_INET6,SOCK_STREAM,0)))==-1)
		{
			x=errno;
			unlock();
			BAD("socket: %s",errstr(x,mem));
		}
		fcntl(s,F_SETFD,FD_CLOEXEC);
		unlock();
	}

	/* make socket reusable */

	x=1;
	if(UNLIKELY(setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&x,sizeof(x))))
	{
		x=errno;
		close(s);
		BAD("setsockopt(SO_REUSEADDR): %s",errstr(x,mem));
	}

	/* keep out of band data out of band */

	x=0;
	if(UNLIKELY(setsockopt(s,SOL_SOCKET,SO_OOBINLINE,&x,sizeof(x))))
	{
		x=errno;
		close(s);
		BAD("setsockopt(SO_OOBINLINE): %s",errstr(x,mem));
	}

	/* enforce IPv6 usage if bound to device */

	if(dev)x=1;
	else x=0;
	if(UNLIKELY(setsockopt(s,IPPROTO_IPV6,IPV6_V6ONLY,&x,sizeof(x))))
	{
		x=errno;
		close(s);
		BAD("setsockopt(IPV6_V6ONLY): %s",errstr(x,mem));
	}

	/* make socket nonblocking */

	x=1;
	if(UNLIKELY(ioctl(s,FIONBIO,&x)))
	{
		x=errno;
		close(s);
		BAD("ioctl(FIONBIO): %s",errstr(x,mem));
	}

	/* bind socket */

	memset(&a,0,sizeof(a));
	a.sin6_family=AF_INET6;
	a.sin6_addr=in6addr_any;;
	a.sin6_port=htons(port);

#ifndef SAFEUID
	shlock();
#endif
	if(UNLIKELY(bind(s,(struct sockaddr *)(&a),sizeof(a))))
	{
		x=errno;
#ifndef SAFEUID
		unlock();
#endif
		close(s);
		BAD("bind: %s",errstr(x,mem));
	}
#ifndef SAFEUID
	unlock();
#endif
	/* make socket listen */

	if(UNLIKELY(listen(s,backlog)))
	{
		x=errno;
		close(s);
		BAD("listen: %s",errstr(x,mem));
	}

	/* return socket descriptor */

	return s;
}
