
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "tcp.h"
#include "log.h"
#include "ifc.h"
#include "lock.h"
#include "safeapi.h"

/*
 * tcpsocket
 *
 * input:  dev      - the listening interface or NULL
 *	   port     - the listening port
 *	   backlock - the listen backlog
 *
 * return: a socket descriptor or -1 in case of error
 *
 * description: This routine creates an IPv4 TCP listening socket.
 */

int tcpsocket(char *dev,int port,int backlog)
{
	int addr;
	int s;
	int x;
	struct sockaddr_in a;
	char mem[256];

	/* sanity ckeck */

	if(UNLIKELY(port<1||port>65535))BAD("internal error");

	/* get interface address if bound to device */

	if(LIKELY(!dev))addr=0;
	else if(UNLIKELY((addr=getifcaddr(dev))==-1))return -1;

	/* create socket */

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((s=socket(AF_INET,SOCK_STREAM|SOCK_CLOEXEC,
			IPPROTO_TCP)))==-1)BAD("socket: %s",errstr(errno,mem));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((s=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP)))==-1)
		{
			x=errno;
			unlock();
			BAD("socket: %s",errstr(x,mem));
		}
		fcntl(s,F_SETFD,FD_CLOEXEC);
		unlock();
	}

	/* make socket reusable */

	x=1;
	if(UNLIKELY(setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&x,sizeof(x))))
	{
		x=errno;
		close(s);
		BAD("setsockopt(SO_REUSEADDR): %s",errstr(x,mem));
	}

	/* keep out of band data out of band */

	x=0;
	if(UNLIKELY(setsockopt(s,SOL_SOCKET,SO_OOBINLINE,&x,sizeof(x))))
	{
		x=errno;
		close(s);
		BAD("setsockopt(SO_OOBINLINE): %s",errstr(x,mem));
	}

	/* make socket nonblocking */

	x=1;
	if(UNLIKELY(ioctl(s,FIONBIO,&x)))
	{
		x=errno;
		close(s);
		BAD("ioctl(FIONBIO): %s",errstr(x,mem));
	}

	/* bind socket */

	memset(&a,0,sizeof(a));
	a.sin_family=AF_INET;
	a.sin_addr.s_addr=addr?htonl(addr):INADDR_ANY;
	a.sin_port=htons(port);

#ifndef SAFEUID
	shlock();
#endif
	if(UNLIKELY(bind(s,(struct sockaddr *)(&a),sizeof(a))))
	{
		x=errno;
#ifndef SAFEUID
		unlock();
#endif
		close(s);
		BAD("bind: %s",errstr(x,mem));
	}
#ifndef SAFEUID
	unlock();
#endif
	/* make socket listen */

	if(UNLIKELY(listen(s,backlog)))
	{
		x=errno;
		close(s);
		BAD("listen: %s",errstr(x,mem));
	}

	/* return socket descriptor */

	return s;
}
