
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "log.h"
#include "term.h"
#include "common.h"
#include "safeapi.h"

static int inited=0;
static int notified;
#if defined(__linux__) || defined(__FreeBSD__)
static int p;
#else
static int p[2];
#endif

/*
 * terminit
 *
 * return: 0 in case of success, -1 in case of failure
 *
 * description:	This routine initializes the termination
 *		notification handling. For modern systems
 *		a kernel event is used, for old systems
 *		a pipe has to be used.
 */

int terminit(void)
{
#if defined(__FreeBSD__)
	int i;
	struct kevent ev;
#endif
	char mem[256];

	/* done if already initialized */

	if(UNLIKELY(inited))return 0;

	/* get either and event fd or create a pipe for the back league */

#if defined(__linux__)
#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((p=eventfd(0,EFD_CLOEXEC))==-1))
			BAD("eventfd: %s",errstr(errno,mem));
	}
	else
#endif
	{
		if(UNLIKELY((p=eventfd(0,0))==-1))
			BAD("eventfd: %s",errstr(errno,mem));
		fcntl(p,F_SETFD,FD_CLOEXEC);
	}
#elif defined(__FreeBSD__)

	if(UNLIKELY((p=kqueue())==-1))BAD("kqueue: %s",errstr(errno,mem));
	EV_SET(&ev,0x100,EVFILT_USER,EV_ADD|EV_ENABLE,NOTE_FFNOP,0,EMPTY);
	if(UNLIKELY(kevent(p,&ev,1,NULL,0,NULL)==-1))
	{
		i=errno;
		close(p);
		BAD("kevent: %s",errstr(i,mem));
	}
#else
#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY(pipe2(p,O_CLOEXEC)))
			BAD("pipe: %s",errstr(errno,mem));
	}
	else
#endif
	{
		if(UNLIKELY(pipe(p)))BAD("pipe: %s",errstr(errno,mem));
		fcntl(p[0],F_SETFD,FD_CLOEXEC);
		fcntl(p[1],F_SETFD,FD_CLOEXEC);
	}
#endif
	/* set up global data */

	notified=0;
	inited=1;

	/* signal success */

	return 0;
}

/*
 * termexit
 *
 * description:	This routine closes the event fd or the pipe fds.
 */

void termexit(void)
{
	/* done if already deinitialized */

	if(UNLIKELY(!inited))return;

	/* set global data and close */

	inited=0;
#if defined(__linux__) || defined(__FreeBSD__)
	close(p);
#else
	close(p[0]);
	close(p[1]);
#endif
}

/*
 * termget
 *
 * return: fd that has POLLIN data when termination is requested
 *
 * description:	This routine returns a pollable fd. If the fd indicates
 *		POLLIN termination is requested.
 */

int termget(void)
{
	if(UNLIKELY(!inited))BAD("internal error");
#if defined(__linux__) || defined(__FreeBSD__)
	return p;
#else
	return p[0];
#endif
}

/*
 * termnotify
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine makes the pollable termination fd
 *		return POLLIN.
 */

int termnotify(void)
{
#if defined(__linux__)
	uint64_t x;
#elif defined(__FreeBSD__)
	struct kevent ev;
#else
	char x;
#endif

	/* sanity check */

	if(UNLIKELY(!inited))BAD("internal error");

	/* done if notification already done */

	if(notified)return 0;

	/* mark as notified */

	notified=1;

	/* set modern kernel event or write to ancient pipe */

#if defined(__linux__)
	x=1;
	x=write(p,&x,sizeof(x));
#elif defined(__FreeBSD__)
	EV_SET(&ev,0x100,EVFILT_USER,0,NOTE_TRIGGER,0,EMPTY);
	kevent(p,&ev,1,NULL,0,NULL);
#else
	x=write(p[1],&x,1);
#endif
	/* signal success */

	return 0;
}
