
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "ifc.h"
#include "log.h"
#include "udp.h"
#include "lock.h"
#include "safeapi.h"

/*
 * udpsocket
 *
 * input:  dev   - the device the socket shall be used on
 *	   port  - the udp port to use
 *	   data  - the udp socket data structure associated with the socket
 *	   mode  - socket mode, either broadacst, local or sitewide multicast
 *	   ttl   - the ttl to be used for the socket
 *	   mcast - the multicast address or NULL if the default shall be used
 *
 * return: 0 in case of success or -1 in case of error
 *
 * description:	This routine creates either a broadcast enabled udp socket
 *		or a udp socket with a multicast group joined.
 *		The socket can also be used for unicasts as usual.
 */

int udpsocket(char *dev,int port,UDP *data,int mode,int ttl,char *mcast)
{
	int i;
	struct sockaddr_in s;
#if defined(__linux__) || defined(__FreeBSD__)
	struct ip_mreqn m;
	int j;
#else
	struct ip_mreq m;
	struct in_addr maddr;
	u_char j;
#endif
	char mem[256];

	/* sanity checks */

	if(UNLIKELY(!dev||!data||ttl<1||ttl>255))BAD("internal error");

	/* get device data */

	if(UNLIKELY((data->addr=getifcaddr(dev))==-1))
		BAD("can't get ip of %s",dev);
	if(UNLIKELY((data->mask=getnetmask(dev))==-1))
		BAD("can't get netmask of %s",dev);
	if(UNLIKELY((data->bcast=getbrdaddr(dev))==-1))
	{
		if(!mode)BAD("can't get broadcast address of %s",dev);
		NOTICE("ignoring missing broadcast address...");
		data->bcast=(0xffffffff&(!data->mask))|(data->addr&data->mask);
	}
	if(UNLIKELY((data->idx=getifindex(dev)))==-1)
		BAD("can't get interface idex for %s",dev);

	/* set up multicast address */

	if(UNLIKELY(inet_pton(AF_INET,mcast?mcast:MCASTADDR4,&data->mcast)!=1))
		BAD("inet_pton() failed");

	/* prepare management data */

	data->net=data->addr&data->mask;
	data->port=port;
	data->mode=mode;
	data->mcast=ntohl(data->mcast);

	/* create socket */

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((data->udp=socket(AF_INET,SOCK_DGRAM|SOCK_CLOEXEC,
			0))==-1))BAD("socket: %s",errstr(errno,mem));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((data->udp=socket(AF_INET,SOCK_DGRAM,0))==-1))
		{
			i=errno;
			unlock();
			BAD("socket: %s",errstr(i,mem));
		}
		fcntl(data->udp,F_SETFD,FD_CLOEXEC);
		unlock();
	}

	/* make socket reusable */

	i=1;
	if(setsockopt(data->udp,SOL_SOCKET,SO_REUSEADDR,&i,sizeof(i)))
	{
		i=errno;
		close(data->udp);
		BAD("setsockopt(SO_REUSEADDR): %s",errstr(i,mem));
	}

	/* set up socket ttl */

	i=UNLIKELY(mode==SITE_MCAST)?ttl:1;
	if(setsockopt(data->udp,IPPROTO_IP,IP_TTL,&i,sizeof(i)))
	{
		i=errno;
		close(data->udp);
		BAD("setsockopt(IP_TTL): %s",errstr(i,mem));
	}

	/* in broadcast mode enable broadcast on socket */

	if(!mode)
	{
		i=1;
		if(UNLIKELY(setsockopt(data->udp,SOL_SOCKET,SO_BROADCAST,&i,
			sizeof(i))))
		{
			i=errno;
			close(data->udp);
			BAD("setsockopt(SO_BROADCAST): %s",errstr(i,mem));
		}
	}

	/* bind socket */

	memset(&s,0,sizeof(s));
	s.sin_family=AF_INET;
	s.sin_addr.s_addr=INADDR_ANY;
	s.sin_port=htons(port);

#ifndef SAFEUID
	shlock();
#endif
	if(UNLIKELY(bind(data->udp,(struct sockaddr *)(&s),sizeof(s))))
	{
		i=errno;
#ifndef SAFEUID
		unlock();
#endif
		close(data->udp);
		BAD("bind: %s",errstr(i,mem));
	}
#ifndef SAFEUID
	unlock();
#endif

	/* multicast mode */

	if(mode)
	{
		/* don't receive own multicast transmissions */

		j=0;
		if(UNLIKELY(setsockopt(data->udp,IPPROTO_IP,IP_MULTICAST_LOOP,
			&j,sizeof(j))))
		{
			i=errno;
			close(data->udp);
			BAD("setsockopt(IP_MULTICAST_LOOP): %s",errstr(i,mem));
		}

		/* set multicast ttl */

		j=UNLIKELY(mode==SITE_MCAST)?ttl:1;
		if(UNLIKELY(setsockopt(data->udp,IPPROTO_IP,IP_MULTICAST_TTL,
			&j,sizeof(j))))
		{
			i=errno;
			close(data->udp);
			BAD("setsockopt(IP_MULTICAST_TTL): %s",errstr(i,mem));
		}

		/* prepare multicast setup */

		memset(&m,0,sizeof(m));
		m.imr_multiaddr.s_addr=htonl(data->mcast);
#if defined(__linux__) || defined(__FreeBSD__)
		m.imr_ifindex=data->idx;
#else
		m.imr_interface.s_addr=htonl(data->addr);
#endif

		/* set outgoing multicast interface */

#if defined(__linux__) || defined(__FreeBSD__)
		if(UNLIKELY(setsockopt(data->udp,IPPROTO_IP,IP_MULTICAST_IF,
			&m,sizeof(m))))
#else
		maddr.s_addr=htonl(data->addr);
		if(UNLIKELY(setsockopt(data->udp,IPPROTO_IP,IP_MULTICAST_IF,
			&maddr,sizeof(maddr))))
#endif
		{
			i=errno;
			close(data->udp);
			BAD("setsockopt(IP_MULTICAST_IF): %s",errstr(i,mem));
		}

		/* join multicast group (sets incoming interface) */

		if(UNLIKELY(setsockopt(data->udp,IPPROTO_IP,IP_ADD_MEMBERSHIP,
			&m,sizeof(m))))
		{
			i=errno;
			close(data->udp);
			BAD("setsockopt(IP_ADD_MEMBERSHIP): %s",errstr(i,mem));
		}
	}

	/* signal success */

	return 0;
}

/*
 * udpclose
 *
 * input:  data - the udp socket data structure associated with the socket
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine leaves the multicast group if required and
 *		then closes the udp socket.
 */

int udpclose(UDP *data)
{
	int i;
#if defined(__linux__) || defined(__FreeBSD__)
	struct ip_mreqn m;
#else
	struct ip_mreq m;
#endif
	char mem[256];

	/* in multicast mode */

	if(data->mode)
	{
		/* leave multicast group */

		memset(&m,0,sizeof(m));
		m.imr_multiaddr.s_addr=htonl(data->mcast);
#if defined(__linux__) || defined(__FreeBSD__)
		m.imr_ifindex=data->idx;
#else
		m.imr_interface.s_addr=htonl(data->addr);
#endif

		if(UNLIKELY(setsockopt(data->udp,IPPROTO_IP,IP_DROP_MEMBERSHIP,
			&m,sizeof(m))))
		{
			i=errno;
			close(data->udp);
			BAD("setsockopt(IP_DROP_MEMBERSHIP): %s",errstr(i,mem));
		}
	}

	/* close socket */

	close(data->udp);

	/* signal success */

	return 0;
}

/*
 * udprecv
 *
 * input:  data - the udp socket data structure associated with the socket
 *	   bfr  - the received data buffer
 *	   len  - the size of the data buffer
 *
 * return: received data length or -1 in case of error
 *
 * description:	This routine receives a udp packet and stores the sender's
 *		address in the udp socket data structure.
 */

int udprecv(UDP *data,void *bfr,int len)
{
	int l;
	socklen_t i;
	struct sockaddr_in s;
	char mem[256];

	/* receive udp packet */

	while(1)
	{
		i=sizeof(s);
		if(UNLIKELY((l=recvfrom(data->udp,bfr,len,0,
			(struct sockaddr *)(&s),&i))==-1))
		{
			if(LIKELY(errno==EINTR))continue;
			DEBUG("recvfrom: %s",errstr(errno,mem));
			return -1;
		}
		break;
	}

	/* sanity checks */

	if(UNLIKELY(i!=sizeof(s)))return -1;
	if(UNLIKELY(s.sin_family!=AF_INET))return -1;

	/* store sender data */

	data->srcaddr=ntohl(s.sin_addr.s_addr);
	data->srcport=ntohs(s.sin_port);

	/* return actual received data length */

	return l;
}

/*
 * udpsend
 *
 * input:  data - the udp socket data structure associated with the socket
 *	   dst  - -1 for multicast/broadcast, else destination address
 *	   bfr  - the data to send
 *	   len  - the length of the data to send
 *
 * return: 0 in case of success, -1 incase of error
 *
 * description:	This routine sends the given data as a udp packet, using
 *		unicast, multicast or broadcast as appropriate.
 */

int udpsend(UDP *data,int dst,void *bfr,int len)
{
	struct sockaddr_in s;
	char mem[256];

	/* set up actual destination address */

	if(dst==-1)dst=htonl(data->mode?data->mcast:data->bcast);
	else dst=htonl(dst);

	/* send data */

	while(1)
	{
		memset(&s,0,sizeof(s));
		s.sin_family=AF_INET;
		s.sin_addr.s_addr=dst;
		s.sin_port=htons(data->port);

		if(UNLIKELY(sendto(data->udp,bfr,len,0,
			(struct sockaddr *)(&s),sizeof(s))==-1))
		{
			if(LIKELY(errno==EINTR))continue;
			DEBUG("sendto: %s",errstr(errno,mem));
			return -1;
		}

		/* signal success */

		return 0;
	}
}

/*
 * udpnewdev
 *
 * input:  dev   - the device the socket shall be used on
 *	   data  - the udp socket data structure associated with the socket
 *
 * return: 1 in case of changes, 0 if no changes, -1 in case of error
 *
 * description:	This routine changes the device data associated with the
 *		socket to the new device. This includes the required
 *		multicast changes if multicast is in use.
 */

int udpnewdev(char *dev,UDP *data)
{
	int i;
	int state;
	int addr;
	int mask;
	int bcast;
	int idx;
#if defined(__linux__) || defined(__FreeBSD__)
	struct ip_mreqn m;
#else
	struct ip_mreq m;
#endif
	char mem[256];

	/* sanity checks */

	if(UNLIKELY(!dev||!data))BAD("internal error");

	/* get new interface data */

	if(UNLIKELY((addr=getifcaddr(dev))==-1))
		BAD("can't get ip of %s",dev);
	if(UNLIKELY((mask=getnetmask(dev))==-1))
		BAD("can't get netmask of %s",dev);
	if(UNLIKELY((bcast=getbrdaddr(dev))==-1))
	{
		if(!data->mode)BAD("can't get broadcast address of %s",dev);
		NOTICE("ignoring missing broadcast address...");
		bcast=(0xffffffff&(!mask))|(addr&mask);
	}
	if(UNLIKELY((idx=getifindex(dev)))==-1)
		BAD("can't get interface idex for %s",dev); 

	/* assume no changes */

	state=0;

	/* multicast mode and interface change */

	if(data->idx!=idx&&data->mode)
	{
		/* prepare multicast join/leave */

		memset(&m,0,sizeof(m));
		m.imr_multiaddr.s_addr=htonl(data->mcast);
#if defined(__linux__) || defined(__FreeBSD__)
		m.imr_ifindex=idx;
#else
		m.imr_interface.s_addr=htonl(addr);
#endif
		/* leave multicast group on old interface */

#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
#if defined(__FreeBSD__)
		m.imr_ifindex=data->idx;
#else
		m.imr_interface.s_addr=htonl(data->addr);
#endif
		if(UNLIKELY(setsockopt(data->udp,IPPROTO_IP,IP_DROP_MEMBERSHIP,
			&m,sizeof(m))))
		{
			i=errno;
			close(data->udp);
			BAD("setsockopt(IP_DROP_MEMBERSHIP): %s",errstr(i,mem));
		}
#endif
		/* join multicast group on new interface */

		if(UNLIKELY(setsockopt(data->udp,IPPROTO_IP,IP_ADD_MEMBERSHIP,
			&m,sizeof(m))))
		{
			i=errno;
			close(data->udp);
			BAD("setsockopt(IP_ADD_MEMBERSHIP): %s",errstr(i,mem));
		}

		/* set new outgoing interface */

		if(UNLIKELY(setsockopt(data->udp,IPPROTO_IP,IP_MULTICAST_IF,&m,
			sizeof(m))))
		{
			i=errno;
			close(data->udp);
			BAD("setsockopt(IP_MULTICAST_IF): %s",errstr(i,mem));
		}

		/* leave multicast group on old interface */

#if defined(__linux__)
		m.imr_ifindex=data->idx;

		if(UNLIKELY(setsockopt(data->udp,IPPROTO_IP,IP_DROP_MEMBERSHIP,
			&m,sizeof(m))))
		{
			i=errno;
			close(data->udp);
			BAD("setsockopt(IP_DROP_MEMBERSHIP): %s",errstr(i,mem));
		}
#endif

		/* changes happened */

		state=1;
	}

	/* if device data have changed */

	if(data->addr!=addr||data->mask!=mask||data->bcast!=bcast)
	{
		/* set up new interface data */

		data->addr=addr;
		data->mask=mask;
		data->bcast=bcast;
		data->idx=idx;
		data->net=data->addr&data->mask;

		/* changes happened */

		state=1;
	}

	/* store new interface index */

	data->idx=idx;

	/* return change state */

	return state;
}
