
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "log.h"
#include "ifc.h"
#include "udp6.h"
#include "lock.h"
#include "safeapi.h"

#if defined(__linux__)
#define JOIN	IPV6_ADD_MEMBERSHIP
#define LEAVE	IPV6_DROP_MEMBERSHIP
#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
#define JOIN	IPV6_JOIN_GROUP
#define LEAVE	IPV6_LEAVE_GROUP
#else
#error need to know how to join/leave multicasts
#endif

/*
 * udp6socket
 *
 * input:  dev   - the device the socket shall be used on
 *	   port  - the udp port to us
 *	   data  - the udp socket data structure associated with the socket
 *	   mode  - socket mode, either local or sitewide multicast
 *	   ttl   - the hoplimit to be used for the socket
 *	   mcast - the multicast address or NULL if the default shall be used
 *
 * return: 0 in case of success or -1 in case of error
 *
 * description: This routine creates an IPv6 udp socket with a
 *		multicast group joined. The socket can also be
 *		used for unicasts as usual.
 */

int udp6socket(char *dev,int port,UDP6 *data,int mode,int ttl,char *mcast)
{
	int udp;
	int i;
	struct sockaddr_in6 s;
	struct ipv6_mreq m;
	char mem[256];

	/* sanity checks */

	if(UNLIKELY(!dev||!data||!mode||ttl<1||ttl>255))BAD("internal error");

	/* set up multicast address */

	if(UNLIKELY(inet_pton(AF_INET6,mcast?mcast:UNLIKELY(mode==SITE_MCAST)?
		MSITEADDR6:MCASTADDR6,&data->mcast)!=1))
		BAD("inet_pton() failed");

	/* get interface index */

	if(UNLIKELY((data->idx=getifindex(dev)))==-1)
		BAD("getifindex for %s failed",dev);

	/* create socket */

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((udp=socket(AF_INET6,SOCK_DGRAM|SOCK_CLOEXEC,0)))
			==-1)BAD("socket: %s",errstr(errno,mem));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((udp=socket(AF_INET6,SOCK_DGRAM,0)))==-1)
		{
			i=errno;
			unlock();
			BAD("socket: %s",errstr(i,mem));
		}
		fcntl(udp,F_SETFD,FD_CLOEXEC);
		unlock();
	}

	/* make socket reusable */

	i=1;
	if(setsockopt(udp,SOL_SOCKET,SO_REUSEADDR,&i,sizeof(i)))
	{
		i=errno;
		close(udp);
		BAD("setsockopt(SO_REUSEADDR): %s",errstr(i,mem));
	}

	/* enforce IPv6 usage */

	i=1;
	if(setsockopt(udp,IPPROTO_IPV6,IPV6_V6ONLY,&i,sizeof(i)))
	{
		i=errno;
		close(udp);
		BAD("setsockopt(IPV6_V6ONLY): %s",errstr(i,mem));
	}

	/* don't receive own multicast transmissions */

	i=0;
	if(setsockopt(udp,IPPROTO_IPV6,IPV6_MULTICAST_LOOP,&i,sizeof(i)))
	{
		i=errno;
		close(udp);
		BAD("setsockopt(IPV6_MULTICAST_LOOP): %s",errstr(i,mem));
	}

	/* set outgoing multicast interface */

	i=data->idx;
	if(setsockopt(udp,IPPROTO_IPV6,IPV6_MULTICAST_IF,&i,sizeof(i)))
	{
		i=errno;
		close(udp);
		BAD("setsockopt(IPV6_MULTICAST_IF): %s",errstr(i,mem));
	}

	/* set unicast hoplimit */

	i=UNLIKELY(mode==SITE_MCAST)?ttl:1;
	if(setsockopt(udp,IPPROTO_IPV6,IPV6_UNICAST_HOPS,&i,sizeof(i)))
	{
		i=errno;
		close(udp);
		BAD("setsockopt(IPV6_UNICAST_HOPS): %s",errstr(i,mem));
	}

	/* set multicast hoplimit */

	i=UNLIKELY(mode==SITE_MCAST)?ttl:1;
	if(setsockopt(udp,IPPROTO_IPV6,IPV6_MULTICAST_HOPS,&i,sizeof(i)))
	{
		i=errno;
		close(udp);
		BAD("setsockopt(IPV6_MULTICAST_HOPS): %s",errstr(i,mem));
	}

	/* bind socket */

	memset(&s,0,sizeof(s));
	s.sin6_family=AF_INET6;
	s.sin6_port=htons(port);
	s.sin6_addr=in6addr_any;

#ifndef SAFEUID
	shlock();
#endif
	if(UNLIKELY(bind(udp,(struct sockaddr *)(&s),sizeof(s))))
	{
		i=errno;
#ifndef SAFEUID
		unlock();
#endif
		close(udp);
		BAD("bind: %s",errstr(i,mem));
	}
#ifndef SAFEUID
	unlock();
#endif

	/* join multicast group (sets incoming interface) */

	memset(&m,0,sizeof(m));
	m.ipv6mr_multiaddr=data->mcast;
	m.ipv6mr_interface=data->idx;

	if(UNLIKELY(setsockopt(udp,IPPROTO_IPV6,JOIN,&m,sizeof(m))))
	{
		i=errno;
		close(udp);
		BAD("setsockopt(multicast join): %s",errstr(i,mem));
	}

	/* fill in socket data structure */

	data->mode=mode;
	data->port=port;
	data->udp=udp;

	/* signal success */

	return 0;
}

/*
 * udp6close
 *
 * input:  data - the udp socket data structure associated with the socket
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description: This routine leaves the multicast group and
 *		then closes the IPv6 udp socket.
 */

int udp6close(UDP6 *data)
{
	int i;
	struct ipv6_mreq m;
	char mem[256];

	/* leave multicast group */

	memset(&m,0,sizeof(m));
	m.ipv6mr_multiaddr=data->mcast;
	m.ipv6mr_interface=data->idx;

	if(UNLIKELY(setsockopt(data->udp,IPPROTO_IPV6,LEAVE,&m,sizeof(m))))
	{
		i=errno;
		close(data->udp);
		BAD("setsockopt(multicast leave): %s",errstr(i,mem));
	}

	/* close socket */

	close(data->udp);

	/* signal success */

	return 0;
}

/*
 * udp6recv
 *
 * input:  data - the udp socket data structure associated with the socket
 *	   bfr  - the received data buffer
 *	   len  - the size of the data buffer
 *
 * return: received data length or -1 incase of error
 *
 * description: This routine receives a udp packet and stores the sender's
 *		address in the udp socket data structure.
 */

int udp6recv(UDP6 *data,void *bfr,int len)
{
	int l;
	socklen_t i;
	struct sockaddr_in6 s;
	char mem[256];

	/* receive udp packet */

	while(1)
	{
		i=sizeof(s);
		if(UNLIKELY((l=recvfrom(data->udp,bfr,len,0,
			(struct sockaddr *)(&s),&i))==-1))
		{
			if(LIKELY(errno==EINTR))continue;
			DEBUG("recvfrom: %s",errstr(errno,mem));
			return -1;
		}
		break;
	}

	/* sanity checks */

	if(UNLIKELY(i!=sizeof(s)))return -1;
	if(UNLIKELY(s.sin6_family!=AF_INET6))return -1;

	/* store sender's address ind incoming interface data */

	data->addr=s.sin6_addr;
	data->srcport=ntohs(s.sin6_port);
	data->srcidx=s.sin6_scope_id;

	/* return message length */

	return l;
}

/*
 * udp6send
 *
 * input:  data - the udp socket data structure associated with the socket
 *	   dst  - 0 for multicast, else unicast
 *	   bfr  - the data to send
 *	   len  - the length of the data to send
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description: This routine sends the given data as a udp packet using
 *		multicast or unicast as appropriate.
 */

int udp6send(UDP6 *data,int dst,void *bfr,int len)
{
	struct sockaddr_in6 s;
	char mem[256];

	/* do until message is sent or something bad happens */

	while(1)
	{
		/* prepare recipient address */

		memset(&s,0,sizeof(s));
		s.sin6_family=AF_INET6;
		s.sin6_addr=dst?data->mcast:data->addr;
		s.sin6_port=htons(data->port);
		s.sin6_scope_id=UNLIKELY(data->mode==SITE_MCAST)?0:data->idx;

		/* send message */

		if(UNLIKELY(sendto(data->udp,bfr,len,0,
			(struct sockaddr *)(&s),sizeof(s))==-1))
		{
			if(LIKELY(errno==EINTR))continue;
			DEBUG("sendto: %s",errstr(errno,mem));
			return -1;
		}

		/* signal success */

		return 0;
	}
}

/*
 * udp6newdev
 *
 * input:  dev   - the device the socket shall be used on
 *	   data  - the udp socket data structure associated with the socket
 *
 * return: 1 in case of changes, 0 if no changes, -1 in case of error
 *
 * description: This routine changes the device data associated with the
 *		socket to the new device. This includes the required
 *		multicast changes.
 */

int udp6newdev(char *dev,UDP6 *data)
{
	int idx;
	struct ipv6_mreq m;
	char mem[256];

	/* sanity checks */

	if(UNLIKELY(!dev||!data))BAD("internal error");

	/* get new interface index */

	if(UNLIKELY((idx=getifindex(dev)))==-1)
		BAD("getifindex for %s failed",dev);

	/* done if interface index didn't change */

	if(data->idx==idx)return 0;

	/* move multicast group and outgoing interface to new interface */

	memset(&m,0,sizeof(m));
	m.ipv6mr_multiaddr=data->mcast;
	m.ipv6mr_interface=idx;

	if(UNLIKELY(setsockopt(data->udp,IPPROTO_IPV6,JOIN,&m,sizeof(m))))
		BAD("setsockopt(multicast join): %s",errstr(errno,mem));

	if(setsockopt(data->udp,IPPROTO_IPV6,IPV6_MULTICAST_IF,
		&idx,sizeof(idx)))
		BAD("setsockopt(IPV6_MULTICAST_IF): %s",errstr(errno,mem));

	/* leave multicast group on old interface */

	m.ipv6mr_interface=data->idx;

	if(UNLIKELY(setsockopt(data->udp,IPPROTO_IPV6,LEAVE,&m,sizeof(m))))
		BAD("setsockopt(multicast leave): %s",errstr(errno,mem));

	/* memorize new interface index */

	data->idx=idx;

	/* signal change happened */

	return 1;
}
