
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "log.h"
#include "unix.h"
#include "lock.h"
#include "safeapi.h"
#if defined(__OpenBSD__)
#include "ipc.h"
#endif

/*
 * unixsocket
 *
 * input:  n - socket pathname
 *	   u - socket user name or NULL
 *	   g - socket group name or NULL
 *	   b - listen backlog
 *
 * return: a socket descriptor or -1 in case of error
 *
 * description:	This routine creates a listening unix domain socket.
 */

int unixsocket(char *n,char *u,char *g,int b)
{
	int s;
	int i;
	mode_t mask;
	uid_t uid;
	gid_t gid;
	struct sockaddr_un a;
	struct stat stb;
	struct passwd *pw;
	struct group *gr;
#if !defined(__OpenBSD__)
	struct passwd pwrk;
	struct group gwrk;
#endif
	char mem[1024];

	/* set default mask */

	mask=0111;

	/* if a group name is specified get gid and adapt mask */

	if(g)
	{
#if defined(__OpenBSD__)
#warning broken getgrnam_r, fix this system, please
		ipc_needex();
		exlock();
		if(UNLIKELY(!(gr=getgrnam(g))))
		{
			unlock();
			ipc_exdone();
			BAD("%s: no such group",g);
		}
		gid=gr->gr_gid;
		unlock();
		ipc_exdone();
#else
		getgrnam_r(g,&gwrk,mem,sizeof(mem),&gr);
		if(UNLIKELY(!gr))BAD("%s: no such group",g);
		gid=gr->gr_gid;
#endif
		mask=0117;
	}
	else gid=0;

	/* if a user name is specified get uid and adapt mask */

	if(u)
	{
#if defined(__OpenBSD__)
#warning broken getpwnam_r, fix this system, please
		ipc_needex();
		exlock();
		if(UNLIKELY(!(pw=getpwnam(u))))
		{
			unlock();
			ipc_exdone();
			BAD("%s: no such user",u);
		}
		uid=pw->pw_uid;
		unlock();
		ipc_exdone();
#else
		getpwnam_r(u,&pwrk,mem,sizeof(mem),&pw);
		if(UNLIKELY(!pw))BAD("%s: no such user",u);
		uid=pw->pw_uid;
#endif
		if(!g)mask=0177;
	}
	else uid=0;

	/* create socket */

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((s=socket(PF_UNIX,SOCK_STREAM|SOCK_CLOEXEC,0))==-1))
			BAD("socket: %s",errstr(errno,mem));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((s=socket(PF_UNIX,SOCK_STREAM,0))==-1))
		{
			i=errno;
			unlock();
			BAD("socket: %s",errstr(i,mem));
		}
		fcntl(s,F_SETFD,FD_CLOEXEC);
		unlock();
	}

	/* make socket nonblocking */

	i=1;
	if(UNLIKELY(ioctl(s,FIONBIO,&i)))
	{
		i=errno;
		close(s);
		BAD("ioctl(FIONBIO): %s",errstr(i,mem));
	}

	/* prepare bind */

	memset(&a,0,sizeof(a));
	a.sun_family=AF_UNIX;
	strncpy(a.sun_path,n,sizeof(a.sun_path)-1);
	a.sun_path[sizeof(a.sun_path)-1]=0;

	/* assert that we can create the socket */

#ifndef SAFEUID
	shlock();
#endif
	if(!lstat(a.sun_path,&stb))
	{
		if(UNLIKELY(!S_ISSOCK(stb.st_mode)))
		{
#ifndef SAFEUID
			unlock();
#endif
			close(s);
			BAD("%s exists and ist not a socket",a.sun_path);
		}

		if(UNLIKELY(unlink(a.sun_path)))
		{
#ifndef SAFEUID
			unlock();
#endif
			close(s);
			BAD("can't unlink %s",a.sun_path);
		}
	}

	/* set up socket creation umask: note that this is a problem for
	   most systems. Either we change the umask here and this is
	   process wide and thus affects all threads to have the
	   unix domain socket created with proper permissions or
	   we chmod the created socket later which leaves a small window
	   for the socket of being accessible to users that shouldn't
	   have access to the socket. Choose your poison. The only
	   system that has a solution is Linux that allows a fchmod
	   on an unbound socket and then uses the permissions during
	   bind */

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY(fchmod(s,mask^0777)))
		{
			i=errno;
#ifndef SAFEUID
			unlock();
#endif
			close(s);
			BAD("fchmod: %s",errstr(i,mem));
		}
	}
	else
#endif
	{
		mask=umask(mask);
	}

	/* bind socket */

	if(UNLIKELY(bind(s,(struct sockaddr *)(&a),sizeof(a))))
	{
		i=errno;
		umask(mask);
#ifndef SAFEUID
		unlock();
#endif
		close(s);
		BAD("bind %s: %s",a.sun_path,errstr(i,mem));
	}

	/* restore old umask value */

#ifdef SAFEAPI
	if(UNLIKELY(!safeapi()))
#endif
	{
		umask(mask);
	}

	/* change owner and group of socket */

	if(UNLIKELY(chown(a.sun_path,u?uid:-1,g?gid:-1)))
	{
		i=errno;
		unlink(a.sun_path);
#ifndef SAFEUID
		unlock();
#endif
		close(s);
		BAD("chown %s: %s",a.sun_path,errstr(i,mem));
	}

	/* make socket listen */

	if(UNLIKELY(listen(s,b)))
	{
		i=errno;
		unlink(a.sun_path);
#ifndef SAFEUID
		unlock();
#endif
		close(s);
		BAD("listen: %s",errstr(i,mem));
	}

#ifndef SAFEUID
	unlock();
#endif

	/* return socket descriptor */

	return s;
}

/*
 * unixclose
 *
 * input: s - the socket descriptor
 *	  n - socket pathname
 *
 * description:	This routine closes and removes a unix domain socket.
 */


void unixclose(int s,char *n)
{
	struct sockaddr_un a;

	/* assert that the same pathname as during creation is used */

	strncpy(a.sun_path,n,sizeof(a.sun_path)-1);
	a.sun_path[sizeof(a.sun_path)-1]=0;

	/* close socket */

	close(s);

	/* delete socket */

#ifdef SAFEUID
	unlink(a.sun_path);
#else
	shlock();
	unlink(a.sun_path);
	unlock();
#endif
}

/*
 * getpeerid
 *
 * input:  s - the socket descriptor
 *
 * output: uid - the peer uid
 *	   gid - the peer gid
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine retrieves the uid and gid of the peer connected
 *		to the given unix domain socket.
 */

int getpeerid(int s,int *uid,int *gid)
{
#if defined(__linux__)
	socklen_t l;
	struct ucred cred;
	char mem[256];

	/* get peer credentials */

	l=sizeof(cred);
	if(UNLIKELY(getsockopt(s,SOL_SOCKET,SO_PEERCRED,&cred,&l)))
		BAD("getsockopt(SO_PEERCRED): %s",errstr(errno,mem));

	/* return credentials to caller */

	*uid=cred.uid;
	*gid=cred.gid;

	/* signal success */

	return 0;
#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	uid_t euid;
	gid_t egid;
	char mem[256];

	/* get peer credentials */

	if(UNLIKELY(getpeereid(s,&euid,&egid)))
		BAD("getpeereid: %s",errstr(errno,mem));

	/* return credentials to caller */

	*uid=euid;
	*gid=egid;

	/* signal success */

	return 0;
#else
#warning Not checking socket peer credentials, this is very bad

	/* fake root credentials */

	*uid=0;
	*gid=0;

	/* signal success */

	return 0;
#endif
}
