
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "fqdn.h"
#include "stats.h"
#include "udp.h"
#include "udp6.h"
#include "ifc.h"
#include "term.h"
#include "log.h"
#include "update.h"
#include "stamp.h"
#include "sigs.h"
#include "auth.h"
#include "ipc.h"
#include "lock.h"
#include "safeapi.h"
#include "caps.h"

/* network message format */

#define NETFORMAT	1

/* operation codes */

#define QUERY		0
#define RESPONSE	1

/* response message, first fields must match request message */

typedef struct
{
	char auth[MD5_SIZE+8];	/* authentication area			*/
	char os[16];		/* operating system			*/
	char arch[16];		/* system architecture			*/
	uint32_t code;		/* operation code			*/
	uint32_t version;	/* compiler version and network format	*/
	uint32_t flags;		/* processing flags			*/
	uint32_t unique;	/* unique id				*/
	uint32_t jobs;		/* running compile jobs			*/
	uint32_t sum;		/* ditto with supervisor adjustment	*/
	uint32_t memfree;	/* current free memory			*/
	uint32_t minfree;	/* minimum free memory			*/
	uint32_t nrcpu;		/* amount of online cores		*/
	uint32_t bogomips;	/* core bogomips (frequency*2)		*/
	uint32_t running;	/* runnable/running processes		*/
	uint32_t pending;	/* queued local compile jobs		*/
	uint32_t rmtqueue;	/* queued remote compile jobs		*/
	char fqdn[0];		/* system fqdn (only for web output)	*/
} RESPONSEMSG;

/* request message */

typedef struct
{
	char auth[MD5_SIZE+8];	/* authentication area			*/
	char os[16];		/* operating system			*/
	char arch[16];		/* system architecture			*/
	uint32_t code;		/* operation code			*/
	uint32_t version;	/* compiler version and network format	*/
	uint32_t flags;		/* processing flags			*/
} QUERYMSG;

/* broadcast/multicast rate limiting message buffer */

typedef struct
{
	int stamp;		/* compiler version association stamp	*/
	int len;		/* message length			*/
	char msg[1500];		/* message buffer			*/
} BCASTMSG;

/* remote service data */

typedef struct _service
{
	struct _service *next;	/* next element in list			*/
	clock_t stamp;		/* receive timestamp			*/
	int addr;		/* remote IPv4 address			*/
	int jobs;		/* running compile jobs			*/
	int sum;		/* ditto with supervisor adjustment	*/
	int submitted;		/* compile jobs submitter to system	*/
	int memfree;		/* current free memory			*/
	int minfree;		/* minimum free memory			*/
	int nrcpu;		/* amount of online cores		*/
	int bogomips;		/* core bogomips (frequency*2)		*/
	int running;		/* runnable/running processed		*/
	int pending;		/* queued local compile jobs		*/
	int rmtqueue;		/* queued remote compile jobs		*/
	int flags;		/* processing flags			*/
	int version;		/* compiler version			*/
	int unique;		/* system unique id			*/
	int broken;		/* broken system (do not use) flag	*/
	struct in6_addr addr6;	/* remote IPv6 address			*/
	char fqdn[0];		/* system fqdn (only for web output)	*/
} SERVICE;

/* biarch cross compile mixing table */

static struct
{
	char *server;	/* server architecture				*/
	char *client;	/* client architecture				*/
	char *compile1;	/* 32bit compile switch				*/
	char *compile2;	/* 64bit compile switch				*/
} mix[]=
{
#if defined(__linux__)
	{"x86_64","i686","-m32","-m64"},
#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	{"amd64","i386","-m32","-m64"},
#else
#warning probably some 32bit compile on 64bit peer definition missing here
#endif
	{NULL,NULL,NULL}
};

static int running=0;
static int active=0;
static int flags=0;
static int site=0;
static int ttl=1;
static int auth=0;
static int nomix=0;
static int window=WINDOW;
static sem_t sem;
static pthread_mutex_t bmtx=PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t mtx=PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t ntx=PTHREAD_MUTEX_INITIALIZER;
static pthread_t b;
static pthread_t h;
static int total;
static int jobchange;
static int srvmix;
static int clnmix;
static int version=0;
#if defined(__linux__) || defined(__FreeBSD__)
static int notify;
#else
static int notify[2];
#endif
static clock_t tmo;
static clock_t tms;
static clock_t now;
static clock_t lastbrd;
static int bogoslow=SLOWCPU<<1;
static int noservice=1;
static int justlocal=0;
static int queued;
static SERVICE *entries[16];
static BCASTMSG bcm[2];
static char os[16];
static char arch[16];
static int stamp=0;
static int ccstamp=0;
static int minfree=MINFREE;
static int slowbcast=0;
static int lightload=0;
static int mcast=0;
static int v6=0;
static int unique=0;
static UDP6 udp6;
static UDP udp;
static MD5HMDATA key;
static char *fqdn=NULL;
static char *mcaddr=NULL;
static int undo=0;
static int remcpu=0;
static int preferlocal=1;
static int rmtqueue;

/*
 * bcasthandler
 *
 * input:  unused - unused
 *
 * return: always NULL
 *
 * description:	This routine ins the broadcast/multicast messages 
 *		sending thread and handles the send ratelimiting.
 */

static void *bcasthandler(void *unused)
{
	struct pollfd p;
	BCASTMSG msg[2];

	/* thread default setup */

	dftblock(THREAD);
	dropcaps(NEED_CAST);

	/* data setup */

	p.fd=termget();
	p.events=POLLIN;

	/* do until process termination time */

	while(LIKELY(running))
	{
		/* wait for next message or wake up */

		sem_wait(&sem);

		/* terminate if process termination is pending */

		if(UNLIKELY(!running))break;

		/* clear all pending notifications, copy pending messages to
		   local buffer */

		pthread_mutex_lock(&bmtx);
		while(!sem_trywait(&sem));
		msg[0].stamp=bcm[0].stamp;
		msg[0].len=bcm[0].len;
		memcpy(msg[0].msg,bcm[0].msg,bcm[0].len);
		bcm[0].len=0;
		msg[1].stamp=bcm[1].stamp;
		msg[1].len=bcm[1].len;
		memcpy(msg[1].msg,bcm[1].msg,bcm[1].len);
		bcm[1].len=0;
		pthread_mutex_unlock(&bmtx);

		/* if there is a message in the first queue */

		if(msg[0].len)
		{
			/* skip message if compiler has changed */

			pthread_mutex_lock(&mtx);
			if(msg[0].stamp!=ccstamp)
			{
				pthread_mutex_unlock(&mtx);
				goto next;
			}
			pthread_mutex_unlock(&mtx);

			/* add authentication to message */

			addauth(&key,msg[0].msg,msg[0].len,auth);

			/* send message */

			pthread_mutex_lock(&ntx);
			if(SWITCHES_SUPPORT_MLDV2(v6))
				udp6send(&udp6,1,msg[0].msg,msg[0].len);
			else udpsend(&udp,-1,msg[0].msg,msg[0].len);
			pthread_mutex_unlock(&ntx);

			/* enforce delay after message sending */

			if(UNLIKELY(poll(&p,1,slowbcast?100:10)))
				if(LIKELY(p.revents&POLLIN))continue;
		}

		/* if there is a message in the second queue */

next:		if(msg[1].len)
		{
			/* skip message if compiler has changed */

			pthread_mutex_lock(&mtx);
			if(msg[1].stamp!=ccstamp)
			{
				pthread_mutex_unlock(&mtx);
				continue;
			}
			pthread_mutex_unlock(&mtx);

			/* add authentication to message */

			addauth(&key,msg[1].msg,msg[1].len,auth);

			/* send message */

			pthread_mutex_lock(&ntx);
			if(SWITCHES_SUPPORT_MLDV2(v6))
				udp6send(&udp6,1,msg[1].msg,msg[1].len);
			else udpsend(&udp,-1,msg[1].msg,msg[1].len);
			pthread_mutex_unlock(&ntx);

			/* enforce delay after message sending */

			if(UNLIKELY(poll(&p,1,slowbcast?100:10)))
				if(LIKELY(p.revents&POLLIN))continue;
		}
	}

	/* terminate thread */

	pthread_exit(NULL);
}

/*
 * selector
 *
 * input:  e1 - first remote service
 *	   e2 - second remote service
 *
 * return: -1 if first service preferred, 0 if equal, 1 if second service
 *
 * description:	This routine compares the two given remote services
 *		and selects the most preferrable one.
 */

static int selector(SERVICE *e1,SERVICE *e2)
{
	int v1;
	int v2;

	/* compiler/network version mismatch (shouldn't happen) */


	if((UNLIKELY(e1->version!=version)))
	{
		if((UNLIKELY(e2->version!=version)))return 0;
		return 1;
	}

	if((UNLIKELY(e2->version!=version)))return -1;

	/* broken (do not use) service */

	if((UNLIKELY(e1->broken)))
	{
		if((UNLIKELY(e2->broken)))return 0;
		return 1;
	}

	if((UNLIKELY(e2->broken)))return -1;

	/* expiring service definition (aged, no new data received) */

	if(UNLIKELY(now-e1->stamp>tms))
	{
		if((UNLIKELY(now-e2->stamp>tms)))return 0;
		else return 1;
	}

	if(UNLIKELY(now-e2->stamp>tms))return -1;

	/* insufficient free memory */

	if(UNLIKELY(e1->memfree<e1->minfree))
	{
		if(UNLIKELY(e2->memfree<e2->minfree))return 0;
		else return 1;
	}

	if(UNLIKELY(e2->memfree<e2->minfree))return -1;

	/* amount of pending local compiles */

	if(e1->pending>=(e1->nrcpu>>1))
	{
		if(e2->pending>=(e2->nrcpu>>1))return 0;
		else return 1;
	}

	if(e2->pending>=(e2->nrcpu>>1))return -1;

	/* amount of pending remote compiles */

	if(e1->rmtqueue>=e1->nrcpu)
	{
		if(e2->rmtqueue>=e2->nrcpu)return 0;
		else return 1;
	}

	if(e2->rmtqueue>=e2->nrcpu)return -1;

	/* data comparison calculation with respect to system light load flag */

	v1=e1->sum+e1->pending+e1->rmtqueue;
	if(e1->bogomips<bogoslow)v1+=e1->nrcpu;
	if(UNLIKELY(e1->flags&HINT_LIGHT))v1<<=1;
	v1+=(UNLIKELY(e1->sum>e1->jobs)?e1->sum-e1->jobs:0)+e1->running+
		e1->submitted;
	v1*=e2->nrcpu;

	v2=e2->sum+e2->pending+e2->rmtqueue;
	if(e2->bogomips<bogoslow)v2+=e2->nrcpu;
	if(UNLIKELY(e2->flags&HINT_LIGHT))v2<<=1;
	v2+=(UNLIKELY(e2->sum>e2->jobs)?e2->sum-e2->jobs:0)+e2->running+
		e2->submitted;
	v2*=e1->nrcpu;

	/* if preferance points differ use them for service selection */

	if(v1<v2)return -1;
	else if(v1>v2)return 1;

	/* differenciate preferring fast systems */

	if(e1->bogomips<bogoslow&&e2->bogomips>=bogoslow)return 1;
	if(e1->bogomips>=bogoslow&&e2->bogomips<bogoslow)return -1;

	/* finally try with the available core count */

	if(e1->nrcpu>e2->nrcpu)return -1;
	else if(e1->nrcpu<e2->nrcpu)return 1;

	/* assume both services are of identical preferance */

	else return 0;
}

/*
 * insert
 *
 * input: e1 - service chain
 *	  e2 - service to be inserted
 *
 * description:	This routine inserts the given service into the given
 *		service chain with priority sorting - the most
 *		preferable service is at the front of the chain.
 */

static inline void insert(SERVICE **e1,SERVICE *e2)
{
	for(e2->next=NULL;;e1=&(*e1)->next)switch((long)(*e1))
	{
	default:
		if(selector(*e1,e2)<0)
			continue;
		e2->next=*e1;
	case 0:
		*e1=e2;
		return;
	}
}

/*
 * enq
 *
 * input: msg - message data
 *	  len - message length
 *	  idx - message queue selector
 *
 * description:	This routine enqueues the given broadcast/multicast message
 *		to the given message queue. It may overwrite a previous
 *		pending message which is ok as older data are replaced
 *		by newer ones in this case.
 */

static inline void enq(void *msg,int len,int idx)
{
	pthread_mutex_lock(&bmtx);
	pthread_mutex_lock(&mtx);
	bcm[idx].stamp=ccstamp;
	pthread_mutex_unlock(&mtx);
	bcm[idx].len=len;
	memcpy(bcm[idx].msg,msg,len);
	sem_post(&sem);
	pthread_mutex_unlock(&bmtx);
}

/*
 * grpcheck
 *
 * input:  msgflags - the message flags field
 *
 * return: 0 if group access is allowed, else -1
 *
 * description:	This routine checks, if the message belongs to a group
 *		we should process.
 */

static inline int grpcheck(int msgflags,int msgcode)
{
	switch(msgcode)
	{
	case QUERY:
		if(msgflags&HINT_G0ACCESS)if(!(flags&HINT_GROUP1))return 0;
		if(msgflags&HINT_G1ACCESS)if(flags&HINT_GROUP1)return 0;
		return -1;

	case RESPONSE:
		if(msgflags&HINT_GROUP1)
		{
			if(!(flags&HINT_G1ACCESS))return -1;
		}
		else if(!(flags&HINT_G0ACCESS))return -1;
		return 0;

	/* sanity only */

	default:return -1;
	}
}

/*
 * msghandler
 *
 * input:  unused - unused
 *
 * return: always NULL
 *
 * description:	This routine is the udp message handling thread.
 */

static void *msghandler(void *unused)
{
	int l;
	int i;
	struct pollfd p[3];
	char bfr[sizeof(RESPONSEMSG)+512];
	RESPONSEMSG *r=(RESPONSEMSG *)bfr;
	int flen;
	clock_t latest=timestamp();
	clock_t curr;
	clock_t tps=sysconf(_SC_CLK_TCK);
	QUERYMSG q;
	SERVICE *e;
	SERVICE *d;
	SERVICE **s;
	SERVICE **t;
#if defined(__linux__)
	uint64_t dummy;
#elif defined(__FreeBSD__)
	struct kevent dummy;
#else
	char dummy;
#endif

	/* thread default setup */

	dftblock(THREAD);
	dropcaps(NEED_CAST);

	/* fqdn setup */

	flen=strlen(fqdn);
	if(UNLIKELY(strlen(fqdn)>511))
	{
		fqdn[511]=0;
		flen=511;
	}

	/* poll data setup */

	if(SWITCHES_SUPPORT_MLDV2(v6))p[0].fd=udp6.udp;
	else p[0].fd=udp.udp;
	p[0].events=POLLIN;
	p[1].fd=termget();
	p[1].events=POLLIN;
	p[2].events=POLLIN;
#if defined(__linux__) || defined(__FreeBSD__)
	p[2].fd=notify;
#else
	p[2].fd=notify[0];
#endif

	/* basic data setup */

	strcpy(q.os,os);
	strcpy(q.arch,arch);
	q.code=htonl(QUERY);

	/* do until the whole process is terminating */

	while(LIKELY(running))
	{
		/* get current time (stamp format) */

		curr=timestamp();

		/* in active mode broadacst/multicast every second */

		pthread_mutex_lock(&mtx);
		if(UNLIKELY(curr-lastbrd>=tps))
		{
			lastbrd+=tps;
			q.version=htonl(version);
			q.flags=htonl(flags|(UNLIKELY(lightload)?HINT_LIGHT:0)|
				(auth?HINT_AUTH:0));
			pthread_mutex_unlock(&mtx);
			if(LIKELY(active&&!justlocal))enq(&q,sizeof(q),0);
		}
		else pthread_mutex_unlock(&mtx);

		/* clean out expired services every second while active */

		if(UNLIKELY(curr-latest>=tps))
		{
			latest+=tps;

			pthread_mutex_lock(&mtx);
			for(l=0;l<16;l++)for(s=&entries[l];*s;)
			{
				if(UNLIKELY(curr-(*s)->stamp>tmo&&
					!(*s)->submitted))
				{
					d=*s;
					*s=d->next;
					remcpu-=d->nrcpu;
					free(d);
					total--;
				}
				else s=&(*s)->next;
			}
			pthread_mutex_unlock(&mtx);
		}

		/* send multicast/broadcast if compile job changed */

		pthread_mutex_lock(&mtx);
		if(jobchange)
		{
			jobchange=0;
			r->jobs=htonl(ipc_get(JOBCOUNT));
			r->sum=htonl(ipc_sum(JOBCOUNT));
			r->pending=htonl(queued);
			r->rmtqueue=htonl(rmtqueue);
			r->flags=htonl(flags|
				(UNLIKELY(lightload)?HINT_LIGHT:0)|
				(auth?HINT_AUTH:0));
			r->version=htonl(version);
			pthread_mutex_unlock(&mtx);
			if(UNLIKELY(!noservice&&!justlocal))
			{
				strcpy(r->os,os);
				strcpy(r->arch,arch);
				r->code=htonl(RESPONSE);
				getstats(MEMFREE_STATS,&l);
				r->memfree=htonl(l);
				r->minfree=htonl(minfree);
				getstats(NUMCPU_STATS,&l);
				r->nrcpu=htonl(l);
				getstats(BOGOMIPS_STATS,&l);
				r->bogomips=htonl(l);
				getstats(RUNNING_STATS,&l);
				r->running=htonl(l);
				r->unique=htonl(unique);
				strcpy(r->fqdn,fqdn);
				enq(r,sizeof(RESPONSEMSG)+flen+1,1);
			}
		}
		else pthread_mutex_unlock(&mtx);

		/* poll for event, wait forever if inactive and no more
		   service definitions around, handle termination notify */

		switch(poll(p,3,!total&&(!active)?-1:100))
		{
		/* event pending (fall through) */

		default:/* termination pennding */

			if(UNLIKELY(p[1].revents&POLLIN))continue;

			/* wake up notification */

			if(p[2].revents&POLLIN)
			{
#if defined(__linux__)
				dummy=read(notify,&dummy,sizeof(dummy));
#elif defined(__FreeBSD__)
				kevent(notify,NULL,0,&dummy,1,NULL);
#else
				dummy=read(notify[0],&dummy,1);
#endif
				continue;
			}

			/* udp message pending */

			if(p[0].revents&POLLIN)break;

		/* no event or error, restart loop */

		case 0:
		case -1:continue;
		}

		/* get udp message */

		if(SWITCHES_SUPPORT_MLDV2(v6))
		{
			if(UNLIKELY((l=udp6recv(&udp6,bfr,sizeof(bfr)))==-1))
				continue;
			if(LIKELY(!site))
			{
				if(UNLIKELY(!linklocal(udp6.addr)))
					continue;
				if(UNLIKELY(udp6.idx!=udp6.srcidx))
					continue;
			}
			if(UNLIKELY(udp6.port!=udp6.srcport))
			{
				WARN("illegal source port %d",udp6.srcport);
				continue;
			}
		}
		else
		{
			if(UNLIKELY((l=udprecv(&udp,bfr,sizeof(bfr)))==-1))
				continue;

			pthread_mutex_lock(&ntx);
			if(UNLIKELY(udp.srcaddr==udp.addr))goto netfail;
			if(LIKELY(!site))
				if(UNLIKELY(udp.net!=(udp.srcaddr&udp.mask)))
					goto netfail;
			if(UNLIKELY(udp.srcport!=udp.port))
			{
				WARN("illegal source port %d",udp.srcport);
netfail:			pthread_mutex_unlock(&ntx);
				continue;
			}
			pthread_mutex_unlock(&ntx);
		}

		/* check message length for required minimum */

		if(UNLIKELY(l<sizeof(QUERYMSG)))
		{
			NOTICE("udp message too short");
			continue;
		}

		/* handle authentication */

		if(!(i=ntohl(r->flags)&HINT_AUTH)&&auth)
			continue;
		else if(i&&!auth)continue;
		else if(chkauth(&key,r,l,i,window))continue;

		/* check, if group permissions allow message processing */

		if(grpcheck(ntohl(r->flags),ntohl(r->code)))continue;

		/* process message according to code */

		switch(ntohl(r->code))
		{
		/* query message */

		case QUERY:

			/* check if we shall reply to query */

			if(UNLIKELY(noservice||justlocal))break;
			if(strcmp(r->os,os))break;
			if(strcmp(r->arch,arch))
			{
				if(srvmix==-1)break;
				if(strcmp(r->arch,mix[srvmix].client))break;
			}

			/* in case of compiler version mismatch remove
			   remote system from service data */

			pthread_mutex_lock(&mtx);
			if(ntohl(r->version)!=version||!r->version)
			{
				if(SWITCHES_SUPPORT_MLDV2(v6))for(l=0,
				    s=&entries[udp6.addr.s6_addr[15]&0xf];*s;)
				{
					if(!memcmp(&(*s)->addr6,&udp6.addr,
						sizeof(udp6.addr)))
					{
						d=*s;
						*s=d->next;
						remcpu-=d->nrcpu;
						free(d);
						total--;
					}
					else s=&(*s)->next;
				}
				else for(l=0,s=&entries[udp.srcaddr&0xf];*s;)
				{
					if((*s)->addr==udp.srcaddr)
					{
						d=*s;
						*s=d->next;
						remcpu-=d->nrcpu;
						free(d);
						total--;
					}
					else s=&(*s)->next;
				}
				pthread_mutex_unlock(&mtx);
				break;
			}

			/* prepare response message */

			r->jobs=htonl(ipc_get(JOBCOUNT));
			r->sum=htonl(ipc_sum(JOBCOUNT));
			r->pending=htonl(queued);
			r->rmtqueue=htonl(rmtqueue);
			r->flags=htonl(flags|
				(UNLIKELY(lightload)?HINT_LIGHT:0)|
				(auth?HINT_AUTH:0));
			pthread_mutex_unlock(&mtx);
			strcpy(r->os,os);
			strcpy(r->arch,arch);
			r->code=htonl(RESPONSE);
			r->version=htonl(version);
			getstats(MEMFREE_STATS,&l);
			r->memfree=htonl(l);
			r->minfree=htonl(minfree);
			getstats(NUMCPU_STATS,&l);
			r->nrcpu=htonl(l);
			getstats(BOGOMIPS_STATS,&l);
			r->bogomips=htonl(l);
			getstats(RUNNING_STATS,&l);
			r->running=htonl(l);
			r->unique=htonl(unique);
			strcpy(r->fqdn,fqdn);

			/* add authentication */

			addauth(&key,r,sizeof(RESPONSEMSG)+flen+1,auth);

			/* send unicast response */

			pthread_mutex_lock(&ntx);
			if(SWITCHES_SUPPORT_MLDV2(v6))
				udp6send(&udp6,0,r,sizeof(RESPONSEMSG)+flen+1);
			else udpsend(&udp,udp.srcaddr,r,
				sizeof(RESPONSEMSG)+flen+1);
			pthread_mutex_unlock(&ntx);
			break;

		/* response message */

		case RESPONSE:

			/* check minimum mesage length */

			if(UNLIKELY(l<sizeof(RESPONSEMSG)+1))
			{
				NOTICE("udp message too short");
				break;
			}

			/* assert that fqdn of message is terminated */

			bfr[l-1]=0;

			/* check if we need to ignore the response */

			if(strcmp(r->os,os))break;
			if(strcmp(r->arch,arch))
			{
				if(clnmix==-1)break;
				if(strcmp(r->arch,mix[clnmix].server))break;
			}

			/* in case of compiler version mismatch remove
			   remote system from service data */

			pthread_mutex_lock(&mtx);
			if(ntohl(r->version)!=version||!r->version)
			{
				if(SWITCHES_SUPPORT_MLDV2(v6))for(l=0,
				    s=&entries[udp6.addr.s6_addr[15]&0xf];*s;)
				{
					if(!memcmp(&(*s)->addr6,&udp6.addr,
						sizeof(udp6.addr)))
					{
						d=*s;
						*s=d->next;
						remcpu-=d->nrcpu;
						free(d);
						total--;
					}
					else s=&(*s)->next;
				}
				else for(l=0,s=&entries[udp.srcaddr&0xf];*s;)
				{
					if((*s)->addr==udp.srcaddr)
					{
						d=*s;
						*s=d->next;
						remcpu-=d->nrcpu;
						free(d);
						total--;
					}
					else s=&(*s)->next;
				}
				pthread_mutex_unlock(&mtx);
				break;
			}
			pthread_mutex_unlock(&mtx);

			/* allocate service definition memory */

			if(UNLIKELY(!(e=malloc(
				sizeof(SERVICE)+l-sizeof(RESPONSEMSG)))))
			{
				NOTICE("malloc failure");
				break;
			}

			/* store set up service definition from message */

			e->next=NULL;
			e->stamp=timestamp();
			if(SWITCHES_SUPPORT_MLDV2(v6))
			{
				e->addr=1;
				e->addr6=udp6.addr;
			}
			else e->addr=udp.srcaddr;
			e->jobs=ntohl(r->jobs);
			e->sum=ntohl(r->sum);
			e->memfree=ntohl(r->memfree);
			e->minfree=ntohl(r->minfree);
			e->nrcpu=ntohl(r->nrcpu);
			e->bogomips=ntohl(r->bogomips);
			e->running=ntohl(r->running);
			e->pending=ntohl(r->pending);
			e->rmtqueue=ntohl(r->rmtqueue);
			e->flags=ntohl(r->flags);
			e->version=ntohl(r->version);
			e->unique=ntohl(r->unique);
			e->broken=0;
			strcpy(e->fqdn,r->fqdn);

			/* remove older service definition of remote system
			   but memorize the amount of submitted jobs to
			   this system */

			pthread_mutex_lock(&mtx);
			if(SWITCHES_SUPPORT_MLDV2(v6))
			    for(l=0,t=s=&entries[udp6.addr.s6_addr[15]&0xf];*s;)
			{
				if(!memcmp(&(*s)->addr6,&udp6.addr,
					sizeof(udp6.addr)))
				{
					d=*s;
					*s=d->next;
					l=d->submitted;
					remcpu-=d->nrcpu;
					free(d);
					total--;
				}
				else s=&(*s)->next;
			}
			else for(l=0,t=s=&entries[udp.srcaddr&0xf];*s;)
			{
				if((*s)->addr==udp.srcaddr)
				{
					d=*s;
					*s=d->next;
					l=d->submitted;
					remcpu-=d->nrcpu;
					free(d);
					total--;
				}
				else s=&(*s)->next;
			}

			/* save submitted job amount */

			e->submitted=l;

			/* update total available core stats */

			remcpu+=e->nrcpu;

			/* insert the service definiton into proper chain */

			insert(t,e);

			/* adjust total service definition count */

			total++;
			pthread_mutex_unlock(&mtx);
			break;
		}
	}

	/* terminate thread */

	pthread_exit(NULL);
}

/*
 * updatemgr
 *
 * input:  mode - th action to be performed
 *	   ptr  - action specific data
 *	   val  - action specific data
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine ist the udp processing module management routine.
 */

int updatemgr(int mode,char *ptr,int val)
{
	int i;
	SERVICE *e;
	clock_t tps;
#if defined(__FreeBSD__)
	struct kevent ev;
#endif
	char mem[256];

	/* handle requested action */

	switch(mode)
	{
	/* star udp processing */

	case UPDATE_START:

		/* sanity handling */

		if(UNLIKELY(running))return 0;
		if(UNLIKELY(!ptr))BAD("internal error");
		if(UNLIKELY(val<1||val>65535))BAD("internal error");
		if(UNLIKELY(!(fqdn=getfqdn())))BAD("can't get fqdn");

		/* create udp socket */

		if(SWITCHES_SUPPORT_MLDV2(v6))
		{
			if(UNLIKELY(udp6socket(ptr,val,&udp6,
				LIKELY(!site)?LAN_MCAST:SITE_MCAST,ttl,mcaddr)))
			{
				free(fqdn);
				BAD("can't create udp socket");
			}
		}
		else
		{
			if(!mcast)i=LAN_BCAST;
			else i=LIKELY(!site)?LAN_MCAST:SITE_MCAST;
			if(UNLIKELY(udpsocket(ptr,val,&udp,i,ttl,mcaddr)))
			{
				free(fqdn);
				BAD("can't create udp socket");
			}
		}

		/* initialize broadcast/multicast thread semaphore */

		sem_init(&sem,0,0);

		/* create wake up event fd */

#if defined(__linux__)
#ifdef SAFEAPI
		if(LIKELY(safeapi()))
		{
			if(UNLIKELY((notify=eventfd(0,EFD_CLOEXEC))==-1))
			{
				i=errno;
				sem_destroy(&sem);
				if(SWITCHES_SUPPORT_MLDV2(v6))udp6close(&udp6);
				else udpclose(&udp);
				free(fqdn);
				BAD("eventfd: %s",errstr(i,mem));
			}
		}
		else
#endif
		{
			shlock();
			if(UNLIKELY((notify=eventfd(0,0))==-1))
			{
				i=errno;
				unlock();
				sem_destroy(&sem);
				if(SWITCHES_SUPPORT_MLDV2(v6))udp6close(&udp6);
				else udpclose(&udp);
				free(fqdn);
				BAD("eventfd: %s",errstr(i,mem));
			}
			fcntl(notify,F_SETFD,FD_CLOEXEC);
			unlock();
		}
#elif defined(__FreeBSD__)
		if(UNLIKELY((notify=kqueue())==-1))
		{
			i=errno;
			sem_destroy(&sem);
			if(SWITCHES_SUPPORT_MLDV2(v6))udp6close(&udp6);
			else udpclose(&udp);
			free(fqdn);
			BAD("kqueue: %s",errstr(i,mem));
		}
		EV_SET(&ev,0x102,EVFILT_USER,EV_ADD|EV_ENABLE|EV_CLEAR,
			NOTE_FFNOP,0,EMPTY);
		if(UNLIKELY(kevent(notify,&ev,1,NULL,0,NULL)==-1))
		{
			i=errno;
			close(notify);
			sem_destroy(&sem);
			if(SWITCHES_SUPPORT_MLDV2(v6))udp6close(&udp6);
			else udpclose(&udp);
			free(fqdn);
			BAD("kevent: %s",errstr(i,mem));
		}
#else
#ifdef SAFEAPI
		if(LIKELY(safeapi()))
		{
			if(UNLIKELY(pipe2(notify,O_CLOEXEC)))
			{
				i=errno;
				sem_destroy(&sem);
				if(SWITCHES_SUPPORT_MLDV2(v6))udp6close(&udp6);
				else udpclose(&udp);
				free(fqdn);
				BAD("pipe2: %s",errstr(i,mem));
			}
		}
		else
#endif
		{
			shlock();
			if(UNLIKELY(pipe(notify)))
			{
				i=errno;
				unlock();
				sem_destroy(&sem);
				if(SWITCHES_SUPPORT_MLDV2(v6))udp6close(&udp6);
				else udpclose(&udp);
				free(fqdn);
				BAD("pipe: %s",errstr(i,mem));
			}
			fcntl(notify[0],F_SETFD,FD_CLOEXEC);
			fcntl(notify[1],F_SETFD,FD_CLOEXEC);
			unlock();
		}
#endif
		/* basic global data setup */

		memset(entries,0,sizeof(entries));
		remcpu=0;
		total=0;
		jobchange=0;
		running=1;
		tps=sysconf(_SC_CLK_TCK);
		tmo=tps<<5;
		tms=tps<<4;
		bcm[0].len=0;
		bcm[1].len=0;
		active=0;
		queued=0;
		rmtqueue=0;
		if(UNLIKELY(nomix))srvmix=clnmix=-1;
		else for(srvmix=-1,clnmix=-1,i=0;mix[i].server;i++)
		{
			if(!strcmp(mix[i].server,arch))srvmix=i;
			if(!strcmp(mix[i].client,arch))clnmix=i;
		}
		lastbrd=timestamp();

		/* start broadcast/multicast handling thread */

		if(UNLIKELY(pthread_create(&b,NULL,bcasthandler,NULL)))
		{
			i=errno;
			running=0;
			sem_destroy(&sem);
			if(SWITCHES_SUPPORT_MLDV2(v6))udp6close(&udp6);
			else udpclose(&udp);
			free(fqdn);
			BAD("pthread_create: %s",errstr(i,mem));
		}

		/* create udp processing thead */

		if(UNLIKELY(pthread_create(&h,NULL,msghandler,NULL)))
		{
			i=errno;
			running=0;
			sem_post(&sem);
			pthread_join(b,NULL);
			sem_destroy(&sem);
			if(SWITCHES_SUPPORT_MLDV2(v6))udp6close(&udp6);
			else udpclose(&udp);
			free(fqdn);
			BAD("pthread_create: %s",errstr(i,mem));
		}

		/* signal success */

		return 0;

	/* stop udp processing */

	case UPDATE_STOP:

		/* sanity handling */

		if(UNLIKELY(!running))return 0;

		/* notify of termination */

		running=0;
		sem_post(&sem);

		/* wait for threads to terminate */

		pthread_join(b,NULL);
		pthread_join(h,NULL);
		pthread_mutex_lock(&mtx);

		/* clean up ipc job stats */

		while(undo--)	
		{
			pthread_mutex_unlock(&mtx);
			ipc_adj(JOBCOUNT,-1);
			ipc_lock(JOBCOND);
			ipc_signal(JOBCOND);
			ipc_unlock(JOBCOND);
			pthread_mutex_lock(&mtx);
		}
		pthread_mutex_unlock(&mtx);

		/* close notification handles */

#if defined(__linux__) || defined(__FreeBSD__)
		close(notify);
#else
		close(notify[0]);
		close(notify[1]);
#endif
		/* free service definitions */

		for(i=0;i<16;i++)while(entries[i])
		{
			e=entries[i];
			entries[i]=e->next;
			free(e);
		}

		/* close udp socket */

		if(SWITCHES_SUPPORT_MLDV2(v6))udp6close(&udp6);
		else udpclose(&udp);

		/* release qorking memory */

		free(fqdn);

		/* clear authentication data */

		clrauth(&key,NULL);

		/* signal success */

		return 0;

	/* network data change */

	case UPDATE_NETWORK:

		/* sanity checks */

		if(UNLIKELY(!running))BAD("internal error");
		pthread_mutex_lock(&ntx);
		pthread_mutex_lock(&mtx);

		/* update network data, act according to result */

		switch(SWITCHES_SUPPORT_MLDV2(v6)?
			udp6newdev(ptr,&udp6):udpnewdev(ptr,&udp))
		{
		/* update error */

		case -1:pthread_mutex_unlock(&mtx);
			pthread_mutex_unlock(&ntx);
			return -1;

		/* data updated (fall through) */

		default:/* get new stamp value */

			stamp=(stamp+1)&0xffff;

			/* reset all remote service data */

			for(i=0;i<16;i++)while(entries[i])
			{
				e=entries[i];
				entries[i]=e->next;
				free(e);
			}
			remcpu=0;
			total=0;

		/* no update happened */

		case 0:	pthread_mutex_unlock(&mtx);
			pthread_mutex_unlock(&ntx);
			return 0;
		}

	/* store system architecture */

	case UPDATE_ARCH:
		if(UNLIKELY(running))BAD("internal error");
		if(UNLIKELY(strlen(ptr)>=sizeof(arch)))BAD("internal error");
		strcpy(arch,ptr);
		return 0;

	/* store system os */

	case UPDATE_OS:
		if(UNLIKELY(running))BAD("internal error");
		if(UNLIKELY(strlen(ptr)>=sizeof(os)))BAD("internal error");
		strcpy(os,ptr);
		return 0;

	/* store minimum free memory */

	case UPDATE_MEMFREE:
		if(UNLIKELY(running))BAD("internal error");
		if(UNLIKELY(val<1))BAD("internal error");
		minfree=val;
		return 0;

	/* store IPv6 mode */

	case UPDATE_USEV6:
		if(UNLIKELY(running))BAD("internal error");
		if(UNLIKELY(val<0||val>1))BAD("internal error");
		v6=val;
		return 0;

	/* store multicast mode */

	case UPDATE_MCAST:
		if(UNLIKELY(running))BAD("internal error");
		if(UNLIKELY(val<0||val>1))BAD("internal error");
		mcast=val;
		return 0;

	/* store sitewide operation mode */

	case UPDATE_SITE:
		if(UNLIKELY(running))BAD("internal error");
		if(UNLIKELY(val<0||val>1))BAD("internal error");
		site=val;
		return 0;

	/* store message ttl */

	case UPDATE_TTL:
		if(UNLIKELY(running))BAD("internal error");
		if(UNLIKELY(val<1||val>255))BAD("internal error");
		ttl=val;
		return 0;

	/* set up authentication */

	case UPDATE_AUTH:
		if(UNLIKELY(running))BAD("internal error");
		if(UNLIKELY(val<1))BAD("internal error");
		setauth(&key,NULL,ptr,val);
		auth=1;
		return 0;

	/* set up authentication window */

	case UPDATE_WINDOW:
		if(UNLIKELY(running))BAD("internal error");
		if(UNLIKELY(val<0||val>300))BAD("internal error");
		window=val;
		return 0;

	/* enable/disable biarch mixing */

	case UPDATE_NOMIX:
		if(UNLIKELY(running))BAD("internal error");
		if(UNLIKELY(val<0||val>1))BAD("internal error");
		nomix=val;
		return 0;

	/* store unique id */

	case UPDATE_UNIQUE:
		if(UNLIKELY(running))BAD("internal error");
		unique=val;
		return 0;

	/* store user defined muticast address */

	case UPDATE_MCADDR:
		if(UNLIKELY(running))BAD("internal error");
		mcaddr=ptr;
		return 0;

	/* store remote compile preferance */

	case UPDATE_USERMT:
		if(UNLIKELY(running))BAD("internal error");
		if(UNLIKELY(val<0||val>1))BAD("internal error");
		preferlocal=val;
		return 0;

	/* sanity check */

	default:BAD("internal error");
	}
}

/*
 * unpend
 *
 * input: unused - unused
 *
 * description:	This routine corrects the queued remote compile count in
 *		case the remote compile thread is cancelled while waiting.
 */

void unpend(void *unused)
{
	pthread_mutex_lock(&mtx);
	rmtqueue--;
	pthread_mutex_unlock(&mtx);
}

/*
 * update_mode
 *
 * input:  how       - action selector
 *	   val       - action dependent data
 *	   currstamp - compiler change mode identifier memory or NULL
 *	   ...       - action dependent parameters
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine handles all runtime data changes affecting
 *		udp message processing.
 */

int update_mode(int how,int val,int *currstamp,...)
{
	int state;
	int nrcpu;
	int vers;
	int flgadd;
	SERVICE *e;
	SERVICE **s;
	SERVICE **t;
	QUERYMSG q;
	va_list ap;
	struct in6_addr *addr6;
#if defined(__linux__)
	uint64_t dummy;
#elif defined(__FreeBSD__)
	struct kevent dummy;
#else
	char dummy;
#endif

	/* sanity check */

	if(UNLIKELY(!running))BAD("internal error");

	/* act according to selected action */

	switch(how)
	{
	/* enable udp updating */

	case UPDATE_ENABLE:

		/* done if already enabled */

		if(active)return 0;

		/* get system data */

		pthread_mutex_lock(&mtx);
		lastbrd=timestamp();
		vers=version;
		flgadd=(UNLIKELY(lightload)?HINT_LIGHT:0)|(auth?HINT_AUTH:0);
		pthread_mutex_unlock(&mtx);

		/* mark mode as active */

		active=1;

		/* done if local only operation */

		if(UNLIKELY(justlocal))return 0;

		/* create query message */

		strcpy(q.os,os);
		strcpy(q.arch,arch);
		q.code=htonl(QUERY);
		q.version=htonl(vers);
		q.flags=htonl(flags|flgadd);

		/* enqueue query message */

		enq(&q,sizeof(q),0);

		/* send wakeup notification */

#if defined(__linux__)
		dummy=1;
		dummy=write(notify,&dummy,sizeof(dummy));
#elif defined(__FreeBSD__)
		EV_SET(&dummy,0x102,EVFILT_USER,0,NOTE_TRIGGER,0,EMPTY);
		kevent(notify,&dummy,1,NULL,0,NULL);
#else
		dummy=write(notify[1],&dummy,1);
#endif
		/* signal success */

		return 0;

	/* disable udp updating */

	case UPDATE_DISABLE:

		/* done if already disabled */

		if(!active)return 0;

		/* mark mode as inactive */

		active=0;

		/* signal success */

		return 0;

	/* remote compile pending (fall through) */

	case JOB_PENDING:

		/* lock job condition mutex */

		ipc_lock(JOBCOND);

		/* reset pending jobs added flag */

		flgadd=0;

		/* in case of remote job wait for compile slot */

		if(LIKELY(!val))while(1)
		{
			/* get online cores */

			getstats(NUMCPU_STATS,&nrcpu);

			/* get job count */

			pthread_mutex_lock(&mtx);
			state=ipc_sum(JOBCOUNT);

			/* exit loop in case of free compile slot */

			if(state<=nrcpu+2)
			{
				if(flgadd)rmtqueue--;
				pthread_mutex_unlock(&mtx);
				break;
			}

			/* increase queued remote job count */

			if(!flgadd)
			{
				rmtqueue++;
				flgadd=1;
			}

			/* wait for other job to be done and then retry */

			pthread_mutex_unlock(&mtx);
			pthread_cleanup_push(unpend,NULL);
			ipc_sigwait(JOBCOND);
			pthread_cleanup_pop(0);
		}

		/* unlock job condition mutex */

		ipc_unlock(JOBCOND);

	/* new compile job */

	case JOB_NEW:

		/* local compile job */

		pthread_mutex_lock(&mtx);
		if(!val)
		{
			/* update data */

			undo++;
			jobchange=1;

			/* notify/wake udp processing thread */

#if defined(__linux__)
			dummy=1;
			dummy=write(notify,&dummy,sizeof(dummy));
#elif defined(__FreeBSD__)
			EV_SET(&dummy,0x102,EVFILT_USER,0,NOTE_TRIGGER,0,EMPTY);
			kevent(notify,&dummy,1,NULL,0,NULL);
#else
			dummy=write(notify[1],&dummy,1);
#endif
			pthread_mutex_unlock(&mtx);

			/* update stats */

			ipc_adj(JOBCOUNT,1);

			/* signal success */

			return 0;
		}

		/* sanity check */

		else if(!currstamp)
		{
			pthread_mutex_unlock(&mtx);
			return -1;
		}

		/* remote compile job */

		else
		{
			/* store current state stamp */

			*currstamp=stamp;

			/* add job to proper service entry */

			if(SWITCHES_SUPPORT_MLDV2(v6))
			{
				va_start(ap,currstamp);
				addr6=va_arg(ap,void *);
				for(t=s=&entries[addr6->s6_addr[15]&0xf];*s;
					s=&(*s)->next)if(!memcmp(&(*s)->addr6,
						addr6,sizeof((*s)->addr6)))
				{
					e=*s;
					*s=e->next;
					e->submitted++;
					insert(t,e);
					break;
				}
				va_end(ap);
			}
			else for(t=s=&entries[val&0xf];*s;s=&(*s)->next)
				if((*s)->addr==val)
			{
				e=*s;
				*s=e->next;
				e->submitted++;
				insert(t,e);
				break;
			}
			pthread_mutex_unlock(&mtx);

			/* signal success */

			return 0;
		}

	/* compile job complete */

	case JOB_DONE:

		/* local job */

		pthread_mutex_lock(&mtx);
		if(!val)
		{
			/* update data */

			undo--;
			jobchange=1;

			/* notify/wake udp processing thread */

#if defined(__linux__)
			dummy=1;
			dummy=write(notify,&dummy,sizeof(dummy));
#elif defined(__FreeBSD__)
			EV_SET(&dummy,0x102,EVFILT_USER,0,NOTE_TRIGGER,0,EMPTY);
			kevent(notify,&dummy,1,NULL,0,NULL);
#else
			dummy=write(notify[1],&dummy,1);
#endif
			pthread_mutex_unlock(&mtx);

			/* update stats */

			ipc_adj(JOBCOUNT,-1);

			/* wake a waiter for a compile slot */

			ipc_lock(JOBCOND);
			ipc_signal(JOBCOND);
			ipc_unlock(JOBCOND);

			/* signal success */

			return 0;
		}

		/* sanity check */

		else if(!currstamp)
		{
			pthread_mutex_unlock(&mtx);
			return -1;
		}

		/* remote job for which we have service data */

		else if(*currstamp==stamp)
		{
			/* remove job from porper service */

			if(SWITCHES_SUPPORT_MLDV2(v6))
			{
				va_start(ap,currstamp);
				addr6=va_arg(ap,void *);
				for(t=s=&entries[addr6->s6_addr[15]&0xf];*s;
					s=&(*s)->next)if(!memcmp(&(*s)->addr6,
						addr6,sizeof((*s)->addr6)))
				{
					e=*s;
					*s=e->next;
					e->submitted--;
					insert(t,e);
					break;
				}
				va_end(ap);
			}
			else for(t=s=&entries[val&0xf];*s;s=&(*s)->next)
				if((*s)->addr==val)
			{
				e=*s;
				*s=e->next;
				e->submitted--;
				insert(t,e);
				break;
			}
			pthread_mutex_unlock(&mtx);

			/* signal success */

			return 0;
		}

	/* adjust queued job count */

	case QUEUED_JOBS:
		pthread_mutex_lock(&mtx);
		queued=val;
		pthread_mutex_unlock(&mtx);
		return 0;

	/* set new fast/slow cpu limit */

	case FAST_CPU_MHZ:
		pthread_mutex_lock(&mtx);
		bogoslow=val<<1;
		pthread_mutex_unlock(&mtx);
		return 0;

	/* set up client only mode */

	case CLIENT_ONLY:
		pthread_mutex_lock(&mtx);
		noservice=val;
		pthread_mutex_unlock(&mtx);
		return 0;

	/* set up local only mode */

	case LOCAL_ONLY:
		pthread_mutex_lock(&mtx);
		justlocal=val;
		pthread_mutex_unlock(&mtx);
		return 0;

	/* set up broadcast/multicast speed */

	case BCAST_SPEED:
		pthread_mutex_lock(&mtx);
		slowbcast=val;
		pthread_mutex_unlock(&mtx);
		return 0;

	/* set up system load mode */

	case SYSTEM_LOAD:
		pthread_mutex_lock(&mtx);
		lightload=val;
		pthread_mutex_unlock(&mtx);
		return 0;

	/* broken remote system notification */

	case SYSTEM_BROKEN:

		/* search remote service entry and mark it as broken */

		pthread_mutex_lock(&mtx);
		if(SWITCHES_SUPPORT_MLDV2(v6))
		{
			va_start(ap,currstamp);
			addr6=va_arg(ap,void *);
			for(t=s=&entries[addr6->s6_addr[15]&0xf];*s;
				s=&(*s)->next)if(!memcmp(&(*s)->addr6,
					addr6,sizeof((*s)->addr6)))
			{
				e=*s;
				*s=e->next;
				e->broken=1;
				insert(t,e);
				break;
			}
			va_end(ap);
		}
		else for(t=s=&entries[val&0xf];*s;s=&(*s)->next)
			if((*s)->addr==val)
		{
			e=*s;
			*s=e->next;
			e->broken=1;
			insert(t,e);
			break;
		}
		pthread_mutex_unlock(&mtx);

		/* signal success */

		return 0;

	/* set up new processing flags */

	case CFG_FLAGS:
		pthread_mutex_lock(&mtx);
		flags=val;
		pthread_mutex_unlock(&mtx);
		return 0;

	/* new compiler version or compiler gone */

	case GCC_VERSION:

		/* add network format to compiler version if compiler avail. */

		if(val)val|=NETFORMAT<<16;
		pthread_mutex_lock(&ntx);
		pthread_mutex_lock(&mtx);

		/* store compiler data and up compiler stamp */

		version=val;
		ccstamp++;

		/* done in local or client only mode */

		if(UNLIKELY(justlocal||noservice))
		{
			pthread_mutex_unlock(&mtx);
			pthread_mutex_unlock(&ntx);
			return 0;
		}

		/* send out query with the new compiler data to flush
		   remote service caches */

		strcpy(q.os,os);
		strcpy(q.arch,arch);
		q.code=htonl(QUERY);
		q.version=htonl(val);
		q.flags=htonl(flags|(UNLIKELY(lightload)?HINT_LIGHT:0)|
			(auth?HINT_AUTH:0));
		addauth(&key,&q,sizeof(QUERYMSG),auth);
		if(SWITCHES_SUPPORT_MLDV2(v6))
			udp6send(&udp6,1,&q,sizeof(QUERYMSG));
		else udpsend(&udp,-1,&q,sizeof(QUERYMSG));
		pthread_mutex_unlock(&mtx);
		pthread_mutex_unlock(&ntx);

		/* signal success */

		return 0;

	/* sanity check */

	default:BAD("internal error");
	}
}

/*
 * select_service
 *
 * input:  pending   . amount of pending compile jobs
 *	   localonly - 1 of local compile is necessary
 *	   addr6     - remote system ipv6 address
 *	   flags     - processing flags to be used
 *	   unique    - remote system unique id
 *
 * return: -1 if no compile is currently possible
 *	    0 in case of local compile
 *	    any other value is either an IPv4 address or a hint to
 *	    use the IPv6 address as remote compile is selected
 *
 * description:	This routine selects a compile destination for the
 *		next compile job.
 */

int select_service(int pending,int localonly,struct in6_addr *addr6,int *flags,
	int *unique)
{
	int totcpu;
	int addr;
	int ladd;
	int radd;
	int ld;
	int rd;
	int i;
	int j;
	int light;
	SERVICE l;
	SERVICE r;

	/* get local stats */

	l.addr=0;
	getstats(MEMFREE_STATS,&l.memfree);
	getstats(NUMCPU_STATS,&l.nrcpu);
	getstats(BOGOMIPS_STATS,&l.bogomips);
	getstats(RUNNING_STATS,&l.running);
	pthread_mutex_lock(&mtx);
	light=lightload;
	l.jobs=ipc_get(JOBCOUNT);
	l.sum=ipc_sum(JOBCOUNT);
	l.pending=queued;
	l.rmtqueue=rmtqueue;
	totcpu=l.nrcpu+remcpu;
	pthread_mutex_unlock(&mtx);
	l.submitted=l.jobs;

	/* set up limits according to light or regular load mode */

	if(UNLIKELY(light))switch(l.nrcpu)
	{
	case 1:	ladd=0;
		ld=l.nrcpu;
		break;

	case 2:	ladd=-1;
		ld=l.nrcpu;
		break;

	default:ladd=-2;
		ld=l.nrcpu-1;
		break;
	}
	else
	{
		ladd=l.bogomips<bogoslow?1:2;
		ld=l.nrcpu<<1;
	}

	/* continue with local processing if no remote compile is possible */

	if(localonly)goto local;

	/* select best remote service available, if none available process
	   as local only compile would have been selected */

	pthread_mutex_lock(&mtx);
	if(!total)goto fail;

	for(j=-1,i=0;i<16;i++)if(entries[i])
	{
		if(j<0)j=i;
		else if(selector(entries[i],entries[j])<0)j=i;
	}

	now=timestamp();

	if(now-entries[j]->stamp>tms||entries[j]->memfree<entries[j]->minfree)
		goto fail;

	if(UNLIKELY(entries[j]->version!=version))goto fail;

	if(UNLIKELY(entries[j]->broken))
	{
		/* local only compile */

fail:		pthread_mutex_unlock(&mtx);

		/* no compile if insufficient free memory */

local:		if(UNLIKELY(l.memfree<minfree))return -1;

		/* local compile if no system overload */

		else if(l.sum<l.nrcpu+ladd)return 0;
		else if(l.sum<ld&&l.running<ld)return 0;

		/* system overload, no compile */

		else return -1;
	}

	/* copy selected remote service entry */

	r=*entries[j];
	pthread_mutex_unlock(&mtx);

	/* no remote compile if remote system is loaded locally */

	if(r.pending>=(r.nrcpu>>1))goto local;

	/* no remote compile if remote system is loaded remotely */

	if(r.rmtqueue>=r.nrcpu)goto local;

	/* set up remote limits according to light or regular load mode */

	if(UNLIKELY(light||(r.flags&HINT_LIGHT)))switch(r.nrcpu)
	{
	case 1:	radd=0;
		rd=r.nrcpu;
		break;

	case 2:	radd=-1;
		rd=r.nrcpu;
		break;

	default:radd=-2;
		rd=r.nrcpu-1;
		break;
	}
	else
	{
		radd=r.bogomips<bogoslow?2:4;
		rd=r.nrcpu<<1;

		if(r.bogomips>=bogoslow)
		{
			rd+=r.nrcpu;
			if(r.nrcpu<=2)rd+=r.nrcpu;
		}
	}

	/* preset result to no compile possible */

	addr=-1;

	/* insufficient free local memory */

	if(UNLIKELY(l.memfree<minfree))
	{
		/* no remote compile possible */

		if(r.submitted>=r.nrcpu+radd)goto out;
		else if(r.running>=rd)goto out;

		/* use remote compile */

		addr=r.addr;
		goto out;
	}

	/* local compile possible */

	if(l.submitted<l.nrcpu+ladd)
	{
		/* remote compile possible */

		if(r.submitted<r.nrcpu+radd)
		{
			/* choose between local and remote compile */

			if(pending<=l.nrcpu&&preferlocal)addr=0;
			else if(l.bogomips<bogoslow&&r.bogomips>=bogoslow)
				addr=r.addr;
			else if(l.bogomips>=bogoslow&&r.bogomips<bogoslow)
				addr=0;
			else if(l.nrcpu+ladd-l.submitted<
				r.nrcpu+radd-r.submitted)addr=r.addr;
			else addr=0;
		}

		/* use local compile */

		else addr=0;
		goto out;
	}

	/* remote compile possible */

	if(r.submitted<r.nrcpu+radd)
	{
		/* use remote compile */

		addr=r.addr;
		goto out;
	}

	/* remote job submission possible and sensible */

	if(r.submitted<rd&&r.sum<rd&&r.running<rd&&
		(r.bogomips>=bogoslow||pending>(totcpu<<2)))
	{
		/* use remote compile */

		addr=r.addr;
		goto out;
	}

	/* local job submission possible and sensible */

	if(l.submitted<ld&&l.running<ld&&
		(l.bogomips>=bogoslow||pending>(totcpu<<2)))
	{
		/* use local compile */

		addr=0;
		goto out;
	}

	/* reset processing flags */

out:	*flags=0;

	/* act according to compile destination */

	switch(addr)
	{
	/* remote compile (fall through) */

	default:*flags=r.flags&(HINT_RMASK|HINT_CMASK);
		*unique=r.unique;
		if(SWITCHES_SUPPORT_MLDV2(v6))*addr6=r.addr6;

	/* local compile or no compile possible */

	case 0:
	case -1:return addr;
	}
}

/*
 * getswitch1
 *
 * return: the 32bit compile selection option or NULL
 *
 * description:	This routine returns the option required for a 32bit
 *		compile on a biarch architecture or NULL if such an
 *		option is not defined for the current architecture.
 */

char *getswitch1(void)
{
	if(UNLIKELY(!running))return NULL;
	if(srvmix==-1)return NULL;
	return mix[srvmix].compile1;
}

/*
 * getswitch2
 *
 * return: the 64bit compile selection option or NULL
 *
 * description:	This routine returns the option required for a 64bit
 *		compile on a biarch architecture or NULL if such an
 *		option is not defined for the current architecture.
 */

char *getswitch2(void)
{
	if(UNLIKELY(!running))return NULL;
	if(srvmix==-1)return NULL;
	return mix[srvmix].compile2;
}

/*
 * isclient
 *
 * return: 1 if we are a possible biarch compile client, else 0
 *
 * description:	This routine checks, if we are a biarch compile client
 *		(e.g. i686).
 */

int isclient(void)
{
	if(clnmix!=-1)return 1;
	return 0;
}

/*
 * safe_cancel
 *
 * intput: h - a pthread handle for the thread to be cancelled
 *
 * description:	This routine cancels a thread with the proper locks
 *		held to assert that cancellation happens only at
 *		acceptable code areas.
 */

void safe_cancel(pthread_t h)
{
		ipc_lock(JOBCOND);
		pthread_mutex_lock(&mtx);
		pthread_cancel(h);
		pthread_mutex_unlock(&mtx);
		ipc_unlock(JOBCOND);
}

/*
 * htmlstats
 *
 * return: html statistics or NULL in case of error
 *
 * description:	This routine collects service statistics as html data.
 *		It is only used by the status mini web server.
 */

char *htmlstats(void)
{
	int i;
	int blen;
	int len;
	int a;
	clock_t curr;
	clock_t tps;
	char *bfr;
	void *tmp;
	SERVICE *e;
	SERVICE l;
	char addr[INET6_ADDRSTRLEN];

	/* get current time */

	tps=sysconf(_SC_CLK_TCK);
	curr=timestamp();

	/* allocate output memory */

	blen=8192;
	if(UNLIKELY(!(bfr=malloc(8192))))
	{
		NOTICE("malloc failure");
		return NULL;
	}

	/* get local system stats */

	pthread_mutex_lock(&mtx);
	l.jobs=ipc_get(JOBCOUNT);
	l.pending=queued;
	l.rmtqueue=rmtqueue;
	pthread_mutex_unlock(&mtx);
	getstats(MEMFREE_STATS,&l.memfree);
	getstats(NUMCPU_STATS,&l.nrcpu);
	getstats(BOGOMIPS_STATS,&l.bogomips);
	getstats(RUNNING_STATS,&l.running);

	/* prepare table header and local stats */

	len=sprintf(bfr,"<html><head><title>netccd Status Page</title></head>"
		"<body><h1>netccd Status:</h1><table border=\"1\"><tr>"
		"<th>Host</th><th>IP</th><th>Update<br>(seconds ago)</th>"
		"<th>CPU<br>Speed (MHz)<th>CPU(s)</th>"
		"<th>Memory<br>free (MB)</th><th>Job(s)<br>submitted</th>"
		"<th>Job(s)<br>running</th><th>Job(s)<br>queued</th>"
		"<th>Job(s)<br>pending</th><th>Process(es)<br>running</th>"
		"</tr><tr><td>%s</td><td>%s</td>"
		"<td>0</td><td>%d</td><td>%d</td><td>%d</td><td>%d</td>"
		"<td>%d</td><td>%d</td><td>%d</td><td>%d</td></tr>",fqdn,
		SWITCHES_SUPPORT_MLDV2(v6)?"::1":"127.0.0.1",l.bogomips>>1,
		l.nrcpu,l.memfree,l.jobs,l.jobs,l.rmtqueue,l.pending,l.running);

	/* for the complete hash table */

	for(i=0;i<16;i++)
	{
		/* process all remote service definitions */

		pthread_mutex_lock(&mtx);
		for(e=entries[i];e;e=e->next)
		{
			/* assert that there's sufficient memory */

			while(blen-len<1024)
			{
				if(UNLIKELY(!(tmp=realloc(bfr,blen+8192))))
				{
					pthread_mutex_unlock(&mtx);
					free(bfr);
					return NULL;
				}
				bfr=tmp;
				blen+=8192;
			}

			/* add service definition to stats */

			a=htonl(e->addr);
			len+=sprintf(bfr+len,"<tr><td>%s</td>"
				"<td>%s</td><td>%d</td><td>%d</td>"
				"<td>%d</td><td>%d</td><td>%d</td><td>%d</td>"
				"<td>%d</td><td>%d</td><td>%d</td></tr>",
				e->fqdn,
				SWITCHES_SUPPORT_MLDV2(v6)?
					inet_ntop(AF_INET6,&e->addr6,addr,
					sizeof(addr)):inet_ntop(AF_INET,&a,
					addr,sizeof(addr)),
				(int)((curr-e->stamp+(tps>>1))/tps),
				e->bogomips>>1,e->nrcpu,e->memfree,
				e->submitted,e->jobs,e->rmtqueue,e->pending,
				e->running);
		}
		pthread_mutex_unlock(&mtx);
	}

	/* terminate table */

	strcpy(bfr+len,"</table></body></html>");

	/* return html data */

	return bfr;
}
