
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "stats.h"
#include "sctp_om.h"
#include "sctp6_om.h"
#include "ipc.h"
#include "term.h"
#include "sigs.h"
#include "log.h"
#include "caps.h"
#include "lock.h"
#include "safeapi.h"
#include "version.h"
#include "stamp.h"

typedef struct
{
	uint32_t jobs;
	uint32_t nrcpu;
} STATMSG;

typedef struct _data
{
	struct _data *next;
	clock_t stamp;
	int jobs;
	int nrcpu;
	int val;
	union
	{
		struct sockaddr_in6 a6;
		struct sockaddr_in a4;
	};
} DATA;

static int running=1;
static int sock=-1;
static int authlen=0;
static int v6;
static int port;
static char *hostaddr;
static char auth[8192];
#if defined(__linux__) || defined(__FreeBSD__)
static int ev;
#else
static int ev[2];
#endif

/*
 * handler
 *
 * input:  depends on how signals are handled
 *
 * return: depends on how signals are handled
 *
 * description:	This routine is the signal handler. It either
 *		get called traditionally when a signal is
 *		invoked or prefessably polls a file
 *		descriptor that becomes readable in case
 *		of a signal. It calls the termination notifier
 *		routine when a signal is catched.
 */

#if defined(__linux__) || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
static void *handler(void *data)
{
	struct pollfd p[2];

	dropcaps(NEED_NONE);

	p[0].fd=(int)((long)data);
	p[0].events=POLLIN;
	p[1].fd=termget();
	p[1].events=POLLIN;

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
	while(poll(p,2,-1)<=0);
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	running=0;
	termnotify();
	ipc_postevent();
	close(p[0].fd);
	pthread_exit(NULL);
}
#else
static void handler(int unused)
{
	running=0;
	termnotify();
	ipc_postevent();
}
#endif

/*
 * evcreate
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine creates a pollable event descriptor.
 */

static inline int evcreate(void)
{
	int i;
#if defined(__FreeBSD__)
	struct kevent evnt;
#endif
	char mem[256];

#if defined(__linux__)
#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((ev=eventfd(0,EFD_CLOEXEC))==-1))
			BAD("eventfd: %s",errstr(errno,mem));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((ev=eventfd(0,0))==-1))
		{
			i=errno;
			unlock();
			BAD("eventfd: %s",errstr(i,mem));
		}
		fcntl(ev,F_SETFD,FD_CLOEXEC);
		unlock();
	}
#elif defined(__FreeBSD__)
	if(UNLIKELY((ev=kqueue())==-1))BAD("kqueue: %s",errstr(errno,mem));
	EV_SET(&evnt,0x100,EVFILT_USER,EV_ADD|EV_ENABLE|EV_CLEAR,NOTE_FFNOP,0,
		EMPTY);
	if(UNLIKELY(kevent(ev,&evnt,1,NULL,0,NULL)==-1))
	{
		i=errno;
		close(ev);
		BAD("kevent: %s",errstr(i,mem));
	}
#else
#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY(pipe2(ev,O_CLOEXEC)))
			BAD("pipe: %s",errstr(errno,mem));
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY(pipe(ev)))
		{
			i=errno;
			unlock();
			BAD("pipe: %s",errstr(i,mem));
		}
		fcntl(ev[0],F_SETFD,FD_CLOEXEC);
		fcntl(ev[1],F_SETFD,FD_CLOEXEC);
		unlock();
	}
#endif
	return 0;
}

/*
 * evdelete
 *
 * description:	This routine deletes the pollcable event descriptor.
 */

static inline void evdelete(void)
{
#if defined(__linux__) || defined(__FreeBSD__)
	close(ev);
#else
	close(ev[0]);
	close(ev[1]);
#endif
}

/*
 * evpost
 *
 * description:	This routine posts an event to the pollable event descriptor.
 */

static inline void evpost(void)
{
#if defined(__linux__)
	uint64_t evnt;
#elif defined(__FreeBSD__)
	struct kevent evnt;
#endif
#if !defined(__FreeBSD__)
	UNUSED(char dummy);
#endif

#if defined(__linux__)
	evnt=1;
	dummy=write(ev,&evnt,sizeof(evnt));
#elif defined(__FreeBSD__)
	EV_SET(&evnt,0x301,EVFILT_USER,0,NOTE_TRIGGER,0,EMPTY);
	kevent(ev,&evnt,1,NULL,0,NULL);
#else
	dummy=write(ev[1],&dummy,1);
#endif
}

/*
 * evget
 *
 * description:	This routine remeves an event from the pollable event
 *		descriptor.
 */

static inline void evget(void)
{
#if defined(__linux__)
	uint64_t evnt;
#elif defined(__FreeBSD__)
	struct kevent evnt;
#endif
#if !defined(__FreeBSD__)
	UNUSED(char dummy);
#endif

#if defined(__linux__)
	dummy=read(ev,&evnt,sizeof(evnt));
#elif defined(__FreeBSD__)
	kevent(ev,NULL,0,&evnt,1,NULL);
#else
	dummy=read(ev[0],&dummy,1);
#endif
}

/*
 * evfd
 *
 * return: the pollable event descriptor
 *
 * description:	This routine returns the pollable event descriptor.
 */

static inline int evfd(void)
{
#if defined(__linux__) || defined(__FreeBSD__)
	return ev;
#else
	return ev[0];
#endif
}

/*
 * eventwaiter
 *
 * input:  unused - unused
 *
 * return: always NULL
 *
 * description:	This routine is a thread that maps SYSV semaphore events
 *		to pollable events.
 */

static void *eventwaiter(void *unused)
{
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	dftblock(THREAD);
	dropcaps(NEED_NONE);

	while(LIKELY(running))
	{
		if(UNLIKELY(ipc_waitevent()))break;
		if(UNLIKELY(!running))break;
		evpost();
	}
	pthread_exit(NULL);
}

/*
 * tx
 *
 * input:  unused - unused
 *
 * return: always NULL
 *
 * description:	This routine waits for netccd job events and sents them
 *		to the master on the physical host.
 */

static void *tx(void *unused)
{
	int tmo;
	int jobs;
	int nrcpu;
	struct pollfd p[2];
	struct sockaddr_in6 a6;
	struct sockaddr_in a4;
	STATMSG msg;
	char mem[256];

	/* default thread setup */

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	dftblock(THREAD);
	dropcaps(NEED_NONE);

	/* fd setup */

	p[0].fd=evfd();
	p[0].events=POLLIN;
	p[1].fd=termget();
	p[1].events=POLLIN;

	/* data setup */

	jobs=0;
	tmo=-1;

	/* destination setup */

	if(SWITCHES_SUPPORT_MLDV2(v6))
	{
		inet_pton(AF_INET6,hostaddr,&a6.sin6_addr);
		a6.sin6_family=AF_INET6;
		a6.sin6_port=htons(port);
	}
	else
	{
		inet_pton(AF_INET,hostaddr,&a4.sin_addr);
		a4.sin_family=AF_INET;
		a4.sin_port=htons(port);
	}

	/* do until process termination */

	do
	{
		/* wait for event or timeout */

		while(poll(p,2,tmo)<0);

		/* termination event */

		if(UNLIKELY(p[1].revents&POLLIN))jobs=0;

		/* job event */

		else if(p[0].revents&POLLIN)
		{
			evget();
			jobs=ipc_get(JOBCOUNT);
		}

		/* set up next poll timeout */

		if(!jobs)tmo=-1;
		else tmo=1000;

		/* get number of online cpus */

		getstats(NUMCPU_STATS,&nrcpu);

		/* prepare message */

		msg.jobs=htonl(jobs);
		msg.nrcpu=htonl(nrcpu);

		/* send message */

		if(UNLIKELY(sendto(sock,&msg,sizeof(msg),MSG_EOR,
			SWITCHES_SUPPORT_MLDV2(v6)?
			(struct sockaddr *)&a6:(struct sockaddr *)&a4,
			SWITCHES_SUPPORT_MLDV2(v6)?sizeof(a6):sizeof(a4))==-1))
				WARN("sendto: %s",errstr(errno,mem));
	}
	while(LIKELY(running));

	/* terminate thread */

	pthread_exit(NULL);
}

/*
 * rx
 *
 * input:  unused - unused
 *
 * return: always NULL
 *
 * description:	This routine received the jobs and cpu data from the
 *		virtual hosts and adjusts the physical host jobs load
 *		accordingly.
 */

static void *rx(void *unused)
{
	int tmo;
	int idx;
	int total;
	int busy;
	int prev;
	int nrcpu;
	int add;
	int v;
	socklen_t len;
	clock_t curr;
	clock_t last;
	clock_t intv;
	clock_t delta;
	struct pollfd p[3];
	struct sockaddr_in6 a6;
	struct sockaddr_in a4;
	STATMSG msg;
	DATA *c;
	DATA *e;
	DATA **d;
	DATA *tab[16];

	/* default thread setup */

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	dftblock(THREAD);
	dropcaps(NEED_NONE);

	/* data setup */

	memset(tab,0,sizeof(tab));
	memset(&a6,0,sizeof(a6));
	memset(&a4,0,sizeof(a4));
	c=NULL;
	delta=sysconf(_SC_CLK_TCK)<<2;
	intv=sysconf(_SC_CLK_TCK);
	last=timestamp();
	total=0;
	busy=0;
	prev=0;
	add=0;

	/* fd setup */

	p[0].fd=evfd();
	p[0].events=POLLIN;
	p[1].fd=termget();
	p[1].events=POLLIN;
	p[2].fd=sock;
	p[2].events=POLLIN;

	/* do until process termination */

	do
	{
		/* assert that we have a free data element */

		if(!c)if(UNLIKELY(!(c=malloc(sizeof(DATA)))))
		{
			WARN("malloc failure");
			usleep(20000);
			continue;
		}

		/* get current time */

		curr=timestamp();

		/* if the cleaning interval is reacted */

		if(curr-last>=intv)
		{
			/* remove all expired entries */

			for(idx=0;idx<16;idx++)for(d=&tab[idx];*d;)
				if(curr-(*d)->stamp>delta)
			{
				e=*d;
				*d=e->next;
				busy-=e->val;
				free(e);
				total--;
			}
			else d=&(*d)->next;
			last=curr;
		}

		/* set up now poll timeout */

		if(total)tmo=500;
		else tmo=-1;

		/* in case the busy cpu amount did change */

		if(prev!=busy)
		{
			/* get online cpu count */

			getstats(NUMCPU_STATS,&nrcpu);

			/* limit busy cpu count to online cpu amount */

			if(busy>nrcpu)v=nrcpu;
			else v=busy;

			/* calculate new virtual load to be used */

			v=(v*(nrcpu+2)+(nrcpu>>1))/nrcpu;

			/* if the virtual load did increase */

			if(v>add)
			{
				/* adjust load, memorize new value */

				ipc_superadj(JOBCOUNT,v-add);
				add=v;
			}

			/* if the virtual load did decrease */

			else if(v<add)
			{
				/* adjust load */

				ipc_superadj(JOBCOUNT,v-add);

				/* wake netccds */

				ipc_lock(JOBCOND);
				ipc_signal(JOBCOND);
				ipc_unlock(JOBCOND);

				/* memorize new value */

				add=v;
			}

			/* memorize current amount of busy cpus */

			prev=busy;
		}

		/* wait for next event */

		while(poll(p,3,tmo)<0);

		/* termination pending */

		if(UNLIKELY(p[1].revents&POLLIN))break;

		/* job change event (ignored on master) */

		if(p[0].revents&POLLIN)evget();

		/* notify from virtual host */

		if(!(p[2].events&POLLIN))continue;

		/* get message, do basic message check */

		len=SWITCHES_SUPPORT_MLDV2(v6)?sizeof(a6):sizeof(a4);

		if(recvfrom(sock,&msg,sizeof(msg),0,
			SWITCHES_SUPPORT_MLDV2(v6)?
			(struct sockaddr *)&a6:(struct sockaddr *)&a4,&len)
			!=sizeof(msg))continue;
		else if(len!=(SWITCHES_SUPPORT_MLDV2(v6)?sizeof(a6):sizeof(a4)))
			continue;
		else if(SWITCHES_SUPPORT_MLDV2(v6))
		{
			if(a6.sin6_family!=AF_INET6)continue;
		}
		else  if(a4.sin_family!=AF_INET)continue;

		/* remove previos stored message from this virtual host */

		if(SWITCHES_SUPPORT_MLDV2(v6))idx=a6.sin6_addr.s6_addr[15]&0xf;
		else idx=ntohl(a4.sin_addr.s_addr)&0xf;

		for(d=&tab[idx];*d;d=&(*d)->next)
		{
			if(SWITCHES_SUPPORT_MLDV2(v6))
			{
				if(memcmp(&a6,&(*d)->a6,sizeof(a6)))continue;
			}
			else if(memcmp(&a4,&(*d)->a4,sizeof(a4)))continue;
			e=*d;
			*d=e->next;
			busy-=e->val;
			free(e);
			total--;
			break;
		}

		/* store virtual host message */

		c->next=tab[idx];
		c->stamp=timestamp();
		c->jobs=ntohl(msg.jobs);
		c->nrcpu=ntohl(msg.nrcpu);
		c->val=c->nrcpu<c->jobs?c->nrcpu:c->jobs;
		if(SWITCHES_SUPPORT_MLDV2(v6))c->a6=a6;
		else c->a4=a4;
		tab[idx]=c;
		total++;
		busy+=c->val;
		c=NULL;
	}
	while(LIKELY(running));

	/* remove any virtual load and notify netccds */

	if(add)
	{
		ipc_superadj(JOBCOUNT,-add);
		ipc_lock(JOBCOND);
		ipc_signal(JOBCOND);
		ipc_unlock(JOBCOND);
	}

	/* free any new element */

	if(c)free(c);

	/* remove all stored elements */

	for(idx=0;idx<16;idx++)for(e=tab[idx];e;)
	{
		c=e;
		e=e->next;
		free(c);
	}

	/* terminate thread */

	pthread_exit(NULL);
}

/*
 * dogetauth
 *
 * input:  src - the authentication source
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine reas up to 8k authentication data from the
 *		given authentication source where '-' means stdin.
 */

static inline int dogetauth(char *src)
{
	int fd;

	/* open authentication source or use stdin */

	if(!strcmp(src,"-"))fd=0;
	else
	{
#ifdef SAFEAPI
		if(LIKELY(safeapi()))
		{
#ifndef SAFEUID
			shlock();
#endif
			if(UNLIKELY((fd=open(src,O_RDONLY|O_CLOEXEC))==-1))
			{
#ifndef SAFEUID
				unlock();
#endif
				BAD("can't access %s",src);
			}
#ifndef SAFEUID
			unlock();
#endif
		}
		else
#endif
		{
			shlock();
			if(UNLIKELY((fd=open(src,O_RDONLY))==-1))
			{
				unlock();
				BAD("can't access %s",src);
			}
			fcntl(fd,F_SETFD,FD_CLOEXEC);
			unlock();
		}
	}

	/* read authentication data */

	if((authlen=read(fd,auth,sizeof(auth)))<=0)
	{
		if(fd>0)close(fd);
		BAD("can't read from %s",src);
	}

	/* close authentication source if not stdin */

	if(fd>0)close(fd);

	/* signal success */

	return 0;
}

/*
 * addrchk
 *
 * input:  addr - address string
 *         v6   - 1 if IPv6 address, 0 if IPv4 address
 *
 * return: 0 if valid address, else -1
 *
 * description:	This routine checks the given IP address for validity.
 *		The address must not be a multicast address.
 */

static inline int addrchk(char *addr,int v6)
{
	union
	{
		struct in6_addr a6;
		struct in_addr a4;
	} a;

	if(UNLIKELY(inet_pton(v6?AF_INET6:AF_INET,addr,&a)!=1))return -1;
	if(SWITCHES_SUPPORT_MLDV2(v6))
	{
		if(a.a6.s6_addr[0]==0xff)return -1;
	}
	else if((ntohl(a.a4.s_addr)&0xe0000000)==0xe0000000)return -1;
	return 0;
}

/*
 * usage
 *
 * description:	This routine prints a usage message and terminates processing.
 */

static NORETURN void usage(void)
{
	/* WARNING: You are *NOT* allowed to remove copyright and version
	  from the usage message - you are also *NOT* allowed to remove
	  the usage function from the program in *ANY* way */

	fprintf(stderr,"Usage: " NETCCSVD " OPTIONS [master-ip]\n"
		COPYRIGHT "\n"
		"Version: " VERSION "\n"
		"master-ip is the IP address of the master daemon\n"
		"OPTIONS:\n"
		"-p port              network port (default " STR(SVPORT) ")\n"
		"-6                   use IPv6 instead of IPv4\n"
		"-f                   do not daemonize, log to stderr too\n"
		"-P pid-file          pid file (default " SVPIDFILE ")\n"
		"-k ipc-key           system wide netccd interworking using ipc-key\n"
		"-b named-pipe        BSD workaround fifo (default " FIFO ")\n"
		"-a auth-source       use authentication from auth-source (- for stdin)\n"
		"-5                   work around SCTP COOKIE ACK kernel auth bug\n"
		"-o cpu-number        manually set the amount of available cpu cores\n"
		"-S cpu-speed         manually set the cpu clock speed in MHz\n"
		"-y                   manually mark the system as not hyperthreaded\n"
		"-Y                   manually mark the system as hyperthreaded\n"
		"-J                   disable lockless operation mode\n"
		"-D dumpdir           enable core dumps in specified directory\n"
		"-h                   this help text\n");
	exit(1);
}

/*
 * main
 *
 * input:  argc - argument count
 *         argv - argument array
 *
 * return: 0 in case of success, a value > 0 in case of error
 *
 * description:	This routine handles the netccsvd daemon startup and shutdown.
 */

int main(int argc,char *argv[])
{
	int rval;
	int val;
	int fg;
	int locking;
	int cpunum;
	int cpufreq;
	int htflag;
	int ipckey;
	int tstflg;
	int kernel_bug_workaround;
	char *bsdpipe;
	char *pidfile;
	char *dumpdir;
	char *authsrc;
	FILE *fp;
	struct rlimit r;
	pthread_t svm;
	pthread_t w;

	rval=1;
	tstflg=0;
	port=SVPORT;
	fg=0;
	locking=0;
	cpunum=-1;
	cpufreq=-1;
	htflag=-1;
	v6=0;
	ipckey=0;
	bsdpipe=FIFO;
	pidfile=SVPIDFILE;
	dumpdir=NULL;
	authsrc=NULL;
	hostaddr=NULL;
	kernel_bug_workaround=0;

	/* signal handling setup */

#if defined(__linux__) || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	dftblock(ALLSIGS);
#else
	sethandler(handler);
	dftblock(PROCESS);
#endif

	/* parse command line options */

	while((val=getopt(argc,argv,"p:fJo:S:yY6D:P:b:k:a:h5"))!=-1)switch(val)
	{
	case 'p':
		port=atoi(optarg);
		break;

	case 'f':
		fg=1;
		break;

	case 'J':
		locking=1;
		break;

	case 'o':
		cpunum=atoi(optarg);
		tstflg|=1;
		break;

	case 'S':
		cpufreq=atoi(optarg);
		tstflg|=2;
		break;

	case 'y':
		htflag=0;
		break;

	case 'Y':
		htflag=1;
		break;

	case '6':
		v6=1;
		break;

	case 'D':
		dumpdir=optarg;
		break;

	case 'P':
		pidfile=optarg;
		break;

	case 'b':
		bsdpipe=optarg;
		break;

	case 'k':
		ipckey=(int)strtol(optarg,NULL,0);
		tstflg|=4;
		break;

	case 'a':
		authsrc=optarg;
		break;

	case '5':
		kernel_bug_workaround=1;
		break;

	case 'h':
	default:usage();
	}

	/* get master ip if specified */

	if(optind<argc)
	{
		hostaddr=argv[optind++];
		if(optind<argc)usage();
	}

	/* open syslog */

	openlog(NETCCSVD,LOG_NDELAY|LOG_PID|(fg?LOG_PERROR:0),LOG_DAEMON);

	/* asert that we're root */

	if(UNLIKELY(getuid()||geteuid()||getgid()||getegid()))
		FAIL("must be root");

	/* yes, we really want that - file access ust use proper rights */

	umask(0);

	/* basic process setup */

	if(UNLIKELY(initlock()))FAIL("can't inizialize locking");
	if(LIKELY(!locking))if(UNLIKELY(enablesafeapi()))
		FAIL("can't select system api");

	/* coredump handling */

	if((UNLIKELY(dumpdir!=NULL)))if(UNLIKELY(chdir(dumpdir)))
		FAIL("can't access core dump directory");
	r.rlim_cur=r.rlim_max=UNLIKELY(dumpdir!=NULL)?RLIM_INFINITY:0;
	if(UNLIKELY(setrlimit(RLIMIT_CORE,&r)))FAIL("can't set core dump mode");

	/* paramter checks */

	if(tstflg&1)if(UNLIKELY(cpunum<1))FAIL("illegal cpu number value");
	if(tstflg&2)if(UNLIKELY(cpufreq<1))FAIL("illegal cpu frequency value");
	if(!(tstflg&4))FAIL("must specify ipc key");
	if(UNLIKELY(!ipckey||ipckey==-1))FAIL("illegal ipc key");
	if(hostaddr)if(UNLIKELY(addrchk(hostaddr,v6)))
		FAIL("illegal host address specified");

	/* init stats gathering */

	if(UNLIKELY(getstats(INIT_STATS,&cpunum,&cpufreq,&htflag)))
		FAIL("cant process /proc/cpuinfo");

	/* get authentication data */

	if(authsrc)if(dogetauth(authsrc))FAIL("can't set up authentication");

	/* initialize termination event */

	if(UNLIKELY(terminit()))
		FJMP(f1,"termination initialization failure");

	/* in daemon mode */

	if(!fg)
	{
		/* open/create pid file */

#ifdef SAFEAPI
		if(LIKELY(safeapi()))
		{
			if(UNLIKELY((val=open(pidfile,
				O_WRONLY|O_CREAT|O_TRUNC|O_EXCL,0644))==-1))
					FJMP(f2,"can't access pid file");
		}
		else
#endif
		{
			shlock();
			if(UNLIKELY((val=open(pidfile,
				O_WRONLY|O_CREAT|O_TRUNC,0644))==-1))
			{
				unlock();
				FJMP(f2,"can't access pid file");
			}
			fcntl(val,F_SETFD,FD_CLOEXEC);
			unlock();
		}
		if(UNLIKELY(!(fp=fdopen(val,"w"))))
		{
			close(val);
			FJMP(f3,"can't access pid file");
		}

		/* daemonize */

		if(UNLIKELY(daemon(UNLIKELY(dumpdir!=NULL)?1:0,0)))
		{
			fclose(fp);
			FJMP(f3,"can't daemonize");
		}

		/* write pid to file and close pid file */

		fprintf(fp,"%d\n",getpid());
		fclose(fp);
	}

	/* join system wide netccd ipc as supervisor */

	if(UNLIKELY(ipc_join(ipckey,bsdpipe,1)))
		FJMP(f3,"shared resource setup failure");

	/* create sctp one to many socket */

	if(SWITCHES_SUPPORT_MLDV2(v6))
	{
		if(UNLIKELY((sock=sctp6_omsocket(NULL,port,255,auth,authlen,
			kernel_bug_workaround))==-1))
				FJMP(f4,"can't create SCTP socket");
	}
	else if(UNLIKELY((sock=sctp_omsocket(NULL,port,255,auth,authlen,
		kernel_bug_workaround))==-1))
			FJMP(f4,"can't create SCTP socket");

	/* clear authentication data */

	memset(auth,0,sizeof(auth));

	/* create evend fd */

	if(UNLIKELY(evcreate()))FJMP(f5,"can't create notify event");

	/* create SYSV to event mapping thread */

	if(UNLIKELY(pthread_create(&svm,NULL,eventwaiter,NULL)))
		FJMP(f6,"can't create event mapping thread");

	/* create proper thread according to processing mode */

	if(UNLIKELY(pthread_create(&w,NULL,hostaddr?tx:rx,NULL)))
	{
		running=0;
		ipc_postevent();
		termnotify();
		FJMP(f7,"can't create event processing thread");
	}

	/* create modern style signal handling thread */

#if defined(__linux__) || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	if(UNLIKELY(sethandler(handler)))
	{
		running=0;
		ipc_postevent();
		termnotify();
		FJMP(f8,"signal handling thread setup failed");
	}
#endif

	/* drop capabilities no longer required */

	dropcaps(NEED_ROOT);

	/* set return value to success */

	rval=0;

	/* wait for threads and clean up */

f8:	pthread_join(w,NULL);
f7:	pthread_join(svm,NULL);
f6:	evdelete();
f5:	close(sock);
f4:	ipc_leave(1);
f3:
#ifndef SAFEUID
	shlock();
#endif
	if(!fg)unlink(pidfile);
#ifndef SAFEUID
	unlock();
#endif
f2:	termexit();
f1:	memset(auth,0,sizeof(auth));
	exitlock();

	/* signal success or error */

	return rval;
}
