
/*
 * This file is part of the netcc project
 *
 * (C) 2011 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2.
 */

#include "headers.h"
#include "md5.h"
#include "common.h"
#include "term.h"
#include "log.h"
#include "lock.h"
#include "watch.h"
#include "find.h"
#include "update.h"
#include "remote.h"
#include "sigs.h"
#include "ipc.h"
#include "safeapi.h"
#include "caps.h"

/******************************************************************************

Race Condition Warning:

The code below expects reasonable system behaviour and no playing around.
It is prone to race conditions if the watched item is moved around
precisely enough. This, however, doesn't matter for the use case. The
watched item typically belongs to root and if root plays around the
system is doomed anyway.
This means: the code is suitable to detect compiler changes due to
installation of a new compiler version. The code can also detect compiler
changes due to distro tools changing the compiler setup. Anything else
first needs to be evaluated.

******************************************************************************/

#if defined(__linux__)
#define FOPTS FIND_FILE|FIND_LINK|FIND_ABSOLUTE
#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
#define FOPTS FIND_FILE|FIND_ABSOLUTE
#else
#error do not know which options to use to get watched file path
#endif

typedef struct
{
	int h;		/* pollable watcher descriptor	*/
	int d;		/* watched item descriptor	*/
	int f;		/* child pipe (reading side)	*/
	pid_t p;	/* child pid			*/
} WATCH;

static int running=0;
static pthread_t h;
static char *pn;
static char *fn;
static char *vo;

/*
 * file_gone
 *
 * description:	This routine handles the case of the gcc binary missing.
 */

static inline void file_gone(void)
{
	/* notify remote compile server and udp processing modules */

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	update_mode(GCC_VERSION,0,NULL);
	remotemgr(RMT_KILL,NULL,0);
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
}

/*
 * file_back
 *
 * input:  w - file watch data that nead cancel cleanup
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine handles the case when the gcc binary is
 *		available (again). It executes the binary to inquire
 *		the compiler version.
 */

static inline int file_back(WATCH *w)
{
	int l;
	int manum;
	int minum;
	int status;
	int p[2];
	char *major;
	char *minor;
	char *save;
	char bfr[256];

	/* create pipe for read from child process */

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY(pipe2(p,O_CLOEXEC)))
		{
			pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
			BAD("pipe failure");
		}
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY(pipe(p)))
		{
			unlock();
			pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
			BAD("pipe failure");
		}
		fcntl(p[0],F_SETFD,FD_CLOEXEC);
		fcntl(p[1],F_SETFD,FD_CLOEXEC);
		unlock();
	}

	/* fork */

#if defined(SAFEAPI) && defined(SAFEUID)
	if(UNLIKELY(!safeapi()))exlock();
#else
#ifndef SAFEUID
	ipc_needex();
#endif
	exlock();
#endif
	switch((w->p=fork()))
	{
	/* fork failure */

	case -1:close(p[0]);
		close(p[1]);
#if defined(SAFEAPI) && defined(SAFEUID)
	if(UNLIKELY(!safeapi()))unlock();
#else
		unlock();
#ifndef SAFEUID
		ipc_exdone();
#endif
#endif
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
		BAD("fork failure");

	/* child process */

	case 0:	/* set up file handles and umask */

		close(p[0]);
		if(p[1]!=1)
		{
#ifdef SAFEAPI
			if(LIKELY(safeapi()))
			{
				dup3(p[1],1,0);
				close(p[1]);
			}
			else
#endif
			{
				dup2(p[1],1);
				fcntl(1,F_SETFD,0);
				close(p[1]);
			}
		}
		else fcntl(1,F_SETFD,0);
		umask(077);

		/* execute gcc with version retrieval option */

		execl(pn,pn,vo,(char *)NULL);
		ERR("can't execute %s",pn);
		exit(255);

	/* parent process */

	default:/* manage pipe handles */

		close(p[1]);
		w->f=p[0];
#if defined(SAFEAPI) && defined(SAFEUID)
		if(UNLIKELY(!safeapi()))unlock();
#else
		unlock();
#ifndef SAFEUID
		ipc_exdone();
#endif
#endif
	}

	/* read version string from gcc */

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
	if(UNLIKELY((l=read(p[0],bfr,sizeof(bfr)-1))<=0))
	{
		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
		kill(w->p,SIGKILL);
		waitpid(w->p,NULL,0);
		w->p=-1;
		close(p[0]);
		w->f=-1;
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
		BAD("read failure");
	}
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);

	/* close pipe to child process, terminate buffer */

	close(p[0]);
	w->f=-1;
	bfr[l]=0;

	/* wait for child process to terminate */

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

	while(UNLIKELY((l=waitpid(w->p,&status,0))<=0))
		if(UNLIKELY(l==-1&&errno!=EINTR))
	{
		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
		kill(w->p,SIGKILL);
		waitpid(w->p,NULL,0);
		w->p=-1;
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
		BAD("waitpid");
	}

	/* handle child process error status */

	if(UNLIKELY(!WIFEXITED(status)))
		BAD("%s did not regularly terminate",pn);
	if(UNLIKELY(WEXITSTATUS(status)))BAD("%s terminated with error",pn);

	/* split major an minor from version number supplied by gcc */

	major=strtok_r(bfr,".\n",&save);
	minor=strtok_r(NULL,".\n",&save);

	/* sanity checks */

	if(UNLIKELY(!major||!minor))BAD("version retrieval for %s failed",pn);

	manum=atoi(major);
	minum=atoi(minor);

	if(UNLIKELY(manum<1||manum>255))
		BAD("major version %d of %s out of range",manum,pn);
	if(UNLIKELY(minum<0||minum>255))
		BAD("minor version %d of %s out of range",manum,pn);

	/* notify udp processing thread of new gcc version number */

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	update_mode(GCC_VERSION,(manum<<8)|minum,NULL);
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

	/* signal success */

	return 0;
}

/*
 * watcherclean
 *
 * input: data - the data to be cleaned up
 *
 * description:	This routine is the cancellation cleanup routine for
 *		the gcc watcher thread.
 */

static void watcherclean(void *data)
{
	WATCH *w=(WATCH *)data;

	/* if a child process is running kill it */

	if(w->p!=-1)
	{
		kill(w->p,SIGKILL);
		waitpid(w->p,NULL,0);
	}

	/* release watched object */

#if defined(__linux__)
	if(w->d!=-1)inotify_rm_watch(w->h,w->d);
#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
	if(w->d!=-1)close(w->d);
#else
#error probably some cleanup is missing here
#endif

	/* release watch descriptor */

	if(w->h!=-1)close(w->h);

	/* close child process pipe */

	if(w->f!=-1)close(w->f);
}

/*
 * watcher
 *
 * input:  unused - unused
 *
 * return: always NULL
 *
 * description:	This routine watches an item (typically the gcc binary)
 *		for modifications. After such a modification the gcc
 *		binary is executed to retrieve the current version.
 */

static void *watcher(void *unused)
{
#if defined(__linux__)

	WATCH w;
	struct pollfd p[2];
	struct inotify_event notused;
	struct stat stb;

	/* data setup */

	w.h=-1;
	w.d=-1;
	w.f=-1;
	w.p=-1;

	/* thread default setup */

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	dftblock(THREAD);
	dropcaps(NEED_NONE);
	pthread_cleanup_push(watcherclean,&w);

	/* create notification socket */

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
		if(UNLIKELY((w.h=inotify_init1(IN_CLOEXEC))==-1))
			GOTO(fail,"inotify_init1 failure");
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((w.h=inotify_init())==-1))
		{
			unlock();
			GOTO(fail,"inotify_init failure");
		}
		fcntl(w.h,F_SETFD,FD_CLOEXEC);
		unlock();
	}

	/* set up watched object */

	if(UNLIKELY((w.d=inotify_add_watch(w.h,fn,IN_DELETE_SELF|
			IN_MODIFY|IN_MOVE_SELF|IN_ONESHOT|IN_DONT_FOLLOW))<0))
			GOTO(fail,"inotify_add_watch failure");

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

	/* preset poll data */

	p[0].fd=w.h;
	p[0].events=POLLIN;
	p[1].fd=termget();
	p[1].events=POLLIN;

	/* do until termination time */

	while(LIKELY(running))
	{
		/* wait for event, exit in case of pending termination */

		switch(poll(p,2,-1))
		{
		case -1:if(errno!=EINTR)GOTO(fail,"poll failure");
		case 0:	continue;
		default:if(UNLIKELY(p[1].revents&POLLIN))goto fail;
			if(UNLIKELY(!p[0].revents&POLLIN))continue;
		}

		/* now see what's happened */

		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
		if(UNLIKELY(read(w.h,&notused,sizeof(notused))==-1))
			GOTO(fail,"read failure");
		if(notused.mask==IN_IGNORED)continue;

		/* something happened, assume watched object went away */

		w.d=-1;
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

		file_gone();

		/* now check, until the watched object is available again */

repeat:		while(1)
		{
			switch(poll(&p[1],1,100))
			{
			case -1:if(errno!=EINTR)GOTO(fail,"poll failure");
				continue;
			case 0:	break;;
			default:if(UNLIKELY(p[0].revents&POLLIN))goto fail;
			}

			if(UNLIKELY(stat(fn,&stb)))continue;
			if(!S_ISREG(stb.st_mode))
				GOTO(fail,"%s is no longer a file",fn);

			break;
		}

		/* set up watch again */

		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
		if(UNLIKELY((w.d=inotify_add_watch(w.h,fn,IN_DELETE_SELF|
			IN_MODIFY|IN_MOVE_SELF|IN_ONESHOT|IN_DONT_FOLLOW))<0))
				GOTO(repeat,"inotify_add_watch failure");
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

		/* do required stuff (gcc version retrieval) */

		if(UNLIKELY(file_back(&w)))
		{
			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
			inotify_rm_watch(w.h,w.d);
			pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
			goto repeat;
		}
	}

	/* clean up */

fail:	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	pthread_cleanup_pop(0);
	if(w.d!=-1)inotify_rm_watch(w.h,w.d);
	if(w.h!=-1)close(w.h);

	/* exit thread */

	pthread_exit(NULL);

#elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)

	int t;
	WATCH w;
	struct kevent k[2];
	struct pollfd p;
	struct stat stb;

	/* thread default setup */

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	dftblock(THREAD);
	dropcaps(NEED_NONE);
	pthread_cleanup_push(watcherclean,&w);

	/* data setup */

	w.h=-1;
	w.d=-1;
	w.f=-1;
	w.p=-1;

	/* get termination notifier file descriptor */

	t=termget();

	/* get descriptor of watched object (this is a deficiency
	   leading to the fact that symlinks can't be directly
	   watched - I refrain from the PITA having to watch a
	   directory and lstat around to handle symlinks) */

#ifdef SAFEAPI
	if(LIKELY(safeapi()))
	{
#ifndef SAFEUID
		shlock();
#endif
		if(UNLIKELY((w.d=open(fn,O_RDONLY|O_CLOEXEC))==-1))
		{
#ifndef SAFEUID
			unlock();
#endif
			GOTO(fail,"can't access %s",fn);
		}
#ifndef SAFEUID
		unlock();
#endif
	}
	else
#endif
	{
		shlock();
		if(UNLIKELY((w.d=open(fn,O_RDONLY))==-1))
		{
			unlock();
			GOTO(fail,"can't access %s",fn);
		}
		fcntl(w.d,F_SETFD,FD_CLOEXEC);
		unlock();
	}

	/* get notification descriptor */

#ifndef SAFEUID
	shlock();
#endif
	if(UNLIKELY((w.h=kqueue())==-1))
	{
#ifndef SAFEUID
		unlock();
#endif
		GOTO(fail,"kqueue failure");
	}
#ifndef SAFEUID
	unlock();
#endif

	/* add watches to the notification descriptor */

	EV_SET(&k[0],w.d,EVFILT_VNODE,EV_ADD|EV_ENABLE|EV_ONESHOT,
		NOTE_DELETE|NOTE_WRITE|NOTE_EXTEND|NOTE_RENAME|NOTE_REVOKE,
		0,EMPTY);
	EV_SET(&k[1],t,EVFILT_READ,EV_ADD|EV_ENABLE,0,0,EMPTY);

#ifndef SAFEUID
	shlock();
#endif
	if(UNLIKELY(kevent(w.h,k,2,NULL,0,NULL)==-1))
	{
#ifndef SAFEUID
		unlock();
#endif
		GOTO(fail,"kevent failure");
	}
#ifndef SAFEUID
	unlock();
#endif

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

	/* do until termination time */

	while(running)
	{
		/* wait for event to occur and exit in case of pending
		   termination - the whoes of BrokenBSD (OpenBSD)
		   need EINTR cluttering */

		if(UNLIKELY(kevent(w.h,NULL,0,k,1,NULL)==-1))
		{
			if(LIKELY(errno==EINTR))continue;
			GOTO(fail,"kevent failure");
		}
		if(UNLIKELY(k[0].flags==EV_ERROR))GOTO(fail,"kevent failure");
		if(UNLIKELY(k[0].ident==t))goto fail;
		if(UNLIKELY(k[0].ident!=w.d))continue;

		/* remove watch */

		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
		EV_SET(&k[0],t,EVFILT_READ,EV_DELETE,0,0,EMPTY);
#ifndef SAFEUID
		shlock();
#endif
		kevent(w.h,k,1,NULL,0,NULL);
#ifndef SAFEUID
		unlock();
#endif
		close(w.d);
		w.d=-1;
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

		/* assume that watched object is gone */

		file_gone();

		/* check, until watched object is available again */

		p.fd=t;
		p.events=POLLIN;

repeat:		while(1)
		{
			switch(poll(&p,1,100))
			{
			case -1:if(errno!=EINTR)GOTO(fail,"poll failure");
				continue;
			case 0:	break;;
			default:if(UNLIKELY(p.revents&POLLIN))goto fail;
			}

			if(UNLIKELY(stat(fn,&stb)))continue;
			if(!S_ISREG(stb.st_mode))
				GOTO(fail,"%s is no longer a file",fn);

			break;
		}

		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);

		/* get descriptor to watched object */

#ifdef SAFEAPI
		if(LIKELY(safeapi()))
		{
#ifndef SAFEUID
			shlock();
#endif
			if(UNLIKELY((w.d=open(fn,O_RDONLY|O_CLOEXEC))==-1))
			{
#ifndef SAFEUID
				unlock();
#endif
				pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,
					NULL);
				GOTO(repeat,"can't access %s",fn);
			}
#ifndef SAFEUID
			unlock();
#endif
		}
		else
#endif
		{
			shlock();
			if(UNLIKELY((w.d=open(fn,O_RDONLY))==-1))
			{
				unlock();
				pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,
					NULL);
				GOTO(repeat,"can't access %s",fn);
			}
			fcntl(w.d,F_SETFD,FD_CLOEXEC);
			unlock();
		}

		/* ad object to watch */

		EV_SET(&k[0],w.d,EVFILT_VNODE,EV_ADD|EV_ENABLE|EV_ONESHOT,
			NOTE_DELETE|NOTE_WRITE|NOTE_EXTEND|NOTE_RENAME|
				NOTE_REVOKE,0,EMPTY);
		EV_SET(&k[1],t,EVFILT_READ,EV_ADD|EV_ENABLE,0,0,EMPTY);

#ifndef SAFEUID
		shlock();
#endif
		if(UNLIKELY(kevent(w.h,k,2,NULL,0,NULL)==-1))
		{
#ifndef SAFEUID
			unlock();
#endif
			GOTO(fail,"kevent failure");
		}
#ifndef SAFEUID
		unlock();
#endif
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);

		/* do required stuff (gcc version retrieval) */

		if(UNLIKELY(file_back(&w)))
		{
			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
			EV_SET(&k[0],t,EVFILT_READ,EV_DELETE,0,0,EMPTY);
			EV_SET(&k[1],w.d,EVFILT_VNODE,EV_DELETE,0,0,EMPTY);
			kevent(w.h,k,2,NULL,0,NULL);
			close(w.d);
			w.d=-1;
			pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
			goto repeat;
		}
	}

	/* clean up */

fail:	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
	pthread_cleanup_pop(0);

	if(w.d!=-1)close(w.d);
	if(w.h!=-1)close(w.h);

	/* exit thread */

	pthread_exit(NULL);

#else
#error do not know how to watch files
#endif
}

/*
 * watchmgr
 *
 * input:  what - requested action
 *	   proc - gcc binary name
 *	   file - object to watch for gcc changes
 *	   opt  - the option to retrieve the compiler version
 *
 * return: 0 in case of success, -1 in case of failure
 *
 * description:	This routine manages the watch of the object that
 *		changes when the compiler changes.
 */

int watchmgr(int what,char *proc,char *file,char *opt)
{
	WATCH w;
	struct stat stb;

	switch(what)
	{
	/* start watch */

	case WATCH_START:

		/* done if already running */

		if(running)return 0;

		/* get absolute pathname of compiler binary */

#ifndef SAFEUID
		shlock();
#endif
		if(UNLIKELY(!(pn=find_path_file(proc,FIND_FILE))))
		{
#ifndef SAFEUID
			unlock();
#endif
			BAD("compiler not found");
		}

		/* get absolute pathname of watched object */

		if(!file)fn=pn;
		else if(UNLIKELY(!(fn=find_path_file(file,FOPTS))))
		{
#ifndef SAFEUID
			unlock();
#endif
			free(pn);
			BAD("watched file path setup failure");
		}

		/* assert that watched object exists ore the watched symlink
		   points to something existing */

		else if(stat(fn,&stb))
		{
#ifndef SAFEUID
			unlock();
#endif
			free(pn);
			free(fn);
			BAD("watched file does not exist");
		}

		/* assert that watched object is a regular file or that the
		   watched symlink points to a regular file */

		else if(!S_ISREG(stb.st_mode))
		{
#ifndef SAFEUID
			unlock();
#endif
			free(pn);
			free(fn);
			BAD("watched file is not a regular file");
		}
#ifndef SAFEUID
		unlock();
#endif
		/* store option to retrieve compiler version */

		if(UNLIKELY(!(vo=strdup(opt))))
		{
			free(pn);
			if(fn!=pn)free(fn);
			BAD("malloc failure");
		}

		/* get version from compiler and update udp processing thread */

		if(UNLIKELY(file_back(&w)))
		{
			free(vo);
			free(pn);
			if(fn!=pn)free(fn);
			BAD("version retrieval failure");
		}

		/* start watcher thread */

		running=1;
		if(UNLIKELY(pthread_create(&h,NULL,watcher,NULL)))
		{
			running=0;
			free(vo);
			free(pn);
			if(fn!=pn)free(fn);
			BAD("watcher thread start failure");
		}

		/* signal success */

		return 0;

	/* end watch */

	case WATCH_STOP:

		/* done if watch is terminated */

		if(!running)return 0;

		/* mark therad as not running, cancel thread */

		running=0;
		pthread_cancel(h);
		pthread_join(h,NULL);

		/* release working memory */

		free(pn);
		if(fn!=pn)free(fn);

		/* signal success */

		return 0;

	/* sanity check */

	default:BAD("internal error");
	}
}
